using System;

namespace Geometry
{
	/// <summary>
	/// Summary description for Bezier.
	/// </summary>
  public class Bezier : Line
  {
    #region nested classes
    private class MyInnerLine : InnerLine
    {
      private Point mp1;
      private Point mp2;

      public MyInnerLine(Point p1, Point p2)
      {
        mp1 = p1;
        mp2 = p2;
      }

      public override Line Complete(Point start, Point end)
      {
        return new Bezier(start,mp1,mp2,end);
      }
    }
    #endregion

    private Point mp1;
    private Point mp2;

    private bool mbParamKnown = false;

    private double[] mdA0;
    private double[] mdA1;
    private double[] mdA2;
    private double[] mdA3;

    public Bezier(Point p0,Point p1,Point p2,Point p3):
      base(p0,p3)
    {
      mp1 = p1;
      mp2 = p2;
    }

    public Point P1
    {
      get
      { return mp1; }
    }
    public Point P2
    {
      get
      { return mp2; }
    }

    public override Box Extent
    {
      get
      {
        Box pBox = base.Extent;
        pBox.Include(mp1);
        pBox.Include(mp2);
        return pBox;
      }
    }

    /// <summary>
    /// Remark: Check Bezier.Box.Intersects(box) beforehand for better overall performance
    /// </summary>
    /// <param name="box"></param>
    /// <returns></returns>
    public override bool Intersects(IBox box)
    {
      int iDim = Math.Min(this.Dimension,box.Dimension);
      return Intersects(box,Geometry.DimensionList(iDim),0);
    }

    public override bool Intersects(IBox box,int[] dimensionList)
    {
      return Intersects(box,dimensionList,0);
    }

    private bool Intersects(IBox box,int[] dimensionList,int pos)
    {
      if (Start.Intersects(box,dimensionList) || 
        End.Intersects(box,dimensionList))
      { return true; }
      if (this.Extent.Intersects(box) == false)
      { return false; }

      AssignParams();
      if (pos >= dimensionList.Length)
      {
        return true;
      }

      int iDim = dimensionList[pos];
      if (
        Start[iDim] >= box.Min[iDim]   && Start[iDim] <= box.Max[iDim] &&
        mp1[iDim]   >= box.Min[iDim]   && mp1[iDim]   <= box.Max[iDim] &&
        mp2[iDim]   >= box.Min[iDim]   && mp2[iDim]   <= box.Max[iDim] &&
        End[iDim]   >= box.Min[iDim]   && End[iDim]   <= box.Max[iDim])
      { // gesamte Ausdehnung ist im Bereich von box
        return Intersects(box,dimensionList,pos + 1);
      }

      bool b01 = Start[iDim] < mp1[iDim];
      bool b12 = mp1[iDim] < mp2[iDim];
      bool b23 = mp2[iDim] < End[iDim];
      if ((b01 == b12 && b12 == b23) ||
        (Start[iDim] == mp1[iDim] && 
        (b12 == b23 || mp1[iDim] == mp2[iDim] || mp2[iDim] == End[iDim])) ||
        (mp1[iDim] == mp2[iDim] && 
        (b01 == b23 || mp2[iDim] == End[iDim])) ||
        (mp2[iDim] == End[iDim] && b01 == b12))
      { // Verlauf von [iDim] monoton fuer t in [0,1]
        return Intervall(box,dimensionList,pos);
      }
      else
      {
        double t0,t1;
        // xmin / xmax
        // dx / dt = a1 + 2 * t * a2 + 3 * t^2 * a3 
        if (NumMath.SolveSqr(3 * mdA3[iDim], 2 * mdA2[iDim], mdA1[iDim],
          out t0,out t1))
        {
          if (t0 > 0 && t0 < 1)
          {
            if (Subpart(0,t0).Intervall(box,dimensionList,pos))
            { return true; }
            if (t1 < 1)
            {
              if (Subpart(t0,t1).Intervall(box,dimensionList,pos))
              { return true; }
              if (Subpart(t1,1).Intervall(box,dimensionList,pos))
              { return true; }
            }
            else if (Subpart(t0,1).Intervall(box,dimensionList,pos))
            { return true; }

            return false;
          }
          else if (t1 > 0 && t1 < 1)
          {
            if (Subpart(0,t1).Intervall(box,dimensionList,pos))
            { return true; }
            if (Subpart(t1,1).Intersects(box,dimensionList,pos))
            { return true; }

            return false;
          }
          else
          { // Nullstellen der Ableitung ausserhalb [0,1] -> monoton in [0,1]
            return Intervall(box,dimensionList,pos);
          }
        }
        else
        { // keine reellen 0-Stellen in der Ableitung -> monoton
          return Intervall(box,dimensionList,pos);
        }
      }
    }

    /// <summary>
    /// Punkt bei Parameter t
    /// </summary>
    /// <param name="t">Wert zwischen 0 und 1</param>
    /// <returns></returns>
    public override Point PointAt(double t)
    {
      AssignParams();
      Point p = Point.Create(Dimension);
      for (int i = 0; i < Dimension; i++)
      {
        p[i] = mdA0[i] + t * (mdA1[i] + t * (mdA2[i] + t * mdA3[i]));
      }

      return p;
    }

    public override double Length()
    {
      AssignParams();
      NumMath pCalc = new NumMath(1e-8);

      return pCalc.Integral(new NumMath.Function(LengthFactor),0,1);
    }

    public override double Length(int dimension)
    {
      // TODO
      return Length();
    }

    public override Line Project(IProjection projection)
    {
      return new Bezier(Start.Project(projection),
        mp1.Project(projection),mp2.Project(projection),
        End.Project(projection));
    }

    public override InnerLine InnerLine()
    {
      return new MyInnerLine(mp1,mp2);
    }

    private bool Intervall(IBox box,int[] dimensionList,int pos)
    { // Annahme:  coord is monoton fuer t in [0,1]
      double t0;
      double t1;
      int iDim = dimensionList[pos];

      bool bSMin = Start[iDim] < box.Min[iDim];
      bool bEMin = End[iDim] < box.Min[iDim];
      bool bSMax = Start[iDim] < box.Max[iDim];
      bool bEMax = End[iDim] < box.Max[iDim];

      if (bSMin == bEMin)
      {
        if (bSMin == true)
        { return false; }
        if (bSMax == bEMax)
        { return bSMax; }
      }

      AssignParams();
      double[] dF = new double[] {mdA3[iDim], mdA2[iDim], mdA1[iDim], 0};
      double[] dA = new double[] {3 * mdA3[iDim], 2 * mdA2[iDim], mdA1[iDim]};
      NumMath p = new NumMath(1e-8);

      NumMath.Polynom pFct = new NumMath.Polynom(dF);
      NumMath.Polynom pAbl = new NumMath.Polynom(dA);

      NumMath.Function f = new NumMath.Function(pFct.Value);
      NumMath.Function a = new NumMath.Function(pAbl.Value);

      if (bSMin != bEMin)
      {
        dF[3] = mdA0[iDim] - box.Min[iDim];
        t0 = p.SolveNewton(f,a,0.5);
      }
      else
      {
        if (bSMax)
        { t0 = 0; }
        else
        { t0 = 1; }
      }

      if (bSMax != bEMax)
      {
        dF[3] = mdA0[iDim] - box.Max[iDim];
        t1 = p.SolveNewton(f,a,0.5);
      }
      else
      {
        if (bSMin)
        { t1 = 1; }
        else
        { t1 = 0; }
      }

      return Subpart(t0,t1).Intersects(box,dimensionList,pos + 1);
    }
    public Bezier Subpart(double t0,double t1)
    {
      double dt = (t1 - t0) / 3;
      int iDim = Dimension;

      Point p0 = this.PointAt(t0);
      Point p3 = this.PointAt(t1);

      Point p1 = Point.Create(iDim);
      Point p2 = Point.Create(iDim);

      for (int i = 0; i < iDim; i++)
      {
        p1[i] = p0[i] + (mdA1[i] + t0 * (2 * mdA2[i] + t0 * 3 * mdA3[i])) * dt;
        p2[i] = p3[i] - (mdA1[i] + t1 * (2 * mdA2[i] + t1 * 3 * mdA3[i])) * dt;
      }
      return new Bezier(p0,p1,p2,p3);
    }

    private double LengthFactor(double t)
    {
      double dx;
      double l = 0;

      for (int i = 0; i < Dimension; i++)
      {
        dx = mdA1[i] + t * (2 * mdA2[i] + t * 3 * mdA3[i]);
        l += dx * dx;
      }

      return Math.Sqrt(l);
    }

    private void AssignParams()
    {
      //               
      // P(t) = (1 - t)^3 P0 + 3 t (1 - t)^2 P1 + 3 t^2 (1 - t) P2 + t^3 P1
      //
      if (mbParamKnown)
      { return; }
      int iDim = Dimension;

      mdA0 = new double[iDim];
      mdA1 = new double[iDim];
      mdA2 = new double[iDim];
      mdA3 = new double[iDim];

      for (int i = 0; i < iDim; i++)
      {
        mdA0[i] = Start[i];
        mdA1[i] = 3 * (mp1[i] - Start[i]);
        mdA3[i] = End[i] - 3 * mp2[i] + mdA1[i] + 2 * mdA0[i];
        mdA2[i] = End[i] - mdA3[i] - mdA1[i] - mdA0[i];
      }

      mbParamKnown = true;
    }
	}
}
