using System;
using System.Collections;
using System.Diagnostics;

namespace Geometry
{
  /// <summary>
  /// Summary description for BoxTree.
  /// </summary>
  public class BoxTree
  {
    private class Enumerator:IEnumerator
    {
      BoxTree mTree;
      BoxTile mCurrentTile;
      Box mCurrentBox;
      Box mSelBox;
      IEnumerator mCurrentList;

      public Enumerator(BoxTree tree,Box selBox)
      {
        mTree = tree;
        mSelBox = selBox;
        Reset();
      }
      #region IEnumerator Members

      public void Reset()
      {
        mCurrentTile = mTree.mMainTile;
        mCurrentList = mCurrentTile.ElemList.GetEnumerator();
        mCurrentList.Reset();
        if (mTree.mMainBox == null)
        { mCurrentBox = null; }
        else
        { mCurrentBox = mTree.mMainBox.Clone(); }
      }

      public object Current
      {
        get 
        { return mCurrentList.Current; }
      }

      public bool MoveNext()
      {
        while (mCurrentList.MoveNext() == true)
        {
          if (((TileEntry) mCurrentList.Current).Box.Intersects(mSelBox))
          { return true; }
        }

        if (mCurrentBox == null)
        { return false; }

        while (mCurrentTile != null) 
        {
          mCurrentTile = mTree.NextSelTile(mCurrentTile,mCurrentBox,mSelBox);
          if (mCurrentTile == null)
          { continue; }

          mCurrentList = mCurrentTile.ElemList.GetEnumerator();
          mCurrentList.Reset();

          while (mCurrentList.MoveNext() == true)
          {
            if (((TileEntry) mCurrentList.Current).Box.Intersects(mSelBox))
            { return true; }
          }
        }

        return false;
      }

      #endregion
    }

    public class TileEntry
    {
      private Box mpBox;
      private object mpValue;
      public TileEntry(Box box,object value)
      {
        mpBox = box;
        mpValue = value;
      }
      public Box Box
      {
        get
        { return mpBox; }
      }
      public object Value
      {
        get
        { return mpValue; }
      }
    }

    public class TileEntryCollection:CollectionBase
    {
      public int Add(TileEntry entry)
      {
        return this.List.Add(entry);
      }
      public void AddRange(TileEntryCollection entries)
      {
        this.InnerList.AddRange(entries);
      }
      public void Insert(int index,TileEntry entry)
      {
        this.List.Insert(index,entry);
      }
      public void Remove(TileEntry entry)
      {
        this.List.Remove(entry);
      }

      public TileEntry this[int index]
      {
        get
        { return (TileEntry) this.InnerList[index]; }
      }
      public bool Contains(TileEntry entry)
      {
        return this.List.Contains(entry);
      }
    }

    internal class BoxTile
    {
      private BoxTile mParent;
      private BoxTile mChild0;
      private BoxTile mChild1;
      private int mSplitDimension;
      private int mCount = 0;
      private bool mbSplitting = false;

      private double mMinSplitDim;
      private double mMaxSplitDim;
      private double mMinParentDim;
      private double mMaxParentDim;

      private TileEntryCollection mElemList = new TileEntryCollection();

      public double MinInParentSplitDim
      {
        get { return mMinParentDim; }
        set { mMinParentDim = value; }
      }
      public double MaxInParentSplitDim
      {
        get { return mMaxParentDim; }
        set { mMaxParentDim = value; }
      }

      public int Count 
      {
        get { return mCount; }
        set { mCount = value; }
      }

      public double MinInSplitDim
      {
        get { return mMinSplitDim; }
        set { mMinSplitDim = value; }
      }
      public double MaxInSplitDim
      {
        get { return mMaxSplitDim; }
        set { mMaxSplitDim = value; }
      }

      public BoxTile()
      {
      }
      public TileEntryCollection ElemList
      {
        get 
        {return mElemList;}
        set 
        {mElemList = value;}
      }
      public BoxTile Parent
      {
        get 
        { return mParent; }
        set 
        { mParent = value; }
      }
      public BoxTile Child0
      {
        get {return mChild0;}
        set 
        {
          mChild0 = value;
          mChild0.Parent = this;
          mChild0.MinInParentSplitDim = mMinSplitDim;
          mChild0.MaxInParentSplitDim = (mMaxSplitDim + mMinSplitDim) / 2.0;
        }
      }
      public BoxTile Child1
      {
        get {return mChild1;}
        set 
        {
          double d;
          mChild1 = value;
          mChild1.Parent = this;
          d = (mMaxSplitDim - mMinSplitDim) / 4.0;
          mChild1.MinInParentSplitDim = mMinSplitDim + d;
          mChild1.MaxInParentSplitDim = mMaxSplitDim - d;
        }
      }
      public int SplitDimension
      {
        get {return mSplitDimension;}
        set {mSplitDimension = value;}
      }
      /// <summary>
      /// used to indicate that splitting is running. 
      /// if true, do not propagate added elements to parent count
      /// </summary>
      public bool Splitting
      {
        set { mbSplitting = value; }
      }

      /// <summary>
      /// Split the tile along the largest dimesion of box
      /// Assumption: extent represents the actual extent of the tile (see method BoxExtent in BoxTree)
      /// </summary>
      public void Split(Box extent)
      {
        int splitDim;
        int dim = extent.Dimension;
        double d0,d;
        double dMin,dMax;
        double tx0,tx1,ux0,ux1;

        TileEntryCollection valueList;
        BoxTile parent = this.Parent;

        Debug.Assert(mChild0 == null);
        Debug.Assert(mChild1 == null);

        d0 =  extent.Max[0] - extent.Min[0];
        splitDim = 0;
        for (int i = 1; i < dim; i++) 
        {
          d = extent.Max[i] - extent.Min[i];
          if (d0 < d) 
          {
            d0 = d;
            splitDim = i;
          }
        }
        mSplitDimension = splitDim;
        mMinSplitDim = extent.Min[splitDim];
        mMaxSplitDim = extent.Max[splitDim];

        this.Child0 = new BoxTile(); // use property so that children are initialised correctly
        this.Child1 = new BoxTile();

        valueList = mElemList;
        mElemList = new TileEntryCollection();

        mChild0.Splitting = true;
        mChild1.Splitting = true;
        this.Splitting = true;

        tx0 = mChild0.MinInParentSplitDim;
        tx1 = mChild0.MaxInParentSplitDim;

        ux0 = mChild1.MinInParentSplitDim;
        ux1 = mChild1.MaxInParentSplitDim;

        for (int i = 0; i < valueList.Count; i++) 
        {
          TileEntry entry = valueList[i];
          dMin = entry.Box.Min[splitDim];
          dMax = entry.Box.Max[splitDim];
          if (dMin < ux0) 
          {
            if (dMax < tx1) 
            { mChild0.Add(entry); } 
            else 
            { this.Add(entry); } // TODO: do not Add directly to mElemList (sorting in future);
          } 
          else 
          { 
            Debug.Assert(dMin <= tx1); 
            if (dMax < ux1) 
            { mChild1.Add(entry); } 
            else 
            { this.Add(entry); }
          }
        }

        mChild0.Splitting = false;
        mChild1.Splitting = false;
        this.Splitting = false;
        mCount = valueList.Count;

      }

      public TileEntry this[int index]
      {
        get 
        {
          if (index < mElemList.Count)
          { return mElemList[index]; }
          else 
          {
            index -= mElemList.Count;
            if (index < mChild0.Count)
            { return mChild0[index]; }

            index -= mChild0.Count;
            return mChild1[index];
          }
        }
      }
      public void Add(Box box,object value)
      {
        this.Add(new TileEntry(box,value));
      }
      internal void Add(TileEntry entry)
      {
        // TODO: implement sorting
        mElemList.Add(entry);

        mCount ++;
        if (mbSplitting) 
        { return; }

        BoxTile parent = this.Parent;
        while (parent != null) 
        {
          parent.Count ++;
          parent = parent.Parent;
        }
      }
      public void AddRange(TileEntryCollection elemList)
      {
        // TODO: implement sorting
        mElemList.AddRange(elemList);
      }
      public void Insert(int index,TileEntry entry)
      {
        mCount++;
        mElemList.Insert(index,entry);

        BoxTile parent = this.Parent;
        while (parent != null) 
        {
          parent.Count ++;
          parent = parent.Parent;
        }
      }
      public void Remove(TileEntry entry)
      {
        mElemList.Remove(entry);

        mCount --;
        BoxTile parent = this.Parent;
        while (parent != null) 
        {
          parent.Count --;
          parent = parent.Parent;
        }
      }
      public void Collapse()
      {
        Debug.Assert(mChild0.Child0 == null);
        Debug.Assert(mChild1.Child0 == null);

        AddRange(mChild0.ElemList);
        AddRange(mChild1.ElemList);

        mChild0 = null;
        mChild1 = null;
      }
    }


    private int mDimension;
    private Box mMainBox;
    private Box mSearchBox;
    private int mMaxElemPerTile;
    private int mNElemJoin;

    private BoxTile mMainTile;

    private bool mDynamic;
    private bool mSort;

    public BoxTree(int dimension)
    {
      mDimension = dimension;

      mMainTile = new BoxTile();
      mMaxElemPerTile = 64;
      mDynamic = true;
      mSort = false;
    }

    public BoxTree(int dimensionm,int nElem, bool dynamic)
    {
      mMainTile = new BoxTile();
      mMaxElemPerTile = nElem;
      mDynamic = dynamic;
      mSort = false;
    }

    public void Init(int dimension,Box box,int maxElemPerTile,int nElemJoin)
    {
      mDimension = dimension;
      mMaxElemPerTile = maxElemPerTile;
      mNElemJoin = nElemJoin;

      if (box != null) 
      { mMainBox = box.Clone(); }
      else 
      { mMainBox = null;}
    }
    public void Init(int dimension,int maxElemPerTile,int nElemJoin)
    {
      Init(dimension,null,maxElemPerTile,nElemJoin);
    }
    public void Init(int dimension,int maxElemPerTile)
    {
      Init(dimension,null,maxElemPerTile,maxElemPerTile / 2);
    }
    public void Init(Box box,int maxElemPerTile,int nElemJoin)
    {
      Init(box.Dimension,box,maxElemPerTile,nElemJoin);
    }
    public void Init(Box box,int maxElemPerTile)
    {
      Init(box.Dimension,box,maxElemPerTile,maxElemPerTile / 2);
    }

    /// <summary>
    /// Enlarge Region. Remark: when data is load, numerical problems may occure
    /// </summary>
    /// <param name="directedDimension">directed dimension to enlarge: (-dim -1) : add negativ (dim + 1) : add positiv </param>
    /// <returns>new parent tile</returns>
    private BoxTile Enlarge(int directedDimension)
    {
      double d;
      int dimension;
      dimension = Math.Abs(directedDimension) - 1;

      BoxTile newTile = new BoxTile();

      d = mMainBox.Max[dimension] - mMainBox.Min[dimension];
      if (directedDimension > 0) 
      {
        mMainBox.Max[dimension] += d;

        newTile.Child0 = mMainTile;
        newTile.Child1 = new BoxTile();
      } 
      else 
      {
        d /= 2.0;
        mMainBox.Min[dimension] -= d;
        mMainBox.Max[dimension] += d;

        newTile.Child0 = new BoxTile();
        newTile.Child1 = mMainTile;
      }
      newTile.SplitDimension = dimension;
      mMainTile = newTile;

      return newTile;
    }

    public void InitSize(IGeometry[] data)
    {
      InitSize((IEnumerable) data);
    }

    private void InitSize()
    {
      InitSize(mMainTile.ElemList);
    }

    private void InitSize(IEnumerable data)
    {
      mMainBox = null;
      foreach (object pRow in data)
      {
        IBox pBox;
        if (pRow is IGeometry)
        { pBox = ((IGeometry) pRow).Extent; }
        else
        {
          Debug.Assert(pRow is TileEntry);
          pBox = ((TileEntry) pRow).Box;
        }
        if (mMainBox == null) 
        {
          mMainBox = new Box(Point.Create(pBox.Min),Point.Create(pBox.Max));
          continue;
        }

        for (int i = 0; i < mDimension; i++) 
        {
          if (pBox.Min[i] < pBox.Min[i]) 
          { mMainBox.Min[i] = pBox.Min[i];}
          if (pBox.Max[i] > mMainBox.Max[i]) 
          { mMainBox.Max[i] = pBox.Max[i];}
        }
      }

      // double the size of each dimension
      for (int i = 0; i < mDimension; i++) 
      {
        mMainBox.Max[i] = 2 * mMainBox.Max[i] - mMainBox.Min[i];
      }
    }

    /// <summary>
    /// find tile where geomExtent fits inside
    /// </summary>
    private BoxTile FindAddTile(BoxTile parent,Box geomExtent,bool dynamic)
    {
      int divDim;
      BoxTile child0,child1;

      child0 = parent.Child0;
      if (child0 != null) 
      {
        child1 = parent.Child1;
        divDim = parent.SplitDimension;

        if (geomExtent.Min[divDim] < child1.MinInParentSplitDim) 
        {
          if (geomExtent.Max[divDim] < child1.MaxInParentSplitDim) 
          { return FindAddTile(child0, geomExtent,dynamic); } 
          else 
          { return parent; }
        } 
        else 
        {
          if (geomExtent.Max[divDim] < child1.MaxInParentSplitDim) 
          { return FindAddTile(child1, geomExtent,dynamic); } 
          else 
          { return parent; }
        }
      } 
      else 
      { // tile has no children
        if (dynamic == false || parent.ElemList.Count < mMaxElemPerTile - 1) 
        { return parent; } 
        else 
        { // split tile and try again
          if (mMainBox == null)
          { this.InitSize(); }
          parent.Split(TileExtent(parent));
          return FindAddTile(parent,geomExtent,false);
        }
      }
    }

    /// <summary>
    /// Get the Extent of Tile
    /// </summary>
    private Box TileExtent(BoxTile tile)
    {
      Box box = mMainBox.Clone();
      BoxTile child = tile;
      BoxTile parent = child.Parent;
      int splitDim;
      double d = -1;
      double dMax = 0;

      // proceed up so long until it is verified that all dimensions of the 
      // box are set to the correct extent
      while (parent != null && d <= dMax) 
      {
        splitDim = parent.SplitDimension;

        d = child.MaxInParentSplitDim - child.MinInParentSplitDim;
        if (box.Min[splitDim] < child.MinInParentSplitDim)
        { box.Min[splitDim] = child.MinInParentSplitDim; }
        if (box.Max[splitDim] > child.MaxInParentSplitDim)
        { box.Max[splitDim] = child.MaxInParentSplitDim; }

        if (dMax == 0) 
        {
          dMax = parent.MaxInSplitDim - parent.MinInSplitDim;
          // dMax was the largest extent when splitting the parent of tile,
          // maybe together with others.
          // So if d > dMax, it must be dimension, that was split already
        }

        child = parent;
        parent = parent.Parent;
      }
      return box;
    }


    public TileEntry this[int index]
    {
      get
      { return mMainTile[index]; }
    }
    public int Add(Box box,object value)
    {
      int dim;
      BoxTile addTile;

      if (mMainTile.Child0 != null)
      { 
        while ((dim = mMainBox.ExceedDimension(box)) != 0) 
        { Enlarge(dim); }
      }

      addTile = FindAddTile(mMainTile,box,mDynamic);
      addTile.Add(box,value);

      return mMainTile.Count - 1;
    }

    /// <summary>
    /// Inserts a geometry at position.
    /// </summary>
    /// <param name="geom"></param>
    /// <param name="position"></param>
    public void Insert(int index,TileEntry entry)
    {
      // TODO , only valid if just one tail exists
      mMainTile.Insert(index,entry);
    }

    public void Remove(TileEntry entry)
    {
      BoxTile tile = FindAddTile(mMainTile,entry.Box,false);
      tile.Remove(entry);

      if (mDynamic && tile.Child0 == null && tile.Parent != null &&
        tile.Parent.Child0.ElemList.Count + tile.Parent.Child1.ElemList.Count +
        tile.Parent.ElemList.Count < mNElemJoin)
      { tile.Parent.Collapse(); }
    }


    /// <summary>
    /// 
    /// </summary>
    /// <param name="currentTile"></param>
    /// <param name="currentBox">begin: box of current tile; end: box of return tile</param>
    /// <param name="extent">extent which must intersect or touch next tile</param>
    /// <returns></returns>
    private BoxTile NextSelTile(BoxTile currentTile,Box currentBox,Box extent)
    {
      int splitDim;
      BoxTile child;

      splitDim = currentTile.SplitDimension;
      child = currentTile.Child0;
      if (child != null) 
      {
        Debug.Assert(currentBox.Min[splitDim] == child.MinInParentSplitDim);
        currentBox.Max[splitDim] = child.MaxInParentSplitDim;
        if (currentBox.Intersects(extent)) 
        { return child; }
      }

      while (currentTile != null) 
      {
        if (child != currentTile.Child1)
        {
          child = currentTile.Child1;
          currentBox.Min[splitDim] = child.MinInParentSplitDim;
          currentBox.Max[splitDim] = child.MaxInParentSplitDim;
          if (currentBox.Intersects(extent)) 
          { return child; }
        }

        // reset box to current extent
        currentBox.Min[splitDim] = currentTile.MinInSplitDim;
        currentBox.Max[splitDim] = currentTile.MaxInSplitDim;

        // one step up
        child = currentTile;
        currentTile = currentTile.Parent;
        if (currentTile != null)
        {	splitDim = currentTile.SplitDimension; }
      }
      // fr genuegend unterteilte Tiles (sonst ev. Min.X == 0 o.ae.):
      // Debug.Assert(currentBox.Equals(mMainBox)); 
      return null;
    }

    public Box SearchBox
    {
      get 
      { return mSearchBox; }
      set 
      { mSearchBox = value; }
    }

    #region IGeometry Members

    public int Dimension
    {
      get { return mDimension; }
    }

    public int Topology
    {
      get { return mDimension; }
    }

    public Box Extent 
    {
      get { return mMainBox; }
    }

    public bool Intersects(Box box)
    { return Extent.Intersects(box); }

    #endregion

    #region IList Members

    public bool Contains(TileEntry value)
    {
      BoxTile tile = FindAddTile(mMainTile, value.Box,false);
      if (tile == null) 
      { return false; }
      return tile.ElemList.Contains(value);
    }

    public void Clear()
    {
      mMainTile = null;
    }

    public bool IsFixedSize
    {
      get { return false; }
    }

    #endregion

    #region IEnumerable Members

    public IEnumerator GetEnumerator() 
    { 
      return new Enumerator(this,this.SearchBox);
    }

    #endregion

    #region ICollection Members

    public int Count
    {
      get { return mMainTile.Count; }
    }

    #endregion
  }
}
