using System;
using System.Collections;

namespace Geometry
{
  /// <summary>
  /// Summary description for LineX.
  /// </summary>
  public class Line : IGeometry
  {
    private Point mpStart;
    private Point mpEnd;

    public Line()
    { }

    public Line(Point start,Point end)
    {
      mpStart = start;
      mpEnd = end;
    }

    public virtual InnerLine InnerLine()
    { return null; }

    public virtual Line[] Generalize(double offset)
    {
      return new Line[] { this };
    }

    public virtual Point Start
    {
      get
      { return mpStart; }
    }

    public virtual Point End
    {
      get
      { return mpEnd; }
    }

    public virtual Point PointAt(double t)
    {
      int iDim = Dimension;
      Point p = Point.Create(iDim);

      for (int i = 0; i < iDim; i++)
      {
        p[i] = mpStart[i] + t * (mpEnd[i] - mpStart[i]);
      }
      return p;
    }

    public virtual double Length()
    {
      return Length(Dimension);
    }

    public virtual double Length(int dimension)
    {
      return Math.Sqrt(mpStart.Dist2(mpEnd,dimension));
    }

    /// <summary>
    /// Findet Punkt wo
    /// this.PointAt(u) = other.PointAt(v)
    /// mit u > Startwert von u
    /// </summary>
    /// <param name="other">Andere Line</param>
    /// <param name="u"></param>
    /// <param name="v"></param>
    /// <returns>weitere Schnittstelle vorhanden</returns>
    public virtual bool Intersect(Line other,ref double u,ref double v)
    {
      if (other.GetType() != typeof(Line))
      { 
        if (this.GetType() == typeof(Line))
        {	return other.Intersect(this,ref v,ref u); }
      }
      // this.PointAt(u) = other.PointAt(v)
      // TODO: Generalize for any LineType
      double x0 = this.Start.X;
      double x1 = other.Start.X;
      double y0 = this.Start.Y;
      double y1 = other.Start.Y;

      double a0 = this.End.X - x0;
      double b0 = this.End.Y - y0;

      double a1 = other.End.X - x1;
      double b1 = other.End.Y - y1;

      double c0 = x1 - x0;
      double c1 = y1 - y0;
      double det = -a0*b1 + b0*a1;
      double epsi = 1e-8;
      if (Math.Abs(det) > epsi * (Math.Abs(x0) + Math.Abs(x1) + Math.Abs(y0) + Math.Abs(y1)) / 1000.0) 
      {
        u = (-c0*b1 + c1*a1) / det;
        v = ( a0*c1 - b0*c0) / det;
        return true;
      } 
      else 
      {
        /* parallel lines */
        det = -a0 * c1 +  b0 *c0;
        v = (det * det) / (a0*a0 + b0*b0);
        return false;
      }
    }

    #region IGeometry Members

    public int Dimension
    {
      get
      { return Math.Min(Start.Dimension,End.Dimension); }
    }

    public int Topology
    {
      get
      { return 1; }
    }

    public virtual bool Intersects(IBox box)
    {
      int iDim = Math.Max(this.Dimension,box.Dimension);
      return Intersects(box,Geometry.DimensionList(iDim));
    }

    public virtual bool Intersects(IBox box,int[] dimensionList)
    {
      double t;
      int iPos = 0;
      int nDim = dimensionList.Length;

      Point pStart = mpStart.Clone();
      Point pEnd = mpEnd.Clone();
      foreach (int iDim in dimensionList)
      { // Start Point
        if (pStart[iDim] < box.Min[iDim])
        { 
          if (pEnd[iDim] < box.Max[iDim])
          { return false; }
          t = (box.Min[iDim] - pStart[iDim]) / (pEnd[iDim] - pStart[iDim]);
          pStart[iDim] = box.Min[iDim];
        }
        else if (pStart[iDim] > box.Max[iDim])
        { 
          if (pEnd[iDim] > box.Max[iDim])
          { return false; }
          t = (box.Max[iDim] - pStart[iDim]) / (pEnd[iDim] - pStart[iDim]);
          pStart[iDim] = box.Max[iDim];
        }
        else
        { t = 0; }

        System.Diagnostics.Debug.Assert(t >= 0 && t <= 1);

        if (t != 0)
        { 
          for (int i = iPos + 1; i < nDim; i++)
          {
            int j = dimensionList[i];
            pStart[j] = pStart[j] + t * (pEnd[j] - pStart[j]);
          }
        }

        // End Point
        // is a little easier because start coordinate is sure to lay in extent
        if (pEnd[iDim] < box.Min[iDim])
        { t = (box.Min[iDim] - pStart[iDim]) / (pEnd[iDim] - pStart[iDim]); }
        else if (pEnd[iDim] > box.Max[iDim])
        { t = (box.Max[iDim] - pStart[iDim]) / (pEnd[iDim] - pStart[iDim]); }
        else
        { t = 0; }

        System.Diagnostics.Debug.Assert(t >= 0 && t <= 1);

        if (t != 0)
        { 
          for (int i = iPos + 1; i < nDim; i++)
          {
            int j = dimensionList[i];
            pEnd[j] = pStart[j] + t * (pEnd[j] - pStart[j]);
          }
        }

        iPos++;
      }
      return true;
      //      double t = -1;
      //      while ( this.Cut(box,ref t) != null && t < 0)
      //      { }
      //      return (t >= 0 && t <= 1);
    }


    public virtual bool Intersects(IGeometry other)
    {
      if (other.Topology > this.Topology)
      { return other.Intersects(this); }

      if (other is IMultipartGeometry)
      {
        IMultipartGeometry pMulti = (IMultipartGeometry) other;
        if (pMulti.HasSubparts)
        {
          foreach (IGeometry pSubpart in pMulti.Subparts(this.Extent))
          {
            if (this.Extent.Intersects(pSubpart.Extent) == false)
            { continue; }

            if (this.Intersects(pSubpart))
            { return true; }
          }
          return false;
        }
      }
      // other has no subparts

      // Point
      if (other.Topology == 0)
      { return this.Intersects((IPoint) other); }

      if (other == this) // same Line (no self intersect !)
      { return false; }

      if (other.GetType() != typeof(Line))
      { throw new System.NotImplementedException(other.GetType().ToString()); }

      double u = -1;
      double v =  0;
      if (this.Intersect((Line) other,ref u,ref v) == false)
      { return false; }

      if (0 <= u && u <= 1 && 0 <= v && v <= 1)
      { return true; }

      return false;
    }
    public virtual bool Intersects(IPoint point)
    {
      throw new System.NotImplementedException(); // TODO
    }
    public Point Cut(Box box,ref double t)
    {
      /*
          (x)     (a)     (x0)   (x1)  (xc)  (xc)
          ( ) + t*( )  =      or     or    or
          (y)     (b)     (yc)   (yc)  (y0)  (y1)

          and t (output) > t (input)               
      */

      double u;
      double tc;
      double tNew = t - 1;
      Point pNew = null;
      Point pc;
      Point pStart = Start;
      Point pEnd = End;
      int dim = this.Dimension;

      if (box.Dimension < dim)
      { dim = box.Dimension; }
      pc = Point.Create(dim);

      for (int iSide = 0; iSide < 2 * dim; iSide++)
      {
        int iDim = iSide / 2;
        double a = pEnd[iDim] - pStart[iDim];

        if (a != 0)
        {
          if ((iSide & 1) == 0)
          { pc[iDim] = box.Min[iDim]; }
          else
          { pc[iDim] = box.Max[iDim]; }

          u = (pc[iDim] - pStart[iDim]) / a;
          if (u > t && (u < tNew || tNew < t))
          { 
            bool bCut = true;
            for (int j = 0; j < dim; j++)
            {
              if (j == iDim)
              { continue; }
              double b = pEnd[j] - pStart[j];
              tc = pStart[j] + b * u;
              if (tc < box.Min[j] || tc > box.Max[j])
              {
                bCut = false;
                break;
              }
              pc[j] = tc;
            }
            if (bCut == true)
            {
              pNew = pc.Clone(); 
              tNew = u;
            }
          }
        }
      }
      if (pNew != null)
      { t = tNew; }
      return pNew;
    }

    public virtual Box Extent
    {
      get
      { return new Box(Start.Clone(),End.Clone(),true); }
    }
    IBox IGeometry.Extent
    {	get { return this.Extent; } }

    public virtual PointCollection Border
    {
      get
      { 
        PointCollection pBorder = new PointCollection();
        pBorder.AddRange(new Point[] {this.Start,this.End}); 
        return pBorder;
      }
    }
    IGeometry IGeometry.Border
    { get { return this.Border;} }

    public virtual Line Project(IProjection projection)
    {
      return new Line(mpStart.Project(projection),mpEnd.Project(projection));
    }
    IGeometry IGeometry.Project(IProjection projection)
    { return Project(projection); }

    #endregion
  }

  public abstract class InnerLine
  {
    public abstract Line Complete(Point start,Point end);
  }
  public class InnerLineCollection : CollectionBase
  {
    public int Add(InnerLine innerLine)
    {
      return InnerList.Add(innerLine);
    }
    public InnerLine this[int index]
    {
      get
      { return (InnerLine) InnerList[index]; }
      set
      { InnerList[index] = value; }
    }
  }
}
