using System;
using System.Collections;
using System.Diagnostics;

namespace Geometry
{
  /// <summary>
  /// Summary description for Mesh.
  /// </summary>
  public class Mesh
  {
    #region Member classes
    public abstract class Tri: IGeometry
    {
      protected Point2D mCenter = null;
      protected double mRadius2 = 0;

      public abstract MeshLine this[int i]
      {
        get;
      }

      public Point2D Center
      {
        get
        { 
          if (mCenter == null)
          {
            Point p0 = this[0].Start;
            mCenter = Geometry.OutCircleCenter(p0, this[1].Start,this[2].Start);
            mRadius2 = mCenter.Dist2(p0);
          }
          return mCenter;
        }
      }
      public double Radius2
      {
        get
        { 
          if (mRadius2 <= 0)
          { Point2D center = this.Center; } // radius is calculated too
          return mRadius2;
        }
      }

      public int GetLineIndex(MeshLine line)
      {
        for (int i = 0; i < 3; i++) 
        {
          if (this[i].HasEqualBaseLine(line) &&
            this[i].Direction == line.Direction)
          { return i; }
        }
        return -1;
      }

      /// <summary>
      /// M2GetNextLine
      /// </summary>
      public abstract MeshLine GetNextLine(MeshLine line);

      /// <summary>
      /// M2GetPreviousLine
      /// </summary>
      public abstract MeshLine GetPreviousLine(MeshLine line);

      public Point GetPoint(int i)
      {
        return this[i].Start;
      }

      public int GetPointIndex(Point2D point)
      {
        if (GetPoint(0) == point) { return 0; }
        if (GetPoint(1) == point) { return 1; }
        if (GetPoint(2) == point) { return 2; }
        return -1;
      }
      #region IGeometry Members

      public int Dimension
      {
        get
        { return 2; }
      }

      public int Topology
      {
        get
        { return 2; }
      }

      public bool Intersects(IBox box)
      {
        IBox pBox = Extent.Clone();
        if (box.IsWithin(pBox))
        { return true; }
        if (box.Intersects(pBox) == false)
        { return false; }
        return (
          this[0].Intersects(box) ||
          this[1].Intersects(box) ||
          this[2].Intersects(box));
      }
      public bool Intersects(IGeometry geometry)
      { // TODO
        return false;
      }

      public abstract Box Extent  { get; }
      IBox IGeometry.Extent
      { get { return this.Extent; } }

      public Polyline Border
      {
        get
        {
          return new Polyline( 
            new Point[] { this[0].Start,this[1].Start,this[2].Start });
        }
      }
      IGeometry IGeometry.Border
      { get { return this.Border; } }

      IGeometry IGeometry.Project(IProjection projection)
      { return null; }


      #endregion
    }


    private class MTri: Tri
    {
      private MMeshLine[] mLine;
      public MTri(MMeshLine l0,MMeshLine l1, MMeshLine l2)
      {
        mLine = new MMeshLine[] {l0,l1,l2};
        l0.MLeftTri = this;
        l1.MLeftTri = this;
        l2.MLeftTri = this;
      }

      public override MeshLine this[int i]
      {
        get
        { return mLine[i]; }
      }
      public MMeshLine MGetLine(int i)
      {
        return mLine[i];
      }
      public void MSetLine(int i,MMeshLine l)
      {
        mLine[i] = l;
        l.MLeftTri = this;
      }

      public MMeshLine MGetNextLine(MeshLine line)
      {
        int index = GetLineIndex(line);
        if (index < 0) 
        { return null; }

        return MGetLine(index < 2 ? index + 1 : 0);
      }
      public override MeshLine GetNextLine(MeshLine line)
      {
        return MGetNextLine(line);
      }

      public MMeshLine MGetPreviousLine(MeshLine line)
      {
        int index = GetLineIndex(line);
        if (index < 0) 
        { return null; }

        return MGetLine(index > 0 ? index - 1 : 2);
      }

      public override MeshLine GetPreviousLine(MeshLine line)
      {
        return MGetPreviousLine(line);
      }

      public void ClearCircle()
      {
        mCenter = null;
        mRadius2 = 0;
      }

      public bool Validate()
      {
        for (int i = 0; i < 3; i++)
        {
          if (mLine[i].LeftTri != this)
          { return false; }
          if (mLine[i].End != mLine[i < 2 ? i + 1 : 0].Start)
          { return false; }
        }

        return true;
      }

      public override Box Extent
      {
        get
        {
          Point p0 = mLine[0].Start.Clone();
          Point p1 = p0.Clone();
          for (int i = 1; i < 3; i++)
          {
            Point p = mLine[i].Start;
            if (p.X < p0.X)
            { p0.X = p.X; }
            else if (p.X > p1.X)
            { p1.X = p.X; }
            if (p.Y < p0.Y)
            { p0.Y = p.Y; }
            else if (p.Y > p1.Y)
            { p1.Y = p.Y; }
          }
          return new Box(p0,p1);
        }
      }
    }

    private class MeshBaseLine
    {
      private MeshPoint mStartPoint;
      private MeshPoint mEndPoint;
      private MTri mLeftTri;
      private MTri mRightTri;
      private object mTag;

      public MeshBaseLine(MeshPoint startPoint,MeshPoint endPoint, object tag)
      {
        mStartPoint = startPoint;
        mEndPoint = endPoint;
        mTag = tag;
      }

      public MeshPoint StartPoint
      {
        get
        { return mStartPoint; }
        set
        { mStartPoint = value; }
      }

      public MeshPoint EndPoint
      {
        get
        { return mEndPoint; }
        set
        { mEndPoint = value; }
      }

      public MTri LeftTri
      {
        get
        { return mLeftTri; }
        set
        { mLeftTri = value; }
      }

      public MTri RightTri
      {
        get
        { return mRightTri; }
        set
        { mRightTri = value; }
      }

      public object Tag
      {
        get { return mTag; }
        set { mTag = value; }
      }
    }

    public abstract class MeshLine:Line
    {
      protected bool mDirection = true;

      public abstract Tri LeftTri
      {
        get;
      }

      public abstract Tri RightTri
      {
        get;
      }

      public abstract object Tag
      {
        get; set;
      }

      public bool Direction
      {
        get { return mDirection; }
      }

      public abstract bool HasEqualBaseLine(MeshLine line);

      public abstract MeshLine Inverse();

      public static MeshLine operator -(MeshLine line)
      {
        return line.Inverse();
      }

      /// <summary>
      /// M2GetNextLine_
      /// </summary>
      public MeshLine GetNextTriLine() 
      {
        Tri tri = LeftTri;
        if (tri == null)
        { return null; }
        return tri.GetNextLine(this);
      }

      /// <summary>
      /// M2GetPreviosLine_
      /// </summary>
      public MeshLine GetPreviousTriLine() 
      {
        Tri tri = LeftTri;
        if (tri == null)
        { return null; }
        return tri.GetPreviousLine(this);
      }

      /// <summary>
      /// return next line after this starting at start point of this 
      /// in counterclockwise direction
      /// </summary>
      public MeshLine GetNextPointLine()
      {
        Tri tri = this.LeftTri;
        if (tri != null) 
        { return - tri.GetPreviousLine(this); }

        MeshLine l = this;
        tri = this.RightTri;
        while (tri != null)
        {
          l = tri.GetPreviousLine(-l);
          tri = l.RightTri;
        }
        return l;
      }

      /// <summary>
      /// return next line after this starting at start point of this 
      /// in clockwise direction
      /// </summary>
      public MeshLine GetPreviousPointLine()
      {
        Tri tri = this.RightTri;
        if (tri != null) 
        { return tri.GetNextLine(-this); }

        MeshLine l = this;
        tri = this.LeftTri;
        while (tri != null)
        {
          l = - tri.GetNextLine(l);
          tri = l.LeftTri;
        }
        return l;
      }

      public abstract void ChangeDiag();
      
      public bool OptionChangeDiag()
      {
        if (this.Tag != null)
        { return false; }

        Tri lTri = LeftTri;

        int i0 = lTri.GetLineIndex(this);
        int i1 = i0 < 2 ? i0 + 1 : 0;

        Point pLeft = lTri[i1].End;
        Tri rTri = RightTri;
        if (rTri == null)
        { return false; }
        if (rTri.Center.Dist2(pLeft,2) >= rTri.Radius2)
        { return false; }

        this.ChangeDiag();
        return true;
      }
      #region ILine Members
      public int Count
      {
        get
        { return 2; }
      }
      public Point this[int index]
      {
        get
        {
          if (index == 0)
          { return Start; }
          else if (index == 1)
          { return End; }
          else
          { throw new System.Exception(string.Format("Index {0} out of Range",index)); }
        }
      }
      #endregion
    }
    

    private class MMeshLine:MeshLine
    {
      private MeshBaseLine mBaseLine;

      public MMeshLine(MeshPoint startPoint,MeshPoint endPoint,object tag, bool direction)
      {
        mBaseLine = new MeshBaseLine(startPoint,endPoint,tag);
        mDirection = direction;
      }
      public MMeshLine(MeshPoint startPoint,MeshPoint endPoint)
      {
        mBaseLine = new MeshBaseLine(startPoint,endPoint,null);
        mDirection = true;
      }
      public MMeshLine(MeshBaseLine baseLine, bool direction)
      {
        mBaseLine = baseLine;
        mDirection = direction;
      }

      public override MeshLine Inverse()
      {
        return new MMeshLine(this.BaseLine, !(this.Direction));
      }

      public MeshPoint MStartPoint
      {
        get
        {
          if (mDirection)
          { return mBaseLine.StartPoint; }
          else
          { return mBaseLine.EndPoint; }
        }
        set 
        {
          if (mDirection)
          { mBaseLine.StartPoint = value; }
          else
          { mBaseLine.EndPoint = value; }
        }
      }

      public override Point Start
      {
        get
        { return MStartPoint.Point; }
      }

      public MeshPoint MEndPoint
      {
        get 
        { 
          if (mDirection)
          { return mBaseLine.EndPoint; }
          else
          { return mBaseLine.StartPoint; }
        }
        set 
        {
          if (mDirection)
          { mBaseLine.EndPoint = value; }
          else
          { mBaseLine.StartPoint = value; }
        }
      }

      public override Point End
      {
        get
        { return MEndPoint.Point; }
      }


      public MTri MLeftTri
      {
        get 
        { 
          if (mDirection)
          { return mBaseLine.LeftTri; }
          else
          { return mBaseLine.RightTri; }
        }
        set
        {
          if (mDirection)
          { mBaseLine.LeftTri = value; }
          else
          { mBaseLine.RightTri = value; }
        }
      }

      public override Tri LeftTri
      {
        get
        { return MLeftTri; }
      }

      public MTri MRightTri
      {
        get 
        { 
          if (mDirection)
          { return mBaseLine.RightTri; }
          else
          { return mBaseLine.LeftTri; }
        }
      }

      public override Tri RightTri
      {
        get
        { return MRightTri; }
      }

      public override object Tag
      {
        get 
        { return mBaseLine.Tag; }
        set
        {
          if (mDirection)
          { mBaseLine.Tag = value; }
          else 
          { 
            if (value == null)
            { 
              mBaseLine.Tag = null;
              return;
            }
            //flip line and set tag
            int index;
            MTri triL = mBaseLine.RightTri;
            MTri triR = mBaseLine.LeftTri;
            MeshPoint pntS = mBaseLine.StartPoint;
            MeshPoint pntE = mBaseLine.EndPoint;

            if (triL != null)
            {
              index = triL.GetLineIndex(this);
              triL.MSetLine(index,this);
            }
            mDirection = !mDirection;
            if (triR != null)
            {
              index = triR.GetLineIndex(this);
              triR.MSetLine(index,-this);
            }
            mBaseLine.StartPoint = pntE;
            mBaseLine.EndPoint = pntS;

            mBaseLine.Tag = value;
          }
        }
      }

      public MeshBaseLine BaseLine
      {
        get
        { return mBaseLine; }
      }

      public override bool HasEqualBaseLine(MeshLine line)
      {
        MMeshLine l = line as MMeshLine;
        if (l == null)
        { return false; }
        return mBaseLine == l.BaseLine;
      }


      public new MMeshLine GetNextTriLine() 
      {
        MTri tri = MLeftTri;
        if (tri == null)
        { return null; }
        return tri.MGetNextLine(this);
      }

      public static MMeshLine operator -(MMeshLine line)
      {
        return new MMeshLine(line.mBaseLine, !(line.Direction));
      }

      /// <summary>
      /// M2GetNextLine_
      /// </summary>
      public MMeshLine MGetNextTriLine() 
      {
        MTri tri = MLeftTri;
        if (tri == null)
        { return null; }
        return tri.MGetNextLine(this);
      }

      /// <summary>
      /// M2GetPreviosLine_
      /// </summary>
      public MMeshLine MGetPreviousTriLine() 
      {
        MTri tri = MLeftTri;
        if (tri == null)
        { return null; }
        return tri.MGetPreviousLine(this);
      }

      /// <summary>
      /// return next line after this starting at start point of this 
      /// in counterclockwise direction
      /// </summary>
      public MMeshLine MGetNextPointLine()
      {
        MTri tri = this.MLeftTri;
        if (tri != null) 
        { return - tri.MGetPreviousLine(this); }

        MMeshLine l = this;
        tri = this.MRightTri;
        while (tri != null)
        {
          l = tri.MGetNextLine(-l);
          tri = l.MRightTri;
        }
        return l;
      }

      /// <summary>
      /// return next line after this starting at start point of this 
      /// in clockwise direction
      /// </summary>
      public MMeshLine MGetPreviousPointLine()
      {
        MTri tri = this.MRightTri;
        if (tri != null) 
        { return tri.MGetNextLine(-this); }

        MMeshLine l = this;
        tri = this.MLeftTri;
        while (tri != null)
        {
          l = - tri.MGetPreviousLine(l);
          tri = l.MLeftTri;
        }
        return l;
      }

      /// <summary>
      /// splits the meshline at point by creating a new meshline from start to point
      /// (return value) and replacing the start point of this by point
      /// </summary>
      /// <param name="point"></param>
      /// <returns>new meshline</returns>
      public MMeshLine Split(MeshPoint point)
      {
        MMeshLine l = new MMeshLine(MStartPoint,point);
        MStartPoint.IndexLine = l;
        MStartPoint = point;
        point.IndexLine = this;
        return l;
      }

      public override void ChangeDiag()
      {
        MTri lTri = MLeftTri;
        MTri rTri = MRightTri;
        int i0 = lTri.GetLineIndex(this);
        int i1 = (i0 < 2 ? i0 + 1 : 0);
        int i2 = (i1 < 2 ? i1 + 1 : 0);
        MMeshLine lL1 = lTri.MGetLine(i1);
        MMeshLine lL2 = lTri.MGetLine(i2);

        MMeshLine rL0 = -this;
        i0 = rTri.GetLineIndex(rL0);
        i1 = (i0 < 2 ? i0 + 1 : 0);
        i2 = (i1 < 2 ? i1 + 1 : 0);
        MMeshLine rL1 = rTri.MGetLine(i1);
        MMeshLine rL2 = rTri.MGetLine(i2);

        lTri.ClearCircle();
        rTri.ClearCircle();

        this.MStartPoint = rL2.MStartPoint;
        this.MEndPoint = lL2.MStartPoint;

        lTri.MSetLine(0,rL2);
        lTri.MSetLine(1,lL1);
        lTri.MSetLine(2,rL0);

        rTri.MSetLine(0,rL1);
        rTri.MSetLine(1,this);
        rTri.MSetLine(2,lL2);

        Debug.Assert(lTri.Validate());
        Debug.Assert(rTri.Validate());

        lL1.MStartPoint.IndexLine = lL1;
        rL1.MStartPoint.IndexLine = rL1;

        rL2.OptionChangeDiag();
        rL1.OptionChangeDiag();
      }
    }
 
    private class MeshPoint:IGeometry
    {
      protected Point2D mPoint;
      protected MMeshLine mIndexLine;

      public MeshPoint(Point2D point)
      {
        mPoint = point;
      }
      public Point2D Point
      {
        get
        { return mPoint; }
      }

      public MMeshLine IndexLine
      {
        get
        { return mIndexLine; }
        set 
        { 
          Debug.Assert(value.MStartPoint == this);
          mIndexLine = value;
        }
      }
      #region IGeometry Members

      public int Dimension
      {
        get
        { return mPoint.Dimension; }
      }

      public int Topology
      {
        get
        { return mPoint.Topology; }
      }

      public bool Intersects(IBox box)
      {
        return box.IsWithin(mPoint);
      }

      public bool Intersects(IGeometry geometry)
      {
        return mPoint.Intersects(geometry);
      }

      public Box Extent
      {
        get
        { return mPoint.Extent; }
      }
      IBox IGeometry.Extent
      { get { return this.Extent; } }

      public IGeometry Border
      { get {return null; } }

      public Point Project(IProjection projection)
      {
        return mPoint.Project(projection);
      }
      IGeometry IGeometry.Project(IProjection projection)
      { return this.Project(projection); }

      #endregion
    }      

    private class LineBoxTree:BaseBoxTree
    {
      public LineBoxTree():base(1)
      {  }
      public int Add(MMeshLine line)
      {
        return BoxTree.Add(line.Extent,line);
      }
      public void Insert(int index,MMeshLine line)
      {
        BoxTree.Insert(index,new BoxTree.TileEntry(line.Extent,line));
      }
      public MMeshLine this[int index]
      {
        get
        { return (MMeshLine) BoxTree[index].Value; }
      }
    }

    private class TriBoxTree:BaseBoxTree
    {
      public TriBoxTree():base(2)
      {  }
      public int Add(Tri tri)
      {
        return BoxTree.Add(tri.Extent,tri);
      }
    }

    #endregion
    // member definitions
    private double mFuzzy2;
    private double mFuzzy;
    private double mEpsi = 1e-12;
    private TriBoxTree mListTri;
    private LineBoxTree mListLine;
    private BoxTree mListPoint;

    public Mesh()
    {
      mListTri = new TriBoxTree();
      mListLine = new LineBoxTree();
      mListPoint = new BoxTree(2);
    }

    public Mesh(int capacity)
    {
      mListTri = new TriBoxTree();
      mListLine = new LineBoxTree();
      mListPoint = new BoxTree(2);
    }

    public BoxTree LineList
    {
      get
      { return mListLine; }
    }
    public double Fuzzy2
    {
      get
      { return mFuzzy2; }
    }
    public double Fuzzy
    {
      get
      { return mFuzzy2; }
      set
      { 
        mFuzzy = value;
        mFuzzy2 = value * value;
      }
    }

    /// <summary>
    /// Adds point to the mesh if not coincident with existing point of mesh
    /// </summary>
    /// <param name="point"></param>
    /// <returns>null if new Point add, coincident point if not added</returns>
    public Point2D Add(Point2D point)
    {
      MeshPoint p = Add(point,null);
      if (p.Point == point)
      { 
        mListPoint.Add(p.Extent,p);
        return null;
      }
      return p.Point; // coincident point
    }

    /// <summary>
    /// returns Line starting in p0 and ending in p1 if existing,
    /// otherwise null
    /// </summary>
    private MeshLine FindPointLine(MeshPoint p0,Point2D p1)
    {
      MeshLine l0 = p0.IndexLine;
      MeshLine l = null;

      l = l0;
      do
      {
        if (l.End == p1)
        { return l; }
        l = l.GetNextPointLine();
      } while (l != l0);

      return null;
    }

    /// <summary>
    /// Find a tri, where point lies within tris circle
    /// </summary>
    private Tri FindCircle(Point2D point,out double radius2)
    {
      Tri tri0 = null;
      double dr2;
    
      radius2 = -1;
      foreach (Tri tri in mListTri)
      {
        dr2 = point.Dist2(tri.Center,2) - tri.Radius2;

        if (dr2 < radius2 || radius2 < 0)
        {
          radius2 = dr2;
          tri0 = tri;
        }
        if (radius2 <= 0) 
        { return tri0; }
      } 
      return tri0;
    }

    private bool FindNextLine(ref MMeshLine l0,Point2D goal)
    {
      Point p0 = l0.Start;
      Point p1 = l0.End;
      Point p2;
      Point dir = p1 - p0;
      MMeshLine l2;
      Point dirGoal = goal - p0;
      double epsi = mEpsi * dirGoal.OrigDist2(2);
      if (dir.VectorProduct(dirGoal) < 0)
      { // goal lies on right side of l0
        l0 = -l0;
        return true;
      }

      l2 = l0.MGetPreviousTriLine();
      if (l2 == null)
      { // goal lies on left side of l0 and l0.LeftTri == null
        l0 = null;
        return true;
      }
      p2 = l2.Start;
      dir = p2 - p0;
      if (dir.VectorProduct(dirGoal) > 0)
      { // goal lies on right side of l2
        l0 = -l2;
        return true;
      }

      dir = p2 - p1;
      dirGoal = goal - p1;
      if (dir.VectorProduct(dirGoal) < 0)
      { // goal lies on right side of l1
        l0 = -(l0.GetNextTriLine());
        return true;
      }

      // goal lies within l0.LeftTri
      return false;
    }

    private MMeshLine FindLine(Point2D goal,MeshPoint start)
    {
      MMeshLine l0;
      MMeshLine l1 = start.IndexLine;
      do
      {
        l0 = l1;
      } while (FindNextLine(ref l1,goal) && l1 != null);
      return l0;
    }

    //
    //    int M2SetLeftTri(Mesh2 *mesh,int l,int t)
    //    {
    //      int sp,ep;
    //
    //      if (l > 0) 
    //      {
    //        mesh->line.lt[l] = t;
    //        sp = mesh->line.sp[l];
    //        ep = mesh->line.ep[l];
    //      } 
    //      else 
    //      {
    //        mesh->line.rt[-l] = t;
    //        sp = mesh->line.sp[-l];
    //        ep = mesh->line.ep[-l];
    //      }
    //      if (t > 0) 
    //      {
    //        mesh->ptri[sp] = t;
    //        mesh->ptri[ep] = t;
    //      }
    //      return 0;
    //    }
    //
    //    int M2SetStartPoint(Mesh2 *mesh,int l,int p)
    //    {
    //      if (l > 0) 
    //      {
    //        mesh->line.sp[l] = p;
    //      } 
    //      else 
    //      {
    //        mesh->line.ep[-l] = p;
    //      }  
    //      return 0;
    //    }
    //

    /// <summary>
    /// add a colinear point while only parallel lines exist
    /// </summary>
    /// <param name="point">point to insert</param>
    /// <returns>Newly insert mesh point or coincident old meshpoint</returns>
    private MeshPoint AddParallel(Point2D point)
    {
      Point start = mListLine[0].Start;
      int m = mListLine.Count - 1;
      Point end = mListLine[m].End;
      Point d = end - start;
      int index;
      MMeshLine newLine;

      if (Math.Abs(d.X) > Math.Abs(d.Y))
      { index = 0; }
      else
      { index = 1; }

      if ((point[index] > start[index]) == (point[index] < end[index]))
      { 
        for (int i = 0;i < mListLine.Count; i++)
        {
          start = mListLine[i].Start;
          end = mListLine[i].End;
          if ((point[index] >= start[index]) == (point[index] <= end[index]))
          { // segment found
            if (point.Dist2(start,2) < mFuzzy2)
            { return mListLine[i].MStartPoint; }
            if (point.Dist2(end,2) < mFuzzy2)
            { return mListLine[i].MEndPoint; }
            MeshPoint mp = new MeshPoint(point);
            newLine = mListLine[i].Split(mp);
            mListLine.Insert(i,newLine);
            return mp;
          }
        }
        throw new System.Exception("no segment found");
      }
      else
      { // point on either side of start or end
        if ((point[index] < start[index]) == (start[index] < end[index]))
        { // insert at start
          if (point.Dist2(start,2) < mFuzzy2)
          { return mListLine[0].MStartPoint; }
          else
          {
            MeshPoint sp = new MeshPoint(point);
            newLine = new MMeshLine(sp,mListLine[0].MStartPoint);
            mListLine.Insert(0,newLine);
            sp.IndexLine = newLine;
            return sp;
          }
        }
        else
        { // insert at end
          if (point.Dist2(end,2) < mFuzzy2)
          { return mListLine[m].MEndPoint; }
          else
          {
            MeshPoint sp = new MeshPoint(point);
            newLine = new MMeshLine(mListLine[m].MEndPoint,sp);
            mListLine.Add(newLine);
            sp.IndexLine = -newLine;
            return sp;
          }
        }
      }
    }

    //
    //    int M2TriChangeLineNr(Mesh2 *mesh,int t,int o,int n)
    //    {
    //      int o0,n0;
    //      int *l;
    //
    //      o0 = abs(o);
    //      if ((o < 0 && n < 0) || (o > 0 && n > 0)) n0 = n;
    //      else                                        n0 = -n;
    //
    //      if (o0 == abs(mesh->tri.l0[t]))      l = &mesh->tri.l0[t];
    //      else if (o0 == abs(mesh->tri.l1[t])) l = &mesh->tri.l1[t];
    //      else if (o0 == abs(mesh->tri.l2[t])) l = &mesh->tri.l2[t];
    //      else 
    //      {
    //        M2ErrorMsg("*** M2TriChangLineNr : no line %i in tri %i\n",o,t);
    //        return -1;
    //      }
    //      if (*l > 0) *l = n0;
    //      else        *l = -n0;
    //      return 0;
    //    }
    //
    //    int M2SetTri(Mesh2 *mesh,int i,
    //      double x,double y,double r2,int l0,int l1,int l2)
    //    {
    //      mesh->tri.x[i] = x;
    //      mesh->tri.y[i] = y;
    //      mesh->tri.r2[i] = r2;
    //      mesh->tri.l0[i] = l0;
    //      mesh->tri.l1[i] = l1;
    //      mesh->tri.l2[i] = l2;
    //      return 0;
    //    }
    //
    //    int M2AddTri0(Mesh2 *mesh,double x,double y,double r2,int l0,int l1,int l2)
    //    {
    //      mesh->nt++;
    //      M2SetTri(mesh,mesh->nt,x,y,r2,l0,l1,l2);
    //      return mesh->nt;
    //    }
    //
    //    double M2GetTriArea(Mesh2 *mesh,int t,double *x,double *y)
    //    {
    //      int l,e0,e1,e2;
    //
    //      l = mesh->tri.l0[t];
    //      e0 = M2GetStartPoint(mesh,l);
    //      e1 = M2GetStartPoint(mesh,-l);
    //      l = mesh->tri.l2[t];
    //      e2 = M2GetStartPoint(mesh,l);
    //      return G2GetTriArea(x[e0],y[e0],x[e1],y[e1],x[e2],y[e2]);
    //    }
    //
    private MeshPoint AddFirstTri(Point2D newPoint)
    {
      int lineMax = mListLine.Count - 1;
      MeshPoint p0,p1,p2;
      MMeshLine l0,l1;
      Point2D center;

      p0 = mListLine[0].MStartPoint;
      p1 = mListLine[lineMax].MEndPoint;
    
      center = Geometry.OutCircleCenter(p0.Point,p1.Point,newPoint);
      if (center == null)
      { return AddParallel(newPoint); }

      // check coincident with all points
      // assumption : parallel lines are sorted so, that each start point is equal to the
      // preceding endpoint
      if (newPoint.Dist2(p0.Point,2) < mFuzzy2)
      { return p0; }
      if (newPoint.Dist2(p1.Point,2) < mFuzzy2)
      { return p1; }
      for (int i = lineMax - 1; i >= 0; i-- )
      {
        p1 = mListLine[i].MEndPoint;
        Debug.Assert(mListLine[i + 1].MStartPoint == p1);

        if (newPoint.Dist2(p1.Point,2) < mFuzzy2)
        { return p1; }
      }

      p2 = new MeshPoint(newPoint);
      if (Geometry.TriArea(p0.Point,p1.Point,newPoint) > 0)
      {
        l0 = new MMeshLine(p1,p2);
        mListLine.Add(l0);
        l1 = new MMeshLine(p2,p0);
        mListLine.Add(l1);
        mListTri.Add(new MTri(mListLine[0],l0,l1));

        p2.IndexLine = l1;

        for (int i = 1; i <= lineMax; i++)
        {
          l1 = -l0;
          p1 = mListLine[i].MEndPoint;
          l0 = new MMeshLine(p1,p2);
          mListLine.Add(l0);
          mListTri.Add(new MTri(mListLine[i],l0,l1));
        }
      } 
      else 
      {
        l0 = new MMeshLine(p2,p1);
        mListLine.Add(l0);
        l1 = new MMeshLine(p0,p2);
        mListLine.Add(l1);
        mListTri.Add(new MTri(l0,-mListLine[0],l1));

        p2.IndexLine = l0;

        for (int i = 1; i <= lineMax; i++)
        {
          l1 = -l0;
          p1 = mListLine[i].MEndPoint;
          l0 = new MMeshLine(p2,p1);
          mListLine.Add(l0);
          mListTri.Add(new MTri(l0,mListLine[i],l1));
        }
      }
      return p2;
    }
    //
    //    int M2ReplacePoint(Mesh2 *mesh,int o_p,int n_p)
    //      /* o_p : old point number
    //         n_p : new point number
    //      */
    //    {
    //      int l0,l;
    //
    //      mesh->ptri[n_p] = mesh->ptri[o_p];
    //      if (mesh->ptri[n_p] < 0) 
    //      {
    //        return -1;
    //      }
    //      l0 = M2GetPointLine(mesh,o_p);
    //      l = l0;
    //      do 
    //      {
    //        M2SetStartPoint(mesh,l,n_p);
    //        l = M2GetNextPointLine(mesh,l);
    //      } while (l != l0);
    //
    //      return 0;
    //    }
    //
    //    int M2ReplaceLine(Mesh2 *mesh,int o_l,int n_l)
    //      /* o_l : old line number
    //         n_l : new line number
    //      */
    //    {
    //      int t1,t2;
    //
    //      t1 = mesh->line.lt[o_l];
    //      t2 = mesh->line.rt[o_l];
    //      M2SetLine(mesh,n_l,mesh->line.sp[o_l],mesh->line.ep[o_l],t1,t2);
    //      M2SetLineId(mesh,n_l,mesh->line.id[o_l]);
    //      if (t1 != 0) M2TriChangeLineNr(mesh,t1,o_l,n_l);
    //      if (t2 != 0) M2TriChangeLineNr(mesh,t2,o_l,n_l);
    //      return 0;
    //    }
    //
    //    int M2ReplaceTri(Mesh2 *mesh,int o_t,int n_t)
    //      /* o_t : old tri number
    //         n_t : new tri number
    //      */
    //    {
    //      MTri *tt;
    //      int l0,l1,l2;
    //
    //      tt = &(mesh->tri);
    //
    //      tt->x[n_t] = tt->x[o_t];
    //      tt->y[n_t] = tt->y[o_t];
    //      tt->r2[n_t] = tt->r2[o_t];
    //      l0 = tt->l0[o_t];
    //      l1 = tt->l1[o_t];
    //      l2 = tt->l2[o_t];
    //      tt->l0[n_t] = l0;
    //      tt->l1[n_t] = l1;
    //      tt->l2[n_t] = l2;
    //
    //      M2SetLeftTri(mesh,l0,n_t);
    //      M2SetLeftTri(mesh,l1,n_t);
    //      M2SetLeftTri(mesh,l2,n_t);
    //
    //      mesh->ptri[M2GetStartPoint(mesh,l0)] = n_t;
    //      mesh->ptri[M2GetStartPoint(mesh,l1)] = n_t;
    //      mesh->ptri[M2GetStartPoint(mesh,l2)] = n_t;
    //      return 0;
    //    }
    //
    //    int M2RmLine(Mesh2 *mesh,int l)
    //    {
    //      /* (Assumptions) : see M2StartAddTri 
    //         and mesh->ptri point to valid triangles
    //      */
    //      int t0;
    //      int n;
    //
    //      t0 = M2GetLeftTri(mesh,l);
    //  
    //      M2SetLeftTri(mesh,mesh->tri.l0[t0],0);
    //      M2SetLeftTri(mesh,mesh->tri.l1[t0],0);
    //      M2SetLeftTri(mesh,mesh->tri.l2[t0],0);
    //
    //      l = abs(l);
    //      n = mesh->nl;
    //      if (l < n) 
    //      {
    //        M2ReplaceLine(mesh,n,l);
    //      }
    //      mesh->nl--;
    //
    //      n = mesh->nt;
    //      if (t0 < n) 
    //      {
    //        M2ReplaceTri(mesh,n,t0);
    //      }
    //      mesh->nt--;
    //      return 0;
    //    }
    //
    //    int M2OptionChangeDiag0(Mesh2 *mesh,int l,Coo2 *co,int i)
    //      /* Assumption : line l is the 1. line of t0 !! */
    //      /* if assumption is not correct : run M2OptionChangeDiag */
    //    {
    //      int t0,t1;
    //      double dx,dy;
    //      int e0,e1,e2,e3;
    //      int l0,l1,l2,l3;
    //      double x,y,r2;
    //      int del;
    //
    //      if (M2GetLineId(mesh,l) != 0) return 0;
    //
    //      t0 = M2GetLeftTri(mesh,l);
    //      t1 = M2GetLeftTri(mesh,-l);
    //
    //      if (t1 == 0) 
    //      {
    //        e0 = M2GetTriPoint(mesh,t0,0);
    //        e1 = M2GetTriPoint(mesh,t0,1);
    //        e2 = M2GetTriPoint(mesh,t0,2);
    //        x = fabs(co->x[e0]) + fabs(co->x[e1]) + fabs(co->x[e2])
    //          + fabs(co->y[e0]) + fabs(co->y[e1]) + fabs(co->y[e2]);
    //        if (G2GetTriArea(co->x[e0],co->y[e0],co->x[e1],co->y[e1],
    //          co->x[e2],co->y[e2]) < 1.e-12 * x) 
    //        {
    //          return l;
    //        } 
    //        else 
    //        {
    //          return 0;
    //        }
    //      }
    //
    //      dx = co->x[i] - mesh->tri.x[t1];
    //      dy = co->y[i] - mesh->tri.y[t1];
    //      if (mesh->tri.r2[t1] <= dx*dx + dy*dy) 
    //      {
    //        return 0;
    //      } 
    //      else 
    //      {
    //        e0 = M2GetTriPoint(mesh,t0,2);
    //        l0 = mesh->tri.l2[t0];
    //        e1 = M2GetTriPoint(mesh,t0,0);
    //        if (-l == mesh->tri.l0[t1]) 
    //        {
    //          l1 = mesh->tri.l1[t1];
    //          e2 = M2GetTriPoint(mesh,t1,2);
    //          l2 = mesh->tri.l2[t1];
    //        } 
    //        else if (-l == mesh->tri.l1[t1]) 
    //        {
    //          l1 = mesh->tri.l2[t1];      
    //          e2 = M2GetTriPoint(mesh,t1,0);
    //          l2 = mesh->tri.l0[t1];
    //        } 
    //        else 
    //        {
    //          l1 = mesh->tri.l0[t1];      
    //          e2 = M2GetTriPoint(mesh,t1,1);
    //          l2 = mesh->tri.l1[t1];
    //        }
    //        e3 = M2GetTriPoint(mesh,t0,1);
    //        l3 = mesh->tri.l1[t0];
    //
    //        l = abs(l);
    //        M2SetLine(mesh,l,e2,e0,t0,t1);
    //
    //        G2GetOutCircle(co->x[e1],co->y[e1],co->x[e2],co->y[e2],
    //          co->x[e0],co->y[e0],&x,&y,&r2);
    //        M2SetTri(mesh,t0,x,y,r2,l1,l,l0);
    //        M2SetLeftTri(mesh,l1,t0);
    //        M2SetLeftTri(mesh,l0,t0);
    //
    //        G2GetOutCircle(co->x[e2],co->y[e2],co->x[e3],co->y[e3],
    //          co->x[e0],co->y[e0],&x,&y,&r2);
    //        M2SetTri(mesh,t1,x,y,r2,l2,l3,-l);
    //        M2SetLeftTri(mesh,l2,t1);
    //        M2SetLeftTri(mesh,l3,t1);
    //
    //        del = 0;
    //        del += M2OptionChangeDiag0(mesh,l1,co,i);
    //        del += M2OptionChangeDiag0(mesh,l2,co,i);
    //        return del;
    //      }
    //    }
    //
    //    int M2OptionChangeDiag(Mesh2 *mesh,int l,Coo2 *co2)
    //    {
    //      int i,t;
    //
    //      /* preparation to run M2OptionChangeDiag0 */
    //
    //      t = M2GetLeftTri(mesh,l);
    //      if (t == 0) 
    //      {
    //        l = -l;
    //        t = M2GetLeftTri(mesh,l);
    //      }
    //
    //      if (l == mesh->tri.l0[t]) 
    //      {
    //        /* already in correct order */
    //      } 
    //      else if (l == mesh->tri.l1[t]) 
    //      {
    //        mesh->tri.l1[t] = mesh->tri.l2[t];
    //        mesh->tri.l2[t] = mesh->tri.l0[t];
    //        mesh->tri.l0[t] = l;
    //      } 
    //      else if (l == mesh->tri.l2[t]) 
    //      {
    //        mesh->tri.l2[t] = mesh->tri.l1[t];
    //        mesh->tri.l1[t] = mesh->tri.l0[t];
    //        mesh->tri.l0[t] = l;
    //      } 
    //      else 
    //      {
    //        M2ErrorMsg("*** M2OptionChangeDiag : no line %i in tri %i\n",l,t);
    //      }
    //      i = M2GetStartPoint(mesh,mesh->tri.l2[t]);
    //      return M2OptionChangeDiag0(mesh,l,co2,i);
    //    }
    //
    //    int M2CompleteRightBord(Mesh2 *mesh,Coo2 *co,int l0)
    //    {
    //      int t;
    //      int l1,l2;
    //      int e0,e1,e2;
    //      double x,y,r2;
    //
    //      t = M2GetLeftTri(mesh,l0);
    //      l1 = l0;
    //      do 
    //      {
    //        l1 = -M2GetNextLine(mesh,t,l1);
    //        t = M2GetLeftTri(mesh,l1);
    //      } while (t != 0);
    //      e0 = M2GetStartPoint(mesh,l0);
    //      e1 = M2GetStartPoint(mesh,-l0);
    //      e2 = M2GetStartPoint(mesh,l1);
    //      x = fabs(co->x[e0]) + fabs(co->x[e1]) + fabs(co->x[e2])
    //        + fabs(co->y[e0]) + fabs(co->y[e1]) + fabs(co->y[e2]);
    //      if (G2GetTriArea(co->x[e2],co->y[e2],co->x[e1],co->y[e1],
    //        co->x[e0],co->y[e0]) > 1.e-12 * x) 
    //      {
    //
    //        if (G2GetOutCircle(co->x[e2],co->y[e2],co->x[e1],co->y[e1],
    //          co->x[e0],co->y[e0],&x,&y,&r2)) 
    //        {
    //          l2 = M2AddLine0(mesh,e0,e2,0,0); /*left,right side completed later */
    //
    //          t = M2AddTri0(mesh,x,y,r2,l1,-l0,l2);
    //          mesh->line.lt[l2] = t;
    //          M2SetLeftTri(mesh,l1,t);
    //          M2SetLeftTri(mesh,-l0,t);
    //
    //          M2OptionChangeDiag0(mesh,l1,co,e0);
    //          M2CompleteRightBord(mesh,co,l2);
    //        }
    //      }
    //      return 0;
    //    }
    //
    //    int M2CompleteLeftBord(Mesh2 *mesh,Coo2 *co,int l0)
    //    {
    //      int t;
    //      int l1,l2;
    //      int e0,e1,e2;
    //      double x,y,r2;
    //
    //      t = M2GetLeftTri(mesh,-l0);
    //      l1 = -l0;
    //      do 
    //      {
    //        l1 = -M2GetPrevLine(mesh,t,l1);
    //        t = M2GetLeftTri(mesh,l1);
    //      } while (t != 0);
    //      e0 = M2GetStartPoint(mesh,l0);
    //      e1 = M2GetStartPoint(mesh,-l0);
    //      e2 = M2GetStartPoint(mesh,-l1);
    //      x = fabs(co->x[e0]) + fabs(co->x[e1]) + fabs(co->x[e2])
    //        + fabs(co->y[e0]) + fabs(co->y[e1]) + fabs(co->y[e2]);
    //      if (G2GetTriArea(co->x[e0],co->y[e0],co->x[e1],co->y[e1],
    //        co->x[e2],co->y[e2]) > 1.e-12 * x) 
    //      {
    //
    //        if (G2GetOutCircle(co->x[e0],co->y[e0],co->x[e1],co->y[e1],
    //          co->x[e2],co->y[e2],&x,&y,&r2)) 
    //        {
    //          l2 = M2AddLine0(mesh,e2,e0,0,0); /*left,right side completed later */
    //
    //          t = M2AddTri0(mesh,x,y,r2,l1,l2,l0);
    //          mesh->line.lt[l2] = t;
    //          M2SetLeftTri(mesh,l0,t);
    //          M2SetLeftTri(mesh,l1,t);
    //
    //          M2OptionChangeDiag0(mesh,l1,co,e0);
    //          M2CompleteLeftBord(mesh,co,-l2);
    //        }
    //      }
    //      return 0;
    //    }
    //
    /// <summary>
    /// Adds a point outside current mesh
    /// </summary>
    private MeshPoint AddBorder(Point2D point,MMeshLine l)
    {
      MeshPoint pNew = new MeshPoint(point);
      MeshPoint p0,p1;
      MMeshLine l0,l1;
      Tri tri;
      Debug.Assert(l.LeftTri == null);

      p0 = l.MStartPoint;
      p1 = l.MEndPoint;

      l0 = new MMeshLine(p0,pNew);
      mListLine.Add(l0);
      l1 = new MMeshLine(p1,pNew);
      mListLine.Add(l1);

      pNew.IndexLine = -l0;

      tri = new MTri(l,l1,-l0);
      mListTri.Add(tri);

      l.OptionChangeDiag();
      CompleteLeftBord(l0);
      CompleteRightBord(l1);

      return pNew;
    }
    
    private void CompleteLeftBord(MMeshLine l0)
    {
      Debug.Assert(l0.MLeftTri == null);

      MMeshLine l1 = l0.MGetNextPointLine();
      MeshPoint p0 = l1.MEndPoint;
      MeshPoint p1 = l0.MStartPoint;
      MeshPoint p2 = l0.MEndPoint;

      if (Geometry.TriArea(p0.Point,p1.Point,p2.Point) <= 0)
      { return; }

      MMeshLine l2 = new MMeshLine(p0,p2);
      mListLine.Add(l2);
      MTri tri = new MTri(-l1,l0,-l2);
      mListTri.Add(tri);

      (-l1).OptionChangeDiag();
      CompleteLeftBord(l2);
    }

    private void CompleteRightBord(MMeshLine l0)
    {
      Debug.Assert(l0.MRightTri == null);

      MMeshLine l1 = l0.MGetPreviousPointLine();
      MeshPoint p0 = l1.MEndPoint;
      MeshPoint p1 = l0.MStartPoint;
      MeshPoint p2 = l0.MEndPoint;

      if (Geometry.TriArea(p0.Point,p2.Point,p1.Point) <= 0)
      { return; }

      MMeshLine l2 = new MMeshLine(p0,p2);
      mListLine.Add(l2);
      MTri tri = new MTri(l1,l2,-l0);
      mListTri.Add(tri);

      l1.OptionChangeDiag();
      CompleteRightBord(l2);
    }

    //
    //    int M2StartAddBorderC(Mesh2 *mesh,Coo2 *co,int p,int t)
    //      /* p: added point 
    //         t: start triangle */
    //    {
    //      int e0,e1,e2;
    //      int l;
    //      double x;
    //
    //      e0 = M2GetTriPoint(mesh,t,0);
    //      e1 = M2GetTriPoint(mesh,t,1);
    //      e2 = M2GetTriPoint(mesh,t,2);
    //      x = fabs(co->x[e0]) + fabs(co->x[e1]) + fabs(co->x[e2])
    //        + fabs(co->y[e0]) + fabs(co->y[e1]) + fabs(co->y[e2]);
    //      if (G2GetTriArea(co->x[e1],co->y[e1],co->x[e0],co->y[e0],
    //        co->x[p],co->y[p]) > 1.e-12 * x) 
    //      {
    //        l = -mesh->tri.l0[t];
    //      } 
    //      else if (G2GetTriArea(co->x[e2],co->y[e2],co->x[e1],co->y[e1],
    //        co->x[p],co->y[p]) > 1.e-12 * x) 
    //      {
    //        l = -mesh->tri.l1[t];
    //      } 
    //      else 
    //      {
    //        l = -mesh->tri.l2[t];
    //      }
    //      M2StartAddBorder0(mesh,co,p,l);
    //      return 0;
    //    }
    //
    //    int M2StartAddTri(Mesh2 *mesh,Coo2 *co,int p,int t)
    //      /* p: added point 
    //         t: start triangle 
    //         return value : deleted Line */
    //    {
    //      int e0,e1,e2;
    //      int l0,l1,l2;
    //      int t1,t2;
    //      double x,y,r2;
    //      int del;
    //
    //      e0 = M2GetTriPoint(mesh,t,0);
    //      e1 = M2GetTriPoint(mesh,t,1);
    //      e2 = M2GetTriPoint(mesh,t,2);
    //      l0 = M2AddLine0(mesh,e0,p,0,t); /* left/right sides completed later */
    //      l1 = M2AddLine0(mesh,e1,p,t,0);
    //      l2 = M2AddLine0(mesh,e2,p,0,0);
    //
    //      G2GetOutCircle(co->x[e1],co->y[e1],co->x[e2],co->y[e2],
    //        co->x[p],co->y[p],&x,&y,&r2);
    //      t1 = M2AddTri0(mesh,x,y,r2,mesh->tri.l1[t],l2,-l1);
    //      M2SetLeftTri(mesh,mesh->tri.l1[t],t1);
    //
    //      G2GetOutCircle(co->x[e2],co->y[e2],co->x[e0],co->y[e0],
    //        co->x[p],co->y[p],&x,&y,&r2);
    //      t2 = M2AddTri0(mesh,x,y,r2,mesh->tri.l2[t],l0,-l2);
    //      M2SetLeftTri(mesh,mesh->tri.l2[t],t2);
    //
    //      G2GetOutCircle(co->x[e0],co->y[e0],co->x[e1],co->y[e1],
    //        co->x[p],co->y[p],&x,&y,&r2);
    //      M2SetTri(mesh,t,x,y,r2,mesh->tri.l0[t],l1,-l0);
    //
    //      /* completing left/right side */
    //      mesh->line.lt[l0] = t2;
    //      mesh->line.rt[l1] = t1;
    //      mesh->line.lt[l2] = t1;
    //      mesh->line.rt[l2] = t2;
    //
    //      /* (Assumption) : 
    //        -maximally 1 triangle ( = left side of del) will be removed,
    //        -del is l0 of this triangle
    //        -right side of del is 0 (del is border) */
    //
    //      del = 0;
    //      del += M2OptionChangeDiag0(mesh,mesh->tri.l0[t],co,p);
    //      del += M2OptionChangeDiag0(mesh,mesh->tri.l0[t1],co,p);
    //      del += M2OptionChangeDiag0(mesh,mesh->tri.l0[t2],co,p);
    //
    //      if (del != 0) 
    //      {
    //        t1 = M2GetLeftTri(mesh,del);
    //        l1 = mesh->tri.l1[t1];
    //        l2 = mesh->tri.l2[t1];
    //        e0 = M2GetStartPoint(mesh,-l1); /* = M2GetStartPoint(mesh,l2) */
    //        e1 = M2GetStartPoint(mesh,l1);
    //        e2 = M2GetStartPoint(mesh,-l2);
    //        mesh->ptri[e0] = M2GetLeftTri(mesh,-l1);
    //        mesh->ptri[e1] = M2GetLeftTri(mesh,-l1);
    //        mesh->ptri[e2] = M2GetLeftTri(mesh,-l2);
    //
    //        M2RmLine(mesh,del);
    //
    //        if (abs(l1) > mesh->nl) 
    //        {
    //          if (l1 > 0) l1 = abs(del);
    //          else        l1 = -abs(del);
    //        } 
    //        else if (abs(l2) > mesh->nl) 
    //        {
    //          if (l2 > 0) l2 = abs(del);
    //          else        l2 = -abs(del);
    //        }
    //        M2CompleteRightBord(mesh,co,-l1);
    //        M2CompleteLeftBord(mesh,co,l2);
    //      }
    //
    //      return del;
    //    }
    //
    /// <summary>
    /// adds a point to the mesh if it is not closer 
    /// than fuzzy distance to an existing meshpoint 
    /// </summary>
    /// <param name="point"></param>
    /// <returns>added meshpoint or coincident meshpoint</returns>
    private MeshPoint Add(Point2D point,MeshPoint startPoint)
    {
      MMeshLine l0,l1,l2;
      MTri tri0,tri1,tri2;
      MeshPoint newPoint;

      #region first point(s)
      if (mListTri.Count == 0)
      {
        if (mListLine.Count == 0)
        {
          MeshPoint p0,p1;
          if (mListPoint.Count == 0)
          { // add first point
            p0 = new MeshPoint(point);
            return p0;
          }

          // try add first line
          p0 = (MeshPoint) mListPoint[0].Value;

          if (point.Dist2(p0.Point,2) < mFuzzy2)
          { return p0; }

          p1 = new MeshPoint(point);
          mListLine.Add(new MMeshLine(p0,p1));
          return p1;
        }

        return AddFirstTri(point);
      }
      #endregion

      // normal points
      if (startPoint == null)
      { startPoint = (MeshPoint) mListPoint[0].Value; }

      l0 = FindLine(point,startPoint);
      tri0 = l0.MLeftTri;

      if (tri0 == null)
      { // point outside of current mesh
        if (point.Dist2(l0.Start,2) <= mFuzzy2)
        { return l0.MStartPoint; }
        if (point.Dist2(l0.End,2) <= mFuzzy2)
        { return l0.MEndPoint; }

        return AddBorder(point,l0);
      }

      for (int i = 0; i < 3; i++) 
      {
        if (point.Dist2(tri0[i].Start,2) <= mFuzzy2)
        { return tri0.MGetLine(i).MStartPoint; }
      }

      newPoint = new MeshPoint(point);
      l0 = new MMeshLine(tri0.MGetLine(0).MStartPoint,newPoint);
      mListLine.Add(l0);
      l1 = new MMeshLine(tri0.MGetLine(1).MStartPoint,newPoint);
      mListLine.Add(l1);
      l2 = new MMeshLine(tri0.MGetLine(2).MStartPoint,newPoint);
      mListLine.Add(l2);

      newPoint.IndexLine = -l0;

      tri1 = new MTri(tri0.MGetLine(1),l2,-l1);
      mListTri.Add(tri1);
      tri2 = new MTri(tri0.MGetLine(2),l0,-l2);
      mListTri.Add(tri2);

      tri0.ClearCircle();
      tri0.MSetLine(1,l1);
      tri0.MSetLine(2,-l0);

      tri0[0].OptionChangeDiag();
      tri1[0].OptionChangeDiag();
      tri2[0].OptionChangeDiag();

      return newPoint;
    }

      //
    //    int M2MovePoint(Mesh2 *mesh,Coo2 *co2,int i,double x,double y)
    //    {
    //      int k,t;
    //      int l0,l1,l2;
    //
    //      co2->x[i] = x;
    //      co2->y[i] = y;
    //
    //      M2WarningMsg("***M2MovePoint : not completely implemented !\n");
    //      M2WarningMsg("***M2MovePoint : current Posistion : %12.1f %12.1f\n",x,y);
    //      t = mesh->ptri[i];
    //      if (t < 0) 
    //      {
    //        M2AddPoint(mesh,co2,i,-t);
    //      } 
    //      else 
    //      {
    //        k = M2GetTriPartPoint(mesh,t,i);
    //        if (k < 0) 
    //        {
    //          M2ErrorMsg("%i %.1f %.1f\n",i,co2->x[i],co2->y[i]);
    //        }
    //        l0 = M2GetTriLine(mesh,t,k);
    //        l2 = l0;
    //        do 
    //        {
    //          if (t > 0) 
    //          {
    //            l1 = M2GetNextLine(mesh,t,l2);
    //            M2OptionChangeDiag(mesh,l1,co2);
    //          }
    //          l2 = M2GetNextPointLine(mesh,l2);
    //          t = M2GetLeftTri(mesh,l2);
    //        } while (l2 != l0);
    //      }
    //      return 0;
    //    }
    //
    //    int M2BorderClean(Mesh2 *mesh,Coo2 *co2,double angle)
    //    {
    //      int i,l0,t;
    //      int e0,e1;
    //      int n;
    //      double sa,f;
    //      double dx0,dx1,dy0,dy1;
    //  
    //      sa = sin(angle);
    //      n = 0;
    //      for (i = 1; i < mesh->nl; i++) 
    //      {
    //        if (mesh->line.id[i] == 0) 
    //        {
    //          l0 = 0;
    //          if (mesh->line.lt[i] == 0) 
    //          {
    //            l0 = i;
    //          } 
    //          else if (mesh->line.rt[i] == 0) 
    //          {
    //            l0 = -i;
    //          }
    //          if (l0 != 0) 
    //          {
    //            t = M2GetLeftTri(mesh,-l0);
    //            e0 = M2GetStartPoint(mesh,l0);
    //            e1 = M2GetStartPoint(mesh,-l0);
    //            dx0 = co2->x[e0] - mesh->tri.x[t];
    //            dy0 = co2->y[e0] - mesh->tri.y[t];
    //            dx1 = co2->x[e1] - mesh->tri.x[t];
    //            dy1 = co2->y[e1] - mesh->tri.y[t];
    //            f = (dx0 * dy1 - dx1 * dy0) / mesh->tri.r2[t];
    //            if (f > 0. && f < sa) 
    //            {  /* remove triangle t , if possible */
    //              int l1,l2,e2;
    //              /*check for double border triangle */
    //              l1 = -M2GetNextLine(mesh,t,-l0);
    //              l2 = -M2GetNextLine(mesh,t,-l1);
    //              if (M2GetLeftTri(mesh,l1) != 0 && M2GetLeftTri(mesh,l2) != 0) 
    //              {
    //                /* validate ptri of edge points */
    //                e2 = M2GetStartPoint(mesh,l1);
    //            
    //                mesh->ptri[e0] = M2GetLeftTri(mesh,l1);
    //                mesh->ptri[e1] = M2GetLeftTri(mesh,l2);
    //                mesh->ptri[e2] = M2GetLeftTri(mesh,l2);
    //            
    //                M2GetTriPartPoint(mesh,M2GetLeftTri(mesh,l1),e0);
    //                M2GetTriPartPoint(mesh,M2GetLeftTri(mesh,l2),e1);
    //                M2GetTriPartPoint(mesh,M2GetLeftTri(mesh,l2),e2);
    //
    //                M2RmLine(mesh,-l0);
    //                n++;
    //                i --; /* check new line assigned to same line # */
    //              }
    //            }
    //          }
    //        }
    //      }
    //      M2StatusMsg("Dropped %i border triangles\n",n);
    //      return n;
    //    }
    //
    //    int M2DropPoint(Mesh2 *mesh,Coo2 *co,int p)
    //    {
    //      int i;
    //      int l,l0,l1,l2,ln;
    //      int t1,t2;
    //      int p1;
    //      double dx,dy,d2,d20;
    //
    //      if (mesh->ptri[p] <= 0) 
    //      {
    //        return mesh->ptri[p];
    //      }
    //
    //      /* find closest point and reset IDs */
    //      l0 = M2GetPointLine(mesh,p);
    //      l = l0;
    //      d20 = -1.;
    //      do 
    //      {
    //        M2SetLineId(mesh,l,0);
    //        /* check for border */
    //        if (M2GetLeftTri(mesh,l) == 0) 
    //        {
    //          return -1;
    //        }
    //        p1 = M2GetStartPoint(mesh,-l);
    //        dx = co->x[p1] - co->x[p];
    //        dy = co->y[p1] - co->y[p];
    //        d2 = dx * dx + dy * dy;
    //        if (d20 < 0. || d2 < d20) 
    //        {
    //          d20 = d2;
    //          ln = l;
    //        }
    //        l = M2GetNextPointLine(mesh,l);
    //      } while (l != l0);
    //
    //      t1 = M2GetLeftTri(mesh,ln);
    //      t2 = M2GetLeftTri(mesh,-ln);
    //      l1 = M2GetNextLine(mesh,t1,ln);
    //      l2 = M2GetPrevLine(mesh,t2,-ln);
    //
    //      /* "move" p to p0, calculate new out circles
    //         drop ln,l0,l2
    //      */
    //      mesh->ptri[p] = 0;
    //      return 1;
    //    }
    //
    //    int M2ListLines(Mesh2 *mesh,FILE *fout)
    //    {
    //      int i;
    //      MLine *l;
    //
    //      l = &mesh->line;
    //      if (fout == 0) fout = stdout;
    //      fprintf(fout,"   #     sp   ep   lt   rt\n");
    //      for (i = 1; i <= mesh->nl; i++) 
    //      {
    //        fprintf(fout,"%4i    %4i %4i  %4i %4i\n",
    //          i,l->sp[i],l->ep[i],l->lt[i],l->rt[i]);
    //      }
    //      return 0;
    //    }
    //
    //    int M2ListTri(Mesh2 *mesh,int i,FILE *fout)
    //    {
    //      int l0,l1,l2;
    //      if (fout == 0) fout = stdout;
    //      fprintf(fout,"   #     l0     l1     l2         s0     s1     s2");
    //      fprintf(fout,"        rt     rt     rt\n");
    //      fprintf(fout,"                                  e0     e1     e2");
    //      fprintf(fout,"        lt     lt     lt\n");
    //      l0 = mesh->tri.l0[i];
    //      l1 = mesh->tri.l1[i];
    //      l2 = mesh->tri.l2[i];
    //      fprintf(fout,"%6i  %6i %6i %6i    %6i %6i %6i    %6i %6i %6i\n",
    //        i,l0,l1,l2,M2GetStartPoint(mesh,l0),M2GetStartPoint(mesh,l1),
    //        M2GetStartPoint(mesh,l2),
    //        M2GetLeftTri(mesh,-l0),M2GetLeftTri(mesh,-l1),M2GetLeftTri(mesh,-l2));
    //      fprintf(fout,"                              %5i %5i %5i    %5i %5i %5i\n",
    //        M2GetStartPoint(mesh,-l0),M2GetStartPoint(mesh,-l1),
    //        M2GetStartPoint(mesh,-l2),
    //        M2GetLeftTri(mesh,l0),M2GetLeftTri(mesh,l1),M2GetLeftTri(mesh,l2));
    //      return 0;
    //    }
    //
    //    int M2ListTris(Mesh2 *mesh,FILE *fout)
    //    {
    //      int i;
    //      if (fout == 0) fout = stdout;
    //      fprintf(fout,"   #     l0    l1    l2\n");
    //      for (i = 1; i <= mesh->nt; i++) 
    //      {
    //        fprintf(fout,"%4i  %5i %5i %5i\n",
    //          i,mesh->tri.l0[i],mesh->tri.l1[i],mesh->tri.l2[i]);
    //      }
    //      return 0;
    //    }
    //
    //    int  (*m2StatusMsg)(char *format,va_list args) = 0;
    //    int (*m2WarningMsg)(char *format,va_list args) = 0;
    //    int   (*m2ErrorMsg)(char *format,va_list args) = 0;
    //
    //    int M2SetStatusMsgFct(int (*fct)(char *format,va_list args))
    //  {
    //    m2StatusMsg = fct;
    //    return 0;
    //  }
    //  int M2SetWarningMsgFct(int (*fct)(char *format,va_list args))
    //{
    //  m2WarningMsg = fct;
    //  return 0;
    //}
    //  int M2SetErrorMsgFct(int (*fct)(char *format,va_list args))
    //{
    //  m2ErrorMsg = fct;
    //  return 0;
    //}
    //  int M2StatusMsg(char *format,...)
    //{
    //  va_list args;
    //  if (m2StatusMsg) 
    //{
    //  va_start(args,format);
    //  m2StatusMsg(format,args);
    //  va_end(args);
    //}
    //  return 0;
    //}
    //
    //  int M2WarningMsg(char *format, ...)
    //{
    //  va_list args;
    //  va_start(args,format);
    //  if (m2WarningMsg) 
    //{
    //  m2WarningMsg(format,args);
    //} 
    //  else 
    //{
    //  vfprintf(stdout,format,args);
    //}
    //  va_end(args);
    //  return 0;
    //}
    //
    //  int M2ErrorMsg(char *format, ...)
    //{
    //  va_list args;
    //  va_start(args,format);
    //  if (m2ErrorMsg) 
    //{
    //  m2ErrorMsg(format,args);
    //} 
    //  else 
    //{
    //  vfprintf(stderr,format,args);
    //}
    //  va_end(args);
    //  return -1;
    //}
  }
}
