﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Globalization;
using System.Windows.Forms;
using System.IO;
using System.Windows.Forms.DataVisualization.Charting;
using System.Windows.Forms.DataVisualization.Charting.Utilities;
using AForge.Neuro;
using AForge.Neuro.Learning;
using System.Threading;
using System.Diagnostics;

namespace GPSLearning
{
  public partial class NeuralNetworkForm : Form
  {
    private TrainingSet _trainingSet;
    private TrainingSet _validationSet;
    private TrainingSet _testSet;
    private VeloActivationNetwork _network;
    private VeloActivationNetworkCommittee _networkCommittee;
    private ISupervisedLearning _teacher;
    private double _epochOfBestNetwork;
    private string lastRunTrainingType = "";

    

    public NeuralNetworkForm()
    {
      InitializeComponent();
      cmbDegradeMethod.DataSource = Enum.GetValues(typeof(DegradeMethod));
      cmbAlgorithm.DataSource = Enum.GetValues(typeof (LearningAlgorithm));
      cmbAlgorithm.SelectedItem = LearningAlgorithm.BackProp;
      cmbInputVector.SelectedIndex = 1; // Map feature groups
      cmbPlot.SelectedIndex = 0; // Error plot
    }

    private void btnAddTrainingFiles_Click(object sender, EventArgs e)
    {
      dlgOpen.Multiselect = true;
      dlgOpen.Filter = "GPX files (*.gpx)|*.gpx";
      if (dlgOpen.ShowDialog(this) == DialogResult.OK)
      {
        foreach (String filename in dlgOpen.FileNames)
        {
          trainingSetList.Items.Add(filename);
        }
      }
      dlgOpen.Multiselect = false;
    }

    private void btnHeight_Click(object sender, EventArgs e)
    {
      dlgOpen.FileName = txtHeight.Text;
      dlgOpen.Filter = "Esri ASCII Grid files (*.asc, *.agr, *.grd)|*.asc;*.agr;*.grd";
      if (dlgOpen.ShowDialog(this) == DialogResult.OK)
      {
        txtHeight.Text = dlgOpen.FileName;
      }
    }

    private void btnForest_Click(object sender, EventArgs e)
    {
      dlgOpen.FileName = txtForest.Text;
      dlgOpen.Filter = "*.tif|*.tif";
      if (dlgOpen.ShowDialog(this) == DialogResult.OK)
      {
        txtForest.Text = dlgOpen.FileName;
      }
    }

    private void txtForest_TextChanged(object sender, EventArgs e)
    {
      if (chbSynchFileNames.Checked)
      {
        String directory = Path.GetDirectoryName(txtForest.Text) + Path.DirectorySeparatorChar;
        String extension = Path.GetExtension(txtForest.Text);
        String forestModelFilename = Path.GetFileNameWithoutExtension(txtForest.Text);
        if (forestModelFilename != null)
        {
          String filenamePrefix = forestModelFilename.Substring(0, forestModelFilename.Length - 7);

          txtWater.Text = directory + filenamePrefix + "_water" + extension;
          txtUndergrowth.Text = directory + filenamePrefix + "_undergrowth" + extension;
          txtPaths.Text = directory + filenamePrefix + "_paths_training" + extension;
          txtGround.Text = directory + filenamePrefix + "_ground" + extension;
          txtPassable.Text = directory + filenamePrefix + "_passable" + extension;
        }
      }
    }

    private void btnCourse_Click(object sender, EventArgs e)
    {
      dlgOpen.FileName = txtCourse.Text;
      dlgOpen.Filter = "*.ocd|*.ocd";
      if (dlgOpen.ShowDialog(this) != DialogResult.OK)
      { return; }

      txtCourse.Text = dlgOpen.FileName;
    }

    private void btnRemoveGpsFiles_Click(object sender, EventArgs e)
    {
      for (int i = trainingSetList.Items.Count - 1; i >= 0; i--)
      {
        if (trainingSetList.SelectedIndices.Contains(i))
        {
          trainingSetList.Items.RemoveAt(i);
        }
      }
    }

    private void btnMap_Click(object sender, EventArgs e)
    {
      dlgOpen.FileName = txtMap.Text;
      dlgOpen.Filter = "*.tif|*.tif";
      if (dlgOpen.ShowDialog(this) == DialogResult.OK)
      {
        txtMap.Text = dlgOpen.FileName;
      }
    }

    private void btnCreate_Click(object sender, EventArgs e)
    {
      btnCreate.Enabled = false;
      lblStatus.Text = "Creating training set. This may take some time...";

      // Put all gpx file names in an array
      int gpxCount = trainingSetList.Items.Count;
      var gpxFiles = new string[gpxCount];
      for (int i = 0; i < gpxCount; i++)
      {
        gpxFiles[i] = trainingSetList.Items[i].ToString();
      }

      // Create a training data "factory"
      var trainingDataCreator = new TrainingDataCreator(txtHeight.Text, txtForest.Text, txtWater.Text,
                                                        txtUndergrowth.Text, txtPaths.Text, txtGround.Text,
                                                        txtPassable.Text, gpxFiles, int.Parse(txtRemoveAtStart.Text),
                                                        int.Parse(txtRemoveAtEnd.Text), int.Parse(txtGpsUpdateInterval.Text),
                                                        (DegradeMethod)cmbDegradeMethod.SelectedItem);

      // Run training data creation in a separate thread
      var createTrainingDataThread = new Thread(trainingDataCreator.CreateTrainingData) {IsBackground = true};
      createTrainingDataThread.Start();
      createTrainingDataThread.Join();
      _trainingSet = trainingDataCreator.TrainingSet;
      lblStatus.Text = "Training set ready: " + _trainingSet.TrainingData.Count 
        + " training instances created. Click \"Save\" to save the data as XML.";
      btnCreate.Enabled = true;
      btnSave.Enabled = true;
    }

    private class TrainingDataCreator
    {
      private readonly string _heightFile, _forestFile, _waterFile, _undergrowthFile, _pathsFile, _groundFile, _passableFile;
      private readonly string[] _gpxFiles;
      private readonly int _removeAtStart;
      private readonly int _removeAtEnd;
      private readonly int _gpsUpdateInterval;
      private readonly DegradeMethod _degradeMethod;
      public TrainingSet TrainingSet { get; private set; }

      public TrainingDataCreator(string heightFile, string forestFile, string waterFile, string undergrowthFile,
        string pathsFile, string groundFile, string passableFile, string[] gpxFiles, int removeAtStart, int removeAtEnd, int gpsUpdateInterval, DegradeMethod degradeMethod)
      {
        _heightFile = heightFile;
        _forestFile = forestFile;
        _waterFile = waterFile;
        _undergrowthFile = undergrowthFile;
        _pathsFile = pathsFile;
        _groundFile = groundFile;
        _passableFile = passableFile;
        _gpxFiles = gpxFiles;
        _removeAtStart = removeAtStart;
        _removeAtEnd = removeAtEnd;
        _gpsUpdateInterval = gpsUpdateInterval;
        _degradeMethod = degradeMethod;
      }

      public void CreateTrainingData()
      {
        TrainingSet = new TrainingSet(_heightFile, _forestFile, _waterFile, _undergrowthFile,
          _pathsFile, _groundFile, _passableFile, _gpxFiles, _removeAtStart, _removeAtEnd, _gpsUpdateInterval, _degradeMethod);
      }
    }

    private void btnSave_Click(object sender, EventArgs e)
    {
      dlgSave.Filter = "*.xml|*.xml";
      if (dlgSave.ShowDialog(this) == DialogResult.OK)
      {
        TrainingSetSerializer.SerializeToXml(dlgSave.FileName, _trainingSet);
        lblStatus.Text = "Training set saved.";
      }
    }

    private void btnBrowseTrainingSet_Click(object sender, EventArgs e)
    {
      dlgOpen.FileName = txtTrainingSet.Text;
      dlgOpen.Filter = "*.xml|*.xml";
      if (dlgOpen.ShowDialog(this) == DialogResult.OK)
      {
        txtTrainingSet.Text = dlgOpen.FileName;
      }
    }

    private double trainNetwork(double outputScaleFactor, TrainingSet trainingSet, TrainingSet validationSet, TrainingSet testSet,
      InputVectorType inputVectorType, bool normalizeInputVector)
    {

      // Prepare training plot
      for (int i = chartTrainingPlot.Series.Count - 1; i >= 0; i--)
      {
        if (chartTrainingPlot.Series[i] == chartTrainingPlot.Series["Training Error"]
          || chartTrainingPlot.Series[i] == chartTrainingPlot.Series["Validation Error"])
        {
          chartTrainingPlot.Series[i].Points.Clear();
        }
        else
        {
          chartTrainingPlot.Series.RemoveAt(i);
        }
      }
      chartTrainingPlot.Series["Training Error"].ToolTip = "Epoch: #VALX, Training Error: #VALY";
      chartTrainingPlot.Series["Validation Error"].ToolTip = "Epoch: #VALX, Validation Error: #VALY";

      // Prepare key features velo development plotting
      KeyFeatureTestSet keyFeatureTestSet = null;
      double[][] keyFeatureTestInputs = null;
      if (txtKeyFeatureTest.Text != "")
      {
        keyFeatureTestSet = TrainingSetSerializer.DeserializeKeyFeatureTestSetFromXml(txtKeyFeatureTest.Text);
        keyFeatureTestInputs = keyFeatureTestSet.GetInput(inputVectorType, normalizeInputVector);
        for (int i = 0; i < keyFeatureTestSet.Tests.Count; i++)
        {
          try
          {
            chartTrainingPlot.Series.Add(keyFeatureTestSet.Tests[i].ToString());
            chartTrainingPlot.Series[keyFeatureTestSet.Tests[i].ToString()].ToolTip = "Epoch: #VALX, Velo: #VALY";
          }
          catch
          {
            ArgumentException exception;
          }
          chartTrainingPlot.Series[keyFeatureTestSet.Tests[i].ToString()].ChartType = SeriesChartType.Line;
        }
        if (cmbPlot.SelectedIndex == 0) // Plot errors
        {
          foreach (Series s in chartTrainingPlot.Series)
          {
            s.Enabled = false;
          }
          chartTrainingPlot.Series["Training Error"].Enabled = true;
          chartTrainingPlot.Series["Validation Error"].Enabled = true;

          chartTrainingPlot.Titles["Header"].Text = "Training and validation error";
          chartTrainingPlot.ChartAreas[0].AxisY.Title = "Mean Abs. Error (m/s)";
        }
        else // Plot velo development
        {
          foreach (Series s in chartTrainingPlot.Series)
          {
            s.Enabled = true;
          }
          chartTrainingPlot.Series["Training Error"].Enabled = false;
          chartTrainingPlot.Series["Validation Error"].Enabled = false;

          chartTrainingPlot.Titles["Header"].Text = "Speed development during training";
          chartTrainingPlot.ChartAreas[0].AxisY.Title = "Output speed";
          if (chbScaleOutput.Checked)
          {
            chartTrainingPlot.ChartAreas[0].AxisY.Title += " scaled to speed range (m/s)";
          }
          else
          {
            chartTrainingPlot.ChartAreas[0].AxisY.Title += " relative to theoretical max. speed";
          }
        }
      }

      lblStatus.Text = "Training network.";

      // Get network and learning parameters
      var learningAlgorithm = (LearningAlgorithm)cmbAlgorithm.SelectedItem;
      double learningRate = double.Parse(txtLearningRate.Text);
      double momentum = double.NaN;
      if (learningAlgorithm == LearningAlgorithm.BackProp)
      {
        momentum = double.Parse(txtMomentum.Text);
      }
      double sigmoidAlphaValue = double.Parse(txtAlpha.Text);
      int neuronsInFirstLayer = int.Parse(txtNeuronsFirstLayer.Text);
      int neuronsInSecondLayer = int.Parse(txtNeuronsSecondLayer.Text);
      int neuronsInThirdLayer = int.Parse(txtNeuronsThirdLayer.Text);
      int epochs = int.Parse(txtEpochs.Text);

      int nofInputNeurons;
      switch (inputVectorType)
      {
        case InputVectorType.OneInputPerFeature:
          nofInputNeurons = TrainingSet.NofMapFeatures + 3;
          break;
        case InputVectorType.OneInputPerGroup:
        default:
          nofInputNeurons = TrainingSet.NofMapFeatureGroups + 3;
          break;
      }

      chartTrainingPlot.Titles["Algorithm"].Text = "Algorithm: " + learningAlgorithm;
      chartTrainingPlot.Titles["InputVector"].Text = "Input vector: " + inputVectorType;
      chartTrainingPlot.Titles["Alpha"].Text = "Alpha: " + txtAlpha.Text;
      chartTrainingPlot.Titles["Layers"].Text = "Hidden layers: " + neuronsInFirstLayer + ", " + neuronsInSecondLayer + ", " + neuronsInThirdLayer;
      chartTrainingPlot.Titles["Epochs"].Text = "Epochs: " + txtEpochs.Text;
      chartTrainingPlot.Titles["LearningRate"].Text = "Learning rate: " + txtLearningRate.Text;
      chartTrainingPlot.ChartAreas[0].AxisX.Title = "Epoch (1 epoch = all " + trainingSet.TrainingData.Count +
                                             " training instances)";

      if (learningAlgorithm == LearningAlgorithm.BackProp)
      {
        chartTrainingPlot.Titles["Momentum"].Visible = true;
        chartTrainingPlot.Titles["Momentum"].Text = "Momentum: " + txtMomentum.Text;
      }
      else
      {
        chartTrainingPlot.Titles["Momentum"].Visible = false;
      }

      logLine("-- NETWORK PARAMS --");
      logLine("Learning algorithm: " + learningAlgorithm);
      logLine("Input vector: " + inputVectorType);
      logLine("Learning rate: " + learningRate);
      if (learningAlgorithm == LearningAlgorithm.BackProp)
      {
        logLine("Momentum: " + momentum);
      }
      logLine("Sigmoid alpha value: " + sigmoidAlphaValue);
      logLine("First hidden layer: " + neuronsInFirstLayer + " neurons");
      logLine("Second hidden layer: " + neuronsInSecondLayer + " neurons");
      logLine("Third hidden layer: " + neuronsInThirdLayer + " neurons");
      logLine("Maximum no. epochs: " + epochs);
      logLine();

      // Create activation network
      if (neuronsInFirstLayer > 0 && neuronsInSecondLayer > 0 && neuronsInThirdLayer > 0)
      {
        // Three hidden layers
        _network = new VeloActivationNetwork(
          new SigmoidFunction(sigmoidAlphaValue),
          nofInputNeurons, neuronsInFirstLayer, neuronsInSecondLayer, neuronsInThirdLayer, 1);
      }
      else if (neuronsInFirstLayer > 0 && neuronsInSecondLayer > 0)
      {
        // Two hidden layers
        _network = new VeloActivationNetwork(
          new SigmoidFunction(sigmoidAlphaValue),
          nofInputNeurons, neuronsInFirstLayer, neuronsInSecondLayer, 1);
      }
      else if (neuronsInFirstLayer > 0)
      {
        // One hidden layer
        _network = new VeloActivationNetwork(
          new SigmoidFunction(sigmoidAlphaValue),
          nofInputNeurons, neuronsInFirstLayer, 1);
      }
      else
      {
        // No hidden layers
        _network = new VeloActivationNetwork(
          new SigmoidFunction(sigmoidAlphaValue),
          nofInputNeurons, 1);
      }

      _network.OutputScaleFactor = outputScaleFactor; // 'outputScaleFactor' is a theoretical max velo (3*avg velo of training set)
      _network.NormalizeInputVector = normalizeInputVector;
      _network.InputVectorType = inputVectorType;

      // Create teacher
      if (learningAlgorithm == LearningAlgorithm.BackProp)
      {
        _teacher = new BackPropagationLearning(_network) { LearningRate = learningRate, Momentum = momentum };
      }
      else
      {
        _teacher = new ResilientBackpropagationLearning(_network) { LearningRate = learningRate };
      }

      // Set up how often the training error should be averaged and plotted
      int nofTrainingInstancesBetweenPlots;
      if (rdbAvgErrorPreviousEpoch.Checked)
      {
        // Plot avg error for the last training epoch
        nofTrainingInstancesBetweenPlots = trainingSet.TrainingData.Count;
      }
      else
      {
        // Plot avg error for the last 'nofTrainingInstancesBetweenPlots' instances
        nofTrainingInstancesBetweenPlots = int.Parse(txtAvgErrorPreviousInstances.Text);
      }


      int epoch = 1;
      //var inputs = new List<double[]>();
      //var inputStrings = new List<string>();

      // loop
      VeloActivationNetwork bestNetwork = _network; // Keeps track of the best network seen during training
      _epochOfBestNetwork = 0;
      double bestValidationError = 10000;
      //int totalTrainingIterations = 0;
      //double errorSum = 0;
      logLine("-- TRAINING EPOCHS --");
      while (epoch <= epochs)
      {
        // run epoch of learning procedure
        Utilities.Shuffle(trainingSet.TrainingData);
        double[][] input = trainingSet.GetInput(_network.InputVectorType, _network.NormalizeInputVector);
        double[][] output = trainingSet.GetOutput(_network.OutputScaleFactor);

        switch (learningAlgorithm)
        {
          case LearningAlgorithm.BackProp:
            for (int i = 0; i < input.Length; i++)
            {
              double error = _teacher.Run(input[i], output[i]); // error = square(output - desired ouput)/2

              // Test of no training at all
              /* double error = maxVelo*maxVelo*Math.Pow(_network.Compute(input[i])[0] - output[i][0], 2); */

              // This is the former training error plotting code. Now using a different approach. See below.
              /*errorSum += error;
              totalTrainingIterations++;
              if (totalTrainingIterations%nofTrainingInstancesBetweenPlots == 0)
              {
                // Plot the average error over the previous 'nofTrainingInstancesBetweenPlots' training instances.
                chartTrainingPlot.Series["Training Error"].Points.AddXY((double)totalTrainingIterations/trainingSet.TrainingData.Count,
                                                                 errorSum/nofTrainingInstancesBetweenPlots);
                errorSum = 0;
              }*/

              /*inputs.Add(input[i]);
              string inputString = "[ ";
              foreach (double inp in input[i])
              {
                inputString += inp.ToString("N2") + ", ";
              }
              inputString += "]";
              inputStrings.Add(inputString);*/
            }
            break;
          default: // Rprop
            {
              double error = _teacher.RunEpoch(input, output) / trainingSet.TrainingData.Count; // Train one epoch and return avg training error during the last epoch
              //chartTrainingPlot.Series["Training Error"].Points.AddXY(epoch, error);
            }
            break;
        }

        chartTrainingPlot.ChartAreas[0].AxisX2.Maximum = chartTrainingPlot.ChartAreas[0].AxisX.Maximum / trainingSet.TrainingData.Count;

        // Run key feature tests to see velo development throughout training
        if (keyFeatureTestSet != null)
        {
          for (int i = 0; i < keyFeatureTestInputs.Length; i++)
          {
            double velo = _network.Compute(keyFeatureTestInputs[i])[0];
            if (chbScaleOutput.Checked)
            {
              velo = velo * _network.OutputScaleFactor;
            }
            chartTrainingPlot.Series[keyFeatureTestSet.Tests[i].ToString()].Points.AddXY(epoch, velo);
          }
        }

        // Run training set to find error over training set after this training epoch
        double meanTrainingError = computeMeanErrorAfterEpoch(trainingSet, "Training Error", epoch);

        // Run validation set to find error over validation set after this training epoch
        double meanValidationError = computeMeanErrorAfterEpoch(validationSet, "Validation Error", epoch);

        // Keep track of best network seen during training
        if (meanValidationError < bestValidationError)
        {
          bestValidationError = meanValidationError;
          bestNetwork = ObjectCopier.Clone<VeloActivationNetwork>(_network);
          _epochOfBestNetwork = epoch;
        }

        // Refresh training plot every 100 epochs. (Refreshing every epoch slows down the process a lot.)
        if (epoch % 100 == 0)
        {
          chartTrainingPlot.Refresh();
        }

        if (chbPrintMeanValidationError.Checked)
        {
          logLine("Completed epoch " + epoch + ": Mean validation error = " + meanValidationError);
        }
        epoch++;

        // Check for keystroke indicating abort
        if (ModifierKeys == Keys.Escape)  // DOES NOT WORK. NEED TO IMPLEMENT TRAINING OPERATION AS A BACKGROUND WORKER TO SUPPORT CANCELLING.
        {
          break;
        }
      }
      // Revert to the best network seen during training (based on validation)
      _network = bestNetwork;
      lblStatus.Text = "Training complete";
      logLine();
      logLine("Best network found after " + _epochOfBestNetwork + " epochs with mean validation error: " + bestValidationError);
      logLine();
      logLine("##### TRAINING COMPLETE #####");
      logLine(); logLine();

      // Place chart cursor at epoch of best network, i.e. lowest validation error
      chartTrainingPlot.ChartAreas[0].CursorX.Position = _epochOfBestNetwork;



      /*
      foreach (var inp in inputs)
      {
        computedOutputs.Add(_network.Compute(inp)[0]);
      }
      var errorFile = new StreamWriter("training_errors.txt");
      for (int i = 0; i < errorList.Count; i++)
      {
        if (epochStartIndeces.Contains(i))
        {
          errorFile.WriteLine();
          errorFile.WriteLine("################################################");
          errorFile.WriteLine("################################################");
          errorFile.WriteLine("################################################");
          errorFile.WriteLine("################################################");
          errorFile.WriteLine("################################################");
          errorFile.WriteLine("############# STARTING NEW EPOCH ###############");
          errorFile.WriteLine("################################################");
          errorFile.WriteLine("################################################");
          errorFile.WriteLine("################################################");
          errorFile.WriteLine("################################################");
          errorFile.WriteLine("################################################");
          errorFile.WriteLine();
        }
        else
        {
          errorFile.WriteLine("Iteration " + i + ": Input = " + inputStrings[i] + "   Computed velo = "
            + computedOutputs[i] + "   Correct velo = " + desiredOutputs[i] + "   Error = " + errorList[i]);
        }
        //Debug.WriteLine(computedOutputs[i]);
      }
      errorFile.Close();
      */

      // Return the mean validation error of the final network
      return bestValidationError;
    }

    private void btnTrainNetwork_Click(object sender, EventArgs e)
    {
      clearLog();
      logLine("###################################");
      logLine("##### ANN TRAINING INITIATED ######");
      logLine("###################################");
      logLine();
      btnTrainNetwork.Enabled = false;
      lastRunTrainingType = "normal";

      // Prepare training set, validation set and test set
      lblStatus.Text = "Reading training data.";
      prepareTrainingData();
      logLine();

      // Set input vector type
      InputVectorType inputVectorType;
      switch (cmbInputVector.SelectedIndex)
      {
        case 0: inputVectorType = InputVectorType.OneInputPerFeature;
          break;
        case 1:
        default: inputVectorType = InputVectorType.OneInputPerGroup;
          break;
      }

      // Set normalization of input vector
      bool normalizeInputVector = chbNormalizeInputVector.Checked;

      // Analyze training, validation and test sets and find theoretical max speed (scaling factor to be used for scaling ANN outputs)
      double trainingMaxVelo, validationMaxVelo, testMaxVelo; // 'trainingMaxVelo' is the theoretical max velo (3*avg velo of training set)
      AnalyzeDataSet(_trainingSet, "TRAINING SET", out trainingMaxVelo);
      AnalyzeDataSet(_validationSet, "VALIDATION SET", out validationMaxVelo);
      if (_testSet != null)
      {
        AnalyzeDataSet(_testSet, "TEST SET", out testMaxVelo);
      }
      logLine("Theoretical max velo of training set is used as the ANN output scaling factor: " + trainingMaxVelo.ToString("N2") + " m/s (" + printMinPerKm(trainingMaxVelo) + ")");
      logLine();

      // Histogram analysis of training set velocities
      PopulateVeloDistribution(_trainingSet, false, "Speed distribution of training set");

      // Train the ANN
      trainNetwork(trainingMaxVelo, _trainingSet, _validationSet, _testSet, inputVectorType, normalizeInputVector);

      // Test ANN using training set
      RunSetThroughANN(_trainingSet, "TRAINING SET", chbPrintAllTrainingSamples.Checked);
      rdbShowTrainingSet.Checked = true;

      // Test ANN using validation set
      RunSetThroughANN(_validationSet, "VALIDATION SET", chbPrintAllValidationSamples.Checked);
      rdbShowValidationSet.Checked = true;

      // Test ANN using test set
      if (_testSet != null)
      {
        RunSetThroughANN(_testSet, "TEST SET", chbPrintAllTestSamples.Checked);
        rdbShowTestSet.Checked = true;
      }

      btnTrainNetwork.Enabled = true;
      btnTestNetwork.Enabled = true;
    }

    private void btnTrainKFold_Click(object sender, EventArgs e)
    {

      clearLog();
      logLine("################################################################");
      logLine("##### ANN TRAINING WITH K-FOLD CROSS-VALIDATION INITIATED ######");
      logLine("################################################################");
      logLine();
      btnTrainKFold.Enabled = false;
      lastRunTrainingType = "k-fold";

      // Prepare training set
      prepareTrainingDataForKFold();
      double trainingMaxVelo; // 'trainingMaxVelo' is the theoretical max velo (3*avg velo of training set)
      AnalyzeDataSet(_trainingSet, "TRAINING SET", out trainingMaxVelo);
      logLine("Theoretical max velo of training set is used as the ANN committee output scaling factor: " + trainingMaxVelo.ToString("N2") + " m/s (" + printMinPerKm(trainingMaxVelo) + ")");
      logLine();

      // Set input vector type
      InputVectorType inputVectorType;
      switch (cmbInputVector.SelectedIndex)
      {
        case 0: inputVectorType = InputVectorType.OneInputPerFeature;
          break;
        case 1:
        default: inputVectorType = InputVectorType.OneInputPerGroup;
          break;
      }

      // Set normalization of input vector
      bool normalizeInputVector = chbNormalizeInputVector.Checked;

      // Train ANN committee
      int trainingSetSize = _trainingSet.TrainingData.Count;
      int folds = int.Parse(txtFolds.Text);
      int networksPerFold = int.Parse(txtNetworksPerFold.Text);
      int foldSize = trainingSetSize / folds;
      bool takeBestEachFold = chbTakeBestOfEachFold.Checked;
      _networkCommittee = new VeloActivationNetworkCommittee(trainingMaxVelo, inputVectorType, normalizeInputVector);
      logLine("Folds: " + folds);
      logLine("ANNs per fold: " + networksPerFold);
      logLine("Fold size: " + foldSize); logLine();
      logLine("##### TRAINING ANNs #####"); logLine();
      for (int i = 0; i < folds; i++)
      {
        TrainingSet trainingSet = ObjectCopier.Clone<TrainingSet>(_trainingSet); // Must use a copy, so that the original set of training data is intact for the next fold.
        List<TrainingInstance> validationData;
        if (i == folds - 1)
        { // Take the rest
          validationData = trainingSet.TrainingData.GetRange(i * foldSize, trainingSet.TrainingData.Count - i * foldSize);
          trainingSet.TrainingData.RemoveRange(i * foldSize, trainingSet.TrainingData.Count - i * foldSize);
        }
        else
        {
          validationData = trainingSet.TrainingData.GetRange(i * foldSize, foldSize);
          trainingSet.TrainingData.RemoveRange(i * foldSize, foldSize);
        }
        _validationSet = new TrainingSet(validationData);

        logLine("###################################################");
        logLine("############## FOLD " + (i+1) + " #################");
        logLine("###################################################");
        logLine();
        logLine("Size of training set: " + trainingSet.TrainingData.Count);
        logLine("Size of validation set: " + _validationSet.TrainingData.Count);
        logLine();

        List<VeloActivationNetwork> networksOfThisFold = new List<VeloActivationNetwork>();
        List<double> meanValidationErrors = new List<double>(); // Contains the mean validation error for each ANN in 'networksOfThisFold'
        for (int j = 0; j < networksPerFold; j++)
        {
          logLine("TRAINING ANN " + (j + 1) + " OF FOLD " + (i + 1)); logLine();

          // Train ANN and return the mean error over the current validation set.
          double meanValidationError = trainNetwork(trainingMaxVelo, trainingSet, _validationSet, _testSet, inputVectorType, normalizeInputVector);
          networksOfThisFold.Add(ObjectCopier.Clone<VeloActivationNetwork>(_network));  // Must store a deep copy of '_network', since '_network' will change in the next iteration of the loop
          meanValidationErrors.Add(meanValidationError);

          // Test ANN using this fold's training set
          RunSetThroughANN(trainingSet, "THIS FOLD'S TRAINING SET", chbPrintAllTrainingSamples.Checked);
          rdbShowTrainingSet.Checked = true;

          // Test ANN using test set
          if (double.Parse(txtTestFraction.Text) > 0)
          {
            RunSetThroughANN(_testSet, "TEST SET", chbPrintAllTrainingSamples.Checked);
            rdbShowTrainingSet.Checked = true;
          }
        }

        if (chbTakeBestOfEachFold.Checked)
        {
          // Add the best ANN of this fold to the committee
          VeloActivationNetwork bestNetworkOfThisFold = networksOfThisFold[0];
          double bestValidationError = meanValidationErrors[0];
          for (int j = 1; j < networksOfThisFold.Count; j++)
          {
            if (meanValidationErrors[j] < bestValidationError)
            {
              bestValidationError = meanValidationErrors[j];
              bestNetworkOfThisFold = networksOfThisFold[j];
            }
          }
          int indexBestNetworkOfThisFold = networksOfThisFold.IndexOf(bestNetworkOfThisFold);
          _networkCommittee.Networks.Add(bestNetworkOfThisFold);
          logLine("ANN " + (indexBestNetworkOfThisFold + 1) + " was the best ANN of this fold and was added to the committee"); logLine();
        }
        else
        {
          // Add all networks of this fold to the committee
          _networkCommittee.Networks.AddRange(networksOfThisFold);
          logLine("All ANNs of this fold were added to the committee."); logLine();
        }
      }

      // Test ANN committee using training set
      RunSetThroughANNCommittee(_trainingSet, "TRAINING SET", chbPrintAllTrainingSamples.Checked);
      rdbShowTrainingSet.Checked = true;

      // Test ANN committee using test set
      if (double.Parse(txtTestFraction.Text) > 0)
      {
        RunSetThroughANNCommittee(_testSet, "TEST SET", chbPrintAllTrainingSamples.Checked);
        rdbShowTestSet.Checked = true;
      }

      logLine("##### K-FOLD CROSS-VALIDATION TRAINING COMPLETED #####"); logLine(); logLine();
      btnTrainKFold.Enabled = true;
    }

    private double computeMeanErrorAfterEpoch(TrainingSet set, string seriesName, int epoch)
    {
      double sumError = 0;
      foreach (TrainingInstance t in set.TrainingData)
      {
        double[] inputVector = TrainingSet.CreateInputVector(t.ClassifiedMapFeatures, t.Climb, t.Slope, _network.InputVectorType, _network.NormalizeInputVector);
        double desiredOutput = t.Velocity / _network.OutputScaleFactor;
        double computedOutput = _network.Compute(inputVector)[0];
        //double error = Math.Pow(computedOutput - desiredOutput, 2) / 2;
        double error = Math.Abs(computedOutput - desiredOutput) * _network.OutputScaleFactor;
        sumError += error;
      }
      double meanError = sumError / set.TrainingData.Count;
      chartTrainingPlot.Series[seriesName].Points.AddXY(epoch, meanError);

      return meanError;
    }

    private void btnTestNetwork_Click(object sender, EventArgs e)
    {
      if (_testSet == null)
      {
        lblStatus.Text = "Cannot run test set. It is empty.";
        return;
      }
      RunSetThroughANN(_testSet, "TEST SET", chbPrintAllTestSamples.Checked);
      rdbShowTestSet.Checked = true;
      lblStatus.Text = "Test completed.";
    }

    private void txtNetworkFile_TextChanged(object sender, EventArgs e)
    {
      btnOpenNetwork.Enabled = txtNetworkFile.Text != "";
    }

    private void btnSaveNetwork_Click(object sender, EventArgs e)
    {
      dlgSave.Filter = "*.bin|*.bin";
      dlgSave.FileName = buildFileName("ANN") + ".bin";
      if (dlgSave.ShowDialog(this) == DialogResult.OK)
      {
        _network.Save(dlgSave.FileName);
        lblStatus.Text = "Neural network saved.";
      }
    }

    private void DlgSaveFileOk(object sender, System.ComponentModel.CancelEventArgs e)
    {

    }

    private void btnBrowseNetworkFile_Click(object sender, EventArgs e)
    {
      dlgOpen.FileName = txtNetworkFile.Text;
      dlgOpen.Filter = "*.bin|*.bin";
      if (dlgOpen.ShowDialog(this) == DialogResult.OK)
      {
        txtNetworkFile.Text = dlgOpen.FileName;
      }
    }

    private void btnOpenNetwork_Click(object sender, EventArgs e)
    {
      try
      {
        _network = (VeloActivationNetwork)Network.Load(txtNetworkFile.Text);
        lblStatus.Text = "Network loaded.";
        btnTestNetwork.Enabled = true;
      }
      catch (Exception)
      {
        lblStatus.Text = "Could not load network. The file could not be found.";
      }
      
    }

    private void chbSeparateValidationSet_CheckedChanged(object sender, EventArgs e)
    {
      if (chbSeparateValidationSet.Checked)
      {
        txtValidationSet.Enabled = true;
        btnBrowseValidationSet.Enabled = true;
        txtValidationFraction.Text = "0";
        lblValidationFraction.Enabled = false;
        txtValidationFraction.Enabled = false;
      }
      else
      {
        txtValidationSet.Enabled = false;
        btnBrowseValidationSet.Enabled = false;
        txtValidationFraction.Text = "0.2";
        lblValidationFraction.Enabled = true;
        txtValidationFraction.Enabled = true;
      }
    }

    private void chbSeparateTestSet_CheckedChanged(object sender, EventArgs e)
    {
      if (chbSeparateTestSet.Checked)
      {
        txtTestSet.Enabled = true;
        btnBrowseTestSet.Enabled = true;
        txtTestFraction.Text = "0";
        lblTestFraction.Enabled = false;
        txtTestFraction.Enabled = false;
      }
      else
      {
        txtTestSet.Enabled = false;
        btnBrowseTestSet.Enabled = false;
        txtTestFraction.Text = "0.1";
        lblTestFraction.Enabled = true;
        txtTestFraction.Enabled = true;
      }
    }

    private void btnBrowseValidationSet_Click(object sender, EventArgs e)
    {
      dlgOpen.FileName = txtValidationSet.Text;
      dlgOpen.Filter = "*.xml|*.xml";
      if (dlgOpen.ShowDialog(this) == DialogResult.OK)
      {
        txtValidationSet.Text = dlgOpen.FileName;
      }
    }

    private void btnBrowseTestSet_Click(object sender, EventArgs e)
    {
      dlgOpen.FileName = txtTestSet.Text;
      dlgOpen.Filter = "*.xml|*.xml";
      if (dlgOpen.ShowDialog(this) == DialogResult.OK)
      {
        txtTestSet.Text = dlgOpen.FileName;
      }
    }

    private void btnSaveErrorChart_Click(object sender, EventArgs e)
    {
      dlgSave.Filter = "*.png|*.png";
      string description;
      if (cmbPlot.SelectedItem.ToString() == "Error")
      {
        description = "Error plot";
      }
      else
      {
        description = "Velo plot";
      }
      dlgSave.FileName = buildFileName(description) + ".png";
      if (dlgSave.ShowDialog(this) == DialogResult.OK)
      {
        chartTrainingPlot.SaveImage(dlgSave.FileName, ChartImageFormat.Png);
        lblStatus.Text = "Chart image saved.";
      }
    }

    private string buildFileName(string description)
    {
      string fileName = description;
      if (fileName != "")
      {
        fileName += " -- ";
      }
      if (lastRunTrainingType == "normal")
      {
        fileName += "normal training -- ";
      }
      else
      {
        fileName += "k-fold training (" + txtFolds.Text + ", " + txtNetworksPerFold.Text + ") -- ";
      }

      fileName += Path.GetFileNameWithoutExtension(txtTrainingSet.Text);
      if (lastRunTrainingType == "normal")
      {
        fileName += " (" + txtValidationFraction.Text + ", " + txtTestFraction.Text + ")";
      }
      else
      {
        fileName += " (" + txtTestFraction.Text + ")";
      }
      fileName += " -- " + txtRemoveTrailingPathPoints.Text + ", " + chbRemoveIfMapFeaturesChanged.Checked.ToString();
      fileName += " -- ";
      if ((LearningAlgorithm)cmbAlgorithm.SelectedItem == LearningAlgorithm.BackProp)
      {
        fileName += "BP, ";
      }
      else
      {
        fileName += "RP, ";
      }
      if (cmbInputVector.SelectedIndex == 0)
      {
        fileName += "mf, ";
      }
      else
      {
        fileName += "mfg, ";
      }
      fileName += txtLearningRate.Text + ", ";
      if ((LearningAlgorithm)cmbAlgorithm.SelectedItem == LearningAlgorithm.BackProp)
      {
        fileName += txtMomentum.Text + ", ";
      }
      fileName += txtAlpha.Text + ", ";
      fileName += "(" + txtNeuronsFirstLayer.Text + ", " + txtNeuronsSecondLayer.Text + ", " + txtNeuronsThirdLayer.Text + "), ";
      fileName += txtEpochs.Text;
      return fileName;
    }

    private void btnAnalyzeTrainingSet_Click(object sender, EventArgs e)
    {
      prepareTrainingData();
      double theoreticalMaxVelo;
      AnalyzeDataSet(_trainingSet, "TRAINING SET", out theoreticalMaxVelo);
      PopulateVeloDistribution(_trainingSet, false, "Speed distribution of training set");
      if (_network != null)
      {
        PopulateVeloDistribution(_trainingSet, true, "Speed distribution of training set as computed by ANN");
      }
      rdbShowTrainingSet.Checked = true;
    }

    private void btnAnalyzeValidationSet_Click(object sender, EventArgs e)
    {
      if (_validationSet == null || chbSeparateValidationSet.Checked)
      {
        prepareTrainingData();
      }
      double theoreticalMaxVelo;
      AnalyzeDataSet(_validationSet, "VALIDATION SET", out theoreticalMaxVelo);
      PopulateVeloDistribution(_validationSet, false, "Speed distribution of validation set");
      if (_network != null)
      {
        PopulateVeloDistribution(_validationSet, true, "Speed distribution of validation set as computed by ANN");
      }
      rdbShowValidationSet.Checked = true;
    }

    private void btnAnalyzeTestSet_Click(object sender, EventArgs e)
    {
      if (_testSet == null || chbSeparateTestSet.Checked)
      {
        prepareTrainingData();
      }
      if (_testSet == null)
      {
        lblStatus.Text = "Cannot analyze test set. It is empty.";
        return;
      }

      double theoreticalMaxVelo;
      AnalyzeDataSet(_testSet, "TEST SET", out theoreticalMaxVelo);
      PopulateVeloDistribution(_testSet, false, "Speed distribution of test set");
      if (_network != null)
      {
        PopulateVeloDistribution(_testSet, true, "Speed distribution of test set as computed by ANN");
      }
      rdbShowTestSet.Checked = true;
    }

    private void AnalyzeDataSet(TrainingSet set, string description, out double theoreticalMaxVelo)
    {
      int nofInstances, nofFlatPathPoints;
      double maxVelo, minVelo, avgVelo, stdDevVelo, maxClimb, minClimb, maxSlope, minSlope, flatPathSpeed;
      set.Analyze(out nofInstances, out maxVelo, out minVelo, out avgVelo, out stdDevVelo, out maxClimb, out minClimb, out maxSlope, out minSlope, out flatPathSpeed, out nofFlatPathPoints);
      theoreticalMaxVelo = 3*avgVelo;
      logLine("---- ANALYSIS OF " + description + " ----");
      logLine("Number of training instances: " + nofInstances);
      logLine("Max velo: " + maxVelo.ToString("N4") + " m/s" + " | " + printMinPerKm(maxVelo));
      logLine("Min velo: " + minVelo.ToString("N4") + " m/s" + " | " + printMinPerKm(minVelo));
      logLine("Avg velo: " + avgVelo.ToString("N4") + " m/s" + " | " + printMinPerKm(avgVelo));
      logLine("Velo standard deviation: " + stdDevVelo.ToString("N4") + " m/s");
      logLine("Theoretical max velo (3*avg): " + (theoreticalMaxVelo).ToString("N2") + " m/s" + " (" + printMinPerKm(theoreticalMaxVelo) + ")");
      logLine("Max climb: " + maxClimb.ToString("N4") + "  |  " + (Math.Atan(maxClimb) * 180 / Math.PI).ToString("N2") + "°");
      logLine("Min climb: " + minClimb.ToString("N4") + "  |  " + (Math.Atan(minClimb) * 180 / Math.PI).ToString("N2") + "°");
      logLine("Max slope: " + maxSlope.ToString("N4") + "  |  " + (Math.Atan(maxSlope) * 180 / Math.PI).ToString("N2") + "°");
      logLine("Min slope: " + minSlope.ToString("N4") + "  |  " + (Math.Atan(minSlope) * 180 / Math.PI).ToString("N2") + "°");
      logLine("Avg. flat path speed: " + flatPathSpeed.ToString("N4") + " m/s" + " | " + printMinPerKm(flatPathSpeed) + " (based on " + nofFlatPathPoints + " instances of flat path");
      logLine();

      if (_network != null)
      {
        
      }
    }

    private void PopulateVeloDistribution(TrainingSet set, bool compute, string description)
    {

      if (compute)
      {
        chartComputedVeloDistribution.Titles[0].Text = description;
        chartComputedVeloDistribution.Series["RawData"].Points.Clear();
        chartComputedVeloDistribution.Series["DataDistribution"].Points.Clear();
        chartComputedVeloDistribution.Series["Histogram"].Points.Clear();

        // Compute ANN outputs and populate chart
        double[][] input = set.GetInput(_network.InputVectorType, _network.NormalizeInputVector);
        for (int i = 0; i < input.Length; i++)
        {
          double[] inputVector = input[i];
          double velocity = _network.Compute(inputVector)[0] * _network.OutputScaleFactor;
          chartComputedVeloDistribution.Series["RawData"].Points.AddY(velocity);
        }
        foreach (DataPoint dataPoint in chartComputedVeloDistribution.Series["RawData"].Points)
        {
          chartComputedVeloDistribution.Series["DataDistribution"].Points.AddXY(dataPoint.YValues[0], 1);
        }

        UpdateVeloDistributionChart(chartComputedVeloDistribution, double.Parse(txtBinSizeComputedVeloDistribution.Text));

        // Make secondary X axis show number of standard deviations from mean
        AddStdDevAxis(chartComputedVeloDistribution);
      }
      else
      {
        chartVeloDistribution.Titles[0].Text = description;
        chartVeloDistribution.Series["RawData"].Points.Clear();
        chartVeloDistribution.Series["DataDistribution"].Points.Clear();
        chartVeloDistribution.Series["Histogram"].Points.Clear();

        foreach (TrainingInstance t in set.TrainingData)
        {
          chartVeloDistribution.Series["RawData"].Points.AddY(t.Velocity);
        }
        foreach (DataPoint dataPoint in chartVeloDistribution.Series["RawData"].Points)
        {
          chartVeloDistribution.Series["DataDistribution"].Points.AddXY(dataPoint.YValues[0], 1);
        }
        UpdateVeloDistributionChart(chartVeloDistribution, double.Parse(txtBinSizeVeloDistribution.Text));

        // Make secondary X axis show number of standard deviations from mean
        AddStdDevAxis(chartVeloDistribution);
      }
    }

    private void AddStdDevAxis(Chart chart)
    {
      
      // Get velo of each data point
      var velocities = new List<double>();
      foreach (DataPoint p in chart.Series["RawData"].Points)
      {
        velocities.Add(p.YValues[0]);
      }

      /*int nofInstances;
      double minVelo, maxVelo, avgVelo, stdDevVelo;
      set.Analyze(out nofInstances, out maxVelo, out minVelo, out avgVelo, out stdDevVelo);*/

      double avgVelo = Utilities.Average(velocities);
      double stdDevVelo = Utilities.StandardDeviation(velocities);

      double primaryXAxisMax = chart.ChartAreas["HistogramArea"].AxisX.Maximum;
      double primaryXAxisMin = chart.ChartAreas["HistogramArea"].AxisX.Minimum;

      chart.ChartAreas["HistogramArea"].AxisX2.Enabled = AxisEnabled.True;
      chart.ChartAreas["HistogramArea"].AxisX2.Maximum =
        (primaryXAxisMax - avgVelo) / stdDevVelo;
      chart.ChartAreas["HistogramArea"].AxisX2.Minimum =
        (primaryXAxisMin - avgVelo) / stdDevVelo;

      chart.ChartAreas["HistogramArea"].AxisX.Interval = 0.5;
      double xMin = chart.ChartAreas["HistogramArea"].AxisX.Minimum;
      chart.ChartAreas["HistogramArea"].AxisX.IntervalOffset = -10 - (xMin - (int)xMin);
      chart.ChartAreas["HistogramArea"].AxisX2.Interval = 0.5;
      chart.ChartAreas["HistogramArea"].AxisX2.MajorGrid.LineColor = Color.Red;
      chart.ChartAreas["HistogramArea"].AxisX2.MajorGrid.LineDashStyle = ChartDashStyle.Dot;
      chart.ChartAreas["HistogramArea"].AxisX2.MajorGrid.Interval = 1;
      double x2Min = chart.ChartAreas["HistogramArea"].AxisX2.Minimum;
      chart.ChartAreas["HistogramArea"].AxisX2.IntervalOffset = -(x2Min - (int)x2Min);
      //chart.ChartAreas["HistogramArea"].AxisX2.MajorGrid.IntervalOffset = -(min - (int)min);
      //chart.ChartAreas["HistogramArea"].AxisX2.MajorTickMark.Interval = 1;
      //chart.ChartAreas["HistogramArea"].AxisX2.MajorTickMark.IntervalOffset = -(min - (int)min);
      //chart.ChartAreas["HistogramArea"].AxisX2.MinorTickMark.Enabled = true;
      //chart.ChartAreas["HistogramArea"].AxisX2.MinorTickMark.Interval = 0.5;
      //chart.ChartAreas["HistogramArea"].AxisX2.MinorTickMark.IntervalOffset = -(min - (int)min);
      chart.ChartAreas["HistogramArea"].AxisX2.Title = "Standard deviations from mean";
      chart.ChartAreas["HistogramArea"].AxisX2.LabelAutoFitStyle = LabelAutoFitStyles.DecreaseFont;
      chart.ChartAreas["HistogramArea"].AxisX.LabelAutoFitStyle = LabelAutoFitStyles.DecreaseFont;
    }

    private double MpsToMinPerKm(double mps)
    {
      if (mps < 0)
      {
        throw new Exception("Speed cannot be negative!");
      }
      if (mps > 0)
      {
        return 1000/(60*mps);
      }
      else
      {
        return double.PositiveInfinity;
      }
    }

    private string printMinPerKm(double mps)
    {
      double minPerKm = MpsToMinPerKm(mps);
      if (double.IsPositiveInfinity(minPerKm))
      {
        return "Infinite min/km";
      }
      var minutes = (int) minPerKm;
      double decimalPart = minPerKm - minutes;
      var seconds = (int) (decimalPart*60);
      return minutes + ":" + seconds.ToString("00") + " min/km";

    }

    private void btnClearGpsFiles_Click(object sender, EventArgs e)
    {
      trainingSetList.Items.Clear();
    }

    private void logLine(string text = "")
    {
      txtTrainingInfo.Text += "\r\n" + text;
      txtTrainingInfo.Select(txtTrainingInfo.Text.Length - 1, 0);
      txtTrainingInfo.ScrollToCaret();
    }

    private void log(string text)
    {
      txtTrainingInfo.Text += text;
    }

    private void clearLog()
    {
      txtTrainingInfo.Clear();
    }

    private void btnClearLog_Click(object sender, EventArgs e)
    {
      clearLog();
    }

    private void btnSaveLog_Click(object sender, EventArgs e)
    {
      dlgSave.Filter = "*.txt|*.txt";
      dlgSave.FileName = buildFileName("Training log");
      if (dlgSave.ShowDialog(this) == DialogResult.OK)
      {
        var logWriter = new StreamWriter(dlgSave.FileName);
        logWriter.Write(txtTrainingInfo.Text);
        logWriter.Close();
      }
    }

    private void btnCreateKeyFeatureTest_Click(object sender, EventArgs e)
    {
      var keyFeatureForm = new KeyFeatureTestCreationForm();
      keyFeatureForm.Show();
    }

    private void UpdateVeloDistributionChart(Chart chart, double segmentIntervalWidth)
    {
      var histogramHelper = new HistogramChartHelper();
      histogramHelper.SegmentIntervalWidth = segmentIntervalWidth;
      histogramHelper.CreateHistogram(chart, "RawData", "Histogram");

      if (chart == chartComputedVeloDistribution)
      {
        // Scale the two distribution charts equivalently
        chartComputedVeloDistribution.ChartAreas["HistogramArea"].AxisX.Minimum = chartVeloDistribution.ChartAreas["HistogramArea"].AxisX.Minimum;
        chartComputedVeloDistribution.ChartAreas["HistogramArea"].AxisX.Maximum = chartVeloDistribution.ChartAreas["HistogramArea"].AxisX.Maximum;
        double veloHistogramMaxCount = 0;
        double computedVeloHistogramMaxCount = 0;
        for (int i = 0; i < chartVeloDistribution.Series["Histogram"].Points.Count; i++)
        {
          double count = chartVeloDistribution.Series["Histogram"].Points[i].YValues[0];
          if (count > veloHistogramMaxCount)
          {
            veloHistogramMaxCount = count;
          }
        }
        for (int i = 0; i < chartComputedVeloDistribution.Series["Histogram"].Points.Count; i++)
        {
          double count = chartComputedVeloDistribution.Series["Histogram"].Points[i].YValues[0];
          if (count > computedVeloHistogramMaxCount)
          {
            computedVeloHistogramMaxCount = count;
          }
        }
        double maxY = Math.Max(veloHistogramMaxCount, computedVeloHistogramMaxCount)*1.1;
        chartComputedVeloDistribution.ChartAreas["HistogramArea"].AxisY.Maximum = maxY;
        chartVeloDistribution.ChartAreas["HistogramArea"].AxisY.Maximum = maxY;
      }

      // Scale raw data distribution chart X axis the same as the histogram
      chart.ChartAreas["Default"].AxisX.Minimum = chart.ChartAreas["HistogramArea"].AxisX.Minimum;
      chart.ChartAreas["Default"].AxisX.Maximum = chart.ChartAreas["HistogramArea"].AxisX.Maximum;
      chart.ChartAreas["Default"].AxisX.Interval = chart.ChartAreas["HistogramArea"].AxisX.Interval;
    }

    private void btnUpdateCharts_Click(object sender, EventArgs e)
    {
      UpdateVeloDistributionChart(chartVeloDistribution, double.Parse(txtBinSizeVeloDistribution.Text));
      UpdateVeloDistributionChart(chartComputedVeloDistribution, double.Parse(txtBinSizeComputedVeloDistribution.Text));

      // Make secondary X axis show number of standard deviations from mean
      AddStdDevAxis(chartVeloDistribution);
      AddStdDevAxis(chartComputedVeloDistribution);
    }

    private void prepareTrainingData()
    {
      logLine("##### PREPARING TRAINING, VALIDATION AND TEST SETS #####");

      // Prepare training set
      _trainingSet = TrainingSetSerializer.DeserializeFromXml(txtTrainingSet.Text);
      logLine("Using training set: \"" + txtTrainingSet.Text + "\"");

      // Filters
      int trailingPathPoints = int.Parse(txtRemoveTrailingPathPoints.Text);
      _trainingSet.Filter(trailingPathPoints, chbRemoveIfMapFeaturesChanged.Checked);

      int trainingSetSize = _trainingSet.TrainingData.Count;

      // Prepare validation set
      if (chbSeparateValidationSet.Checked)
      {
        logLine("Using separate validation set: \"" + txtValidationSet.Text + "\"");
        _validationSet = TrainingSetSerializer.DeserializeFromXml(txtValidationSet.Text);
        _validationSet.Filter(trailingPathPoints, chbRemoveIfMapFeaturesChanged.Checked);
      }
      else
      {
        double validationFraction;
        bool tryParse = double.TryParse(txtValidationFraction.Text, out validationFraction);
        if (!tryParse)
        {
          lblStatus.Text = "Error: Validation fraction must be a real number between 0 and 1.";
          btnTrainNetwork.Enabled = true;
          return;
        }
        logLine("Using validation set: " + validationFraction * 100 + " % of training set");
        int validationSetSize = (int)Math.Round(trainingSetSize * validationFraction);
        Utilities.Shuffle(_trainingSet.TrainingData);
        List<TrainingInstance> validationData = _trainingSet.TrainingData.GetRange(0, validationSetSize);
        _trainingSet.TrainingData.RemoveRange(0, validationSetSize);
        _validationSet = new TrainingSet(validationData);
      }

      // Prepare test set
      if (chbSeparateTestSet.Checked)
      {
        logLine("Using separate test set: \"" + txtTestSet.Text + "\"");
        _testSet = TrainingSetSerializer.DeserializeFromXml(txtTestSet.Text);
        _testSet.Filter(trailingPathPoints, chbRemoveIfMapFeaturesChanged.Checked);
      }
      else
      {
        double testFraction;
        bool tryParse = double.TryParse(txtTestFraction.Text, out testFraction);
        if (!tryParse)
        {
          lblStatus.Text = "Error: Test fraction must be a real number between 0 and 1.";
          btnTrainNetwork.Enabled = true;
          return;
        }
        int testSetSize = (int)Math.Round(trainingSetSize * testFraction);
        if (testSetSize > 0)
        {
          logLine("Using test set: " + testFraction * 100 + " % of training set");
          Utilities.Shuffle(_trainingSet.TrainingData);
          List<TrainingInstance> testData = _trainingSet.TrainingData.GetRange(0, testSetSize);
          _trainingSet.TrainingData.RemoveRange(0, testSetSize);
          _testSet = new TrainingSet(testData);
          rdbShowTestSet.Enabled = true;
        }
        else
        {
          _testSet = null;
          rdbShowTestSet.Enabled = false;
          logLine("Using no test set");
        }
      }
      logLine();
    }

    private void prepareTrainingDataForKFold()
    {
      // Prepare training set
      _trainingSet = TrainingSetSerializer.DeserializeFromXml(txtTrainingSet.Text);
      logLine("Using training set: \"" + txtTrainingSet.Text + "\"");

      // Filters
      int trailingPathPoints = int.Parse(txtRemoveTrailingPathPoints.Text);
      _trainingSet.Filter(trailingPathPoints, chbRemoveIfMapFeaturesChanged.Checked);

      int trainingSetSize = _trainingSet.TrainingData.Count;

      // Prepare test set
      if (chbSeparateTestSet.Checked)
      {
        logLine("Using separate test set: \"" + txtTestSet.Text + "\"");
        _testSet = TrainingSetSerializer.DeserializeFromXml(txtTestSet.Text);
        _testSet.Filter(trailingPathPoints, chbRemoveIfMapFeaturesChanged.Checked);
      }
      else
      {
        double testFraction;
        bool tryParse = double.TryParse(txtTestFraction.Text, out testFraction);
        if (!tryParse)
        {
          lblStatus.Text = "Error: Test fraction must be a real number between 0 and 1.";
          btnTrainNetwork.Enabled = true;
          return;
        }
        int testSetSize = (int)Math.Round(trainingSetSize * testFraction);
        if (testSetSize > 0)
        {
          logLine("Using test set: " + testFraction * 100 + " % of training set");
          Utilities.Shuffle(_trainingSet.TrainingData);
          List<TrainingInstance> testData = _trainingSet.TrainingData.GetRange(0, testSetSize);
          _trainingSet.TrainingData.RemoveRange(0, testSetSize);
          _testSet = new TrainingSet(testData);
          rdbShowTestSet.Enabled = true;
        }
        else
        {
          _testSet = null;
          rdbShowTestSet.Enabled = false;
          logLine("Using no test set");
        }
      }
      logLine();
    }

    private void rdbShowTrainingSet_CheckedChanged(object sender, EventArgs e)
    {
      if (rdbShowTrainingSet.Checked)
      {
        PopulateVeloDistribution(_trainingSet, false, "Speed distribution of training set");
        if (_network != null)
        {
          if (lastRunTrainingType == "normal")
          {
            PopulateVeloDistribution(_trainingSet, true, "Speed distribution of training set as computed by ANN");
          }
          else
          {
            PopulateVeloDistribution(_trainingSet, true, "Speed distribution of training set as computed by ANN Committee");
          }
        }
      }
    }

    private void rdbShowValidationSet_CheckedChanged(object sender, EventArgs e)
    {
      if (rdbShowValidationSet.Checked)
      {
        PopulateVeloDistribution(_validationSet, false, "Speed distribution of validation set");
        if (_network != null)
        {
          PopulateVeloDistribution(_validationSet, true, "Speed distribution of validation set as computed by ANN");
        }
      }
    }

    private void rdbShowTestSet_CheckedChanged(object sender, EventArgs e)
    {
      if (rdbShowTestSet.Checked)
      {
        if (_testSet == null)
        {
          rdbShowTestSet.Checked = false;
          rdbShowValidationSet.Checked = true;
          lblStatus.Text = "Cannot show test set. It is empty.";
          return;
        }
        PopulateVeloDistribution(_testSet, false, "Speed distribution of test set");
        if (_network != null)
        {
          if (lastRunTrainingType == "normal")
          {
            PopulateVeloDistribution(_testSet, true, "Speed distribution of test set as computed by ANN");
          }
          else
          {
            PopulateVeloDistribution(_testSet, true, "Speed distribution of test set as computed by ANN Committee");
          }
        }
      }
    }

    private void btnSaveVeloDistributionChart_Click(object sender, EventArgs e)
    {
      dlgSave.Filter = "*.png|*.png";
      if (rdbShowTrainingSet.Checked)
      {
        dlgSave.FileName = buildFileName("target histogram trainingset (" + txtBinSizeVeloDistribution.Text + ")");
      }
      else if (rdbShowValidationSet.Checked)
      {
        dlgSave.FileName = buildFileName("target histogram validationset (" + txtBinSizeVeloDistribution.Text + ")");
      }
      else // Test set is showing
      {
        dlgSave.FileName = buildFileName("target histogram testset (" + txtBinSizeVeloDistribution.Text + ")");
      }
      if (dlgSave.ShowDialog(this) == DialogResult.OK)
      {
        chartVeloDistribution.SaveImage(dlgSave.FileName, ChartImageFormat.Png);
        lblStatus.Text = "Chart image saved.";
      }
    }

    private void btnSaveComputedVeloDistributionChart_Click(object sender, EventArgs e)
    {
      dlgSave.Filter = "*.png|*.png";
      if (rdbShowTrainingSet.Checked)
      {
        dlgSave.FileName = buildFileName("output histogram trainingset (" + txtBinSizeComputedVeloDistribution.Text + ")");
      }
      else if (rdbShowValidationSet.Checked)
      {
        dlgSave.FileName = buildFileName("output histogram validationset (" + txtBinSizeComputedVeloDistribution.Text + ")");
      }
      else // Test set is showing
      {
        dlgSave.FileName = buildFileName("output histogram testset (" + txtBinSizeComputedVeloDistribution.Text + ")");
      }
      if (dlgSave.ShowDialog(this) == DialogResult.OK)
      {
        chartComputedVeloDistribution.SaveImage(dlgSave.FileName, ChartImageFormat.Png);
        lblStatus.Text = "Chart image saved.";
      }
    }

    private void btnBrowseKeyFeatureTestFile_Click(object sender, EventArgs e)
    {
      dlgOpen.FileName = txtKeyFeatureTest.Text;
      dlgOpen.Filter = "*.xml|*.xml";
      if (dlgOpen.ShowDialog(this) == DialogResult.OK)
      {
        txtKeyFeatureTest.Text = dlgOpen.FileName;
      }
    }

    private void btnEditKeyFeatureTest_Click(object sender, EventArgs e)
    {
      var keyFeatureForm = new KeyFeatureTestCreationForm(txtKeyFeatureTest.Text);
      keyFeatureForm.Show();
    }

    private void btnRunKeyFeatureTest_Click(object sender, EventArgs e)
    {
      logLine("###################################################");
      logLine("##### TESTING KEY FEATURE COMBINATIONS ON ANN #####");
      logLine("###################################################");
      logLine();

      KeyFeatureTestSet keyFeatureTestSet = TrainingSetSerializer.DeserializeKeyFeatureTestSetFromXml(txtKeyFeatureTest.Text);

      logLine("Input vector: " + _network.InputVectorType);
      logLine();

      double[][] input = keyFeatureTestSet.GetInput(_network.InputVectorType, _network.NormalizeInputVector);

      logLine("TEST RESULTS:"); logLine();
      logLine();
      logLine("(The features of each test are: Forest, Marsh, Undergrowth, Path, Climb, Slope)"); // Left out Ground (passable rock faces, stony ground etc). Ignoring it for now.
      logLine();

      var scaledVelos = new List<double>();
      for (int i = 0; i < input.Length; i++)
      {
        double[] inputVector = input[i];
        double computedVelo = _network.Compute(inputVector)[0];
        double scaledVelo = computedVelo * _network.OutputScaleFactor;
        scaledVelos.Add(scaledVelo);
        //keyFeatureTestSet.Tests[i].Velocity = computedVelo;
        logLine("Test " + i + ":");
        logLine("Features: " + keyFeatureTestSet.Tests[i]);
        logLine("Computed speed: " + computedVelo.ToString("N4") + "  |  " + scaledVelo.ToString("N4") + " m/s  |  " + printMinPerKm(scaledVelo));
        logLine();
      }

      logLine("ALL VELOS FOR EASY EXPORT TO EXCEL:");
      foreach (double velo in scaledVelos)
      {
        logLine(velo.ToString("N4"));
      }
      logLine();

      logLine("##### KEY FEATURE TEST COMPLETED #####");
      logLine(); logLine();
    }

    private void btnRunKeyFeatureTestOnCommittee_Click(object sender, EventArgs e)
    {
      logLine("#############################################################");
      logLine("##### TESTING KEY FEATURE COMBINATIONS ON ANN COMMITTEE #####");
      logLine("#############################################################");
      logLine();

      KeyFeatureTestSet keyFeatureTestSet = TrainingSetSerializer.DeserializeKeyFeatureTestSetFromXml(txtKeyFeatureTest.Text);

      logLine("Input vector: " + _networkCommittee.InputVectorType);
      logLine();

      double[][] input = keyFeatureTestSet.GetInput(_networkCommittee.InputVectorType, _networkCommittee.NormalizeInputVector);

      logLine("TEST RESULTS:"); logLine();
      logLine();
      logLine("(The features of each test are: Forest, Marsh, Undergrowth, Path, Climb, Slope)"); // Left out Ground (passable rock faces, stony ground etc). Ignoring it for now.
      logLine();

      var scaledVelos = new List<double>();
      var scaledVeloDeviations = new List<double>();
      for (int i = 0; i < input.Length; i++)
      {
        double[] inputVector = input[i];
        double committeeVeloDeviation;
        double computedVelo = _networkCommittee.Compute(inputVector, out committeeVeloDeviation);
        double scaledVelo = computedVelo * _networkCommittee.OutputScaleFactor;
        double scaledVeloDeviation = committeeVeloDeviation * _networkCommittee.OutputScaleFactor;
        scaledVelos.Add(scaledVelo);
        scaledVeloDeviations.Add(scaledVeloDeviation);
        //keyFeatureTestSet.Tests[i].Velocity = computedVelo;
        logLine("Test " + i + ":");
        logLine("Features: " + keyFeatureTestSet.Tests[i]);
        logLine("Computed speed: " + computedVelo.ToString("N4") + "  |  " + scaledVelo.ToString("N4") + " m/s  |  " + printMinPerKm(scaledVelo));
        logLine("Committee speed deviation: " + committeeVeloDeviation.ToString("N4") + "  |  " + scaledVeloDeviation.ToString("N4") + " m/s");
        logLine();
      }
      logLine("ALL VELOS FOR EASY EXPORT TO EXCEL:");
      foreach (double velo in scaledVelos)
      {
        logLine(velo.ToString("N4"));
      }
      logLine();
      logLine("ALL VELO DEVIATIONS FOR EASY EXPORT TO EXCEL:");
      foreach (double dev in scaledVeloDeviations)
      {
        logLine(dev.ToString("N4"));
      }

      logLine("##### KEY FEATURE TEST COMPLETED #####");
      logLine(); logLine();
    }

    private void chbShowValidationErrorMarker_CheckedChanged(object sender, EventArgs e)
    {
      if (chbShowValidationErrorMarker.Checked)
      {
        chartTrainingPlot.Series["Validation Error"].MarkerStyle = MarkerStyle.Circle;
      }
      else
      {
        chartTrainingPlot.Series["Validation Error"].MarkerStyle = MarkerStyle.None;
      }
    }

    private void RunSetThroughANNCommittee(TrainingSet set, string setDescription, bool printAllExamples)
    {
      logLine("###############################################");
      logLine("##### TESTING ANN COMMITTEE WITH " + setDescription.ToUpper() + " ######");
      logLine("###############################################");
      logLine();

      double testSetTheoreticalMaxVelo;
      AnalyzeDataSet(set, setDescription, out testSetTheoreticalMaxVelo);
      PopulateVeloDistribution(set, false, "Speed distribution of " + setDescription.ToLower());

      logLine("Theoretical max speed of ANN committe (scaling factor):" + _networkCommittee.OutputScaleFactor.ToString("N4") + " m/s");
      logLine("Input vector type: " + _networkCommittee.InputVectorType);
      logLine();


      if (printAllExamples)
      {
        logLine("-- ALL TRAINING EXAMPLES --");
        logLine();
      }

      double[][] input = set.GetInput(_networkCommittee.InputVectorType, _networkCommittee.NormalizeInputVector);
      double[][] output = set.GetOutput(_networkCommittee.OutputScaleFactor);

      var computedOutputs = new List<double>();
      var scaledComputedOutputs = new List<double>();
      var absErrors = new List<double>();
      var squaredErrors = new List<double>();
      var absScaledErrors = new List<double>();
      var squaredScaledErrors = new List<double>();

      double minOutput = 999999;
      double maxOutput = -999999;
      for (int i = 0; i < input.Length; i++)
      {
        double[] inputVector = input[i];
        double computedOutput = _networkCommittee.Compute(inputVector);
        double desiredOutput = output[i][0];
        /*string inputString = "[ ";
        foreach (var inp in input[i])
        {
          inputString += inp.ToString("N2") + ", ";
        }
        inputString += "]";
        inputs.Add(inputString);
        */

        double error = Math.Abs(desiredOutput - computedOutput);
        absErrors.Add(error);
        squaredErrors.Add(error * error);
        absScaledErrors.Add(error * _networkCommittee.OutputScaleFactor);
        squaredScaledErrors.Add(error * error * _networkCommittee.OutputScaleFactor * _networkCommittee.OutputScaleFactor);
        computedOutputs.Add(computedOutput);
        scaledComputedOutputs.Add(computedOutput * _networkCommittee.OutputScaleFactor);

        if (computedOutput < minOutput)
        {
          minOutput = computedOutput;
        }
        if (computedOutput > maxOutput)
        {
          maxOutput = computedOutput;
        }

        if (printAllExamples)
        {
          logLine(set.TrainingData[i].ToString());
          string inputString = "[ ";
          foreach (var inp in inputVector)
          {
            inputString += inp.ToString("N2") + ", ";
          }
          inputString += "]";
          logLine(inputString);
          logLine("Target: " + desiredOutput.ToString("N4") + "  |  " +
                  (desiredOutput * _networkCommittee.OutputScaleFactor).ToString("N2") + " m/s");
          logLine("Output: " + computedOutput.ToString("N4") + "  |  " +
                  (computedOutput * _networkCommittee.OutputScaleFactor).ToString("N2") + " m/s");
          logLine("Absolute error: " + error.ToString("N4") + "  |  " +
                  (error * _networkCommittee.OutputScaleFactor).ToString("N2") + " m/s");
          logLine();
        }
      }
      logLine();

      double minScaledOutput = minOutput * _networkCommittee.OutputScaleFactor;
      double maxScaledOutput = maxOutput * _networkCommittee.OutputScaleFactor;

      double meanAbsError = Utilities.Average(absErrors);
      double meanSquaredError = Utilities.Average(squaredErrors);
      double meanAbsScaledError = Utilities.Average(absScaledErrors);
      double meansSqaredScaledError = Utilities.Average(squaredScaledErrors);
      double avgOutput = Utilities.Average(computedOutputs);
      double avgScaledOutput = Utilities.Average(scaledComputedOutputs);
      double stdevOutputs = Utilities.StandardDeviation(computedOutputs);
      double stdevScaledOutputs = Utilities.StandardDeviation(scaledComputedOutputs);

      /*
      var errorFile = new StreamWriter("test_errors.txt");
      for (int i = 0; i < squaredErrors.Count; i++)
      {
        errorFile.WriteLine("Test instance " + i + ": " + inputs[i] + "   Computed velo = " + computedOutputs[i] + "   Correct velo = " + output[i][0] + "   SquaredError = " + squaredErrors[i]
          + "   Absolute error = " + absErrors[i]);
      }
      errorFile.WriteLine();
      errorFile.WriteLine("-- SUMMARY --");
      errorFile.WriteLine();
      errorFile.WriteLine("Mean Squared Error = " + meanSquaredError);
      errorFile.WriteLine();
      errorFile.WriteLine("Mean Absolute Error = " + meanAbsError);
      errorFile.Close();
      */

      logLine("-- TEST RESULTS (USING " + setDescription.ToUpper() + ") --");
      logLine("Mean Squared Error: " + meanSquaredError.ToString("N6"));
      logLine("Mean Squared Error (scaled): " + meansSqaredScaledError.ToString("N6") + " (m/s)^2");
      logLine("Mean Absolute Error: " + meanAbsError.ToString("N4"));
      logLine("Mean Absolute Error (scaled): " + meanAbsScaledError.ToString("N4") + " m/s");
      logLine();
      logLine("Maximum output: " + maxOutput.ToString("N4") + "  |  " + maxScaledOutput.ToString("N4") + " m/s  |  " + printMinPerKm(maxScaledOutput));
      logLine("Minimum output:  " + minOutput.ToString("N4") + "  |  " + minScaledOutput.ToString("N4") + " m/s  |  " + printMinPerKm(minScaledOutput));
      logLine("Avg. output:       " + avgOutput.ToString("N4") + "  |  " + avgScaledOutput.ToString("N4") + " m/s  |  " + printMinPerKm(avgScaledOutput));
      logLine("Standard deviation of outputs: " + stdevOutputs.ToString("N4") + "  |  " + stdevScaledOutputs.ToString("N4") + " m/s");
      logLine();
      logLine("##### TEST ON " + setDescription.ToUpper() + " COMPLETED #####");
      logLine(); logLine();

      PopulateVeloDistribution(set, true, "Speed distribution of " + setDescription.ToLower() + " as computed by ANN Committe");
    }

    private void RunSetThroughANN(TrainingSet set, string setDescription, bool printAllExamples)
    {
      //logLine("########################################");
      logLine("##### TESTING ANN WITH " + setDescription.ToUpper() + " ######");
      //logLine("########################################");
      logLine();

      double testSetTheoreticalMaxVelo;
      AnalyzeDataSet(set, setDescription, out testSetTheoreticalMaxVelo);
      PopulateVeloDistribution(set, false, "Speed distribution of " + setDescription.ToLower());

      logLine("Theoretical max speed of ANN (scaling factor):" + _network.OutputScaleFactor.ToString("N4") + " m/s");
      logLine("Input vector: " + _network.InputVectorType);
      logLine();


      if (printAllExamples)
      {
        logLine("---- ALL TRAINING EXAMPLES ----");
        logLine();
      }

      double[][] input = set.GetInput(_network.InputVectorType, _network.NormalizeInputVector);
      double[][] output = set.GetOutput(_network.OutputScaleFactor);

      var computedOutputs = new List<double>();
      var scaledComputedOutputs = new List<double>();
      var absErrors = new List<double>();
      var squaredErrors = new List<double>();
      var absScaledErrors = new List<double>();
      var squaredScaledErrors = new List<double>();

      double minOutput = 999999;
      double maxOutput = -999999;
      for (int i = 0; i < input.Length; i++)
      {
        double[] inputVector = input[i];
        double computedOutput = _network.Compute(inputVector)[0];
        double desiredOutput = output[i][0];
        /*string inputString = "[ ";
        foreach (var inp in input[i])
        {
          inputString += inp.ToString("N2") + ", ";
        }
        inputString += "]";
        inputs.Add(inputString);
        */

        double error = Math.Abs( desiredOutput - computedOutput );
        double squaredError = error * error / 2;
        absErrors.Add(error);
        squaredErrors.Add(squaredError);
        absScaledErrors.Add( error * _network.OutputScaleFactor );
        squaredScaledErrors.Add( squaredError * _network.OutputScaleFactor * _network.OutputScaleFactor );
        computedOutputs.Add(computedOutput);
        scaledComputedOutputs.Add(computedOutput * _network.OutputScaleFactor);

        if (computedOutput < minOutput)
        {
          minOutput = computedOutput;
        }
        if (computedOutput > maxOutput)
        {
          maxOutput = computedOutput;
        }

        if (printAllExamples)
        {
          logLine(set.TrainingData[i].ToString());
          string inputString = "[ ";
          foreach (var inp in inputVector)
          {
            inputString += inp.ToString("N2") + ", ";
          }
          inputString += "]";
          logLine(inputString);
          logLine("Target: " + desiredOutput.ToString("N4") + "  |  " +
                  (desiredOutput*_network.OutputScaleFactor).ToString("N2") + " m/s");
          logLine("Output: " + computedOutput.ToString("N4") + "  |  " +
                  (computedOutput*_network.OutputScaleFactor).ToString("N2") + " m/s");
          logLine("Absolute error: " + error.ToString("N4") + "  |  " +
                  (error*_network.OutputScaleFactor).ToString("N2") + " m/s");
          logLine();
        }
      }
      logLine();

      double minScaledOutput = minOutput * _network.OutputScaleFactor;
      double maxScaledOutput = maxOutput * _network.OutputScaleFactor;

      double meanAbsError = Utilities.Average(absErrors);
      double meanSquaredError = Utilities.Average(squaredErrors);
      double meanAbsScaledError = Utilities.Average(absScaledErrors);
      double meansSqaredScaledError = Utilities.Average(squaredScaledErrors);
      double avgOutput = Utilities.Average(computedOutputs);
      double avgScaledOutput = Utilities.Average(scaledComputedOutputs);
      double stdevOutputs = Utilities.StandardDeviation(computedOutputs);
      double stdevScaledOutputs = Utilities.StandardDeviation(scaledComputedOutputs);

      /*
      var errorFile = new StreamWriter("test_errors.txt");
      for (int i = 0; i < squaredErrors.Count; i++)
      {
        errorFile.WriteLine("Test instance " + i + ": " + inputs[i] + "   Computed velo = " + computedOutputs[i] + "   Correct velo = " + output[i][0] + "   SquaredError = " + squaredErrors[i]
          + "   Absolute error = " + absErrors[i]);
      }
      errorFile.WriteLine();
      errorFile.WriteLine("-- SUMMARY --");
      errorFile.WriteLine();
      errorFile.WriteLine("Mean Squared Error = " + meanSquaredError);
      errorFile.WriteLine();
      errorFile.WriteLine("Mean Absolute Error = " + meanAbsError);
      errorFile.Close();
      */

      logLine("---- TEST RESULTS (USING " + setDescription.ToUpper() + ") ----");
      logLine("Mean Squared Error: " + meanSquaredError.ToString("N6"));
      logLine("Mean Squared Error (scaled): " + meansSqaredScaledError.ToString("N6") + " (m/s)^2");
      logLine("Mean Absolute Error: " + meanAbsError.ToString("N4"));
      logLine("Mean Absolute Error (scaled): " + meanAbsScaledError.ToString("N4") + " m/s");
      logLine();
      logLine("Maximum output: " + maxOutput.ToString("N4") + "  |  " + maxScaledOutput.ToString("N4") + " m/s  |  " + printMinPerKm(maxScaledOutput));
      logLine("Minimum output:  " + minOutput.ToString("N4") + "  |  " + minScaledOutput.ToString("N4") + " m/s  |  " + printMinPerKm(minScaledOutput));
      logLine("Avg. output:       " + avgOutput.ToString("N4") + "  |  " + avgScaledOutput.ToString("N4") + " m/s  |  " + printMinPerKm(avgScaledOutput));
      logLine("Standard deviation of outputs: " + stdevOutputs.ToString("N4") + "  |  " + stdevScaledOutputs.ToString("N4") + " m/s");
      logLine();
      logLine("##### TEST ON " + setDescription.ToUpper() + " COMPLETED #####");
      logLine(); logLine();

      PopulateVeloDistribution(set, true, "Speed distribution of " + setDescription.ToLower() + " as computed by ANN");
    }

    private void splitContainer1_Panel1_MouseEnter(object sender, EventArgs e)
    {
      splitContainer1.Panel1.Focus();
    }

    private void splitContainer1_Panel2_MouseEnter(object sender, EventArgs e)
    {
      splitContainer1.Panel2.Focus();
    }

    private void cmbAlgorithm_SelectedIndexChanged(object sender, EventArgs e)
    {
      if ((LearningAlgorithm)cmbAlgorithm.SelectedItem == LearningAlgorithm.BackProp)
      {
        lblMomentum.Enabled = true;
        txtMomentum.Enabled = true;
        rdbAvgErrorPreviousInstances.Enabled = true;
      }
      else
      {
        lblMomentum.Enabled = false;
        txtMomentum.Enabled = false;
        rdbAvgErrorPreviousEpoch.Checked = true;
        rdbAvgErrorPreviousInstances.Enabled = false;
      }
    }

    private enum LearningAlgorithm
    {
      BackProp,
      RProp
    }

    private void rdbAvgErrorPreviousInstances_CheckedChanged(object sender, EventArgs e)
    {
      if (rdbAvgErrorPreviousInstances.Checked)
      {
        txtAvgErrorPreviousInstances.Enabled = true;
      }
      else
      {
        txtAvgErrorPreviousInstances.Enabled = false;
      }
    }

    private void cmbPlot_SelectedIndexChanged(object sender, EventArgs e)
    {
      if (cmbPlot.SelectedIndex == 0) // Error plot
      {
        foreach (Series s in chartTrainingPlot.Series)
        {
          s.Enabled = false;
        }
        chartTrainingPlot.Series["Training Error"].Enabled = true;
        chartTrainingPlot.Series["Validation Error"].Enabled = true;
        chartTrainingPlot.Titles["Header"].Text = "Training and validation error";
        chartTrainingPlot.ChartAreas[0].AxisY.Title = "Mean Abs. Error (m/s)";

        chbScaleOutput.Enabled = false;
        lblAvgTrainingErrorOver.Enabled = true;
        rdbAvgErrorPreviousEpoch.Enabled = true;
        rdbAvgErrorPreviousInstances.Enabled = true;
        txtAvgErrorPreviousInstances.Enabled = true;
        lblShowMarkers.Enabled = true;
        chbShowTrainingErrorMarker.Enabled = true;
        chbShowValidationErrorMarker.Enabled = true;
      }
      else // Speed development plot
      {
        foreach (Series s in chartTrainingPlot.Series)
        {
          s.Enabled = true;
        }
        chartTrainingPlot.Series["Training Error"].Enabled = false;
        chartTrainingPlot.Series["Validation Error"].Enabled = false;
        chartTrainingPlot.Titles["Header"].Text = "Speed development during training";
        chartTrainingPlot.ChartAreas[0].AxisY.Title = "Output speed";
        if (chbScaleOutput.Checked)
        {
          chartTrainingPlot.ChartAreas[0].AxisY.Title += " scaled to speed range (m/s)";
        }
        else
        {
          chartTrainingPlot.ChartAreas[0].AxisY.Title += " relative to theoretical max. speed";
        }

        chbScaleOutput.Enabled = true;
        lblAvgTrainingErrorOver.Enabled = false;
        rdbAvgErrorPreviousEpoch.Enabled = false;
        rdbAvgErrorPreviousInstances.Enabled = false;
        txtAvgErrorPreviousInstances.Enabled = false;
        lblShowMarkers.Enabled = false;
        chbShowTrainingErrorMarker.Enabled = false;
        chbShowValidationErrorMarker.Enabled = false;
      }
      chartTrainingPlot.ChartAreas[0].RecalculateAxesScale();
    }

    private void chbShowTrainingErrorMarker_CheckedChanged(object sender, EventArgs e)
    {
      if (chbShowTrainingErrorMarker.Checked)
      {
        chartTrainingPlot.Series["Training Error"].MarkerStyle = MarkerStyle.Triangle;
      }
      else
      {
        chartTrainingPlot.Series["Training Error"].MarkerStyle = MarkerStyle.None;
      }
    }

    private void btnToggleTrainingPlotLegend_Click(object sender, EventArgs e)
    {
      chartTrainingPlot.Legends[0].Enabled = !chartTrainingPlot.Legends[0].Enabled;
      if (chartTrainingPlot.Legends[0].Enabled)
      {
        chartTrainingPlot.ChartAreas[0].Position.Width = 79;
        btnToggleTrainingPlotLegend.Text = "Hide legend";
      }
      else
      {
        chartTrainingPlot.ChartAreas[0].Position.Width = 95;
        btnToggleTrainingPlotLegend.Text = "Show legend";
      }

    }

    private void chbScaleOutput_CheckedChanged(object sender, EventArgs e)
    {
      for (int i = 0; i < chartTrainingPlot.Series.Count; i++)
      {
        if (chartTrainingPlot.Series[i] != chartTrainingPlot.Series["Training Error"]
          && chartTrainingPlot.Series[i] != chartTrainingPlot.Series["Validation Error"])
        {
          for (int j = 0; j < chartTrainingPlot.Series[i].Points.Count; j++)
          {
            if (chbScaleOutput.Checked)
            {
              chartTrainingPlot.Series[i].Points[j].YValues[0] *= _network.OutputScaleFactor; // Scale velo to velocity range
            }
            else
            {
              chartTrainingPlot.Series[i].Points[j].YValues[0] /= _network.OutputScaleFactor; // Scale velo to <0,1> range
            }
          }
        }
      }
      chartTrainingPlot.ChartAreas[0].AxisY.Title = "Output speed";
      if (chbScaleOutput.Checked)
      {
        chartTrainingPlot.ChartAreas[0].AxisY.Title += " scaled to speed range (m/s)";
      }
      else
      {
        chartTrainingPlot.ChartAreas[0].AxisY.Title += " relative to theoretical max. speed";
      }

      chartTrainingPlot.ChartAreas[0].RecalculateAxesScale();
      chartTrainingPlot.Refresh();
    }

    private void txtLineWidth_TextChanged(object sender, EventArgs e)
    {
      for (int i = 0; i < chartTrainingPlot.Series.Count; i++)
      {
        try
        {
          chartTrainingPlot.Series[i].BorderWidth = int.Parse(txtLineWidth.Text);
        }
        catch
        {
          lblStatus.Text = "ERROR: Line width must be an integer.";
        }
      }
    }

    private void btnToggleShowParams_Click(object sender, EventArgs e)
    {
      for (int i = 0; i < chartTrainingPlot.Titles.Count; i++)
      {
        chartTrainingPlot.Titles[i].Visible = !chartTrainingPlot.Titles[i].Visible;
      }
      chartTrainingPlot.Titles["Header"].Visible = true; // Header should always be visible
      if (cmbAlgorithm.SelectedIndex == 1)  // Using RProp. Momentum is not used
      {
        chartTrainingPlot.Titles["Momentum"].Visible = false;
      }
      if (btnToggleShowParams.Text == "Hide params")
      {
        btnToggleShowParams.Text = "Show params";
      }
      else
      {
        btnToggleShowParams.Text = "Hide params";
      }
    }

    private void chbHideTrainingError_CheckedChanged(object sender, EventArgs e)
    {
      if (chbHideTrainingError.Checked)
      {
        chartTrainingPlot.Series["Training Error"].Enabled = false;
        if (chartTrainingPlot.Series["Validation Error"].Enabled)
        {
          chartTrainingPlot.Titles["Header"].Text = "Validation error";
        }
        else
        {
          chartTrainingPlot.Titles["Header"].Text = "";
        }
        
      }
      else
      {
        chartTrainingPlot.Series["Training Error"].Enabled = true;
        if (chartTrainingPlot.Series["Validation Error"].Enabled)
        {
          chartTrainingPlot.Titles["Header"].Text = "Training and validation error";
        }
        else
        {
          chartTrainingPlot.Titles["Header"].Text = "Training error";
        }
      }
    }

    private void chbHideValidationError_CheckedChanged(object sender, EventArgs e)
    {
      if (chbHideValidationError.Checked)
      {
        chartTrainingPlot.Series["Validation Error"].Enabled = false;
        if (chartTrainingPlot.Series["Training Error"].Enabled)
        {
          chartTrainingPlot.Titles["Header"].Text = "Training error";
        }
        else
        {
          chartTrainingPlot.Titles["Header"].Text = "";
        }
      }
      else
      {
        chartTrainingPlot.Series["Validation Error"].Enabled = true;
        if (chartTrainingPlot.Series["Training Error"].Enabled)
        {
          chartTrainingPlot.Titles["Header"].Text = "Training and validation error";
        }
        else
        {
          chartTrainingPlot.Titles["Header"].Text = "Validation error";
        }
      }
    }

    private void chartTrainingPlot_AxisViewChanged(object sender, ViewEventArgs e)
    {
      chartTrainingPlot.ChartAreas[0].CursorX.Position = _epochOfBestNetwork;
    }

    private void btnOpenANNCommittee_Click(object sender, EventArgs e)
    {
      try
      {
        _networkCommittee = (VeloActivationNetworkCommittee)VeloActivationNetworkCommittee.Load(txtANNCommiteeFile.Text);
        lblStatus.Text = "ANN committee loaded.";
        btnTestNetwork.Enabled = true;
      }
      catch (Exception)
      {
        lblStatus.Text = "Could not load ANN committee. The file could not be found.";
      }
    }

    private void btnSaveANNCommittee_Click(object sender, EventArgs e)
    {
      dlgSave.Filter = "*.bin|*.bin";
      dlgSave.FileName = buildFileName("ANNCommittee") + ".bin";
      if (dlgSave.ShowDialog(this) == DialogResult.OK)
      {
        _networkCommittee.Save(dlgSave.FileName);
        lblStatus.Text = "ANN committee saved.";
      }
    }

    private void btnBrowseANNCommitteeFile_Click(object sender, EventArgs e)
    {
      dlgOpen.FileName = txtANNCommiteeFile.Text;
      dlgOpen.Filter = "*.bin|*.bin";
      if (dlgOpen.ShowDialog(this) == DialogResult.OK)
      {
        txtANNCommiteeFile.Text = dlgOpen.FileName;
      }
    }

    private void chbDegradeToSevenSeconds_CheckedChanged(object sender, EventArgs e)
    {
      if (chbDegradeToSevenSeconds.Checked)
      {
        cmbDegradeMethod.Enabled = true;
      }
      else
      {
        cmbDegradeMethod.Enabled = false;
        cmbDegradeMethod.SelectedItem = DegradeMethod.None;
      }
    }

  }
}
