﻿using System;
using System.Collections.Generic;
using Grid;
using Tracking;
using System.Diagnostics;
using System.Runtime.Serialization;

namespace GPSLearning
{
  [Serializable]
  public class TrainingSet : ISerializable
  {
    public const int NofForestClasses = 6; // VeryDifficultToRun, DifficultToRun, SlowRunning, RoughOpenScatteredTrees, RoughOpen, Open/OpenScatteredTrees
    public const int NofWaterClasses = 2; // Marsh, IndistinctMarsh/NarrowMarsh
    public const int NofUndergrowthClasses = 2; // DifficultToRun, SlowRunning
    public const int NofPathsClasses = 5; // Road/VehicleTrack, Footpath, NarrowRide, SmallPath, LessDistinctSmallPath
    public const int NofGroundClasses = 2; // BoulderField, StonyGround
    public const int NofMapFeatures = NofForestClasses + NofWaterClasses +
      NofUndergrowthClasses + NofPathsClasses /*+ NofGroundClasses*/; // Ignoring ground for now
    //public const int NofMapFeatureGroups = 5;
    public const int NofMapFeatureGroups = 4; // Ignoring ground
    public List<TrainingInstance> TrainingData;

    public TrainingSet(string heightFile, string forestFile, string waterFile,
      string undergrowthFile, string pathsFile, string groundFile,
      string passableFile, IEnumerable<string> gpxFiles, int removeAtStart, int removeAtEnd, int gpsUpdateInterval, DegradeMethod degradeMethod)
    {
      TrainingData = new List<TrainingInstance>();
      DoubleGrid heightGrid = DoubleGrid.FromAsciiFile(heightFile, 0, 0.01,
        typeof(int));
      var mapGrid =
        new FeatureMap(forestFile, waterFile, undergrowthFile, pathsFile,
          groundFile, passableFile);

      // Read all GPX files
      foreach (string file in gpxFiles)
      {
        Debug.WriteLine("Reading GPX file: " + file);
        List<TrackPoint> trackPoints = (List<TrackPoint>)GpxIO.ReadData(file);

        // Remove some points at the beginning and end of the track
        trackPoints.RemoveRange(0, removeAtStart);
        trackPoints.RemoveRange(trackPoints.Count - removeAtEnd, removeAtEnd);

        // Set the height of each track point, based on heightgrid
        Track.AssignHeight(trackPoints, heightGrid);

        // Possibly degrade to 7 sec resolution
        List<List<TrackPoint>> trackPointLists = new List<List<TrackPoint>>();
        if (degradeMethod != DegradeMethod.None)
        {
          int max;
          if (degradeMethod == DegradeMethod.Simple)
          {
            max = 1;
          }
          else
          {
            max = 7;
          }
          {
            for (int i = 0; i < max; i++)
            {
              Debug.WriteLine("NEWLIST");
              Debug.WriteLine("");
              List<TrackPoint> currentTrackPoints = new List<TrackPoint>();
              for (int j = 0; j < trackPoints.Count; j++)
              {
                if (j % 7 == 0)
                {
                  if (j + i < trackPoints.Count)
                  {
                    currentTrackPoints.Add(trackPoints[j + i]);
                    Debug.WriteLine("Added " + (j + i));
                  }
                }
              }
              trackPointLists.Add(currentTrackPoints);
            }
          }
        }
        else
        {
          trackPointLists.Add(trackPoints);
        }

        // Take all trackpoints from GPX file and add map information
        for (int j = 0; j < trackPointLists.Count; j++)
        {
          trackPoints = trackPointLists[j];
          for (int i = 1; i < trackPoints.Count; i++)
          {
            TrackPoint p = trackPoints[i];
            TrackPoint pPrev = trackPoints[i - 1];

            double distance = Math.Sqrt(Math.Pow(p.X - pPrev.X, 2) + Math.Pow(p.Y - pPrev.Y, 2));
            TimeSpan timeDiff = p.Time - pPrev.Time;
            p.Velo = distance / timeDiff.TotalSeconds;
            if (timeDiff.TotalSeconds > gpsUpdateInterval + 0.01 && degradeMethod == DegradeMethod.None)
            {
              // Do not make a training instance of this track point, as it follows a gap in the GPX data
              continue;
            }

            // Compute climb and slope. Climb is the height difference since the previous trackpoint divided by distance since that point, i.e. the slope in running direction.
            // Slope is the average terrain gradient of the current and the previous track points.
            double heightDiff = p.H - pPrev.H;
            double climb = heightDiff / distance;
            double slopeCurrent = Math.Sqrt(heightGrid.Slope2(p.X, p.Y));
            double slopePrev = Math.Sqrt(heightGrid.Slope2(pPrev.X, pPrev.Y));
            double slope = (slopeCurrent + slopePrev) / 2;

            var mapFeatures = (MapPointFeatures)mapGrid.GetMapPoint(p.X, p.Y);
            var classifiedMapFeatures = new ClassifiedMapPointFeatures(mapFeatures);

            // Check whether the map features have changed since the last trackpoint
            bool mapFeaturesChanged;
            if (TrainingData.Count == 0 || i == 0)
            {
              mapFeaturesChanged = false;
            }
            else
            {
              mapFeaturesChanged = !classifiedMapFeatures.Equals(TrainingData[TrainingData.Count - 1].ClassifiedMapFeatures);
            }


            // Skip training instances that are located in forbidden areas.
            if (classifiedMapFeatures.Passable == Passable.Forbidden
              || (classifiedMapFeatures.Passable == Passable.ForbiddenButPathsAllowed && classifiedMapFeatures.Paths == Paths.None))
            {
              continue;
            }

            // Skip training instances with invalid velocity, climb or slope
            if (double.IsNaN(p.Velo) || p.Velo < 0 || p.Velo > 10
              || double.IsNaN(climb) || climb < -9999 || climb > 9999
              || double.IsNaN(slope) || slope < 0 || slope > 9999)
            {
              continue;
            }

            TrainingData.Add(new TrainingInstance(p.X, p.Y, climb, slope, classifiedMapFeatures, p.Velo, mapFeaturesChanged));
          }
        }
      }
    }

    public TrainingSet(List<TrainingInstance> trainingData)
    {
      TrainingData = trainingData;
    }

    public TrainingSet()
    {
    }

    public double[][] GetInput(InputVectorType inputVectorType, bool normalizeInputVector)
    {
      var input = new double[TrainingData.Count][];
      for (int i = 0; i < TrainingData.Count; i++)
      {
        ClassifiedMapPointFeatures cmf = TrainingData[i].ClassifiedMapFeatures;
        double climb = TrainingData[i].Climb;
        double slope = TrainingData[i].Slope;
        input[i] = CreateInputVector(cmf, climb, slope, inputVectorType, normalizeInputVector);
      }
      return input;
    }

    /// <summary>
    /// Constructs the input vector for the ANN. The three last inputs are climb (ascent/descent),
    /// terrain slope squared and bias.
    /// The other inputs are the map features. There is one input for each possible map feature. An input
    /// value of 1 means that the corresponding map feature is present. 0 means that it is not present.
    /// </summary>
    /// <param name="cmf">The map features as enum values</param>
    /// <param name="climb">Climb (height1 - height0)</param>
    /// <param name="slope">Slope squared</param>
    /// <param name="vectorType">The type of input vector to create</param>
    /// <param name="normalized">Whether the input vector should be normalized, ie. each input is divided by the root of the sum of the squares of all inputs.</param>
    /// <returns></returns>
    public static double[] CreateInputVector(ClassifiedMapPointFeatures cmf, double climb, double slope, InputVectorType vectorType, bool normalized)
    {
      double[] inputVector;

      switch (vectorType)
      {
        case InputVectorType.OneInputPerFeature:

          inputVector = new double[NofMapFeatures + 3];

          // Initialize all map feature inputs to 0
          for (int i = 0; i < NofMapFeatures; i++)
          {
            inputVector[i] = 0;
          }

          // Forest
          switch(cmf.Forest)
          {
            case Forest.VeryDifficultToRun:
              inputVector[0] = 1;
              break;
            case Forest.DifficultToRun:
              inputVector[1] = 1;
              break;
            case Forest.SlowRunning:
              inputVector[2] = 1;
              break;
            case Forest.RoughOpenScatteredTrees:
              inputVector[3] = 1;
              break;
            case Forest.RoughOpen:
              inputVector[4] = 1;
              break;
            case Forest.OpenScatteredTrees:
            case Forest.Open:
              inputVector[5] = 1;
              break;
          }

          // Water (just marshes)
          switch(cmf.Water)
          {
            case Water.Marsh:
              inputVector[6] = 1;
              break;
            case Water.IndistinctMarsh:
            case Water.NarrowMarsh:
              inputVector[7] = 1;
              break;
          }

          // Undergrowth
          switch(cmf.Undergrowth)
          {
            case Undergrowth.DifficultToRun:
              inputVector[8] = 1;
              break;
            case Undergrowth.SlowRunning:
              inputVector[9] = 1;
              break;
          }

          // Paths
          switch(cmf.Paths)
          {
            case Paths.LessDistinctSmallPath:
              inputVector[10] = 1;
              break;
            case Paths.SmallPath:
              inputVector[11] = 1;
              break;
            case Paths.NarrowRide:
              inputVector[12] = 1;
              break;
            case Paths.Footpath:
              inputVector[13] = 1;
              break;
            case Paths.VehicleTrack:
            case Paths.Road:
              inputVector[14] = 1;
              break;
          }

          /* 
          // Ground (Ignoring this for now. Remember to update indices if reactivating.)
          switch(cmf.Ground)
          {
            case Ground.BoulderField:
              inputVector[15] = 1;
              break;
            case Ground.StonyGround:
              inputVector[16] = 1;
              break;
          }
          */

          // Climb
          inputVector[15] = climb;

          // Slope
          inputVector[16] = slope;

          // Bias
          inputVector[17] = -1.0;
          
          /*
          // Use index of enum values as input index for each map feature
          var forest = (int)cmf.Forest;
          var water = (int)cmf.Water;
          var undergrowth = (int)cmf.Undergrowth;
          var paths = (int)cmf.Paths;
          var ground = (int)cmf.Ground;

          int offset = 0;
      
          if (forest < NofForestClasses)
          {
            inputVector[forest] = 1.0;
          }
          offset += NofForestClasses;
          if (water < NofWaterClasses)
          {
            inputVector[water + offset] = 1.0;
          }
          offset += NofWaterClasses;
          if (undergrowth < NofUndergrowthClasses)
          {
            inputVector[undergrowth + offset] = 1.0;
          }
          offset += NofUndergrowthClasses;
          if (paths < NofPathsClasses)
          {
            inputVector[paths + offset] = 1.0;
          }
          offset += NofPathsClasses;
          if (ground < NofGroundClasses)
          {
            inputVector[ground + offset] = 1.0;
          }
          offset += NofGroundClasses;
      
          // Climb
          inputVector[offset] = climb;
        
          // Slope
          offset++;
          inputVector[offset] = Slope;

          // Bias
          offset++;
          inputVector[offset] = -1.0;
          */

          break;

        case InputVectorType.OneInputPerGroup:
        default:
          inputVector = new double[NofMapFeatureGroups + 3];
          int index = 0;

          // Forest
          switch (cmf.Forest)
          {
            case Forest.VeryDifficultToRun:
              inputVector[index] = 0;
              break;
            case Forest.DifficultToRun:
              inputVector[index] = 1.0 / 6;
              break;
            case Forest.SlowRunning:
              inputVector[index] = 2.0 / 6;
              break;
            case Forest.EasyRunning:
            case Forest.OpenSandyGround:  // Not necessarily equivalent with EasyRunning in general. Create separate case if there are enough sandy areas to provide enough data samples.
            case Forest.BareRock:         // Not necessarily equivalent with EasyRunning in general. Create separate case if there are enough bare rock areas to provide enough data samples.  
            inputVector[index] = 3.0 / 6;
              break;
            case Forest.RoughOpenScatteredTrees:
              inputVector[index] = 4.0 / 6;
              break;
            case Forest.RoughOpen:

              inputVector[index] = 5.0 / 6;
              break;
            case Forest.OpenScatteredTrees:
            case Forest.Open:
              inputVector[index] = 1;
              break;
          }
          index++;

          // Water (marshes)
          switch (cmf.Water)
          {
            case Water.Marsh:
              inputVector[index] = 0;
              break;
            case Water.IndistinctMarsh:
            case Water.NarrowMarsh:
              inputVector[index] = 0.5;
              break;
            default:  // No marsh. Ignores all other water features.
              inputVector[index] = 1;
              break;
          }
          index++;

          // Undergrowth
          switch (cmf.Undergrowth)
          {
            case Undergrowth.DifficultToRun:
              inputVector[index] = 0;
              break;
            case Undergrowth.SlowRunning:
              inputVector[index] = 0.5;
              break;
            case Undergrowth.None:
              inputVector[index] = 1;
              break;
          }
          index++;

          // Paths
          switch (cmf.Paths)
          {
            case Paths.None:
              inputVector[index] = 0;
              break;
            case Paths.LessDistinctSmallPath:
              inputVector[index] = 1.0 / 6;
              break;
            case Paths.SmallPath:
              inputVector[index] = 2.0 / 6;
              break;
            case Paths.NarrowRide:
              inputVector[index] = 3.0 / 6;
              break;
            case Paths.Footpath:
              inputVector[index] = 4.0 / 6;
              break;
            case Paths.VehicleTrack:
              inputVector[index] = 5.0 / 6;
              break;
            case Paths.Road:
              inputVector[index] = 1;
              break;
          }
          index++;

          // Ground (boulder field/stony ground). Ignoring this for now.
          /*switch (cmf.Ground)
          {
            case Ground.BoulderField:
              inputVector[index] = 0;
              break;
            case Ground.StonyGround:
              inputVector[index] = 0.5;
              break;
            default:
              inputVector[index] = 1;
              break;
          }
          index++;
          */

          // Climb
          inputVector[index] = climb;
          index++;

          // Slope
          inputVector[index] = slope;
          index++;

          // Bias
          inputVector[index] = -1;

          break;
      }

      // Normalize input vector
      if (normalized)
      {
        double sumOfSquares = 0;
        foreach (double i in inputVector)
        {
          sumOfSquares += i * i;
        }
        double root = Math.Sqrt(sumOfSquares);
        for (int i = 0; i < inputVector.Length; i++)
        {
          inputVector[i] = inputVector[i] / root;
        }
      }

      return inputVector;
    }

    public double[][] GetOutput(double maxVelo)
    {
      var output = new double[TrainingData.Count][];
      for ( int i = 0; i < TrainingData.Count; i++ )
      {
        output[i] = new double[1];
        output[i][0] = TrainingData[i].Velocity;
      }

      // Scale output velocity into output range of activation function
      foreach (double[] o in output)
      {
        double velo = o[0];
        o[0] = velo / maxVelo;
      }
      return output;
    }

    public void Analyze(out int nofInstances, out double maxVelo, out double minVelo, out double avgVelo, out double stdDevVelo,
      out double maxClimb, out double minClimb, out double maxSlope, out double minSlope, out double flatPathSpeed, out int nofFlatPathPoints)
    {
      maxVelo = 0;
      minVelo = 9999;
      maxClimb = -9999;
      minClimb = 9999;
      maxSlope = 0;
      minSlope = 9999;
      nofInstances = TrainingData.Count;
      nofFlatPathPoints = 0;
      flatPathSpeed = 0;
      var velocities = new List<double>();
      for (int i = 0; i < nofInstances; i++)
      {
        double velo = TrainingData[i].Velocity;
        velocities.Add(velo);
        if (velo > maxVelo)
        {
          maxVelo = velo;
        }
        if (velo < minVelo)
        {
          minVelo = velo;
        }

        double climb = TrainingData[i].Climb;
        if (climb > maxClimb)
        {
          maxClimb = climb;
        }
        if (climb < minClimb)
        {
          minClimb = climb;
        }

        double slope = TrainingData[i].Slope;
        if (slope > maxSlope)
        {
          maxSlope = slope;
        }
        if (slope < minSlope)
        {
          minSlope = slope;
        }
        if ((TrainingData[i].ClassifiedMapFeatures.Paths == Paths.Road || TrainingData[i].ClassifiedMapFeatures.Paths == Paths.VehicleTrack)
          && TrainingData[i].Climb < 0.02 && TrainingData[i].Climb > -0.02)
        {
          nofFlatPathPoints++;
          flatPathSpeed += TrainingData[i].Velocity;
        }
      }
      flatPathSpeed = flatPathSpeed / nofFlatPathPoints;
      avgVelo = Utilities.Average(velocities);
      stdDevVelo = Utilities.StandardDeviation(velocities);
    }

    public void Filter(int trailingPathPoints, bool removeFirstAfterMapFeatureChange)
    {
      if (removeFirstAfterMapFeatureChange)
      {
        // Remove first point after each map feature change
        for (int i = TrainingData.Count - 1; i >= 0; i--)
        {
          if (TrainingData[i].MapFeaturesChanged)
          {
            TrainingData.RemoveAt(i);
          }
        }
      }

      // Filter out the first 'trailingPathPoints' points after entering a path, and the last 'trailingPathPoints' before exiting a path,
      // as these are usually erroneous due to exagerated width of paths during training.
      int consecutivePointsOnPath = 0;
      for (int i = TrainingData.Count - 2; i >= 0; i--)
      {
        TrainingInstance prevInstance = TrainingData[i+1];
        TrainingInstance currentInstance = TrainingData[i];
        bool prevPointOnPath = prevInstance.ClassifiedMapFeatures.Paths != Paths.None;
        bool currentPointOnPath = currentInstance.ClassifiedMapFeatures.Paths != Paths.None;
        if (currentPointOnPath)
        {
          consecutivePointsOnPath++;
          if (consecutivePointsOnPath == trailingPathPoints)
          {
            // Recently entered a path. Remove the first few path points.
            /*Debug.WriteLine("Entered. Start: (" + TrainingData[i].X + "," + TrainingData[i].Y + ")"
                            + ", Slutt: (" + TrainingData[i + trailingPathPoints - 1].X + "," + TrainingData[i + trailingPathPoints - 1].Y + ")");*/
            TrainingData.RemoveRange(i, trailingPathPoints);
          }
        }
        if (prevPointOnPath && !currentPointOnPath)
        {
          // Just left path. Remove the last few path points.
          consecutivePointsOnPath = 0;
          int nofInstancesToRemove = Math.Min(trailingPathPoints, TrainingData.Count - 1 - i);
          for (int j = 0; j < nofInstancesToRemove; j++)
          {
            if (TrainingData[i+1].ClassifiedMapFeatures.Paths != Paths.None)
            {
              /*Debug.WriteLine("Exit. (" + TrainingData[i+1].X + "," + TrainingData[i+1].Y + ")");*/
              TrainingData.RemoveAt(i+1);
            }
            else
            {
              break;
            }
          }
        }
      }
        

                

    }

    public TrainingSet(SerializationInfo info, StreamingContext context)
    {
      TrainingData = (List<TrainingInstance>)info.GetValue("TrainingData", typeof(List<TrainingInstance>));
    }

    public void GetObjectData(SerializationInfo info, StreamingContext context)
    {
      info.AddValue("TrainingData", TrainingData);
    }
  }

  [Serializable]
  public struct TrainingInstance : ISerializable
  {
    public double X;
    public double Y;
    public double Climb;
    public double Slope;
    public ClassifiedMapPointFeatures ClassifiedMapFeatures;
    public double Velocity;
    public bool MapFeaturesChanged;

    public TrainingInstance(double x, double y, double climb, double slope, ClassifiedMapPointFeatures cmf, double velo, bool mapFeaturesChanged)
    {
      X = x;
      Y = y;
      Climb = climb;
      Slope = slope;
      ClassifiedMapFeatures = cmf;
      Velocity = velo;
      MapFeaturesChanged = mapFeaturesChanged;
    }

    public TrainingInstance(double climb, double slope, ClassifiedMapPointFeatures cmf)
    {
      X = 0;
      Y = 0;
      Climb = climb;
      Slope = slope;
      ClassifiedMapFeatures = cmf;
      Velocity = 0;
      MapFeaturesChanged = false;
    }

    public TrainingInstance(SerializationInfo info, StreamingContext context)
    {
      X = (double)info.GetValue("X", typeof(double));
      Y = (double)info.GetValue("Y", typeof(double));
      Climb = (double)info.GetValue("Climb", typeof(double));
      Slope = (double)info.GetValue("Slope", typeof(double));
      ClassifiedMapFeatures = (ClassifiedMapPointFeatures)info.GetValue("ClassifiedMapFeatures", typeof(ClassifiedMapPointFeatures));
      Velocity = (double)info.GetValue("Velocity", typeof(double));
      MapFeaturesChanged = (bool)info.GetValue("MapFeaturesChanged", typeof(bool));
    }

    public void GetObjectData(SerializationInfo info, StreamingContext context)
    {
      info.AddValue("X", X);
      info.AddValue("Y", Y);
      info.AddValue("Climb", Climb);
      info.AddValue("Slope", Slope);
      info.AddValue("ClassifiedMapFeatures", ClassifiedMapFeatures);
      info.AddValue("Velocity", Velocity);
      info.AddValue("MapFeaturesChanged", MapFeaturesChanged);
    }

    public override string ToString()
    {
      string forest = ClassifiedMapFeatures.Forest.ToString();
      string water = ClassifiedMapFeatures.Water.ToString();
      string undergrowth = ClassifiedMapFeatures.Undergrowth.ToString();
      string paths = ClassifiedMapFeatures.Paths.ToString();
      //string ground = ClassifiedMapFeatures.Ground.ToString();
      string climb = Climb.ToString("N2");
      string slope = Slope.ToString("N2");
      return forest + ", " + water + ", " + undergrowth + ", " + paths + ", " /* + ground + ", " */ + climb + ", " + slope;
    }
  }

  public enum InputVectorType
  {
    OneInputPerFeature,
    OneInputPerGroup
  }

  public enum DegradeMethod
  {
    None,
    Simple,
    Advanced
  }

}
