﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using AForge.Neuro;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using System.IO;

namespace GPSLearning
{
  [Serializable]
  public class VeloActivationNetworkCommittee
  {
    public List<VeloActivationNetwork> Networks;
    private double _outputScaleFactor;
    public InputVectorType InputVectorType;
    public bool NormalizeInputVector;

    public VeloActivationNetworkCommittee(double outputScaleFactor, InputVectorType inputVectorType, bool normalizeInputVector)
    {
      Networks = new List<VeloActivationNetwork>();
      _outputScaleFactor = outputScaleFactor;
      InputVectorType = inputVectorType;
      NormalizeInputVector = normalizeInputVector;
    }

    public double OutputScaleFactor
    {
      get
      {
        return _outputScaleFactor;
      }
      set
      {
        _outputScaleFactor = value;
        for (int i = 0; i < Networks.Count; i++)
        {
          Networks[i].OutputScaleFactor = _outputScaleFactor;
        }
      }
    }

    public double Compute(double[] input)
    {
      double outputSum = 0;
      foreach (VeloActivationNetwork network in Networks)
      {
        outputSum += network.Compute(input)[0];
      }
      return outputSum / Networks.Count;
    }

    public double Compute(double[] input, out double meanAbsDeviation)
    {
      double outputSum = 0;
      var annOutputs = new List<double>();
      foreach (VeloActivationNetwork network in Networks)
      {
        double output = network.Compute(input)[0];
        annOutputs.Add(output);
        outputSum += output;
      }
      double meanOutput = outputSum / Networks.Count;

      double totalAbsDeviation = 0;
      foreach (double output in annOutputs)
      {
        totalAbsDeviation += Math.Abs(output - meanOutput);
      }
      meanAbsDeviation = totalAbsDeviation / annOutputs.Count;
      return meanOutput;
    }

    /// <summary>
    /// Save ANN committee to specified file.
    /// </summary>
    /// 
    /// <param name="fileName">File name to save ANN committee into.</param>
    /// 
    /// <remarks><para>The ANN committee is saved using .NET serialization (binary formatter is used).</para></remarks>
    /// 
    public void Save(string fileName)
    {
      FileStream stream = new FileStream(fileName, FileMode.Create, FileAccess.Write, FileShare.None);
      Save(stream);
      stream.Close();
    }

    /// <summary>
    /// Save ANN committee to specified file.
    /// </summary>
    /// 
    /// <param name="stream">Stream to save ANN committee into.</param>
    /// 
    /// <remarks><para>The ANN committee is saved using .NET serialization (binary formatter is used).</para></remarks>
    /// 
    public void Save(Stream stream)
    {
      IFormatter formatter = new BinaryFormatter();
      formatter.Serialize(stream, this);
    }

    /// <summary>
    /// Load ANN committee from specified file.
    /// </summary>
    /// 
    /// <param name="fileName">File name to load ANN committee from.</param>
    /// 
    /// <returns>Returns instance of VeloActivationNetworkComittee class with all properties initialized from file.</returns>
    /// 
    /// <remarks><para>ANN committee is loaded from file using .NET serialization (binary formater is used).</para></remarks>
    /// 
    public static VeloActivationNetworkCommittee Load(string fileName)
    {
      FileStream stream = new FileStream(fileName, FileMode.Open, FileAccess.Read, FileShare.Read);
      VeloActivationNetworkCommittee committee = Load(stream);
      stream.Close();

      return committee;
    }

    /// <summary>
    /// Load ANN committee from specified file.
    /// </summary>
    /// 
    /// <param name="stream">Stream to load ANN committee from.</param>
    /// 
    /// <returns>Returns instance of VeloActivationNetworkComittee class with all properties initialized from file.</returns>
    /// 
    /// <remarks><para>ANN committee is loaded from file using .NET serialization (binary formater is used).</para></remarks>
    /// 
    public static VeloActivationNetworkCommittee Load(Stream stream)
    {
      IFormatter formatter = new BinaryFormatter();
      VeloActivationNetworkCommittee committee = (VeloActivationNetworkCommittee)formatter.Deserialize(stream);
      return committee;
    }
  }
}
