using System;
using System.ComponentModel;
using System.Drawing;
using System.IO;
using System.Reflection;
using System.Windows.Forms;
using Grid;
using Mathe.Geom;
using Tracking;

namespace LeastCostPathUI
{
  public partial class CntConfig : UserControl
  {
    private class StepImage
    {
      public Image Image;
      public Step Step;
      public StepImage(Step step, Image img)
      {
        Step = step;
        Image = img;
      }
    }

    public CntConfig()
    {
      InitializeComponent();

      txtResol.Text = string.Format("{0:N1}", 1.0);
      lstStep.Items.AddRange(new object[]{
        new StepImage(Step.Step16, Images.Step16),
        new StepImage(Step.Step8, Images.Step8),
        new StepImage(Step.Step4, Images.Step4),
        });
      lstStep.SelectedIndex = 0;

      SetStepCost(Common.CostPath_StepCost);
    }

    public StepCostHandler StepCost
    {
      get { return _stepCost; }
    }
    private void SetStepCost(StepCostHandler stepCost)
    {
      _stepCost = stepCost;
      txtCost.Text = Common.GetMethodName(_stepCost.Method);
      ttp.SetToolTip(txtCost, txtCost.Text);
    }

    public string HeightName
    { get { return txtHeight.Text; } }
    public DoubleGrid HeightGrid
    {
      get
      {
        if (_grdHeight == null)
        {
          if (File.Exists(txtHeight.Text))
          {
            _grdHeight = DoubleGrid.FromAsciiFile(txtHeight.Text, 0, 0.01, typeof(double));
          }
        }
        return _grdHeight;
      }
    }

    public string MapName
    { get { return txtMap.Text; } }

    public Map Map
    {
      get
      {
        if (_map == null)
        {
          if (File.Exists(txtMap.Text))
          {
            if (rdbVelocity.Checked)
            {
              _map = new VelocityMap(txtMap.Text);
            }
            else if (rdbFeatures.Checked)
            {
              const string forestSuffix = "_forest";
              const string waterSuffix = "_water";
              const string undergrowthSuffix = "_undergrowth";
              const string pathsSuffix = "_paths";
              const string groundSuffix = "_ground";
              const string passableSuffix = "_passable";

              string forest = txtMap.Text;

              string directory = Path.GetDirectoryName(forest) + Path.DirectorySeparatorChar;
              string extension = Path.GetExtension(forest);
              string ForestFileName = Path.GetFileNameWithoutExtension(forest);
              string filenamePrefix = ForestFileName.Substring(0, ForestFileName.Length - forestSuffix.Length);

              string water = directory + filenamePrefix + waterSuffix + extension;
              string undergrowth = directory + filenamePrefix + undergrowthSuffix + extension;
              string paths = directory + filenamePrefix + pathsSuffix + extension;
              string ground = directory + filenamePrefix + groundSuffix + extension;
              string passable = directory + filenamePrefix + passableSuffix + extension;

              _map = new FeatureMap(forest, water, undergrowth, paths, ground, passable);
            }
          }
        }
        return _map;
      }
    }

    public double Resolution
    {
      get { return double.Parse(txtResol.Text); }
      set { txtResol.Text = string.Format("{0:N1}", value); }
    }

    public Step Step
    {
      get
      {
        StepImage step = (StepImage)lstStep.SelectedItem;
        return step.Step;
      }
    }

    public bool CorridorMode
    {
      get { return chbEnableCorridor.Checked; }
    }

    public string GpxFile
    {
      get { return txtGPS.Text; }
    }

    public double CorridorWidth
    {
      get
      {
        double width;
        bool success = double.TryParse(txtWidth.Text, out width);
        if (success)
        {
          return width;
        }
        return 0;
      }
    }

    public Polyline NearLine
    {
      get
      {
        if (_nearLine == null)
        {
          if (CorridorMode)
          {
            if (File.Exists(GpxFile))
            {
              _nearLine = Tracking.Track.GpxToPolyline(GpxFile);
            }
            else
            {
              throw new FileNotFoundException("Could not find GPX file.");
            }
          }
        }
        return _nearLine;
      }
    }

    private DoubleGrid _grdHeight;
    private Map _map;
    private Polyline _nearLine;

    private StepCostHandler _stepCost;

    private void txtResol_Validating(object sender, CancelEventArgs e)
    {
      try
      {
        double d = Convert.ToDouble(txtResol.Text);
        if (d < 0)
        { e.Cancel = true; }
      }
      catch
      { e.Cancel = true; }
    }

    private void lstStep_DrawItem(object sender, DrawItemEventArgs e)
    {
      StepImage step = (StepImage)lstStep.Items[e.Index];
      e.Graphics.DrawImage(step.Image, 4, 2 + e.Bounds.Y);
    }

    private void btnStepCost_Click(object sender, EventArgs e)
    {
      WdgCustom wdg = new WdgCustom();
      if (wdg.ShowDialog(this) != DialogResult.OK)
      { return; }

      MethodInfo method = wdg.StepCostMethod;
      SetStepCost((StepCostHandler)Delegate.CreateDelegate(typeof(StepCostHandler), method));
    }

    private void btnHeight_Click(object sender, EventArgs e)
    {
      try
      {
        dlgOpen.Filter = "height grid files|*.asc;*.agr;*.grd;*.txt|*.asc|*.asc|*.agr|*.agr|*.grd|*.grd|All Files|*.*";
        if (dlgOpen.ShowDialog() == DialogResult.OK)
        {
          txtHeight.Text = dlgOpen.FileName;
        }
      }
      catch (Exception exp)
      { MessageBox.Show(exp.Message + "\n" + exp.StackTrace); }
    }

    private void btnMap_Click(object sender, EventArgs e)
    {
      try
      {
        dlgOpen.Filter = "map grid files (*.tif)|*.tif";
        if (dlgOpen.ShowDialog() == DialogResult.OK)
        {
          txtMap.Text = dlgOpen.FileName;
        }
      }
      catch (Exception exp)
      { MessageBox.Show(exp.Message + "\n" + exp.StackTrace); }
    }

    private void txtHeight_TextChanged(object sender, EventArgs e)
    {
      _grdHeight = null;
    }

    private void txtMap_TextChanged(object sender, EventArgs e)
    {
      _map = null;
    }

    private void chbEnableGPS_CheckedChanged(object sender, EventArgs e)
    {
      if (chbEnableCorridor.Checked)
      {
        lblWidth.Enabled = true;
        txtWidth.Enabled = true;
        txtWidth.Text = "5";
        txtGPS.Enabled = true;
        btnGPS.Enabled = true;
        txtGPS.Text = "";
      }
      else
      {
        lblWidth.Enabled = false;
        txtWidth.Enabled = false;
        txtWidth.Text = "";
        txtGPS.Enabled = false;
        btnGPS.Enabled = false;
        txtGPS.Text = "";
        _nearLine = null;
      }
    }

    private void btnGPS_Click(object sender, EventArgs e)
    {
      try
      {
        dlgOpen.Filter = "GPS track files (*.gpx)|*.gpx";
        if (dlgOpen.ShowDialog() == DialogResult.OK)
        {
          txtGPS.Text = dlgOpen.FileName;
        }
      }
      catch (Exception exp)
      { MessageBox.Show(exp.Message + "\n" + exp.StackTrace); }
    }

    private void txtGPS_TextChanged(object sender, EventArgs e)
    {
      _nearLine = null;
    }

  }
}
