using System;
using System.ComponentModel;
using System.Windows.Forms;
using Grid;
using Mathe.Geom;

namespace LeastCostPathUI
{
  public partial class CntOutput : UserControl
  {
    private LeastCostPath _costPath;
    private bool _cancelled;
    private bool _disposed;

    public CntOutput()
    {
      InitializeComponent();
      cntRoute.CheckCostImage(true);
    }

    public void SetExtent(IBox box)
    {
      txtXMin.Text = box.Min.X.ToString("N1");
      txtYMin.Text = box.Min.Y.ToString("N1");
      txtXMax.Text = box.Max.X.ToString("N1");
      txtYMax.Text = box.Max.Y.ToString("N1");
    }

    public void SetStart(IPoint start)
    {
      cntFrom.SetPoint(start);
    }

    public void SetEnd(IPoint end)
    {
      cntTo.SetPoint(end);
    }

    public void SetMainPath(string mainPath)
    {
      cntFrom.SetMainPath(mainPath);
      cntTo.SetMainPath(mainPath);
      cntRoute.SetMainPath(mainPath);
    }

    public void SetNames(string from, string to)
    {
      cntFrom.SetName(from + "_");
      cntTo.SetName("_" + to);

      cntRoute.SetName(from + "_" + to);
      txtRouteShp.Text = from + "_" + to + "r.shp";
      txtRoute.Text = from + "_" + to + "r.tif";
    }

    private DoubleGrid _grdHeight;
    private Map _map;

    // These two are used when calculating the cost of a GPS recorded route choice
    // using the corridor approach.
    private Polyline _nearLine;
    private double _corridorWidth;

    private DoubleGrid _fromCost;
    private IntGrid _fromDir;

    private DoubleGrid _toCost;
    private IntGrid _toDir;

    private Point2D _from;
    private Point2D _to;

    public void Calc(DoubleGrid grHeight, Map map, 
      double resolution, Step step, StepCostHandler stepCost,
      Polyline nearLine, double corridorWidth)
    {
      try
      {
        _cancelled = false;
        Enable(false);

        double dX0 = Convert.ToDouble(txtXMin.Text);
        double dY0 = Convert.ToDouble(txtYMin.Text);
        double dX1 = Convert.ToDouble(txtXMax.Text);
        double dY1 = Convert.ToDouble(txtYMax.Text);
        double dResol = resolution;
        if (dY0 > dY1)
        {
          double t;

          t = dX0;
          dX0 = dX1;
          dX1 = t;
          txtXMin.Text = dX0.ToString();
          txtXMax.Text = dX1.ToString();
        }
        if (dY0 > dY1)
        {
          double t;

          t = dX0;
          dX0 = dX1;
          dX1 = t;
          txtXMin.Text = dX0.ToString();
          txtXMax.Text = dX1.ToString();
        }

        Box box = new Box(new Point2D(dX0, dY0), new Point2D(dX1, dY1));

        LeastCostPath costPath = new LeastCostPath(box, dResol, step);
        costPath.StepCost = stepCost;

        if (!costPath.EqualSettings(_costPath) ||
          map != _map || grHeight != _grdHeight ||
          nearLine != _nearLine || Math.Abs(corridorWidth - _corridorWidth) > 0.001)
        {
          _fromCost = null;
          _fromDir = null;
          _toCost = null;
          _toDir = null;

          _costPath = costPath;

          _map = map;
          _grdHeight = grHeight;
          _nearLine = nearLine;
          _corridorWidth = corridorWidth;
        }
        try
        {
          _costPath.Status -= costPath_Status;
          _costPath.Status += costPath_Status;

          if (_cancelled)
          { return; }

          Step fromStep = _costPath.Step;
          Step toStep = _costPath.Step;

          lblStep.Text = "FROM:";
          Point2D from = new Point2D(cntFrom.X, cntFrom.Y);
          if (_from == null || from.Dist2(_from) != 0 || _fromCost == null)
          {
            if (nearLine != null)
            {
              _costPath.CalcCost(from, _grdHeight, _map, out _fromCost, out _fromDir, false, nearLine, corridorWidth);
            }
            else
            {
              _costPath.CalcCost(from, _grdHeight, _map, out _fromCost, out _fromDir, false);
            }
            _from = from;
          }

          if (_cancelled)
          { return; }

          lblStep.Text = "TO:";
          Point2D to = new Point2D(cntTo.X, cntTo.Y);
          if (_to == null || to.Dist2(_to) != 0 || _toCost == null)
          {
            if (nearLine != null)
            {
              _costPath.CalcCost(to, _grdHeight, _map, out _toCost, out _toDir, true, nearLine, corridorWidth);
            }
            else
            {
              _costPath.CalcCost(to, _grdHeight, _map, out _toCost, out _toDir, true);
            }
            _to = to;
          }

          Step.ReverseEngineer(_toDir, _toCost);

          lblStep.Text = "SUM:";
          costPath_Status(this, new StatusEventArgs(null, null, 0, 0, 0, 0));
          DoubleGrid sum = _fromCost + _toCost;
          DoubleGrid route = sum - sum.Min();

          if (_cancelled) { return; }

          if (chkRoute.Checked || chkRouteShp.Checked)
          {
            lblStep.Text = "ROUTES:";
            costPath_Status(this, new StatusEventArgs(null, null, 0, 0, 0, 0));

            double lengthFact = 1.2;
            double lengthPct;
            if (double.TryParse(txtSlower.Text, out lengthPct))
            {
              lengthFact = 1 + lengthPct / 100.0;
            }
            double offset = 0.05;
            double offsetPct;
            if (double.TryParse(txtMinOffset.Text, out offsetPct))
            {
              offset = offsetPct / 100.0;
            }
            //CreateRouteImage(sum, cntRoute.FullName(txtRoute.Text));
            RouteTable routes = LeastCostPath.CalcBestRoutes(sum, _fromCost, _fromDir, fromStep,
              _toCost, _toDir, toStep, lengthFact, offset, costPath_Status);
            if (chkRoute.Checked)
            {
              IO.CreateRouteImage(routes, sum.Extent, cntRoute.FullName(txtRoute.Text));
            }
            if (chkRouteShp.Checked)
            {
              IO.CreateRouteShapes(routes, sum, _grdHeight, cntRoute.FullName(txtRouteShp.Text));
            }
          }

          if (_cancelled) { return; }

          lblStep.Text = "EXPORT:";
          costPath_Status(this, new StatusEventArgs(null, null, 0, 0, 0, 0));

          cntFrom.Export(_fromCost, _fromDir, _costPath.Step);
          cntTo.Export(_toCost, _toDir, _costPath.Step);
          cntRoute.Export(route, null, null);

          lblStep.Text = "< >";
          costPath_Status(this, new StatusEventArgs(null, null, 0, 0, 0, 0));
        }
        finally
        {
          _costPath.Status -= costPath_Status;
        }
      }
      finally
      {
        Enable(true);
      }
    }

    public void Cancel()
    {
      _cancelled = true;
      lblProgress.Text = "CANCELLED";
    }

    private void Enable(bool enable)
    {
      if (_disposed == false)
      {
        grpExtent.Enabled = enable;
        grpFrom.Enabled = enable;
        grpTo.Enabled = enable;
        grpRoute.Enabled = enable;
      }
    }

    void costPath_Status(object sender, StatusEventArgs args)
    {
      if (_cancelled)
      {
        args.Cancel = true;
        return;
      }
      if (args.CurrentStep % 1000 == 0)
      {
        if (TopLevelControl.Visible == false)
        {
          _cancelled = true;
        }
        else
        {
          if (args.TotalStep > 0)
          {
            lblProgress.Text = string.Format("Step {0:N0} of {1:N0}", args.CurrentStep, args.TotalStep);
          }
          else
          {
            lblProgress.Text = string.Format("...");
          }
          Application.DoEvents();
        }
      }
      if (_cancelled)
      {
        args.Cancel = true;
        return;
      }
    }

    private void txtDouble_Validating(object sender, CancelEventArgs e)
    {
      try
      {
        TextBox txt = (TextBox)sender;
        double d;
        if (double.TryParse(txt.Text, out d) == false)
        { e.Cancel = true; }
      }
      catch
      { e.Cancel = true; }
    }

    private void btnRoute_Click(object sender, EventArgs e)
    {
      cntRoute.dlgSave.Filter = "route image files (*r.tif)|*r.tif";
      if (cntRoute.dlgSave.ShowDialog() == DialogResult.OK)
      {
        txtRoute.Text = cntRoute.dlgSave.FileName;
        cntRoute.Synchrone(txtRoute);
      }
    }

    private void btnRouteShp_Click(object sender, EventArgs e)
    {
      cntRoute.dlgSave.Filter = "route shape files (*r.shp)|*r.shp";
      if (cntRoute.dlgSave.ShowDialog() == DialogResult.OK)
      {
        txtRouteShp.Text = cntRoute.dlgSave.FileName;
        cntRoute.Synchrone(txtRouteShp);
      }
    }
  }
}
