﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.Serialization;


namespace Grid
{
  /// <summary>
  /// Base class representing the orienteering map, ie. contains all the information
  /// that might affect running velocity, except for height/slope information.
  /// </summary>
  public abstract class Map : ILockable
  {
    /// <summary>
    /// Get the MapPoint at the given coordinates on the map
    /// </summary>
    /// <param name="x"></param>
    /// <param name="y"></param>
    /// <returns></returns>
    public abstract MapPoint this[int x, int y] { get; }

    public abstract MapPoint GetMapPoint(double x, double y);

    public abstract void LockBits();
    public abstract void UnlockBits();
  }

  /// <summary>
  /// Orienteering map represented by six FeatureGrids, each of which encodes
  /// a subset of all possible map features at every point on the map.
  /// </summary>
  public class FeatureMap : Map
  {
    public FeatureGrid ImpassableGrid { get; set; }
    public FeatureGrid ForestGrid { get; set; }
    public FeatureGrid WaterGrid { get; set; }
    public FeatureGrid UndergrowthGrid { get; set; }
    public FeatureGrid PathsGrid { get; set; }
    public FeatureGrid GroundGrid { get; set; }

    /// <summary>
    /// Constructs a FeatureMap with map features from a set of image files, each file describing
    /// a subset of all possible map features at each point on the map.
    /// </summary>
    /// <param name="forestName"></param>
    /// <param name="waterName"></param>
    /// <param name="undergrowthName"></param>
    /// <param name="pathsName"></param>
    /// <param name="groundName"></param>
    /// <param name="impassableName"></param>
    public FeatureMap(string forestName, string waterName, string undergrowthName,
      string pathsName, string groundName, string impassableName)
    {
        ForestGrid = FeatureGrid.FromImage(forestName);
        WaterGrid = FeatureGrid.FromImage(waterName);
        UndergrowthGrid = FeatureGrid.FromImage(undergrowthName);
        PathsGrid = FeatureGrid.FromImage(pathsName);
        GroundGrid = FeatureGrid.FromImage(groundName);
        ImpassableGrid = FeatureGrid.FromImage(impassableName);
    }

    /// <summary>
    /// Get the map features at the given coordinates on the map
    /// </summary>
    /// <param name="x"></param>
    /// <param name="y"></param>
    /// <returns></returns>
    public override MapPoint this[int x, int y]
    {
      get
      {
        double impassable = ImpassableGrid[x, y];
        double forest = ForestGrid[x, y];
        double water = WaterGrid[x, y];
        double undergrowth = UndergrowthGrid[x, y];
        double paths = PathsGrid[x, y];
        double ground = GroundGrid[x, y];

        return new MapPointFeatures(impassable, forest, water, undergrowth, paths, ground);
      }
    }

    /// <summary>
    /// Get the map features at the given coordinates on the map
    /// </summary>
    /// <param name="x"></param>
    /// <param name="y"></param>
    /// <returns></returns>
    public override MapPoint GetMapPoint(double x, double y)
    {
      double impassable = ImpassableGrid.Value(x, y, EGridInterpolation.nearest);
      double forest = ForestGrid.Value(x, y, EGridInterpolation.nearest);
      double water = WaterGrid.Value(x, y, EGridInterpolation.nearest);
      double undergrowth = UndergrowthGrid.Value(x, y, EGridInterpolation.nearest);
      double paths = PathsGrid.Value(x, y, EGridInterpolation.nearest);
      double ground = GroundGrid.Value(x, y, EGridInterpolation.nearest);

      return new MapPointFeatures(impassable, forest, water, undergrowth, paths, ground);
    }

    public override void LockBits()
    {
      ForestGrid.LockBits();
      WaterGrid.LockBits();
      UndergrowthGrid.LockBits();
      PathsGrid.LockBits();
      GroundGrid.LockBits();
      ImpassableGrid.LockBits();
    }

    public override void UnlockBits()
    {
      ForestGrid.UnlockBits();
      WaterGrid.UnlockBits();
      UndergrowthGrid.UnlockBits();
      PathsGrid.UnlockBits();
      GroundGrid.UnlockBits();
      ImpassableGrid.UnlockBits();
    }
  }

  /// <summary>
  /// Orienteering map represented by a VelocityGrid, which encodes the flat terrain
  /// velocity at each point on the map.
  /// </summary>
  public class VelocityMap : Map
  {
    public VelocityGrid VelocityGrid { get; set; }

    /// <summary>
    /// Constructs a Map with velocities from the supplied image file.
    /// </summary>
    /// <param name="velocityName"></param>
    public VelocityMap(string velocityName)
    {
      VelocityGrid = VelocityGrid.FromImage(velocityName);
    }

    /// <summary>
    /// Get the flat terrain velocity at the given coordinates on the map
    /// </summary>
    /// <param name="x"></param>
    /// <param name="y"></param>
    /// <returns></returns>
    public override MapPoint this[int x, int y]
    {
      get
      {
        double velocity = VelocityGrid[x, y];
        return new MapPointVelocity(velocity);
      }
    }

    /// <summary>
    /// Get the flat terrain velocity at the given coordinates on the map
    /// </summary>
    /// <param name="x"></param>
    /// <param name="y"></param>
    /// <returns></returns>
    public override MapPoint GetMapPoint(double x, double y)
    {
      double velocity = VelocityGrid.Value(x, y, EGridInterpolation.bilinear);
      return new MapPointVelocity(velocity);
    }

    public override void LockBits()
    {
      VelocityGrid.LockBits();
    }

    public override void UnlockBits()
    {
      VelocityGrid.UnlockBits();
    }
  }

  /// <summary>
  /// Base class representing map data at some point on the map, ie. the information
  /// necessary to determine velocity in any direction at that point, except for
  /// height/slope information.
  /// </summary>
  public abstract class MapPoint
  {
    /// <summary>
    /// Make the map point impassable (not possible to run through)
    /// </summary>
    public abstract void MakeImpassable();
  }

  /// <summary>
  /// Represents the map features at a point on the map as six double values,
  /// each encoding a subset of all possible map features.
  /// </summary>
  public class MapPointFeatures : MapPoint
  {
    public double Passable { get; set; }
    public double Forest { get; set; }
    public double Water { get; set; }
    public double Undergrowth { get; set; }
    public double Paths { get; set; }
    public double Ground { get; set; }

    public MapPointFeatures (double passable, double forest, double water,
      double undergrowth, double paths, double ground)
    {
      Passable = passable;
      Forest = forest;
      Water = water;
      Undergrowth = undergrowth;
      Paths = paths;
      Ground = ground;
    }

    public override void MakeImpassable()
    {
      Passable = 0;
    }
  }

  // Represents the flat terrain running velocity at a point on the map
  public class MapPointVelocity : MapPoint
  {
    public double Velocity { get; set; }

    public MapPointVelocity(double velocity)
    {
      Velocity = velocity;
    }

    public override void MakeImpassable()
    {
      Velocity = 0;
    }
  }


  /// <summary>
  /// Contains the map features present at a point in the search grid. The map features
  /// are represented as a number of enum values.
  /// </summary>

  [Serializable()]
  public struct ClassifiedMapPointFeatures : ISerializable
  {
    public Passable Passable { get; set; }
    public Forest Forest { get; set; }
    public Water Water { get; set; }
    public Undergrowth Undergrowth { get; set; }
    public Paths Paths { get; set; }
    public Ground Ground { get; set; }

        /// <summary>
    /// Classifies the double values of a MapFeatures object as actual map features (enum values)
    /// </summary>
    /// <param name="mf">The MapFeatures object to be classified</param>
    public ClassifiedMapPointFeatures(MapPointFeatures mf)
      : this()
    {
      // Classify passable
      if (mf.Passable < 0.25)
      {
        this.Passable = Passable.Forbidden;
      }
      else if (mf.Passable < 0.75)
      {
        this.Passable = Passable.ForbiddenButPathsAllowed;
      }
      else
      {
        this.Passable = Passable.Passable;
      }

      // Classify forest
      if (mf.Forest < 0.05)
      {
        this.Forest = Forest.VeryDifficultToRun;
      }
      else if (mf.Forest < 0.15)
      {
        this.Forest = Forest.DifficultToRun;
      }
      else if (mf.Forest < 0.25)
      {
        this.Forest = Forest.SlowRunning;
      }
      else if (mf.Forest < 0.35)
      {
        this.Forest = Forest.RoughOpenScatteredTrees;
      }
      else if (mf.Forest < 0.45)
      {
        this.Forest = Forest.RoughOpen;
      }
      else if (mf.Forest < 0.55)
      {
        this.Forest = Forest.OpenScatteredTrees;
      }
      else if (mf.Forest < 0.65)
      {
        this.Forest = Forest.Open;
      }
      else if (mf.Forest < 0.75)
      {
        this.Forest = Forest.BareRock;
      }
      else if (mf.Forest < 0.85)
      {
        this.Forest = Forest.OpenSandyGround;
      }
      else
      {
        this.Forest = Forest.EasyRunning;
      }

      // Classify water
      if (mf.Water < 0.1)
      {
        this.Water = Water.Marsh;
      }
      else if (mf.Water < 0.3)
      {
        this.Water = Water.IndistinctMarsh;
      }
      else if (mf.Water < 0.45)
      {
        this.Water = Water.NarrowMarsh;
      }
      else if (mf.Water < 0.55)
      {
        this.Water = Water.LakeOrRiver;
      }
      else if (mf.Water < 0.65)
      {
        this.Water = Water.Pond;
      }
      else if (mf.Water < 0.75)
      {
        this.Water = Water.CrossableWatercourse;
      }
      else if (mf.Water < 0.85)
      {
        this.Water = Water.CrossableSmallWatercourse;
      }
      else if (mf.Water < 0.95)
      {
        this.Water = Water.MinorWaterChannel;
      }
      else
      {
        this.Water = Water.None;
      }

      // Classify underbrush
      if (mf.Undergrowth < 0.25)
      {
        this.Undergrowth = Undergrowth.DifficultToRun;
      }
      else if (mf.Undergrowth < 0.75)
      {
        this.Undergrowth = Undergrowth.SlowRunning;
      }
      else
      {
        this.Undergrowth = Undergrowth.None;
      }

      // Classify path
      if (mf.Paths < 0.1)
      {
        this.Paths = Paths.Road;
      }
      else if (mf.Paths < 0.3)
      {
        this.Paths = Paths.VehicleTrack;
      }
      else if (mf.Paths < 0.5)
      {
        this.Paths = Paths.Footpath;
      }
      else if (mf.Paths < 0.7)
      {
        this.Paths = Paths.SmallPath;
      }
      else if (mf.Paths < 0.85)
      {
        this.Paths = Paths.LessDistinctSmallPath;
      }
      else if (mf.Paths < 0.95)
      {
        this.Paths = Paths.NarrowRide;
      }
      else
      {
        this.Paths = Paths.None;
      }

      // Classify ground
      if (mf.Ground < 0.1)
      {
        this.Ground = Ground.PassableRockFace;
      }
      else if (mf.Ground < 0.3)
      {
        this.Ground = Ground.EarthBank;
      }
      else if (mf.Ground < 0.5)
      {
        this.Ground = Ground.BoulderField;
      }
      else if (mf.Ground < 0.7)
      {
        this.Ground = Ground.StonyGround;
      }
      else if (mf.Ground < 0.9)
      {
        this.Ground = Ground.BrokenGround;
      }
      else
      {
        this.Ground = Ground.None;
      }
    }

    public ClassifiedMapPointFeatures(Forest forest, Water water, Undergrowth undergrowth,
      Paths paths, Ground ground, Passable passable) : this()
    {
      Forest = forest;
      Water = water;
      Undergrowth = undergrowth;
      Paths = paths;
      Ground = ground;
      Passable = passable;
    }

    public ClassifiedMapPointFeatures(SerializationInfo info, StreamingContext context) : this()
    {
      this.Forest = (Forest)info.GetValue("Forest", typeof(Forest));
      this.Water = (Water)info.GetValue("Water", typeof(Water));
      this.Undergrowth = (Undergrowth)info.GetValue("Undergrowth", typeof(Undergrowth));
      this.Paths = (Paths)info.GetValue("Paths", typeof(Paths));
      this.Ground = (Ground)info.GetValue("Ground", typeof(Ground));
      this.Passable = (Passable)info.GetValue("Passable", typeof(Passable));
    }

    public void GetObjectData(SerializationInfo info, StreamingContext context)
    {
      info.AddValue("Forest", this.Forest);
      info.AddValue("Water", this.Water);
      info.AddValue("Undergrowth", this.Undergrowth);
      info.AddValue("Paths", this.Paths);
      info.AddValue("Ground", this.Ground);
      info.AddValue("Passable", this.Passable);
    }

    public bool Equals(ClassifiedMapPointFeatures other)
    {
      bool equal;
      if (Forest != other.Forest)
      {
        equal = false;
        // Hack to ignore BareRock and OpenSandyGround (consider them the same as EasyRunning)
        if ((Forest == Forest.EasyRunning || Forest == Forest.BareRock || Forest == Forest.OpenSandyGround)
            && (other.Forest == Forest.EasyRunning || other.Forest == Forest.BareRock || other.Forest == Forest.OpenSandyGround))
        {
          equal = true;
        }
        if (!equal)
        {
          return false;
        }
      }

      if (Water != other.Water)
      {
        equal = false;
        // Hack to ignore CrossableWatercourse, CrossableSmallWatercourse, MinorWaterChanner (consider them the same as None)
        if ((Water == Water.None || Water == Water.CrossableWatercourse || Water == Water.CrossableSmallWatercourse || Water == Water.MinorWaterChannel)
          && (other.Water == Water.None || other.Water == Water.CrossableWatercourse || other.Water == Water.CrossableSmallWatercourse || other.Water == Water.MinorWaterChannel))
        {
          equal = true;
        }
        if (!equal)
        {
          return false;
        }
      }

      if (Undergrowth != other.Undergrowth)
      {
        return false;
      }

      if (Paths != other.Paths)
      {
        if (Paths != other.Paths)
        {
          equal = false;
          // Hack to consider VehicleTrack and Road equally good
          if ((Paths == Paths.Road || Paths == Paths.VehicleTrack)
              && (other.Paths == Paths.Road || other.Paths == Paths.VehicleTrack))
          {
            equal = true;
          }
          if (!equal)
          {
            return false;
          }
        }
      }
      /* // Ignore ground for now
      if (Ground != other.Ground)
      {
        equal = false;
        // Hack to ignore Earthbank, PassableRockFace and BrokeGround (consider them the same as None)
        if ((Ground == Ground.None || Ground == Ground.EarthBank || Ground == Ground.PassableRockFace || Ground == Ground.BrokenGround)
          && (other.Ground == Ground.None || other.Ground == Ground.EarthBank || other.Ground == Ground.PassableRockFace || other.Ground == Ground.BrokenGround))
        {
          equal = true;
        }
        if (!equal)
        {
          return false;
        }
      }
       */
      return true;
    }
  }
}
