using System;
using System.Collections.Generic;

namespace Grid
{
  public class Step
  {
    private static readonly int[,] _step16 = new int[,]
      {         {-1, 2},        { 1, 2},
        {-2, 1},{-1, 1},{ 0, 1},{ 1, 1},{ 2, 1},
                {-1, 0},        { 1, 0},
        {-2,-1},{-1,-1},{ 0,-1},{ 1,-1},{ 2,-1},
                {-1,-2},        { 1,-2}
      };

    private static readonly int[,] _step8 = new int[,]
    {         
              {-1, 1},{ 0, 1},{ 1, 1},
              {-1, 0},        { 1, 0},
              {-1,-1},{ 0,-1},{ 1,-1}
    };

    private static readonly int[,] _step4 = new int[,]
    {         
                      { 0, 1},
              {-1, 0},        { 1, 0},
                      { 0,-1}
    };


    public static Step Step16
    {
      get
      {
        return new Step((int[,])_step16.Clone());
      }
    }

    public static Step Step8
    {
      get
      {
        return new Step((int[,])_step8.Clone());
      }
    }

    public static Step Step4
    {
      get
      {
        return new Step((int[,])_step4.Clone());
      }
    }


    private int _count;
    private int[,] _steps;
    private double[] _angles;
    private double[] _distance;

    public int this[int i, int dir]
    {
      get { return _steps[i, dir]; }
    }

    public Step(int[,] steps)
    {
      _steps = steps;

      _count = _steps.GetLength(0);
      InitDistance();
      InitAngles();
    }
    public int Count
    {
      get { return _count; }
    }

    private void InitDistance()
    {
      double[] distance = new double[Count];
      for (int i = 0; i < Count; i++)
      {
        distance[i] =
          Math.Sqrt(this[i, 0] * this[i, 0] + this[i, 1] * this[i, 1]);
      }
      _distance = distance;
    }

    public double GetDistance(int i)
    {
      return _distance[i];
    }

    private void InitAngles()
    {
      double[] angles = new double[Count];
      for (int i = 0; i < Count; i++)
      {
        angles[i] = Math.Atan2(this[i, 1], this[i, 0]);
      }
      _angles = angles;
    }

    public double GetAngle(int i)
    {
      return _angles[i];
    }
    public DoubleGrid this[IntGrid grd]
    {
      get { return new DoubleGrid(_angles, grd); }
    }

    private class StepInfo
    {
      public int Ix;
      public int Iy;
      public int Dir;
      public double Cost;

      public StepInfo(int ix, int iy, int dir, double cost)
      {
        Ix = ix;
        Iy = iy;
        Dir = dir;
        Cost = cost;
      }
      public override string ToString()
      {
        string s = string.Format("{0,2}: {1} {2}; {3:N1}", Dir, Ix, Iy, NormedCost());
        return s;
      }

      public double NormedCost()
      {
        return Cost / Math.Sqrt(Ix * Ix + Iy * Iy);
      }
    }
    private class StepInfoBuilder
    {
      public readonly List<StepInfo> Candidates = new List<StepInfo>();
      StepInfo _assigned = new StepInfo(0, 0, 0, 0);
      public bool IsAssigned()
      {
        return (_assigned.Ix != 0 || _assigned.Iy != 0);
      }
      public override string ToString()
      {
        return _assigned.ToString();
      }
      public bool IsInitialized()
      {
        return Candidates.Count > 0;
      }
      public bool TryAssignDir()
      {
        if (Candidates.Count == 1)
        {
          if (_assigned != Candidates[0])
          {
            _assigned = Candidates[0];
          }
          return true;
        }
        return false;
      }

      public void Analyze(StepInfo step, DoubleGrid costGrid)
      {
        for (int iDir = 0; iDir < Candidates.Count; iDir++)
        {
          StepInfo candidate = Candidates[iDir];
          int ix = step.Ix - candidate.Ix;
          int iy = step.Iy - candidate.Iy;
          if (ix < 0 || ix >= costGrid.Extent.Nx ||
            iy < 0 || iy >= costGrid.Extent.Ny)
          {
            Candidates.RemoveAt(iDir);
            continue;
          }
          double cost = step.Cost - costGrid[ix, iy];
          if (cost <= 0)
          {
            Candidates.RemoveAt(iDir);
            continue;
          }
          else
          {
            candidate.Cost += cost;
          }
        }

      }

      public void Initialize(StepInfo step, List<StepInfo> pres, Dictionary<int, StepInfoBuilder> dirs)
      {
        foreach (StepInfo pre in pres)
        {
          int dx = step.Ix - pre.Ix;
          int dy = step.Iy - pre.Iy;

          if (Mathe.Utils.Gcd(dx, dy) != 1)
          { continue; }

          bool isAssigned = false;
          foreach (StepInfoBuilder t in dirs.Values)
          {
            if (t.IsAssigned() && t._assigned.Ix == dx && t._assigned.Iy == dy)
            {
              isAssigned = true;
              break;
            }
          }
          if (!isAssigned)
          {
            Candidates.Add(new StepInfo(dx, dy, step.Dir, step.Cost - pre.Cost));
          }
        }
      }

      public StepInfo AssignDirFromMaxCost()
      {
        StepInfo max = null;
        foreach (StepInfo candidate in Candidates)
        {
          if (max == null || candidate.NormedCost() > max.NormedCost())
          { max = candidate; }
        }
        Candidates.Clear();
        Candidates.Add(max);
        _assigned = max;
        return max;
      }
    }
    public static Step ReverseEngineer(IntGrid dirGrid, DoubleGrid costGrid)
    {
      const int size = 1024;
      SortedList<double, IList<StepInfo>> costs = new SortedList<double, IList<StepInfo>>();
      Dictionary<int, StepInfoBuilder> dirs = new Dictionary<int, StepInfoBuilder>();
      LoadAnalyzeData(dirGrid, costGrid, dirs, costs, size);

      GridExtent extent = dirGrid.Extent;

      int nCost = costs.Count;
      List<StepInfo> pres = new List<StepInfo>(size);
      for (int iCost = 0; iCost < nCost; iCost++)
      {
        IList<StepInfo> steps = costs.Values[iCost];
        foreach (StepInfo step in steps)
        {
          StepInfoBuilder dir_ = dirs[step.Dir];
          if (dir_.IsAssigned())
          { continue; }

          if (dir_.IsInitialized())
          {
            dir_.Analyze(step, costGrid);
          }
          else
          {
            dir_.Initialize(step, pres, dirs);
          }
          dir_.TryAssignDir();
        }

        foreach (StepInfo step in steps)
        {
          pres.Add(step);
        }
      }
      //
      for (int ix = 0; ix < extent.Nx; ix++)
      {
        for (int iy = 0; iy < extent.Ny; iy++)
        {
          StepInfo step = new StepInfo(ix, iy, dirGrid[ix, iy], costGrid[ix, iy]);
          StepInfoBuilder dir = dirs[step.Dir];
          if (!dir.IsAssigned() && dir.Candidates.Count > 0)
          {
            dir.Analyze(step, costGrid);
            if (dir.TryAssignDir())
            {
              foreach (StepInfoBuilder other in dirs.Values)
              {
                if (other == dir)
                { continue; }
                for (int i = 0; i < other.Candidates.Count; i++)
                {
                  StepInfo otherCand = other.Candidates[i];
                  if (otherCand.Ix == dir.Candidates[0].Ix && otherCand.Iy == dir.Candidates[0].Iy)
                  {
                    other.Candidates.RemoveAt(i);
                  }
                }
              }
            }
          }
        }
      }

      List<int> stepIndices = new List<int>(dirs.Count - 1);
      foreach (int stepIndex in dirs.Keys)
      {
        if (stepIndex < 0 || stepIndex > dirs.Count)
        { continue; }
        stepIndices.Add(stepIndex);
      }
      stepIndices.Sort();
      List<StepInfoBuilder> sortedSteps = new List<StepInfoBuilder>(stepIndices.Count);
      foreach (int i in stepIndices)
      {
        sortedSteps.Add(dirs[i]);
      }
      foreach (StepInfoBuilder step in sortedSteps)
      {
        if (step.IsAssigned())
        { continue; }

        step.AssignDirFromMaxCost();
      }
      int nSteps = sortedSteps.Count;
      int[,] result = new int[sortedSteps.Count, 2];
      for (int iStep = 0; iStep < nSteps; iStep++)
      {
        StepInfo step = sortedSteps[iStep].Candidates[0];
        if (step.Dir != iStep + 1) throw new InvalidOperationException(string.Format("Direction {0} is missing", iStep + 1));
        result[iStep, 0] = step.Ix;
        result[iStep, 1] = step.Iy;
      }
      return new Step(result);
    }

    private static void LoadAnalyzeData(IntGrid dirGrid, DoubleGrid costGrid,
      Dictionary<int, StepInfoBuilder> dirs,
      SortedList<double, IList<StepInfo>> costs, int size)
    {
      GridExtent extent = dirGrid.Extent;

      for (int ix = 0; ix < extent.Nx; ix++)
      {
        for (int iy = 0; iy < extent.Ny; iy++)
        {
          double cost = costGrid[ix, iy];
          int dir = dirGrid[ix, iy];
          if (dirs.ContainsKey(dir) == false)
          {
            StepInfoBuilder list = new StepInfoBuilder();
            dirs.Add(dir, list);
          }

          if (costs.Count < size || cost <= costs.Keys[size - 1])
          {
            IList<StepInfo> steps;
            if (costs.TryGetValue(cost, out steps) == false)
            {
              if (costs.Count == size)
              {
                costs.RemoveAt(size - 1);
              }
              steps = new List<StepInfo>();
              costs.Add(cost, steps);
            }
            steps.Add(new StepInfo(ix, iy, dir, cost));
          }
        }
      }
    }
  }
}