using System;
using System.Collections.Generic;
using Mathe.Geom;
using Grid;

namespace GridTest
{
  public abstract class Step
  {
    private readonly int[,] _step;

    private readonly int _nSteps;

    public Step(int[,] step)
    {
      _step = step;
      _nSteps = step.GetLength(0);
    }

    public int this[int i, int dir]
    {
      get { return _step[i, dir]; }
    }

    public int Count
    {
      get { return _nSteps; }
    }

  }

  public class Step16 : Step
  {
    private static readonly int[,] _step = new int[,]
    {         {-1, 2},        { 1, 2},
      {-2, 1},{-1, 1},{ 0, 1},{ 1, 1},{ 2, 1},
              {-1, 0},        { 1, 0},
      {-2,-1},{-1,-1},{ 0,-1},{ 1,-1},{ 2,-1},
              {-1,-2},        { 1,-2}
    };
    public Step16()
      :
      base(_step)
    { }
  }

  public class Step8 : Step
  {
    private static readonly int[,] _step = new int[,]
    {         
              {-1, 1},{ 0, 1},{ 1, 1},
              {-1, 0},        { 1, 0},
              {-1,-1},{ 0,-1},{ 1,-1}
    };
    public Step8()
      :
      base(_step)
    { }
  }

  public class Step4 : Step
  {
    private static readonly int[,] _step = new int[,]
    {         
                      { 0, 1},
              {-1, 0},        { 1, 0},
                      { 0,-1}
    };
    public Step4()
      : base(_step)
    { }
  }

  public class StepDistance
  {
    private double[] _stepDistValues = null;

    public StepDistance(Step step)
    {
      int n = step.Count;
      _stepDistValues = new double[n];
      for (int i = 0; i < n; i++)
      {
        _stepDistValues[i] =
          Math.Sqrt(step[i, 0] * step[i, 0] + step[i, 1] * step[i, 1]);
      }
    }
    public double this[int i]
    {
      get { return _stepDistValues[i]; }
    }
  }

  public class StepAngle
  {
    private double[] _stepAngleValues = null;

    public StepAngle(Step step)
    {
      int n = step.Count;
      _stepAngleValues = new double[n];
      for (int i = 0; i < n; i++)
      {
        _stepAngleValues[i] = Math.Atan2(step[i, 1], step[i, 0]);
      }
    }
    public double this[int i]
    {
      get { return _stepAngleValues[i]; }
    }

    public DoubleGrid this[IntGrid grd]
    {
      get { return new DoubleGrid(_stepAngleValues, grd); }
    }
  }

  public class LeastCostPath
  {
    public class NodeInfo
    {
      public int X, Y;
      LeastCostPath _env;
      private double _tx, _ty;
      private double _height;
      private double _velocity;
      private double _slope2;
      public int IdDir;
      public int CurrentStep;

      private double _cost;
      private NodeInfo _from;

      public double Cost
      {
        get { return _cost; }
        set { _cost = value; }
      }

      public NodeInfo(LeastCostPath environment)
      {
        _env = environment;
        _velocity = -1;
        _slope2 = -1;
      }

      public NodeInfo FromNode
      {
        get { return _from; }
        set
        {
          IdDir = value.CurrentStep;
        }
      }

      public double Height
      {
        get
        {
          if (_velocity < 0)
          { GetHeightVelocity(); }
          return _height;
        }
      }

      public double Velocity
      {
        get
        {
          if (_velocity < 0)
          { GetHeightVelocity(); }
          return _velocity;
        }
      }

      public double Slope2
      {
        get
        {
          if (_slope2 < 0)
          {
            if (_velocity < 0)
            {
              _tx = _env._dX0 + X * _env._dDx;
              _ty = _env._dY0 + Y * _env._dDy;
            }

            _slope2 = _env._heightGrid.Slope2(_tx, _ty);
          }

          return _slope2;
        }
      }
      private void GetHeightVelocity()
      {
        if (_slope2 < 0)
        {
          _tx = _env._dX0 + X * _env._dDx;
          _ty = _env._dY0 + Y * _env._dDy;
        }

        _height = _env._heightGrid.Value(_tx, _ty);
        _velocity = _env._veloGrid.Value(_tx, _ty);
      }

      public int CompareWithoutCostTo(NodeInfo other)
      {
        if (X < other.X)
        { return -1; }
        else if (X > other.X)
        { return 1; }
        else if (Y < other.Y)
        { return -1; }
        else if (Y > other.Y)
        { return 1; }

        return 0;
        //int i;
        //i = X.CompareTo(other.X);
        //if (i != 0)
        //{ return i; }

        //i = Y.CompareTo(other.Y);

        //return i;
      }

      public IEnumerable<NodeInfo> RawNeighbors_
      {
        get
        {
          Step step = _env._step;
          int n = step.Count;
          for (int iStep = 0; iStep < n; iStep++)
          {
            NodeInfo rawNeighbor = new NodeInfo(_env);

            rawNeighbor.X = X + step[iStep, 0];
            rawNeighbor.Y = Y + step[iStep, 1];

            // check range
            if (rawNeighbor.X < 0 || rawNeighbor.X >= _env._dXMax ||
                rawNeighbor.Y < 0 || rawNeighbor.Y >= _env._dYMax)
            {
              continue;
            }

            CurrentStep = iStep;
            yield return rawNeighbor;
          }
        }
      }
    }

    public StepCostHandler StepCost;
    public event StatusEventHandler Status;

    private Step _step;
    private StepDistance _stepDistance;
    private StepAngle _stepAngle;

    private int _dXMax, _dYMax;
    private double _dX0, _dY0, _dDx, _dDy;

    private DoubleGrid _costGrid;
    private IntGrid _dirGrid;
    private DoubleGrid _veloGrid;
    private DoubleGrid _heightGrid;

    private bool _invers;

    private IList<int[]> _stopList;
    private int _currentStep;
    private int _maxSteps;

    public class NodeComparer : IComparer<NodeInfo>
    {
      public int Compare(NodeInfo field0, NodeInfo field1)
      { return field0.CompareWithoutCostTo(field1); }
    }

    public class CostComparer : IComparer<NodeInfo>
    {
      #region IComparer Members

      public int Compare(NodeInfo field0, NodeInfo field1)
      {
        if (field0.Cost < field1.Cost)
        { return -1; }
        else if (field0.Cost > field1.Cost)
        { return 1; }
        //int i;
        //i = field0.Cost.CompareTo(field1.Cost);
        //if (i != 0)
        //{ return i; }

        return field0.CompareWithoutCostTo(field1);
      }

      #endregion
    }

    protected void BuildTree(NodeInfo startNode)
    {
      SortedDictionary<NodeInfo, NodeInfo> nodes =
        new SortedDictionary<NodeInfo, NodeInfo>(new NodeComparer());
      SortedDictionary<NodeInfo, double> costs =
        new SortedDictionary<NodeInfo, double>(new CostComparer());

      startNode.Cost = 0;

      nodes.Add(startNode, startNode);
      costs.Add(startNode, 0);

      NodeInfo minCost;
      minCost = startNode;

      while (minCost != null)
      {
        costs.Remove(minCost);
        nodes.Remove(minCost);

        if (PreAssign(minCost) == false)
        {
          break;
        }
        AssignNewCosts(minCost, nodes, costs);

        minCost = null;
        foreach (NodeInfo cost in costs.Keys)
        {
          minCost = cost;
          break;
        }
      }
      nodes.Clear();
      costs.Clear();
    }

    private void AssignNewCosts(NodeInfo startNode,
      SortedDictionary<NodeInfo, NodeInfo> nodes,
      SortedDictionary<NodeInfo, double> costs)
    {
      Step step = _step;
      int n = step.Count;
      for (int iStep = 0; iStep < n; iStep++)
      {
        NodeInfo rawNeighbor = new NodeInfo(this);

        rawNeighbor.X = startNode.X + step[iStep, 0];
        rawNeighbor.Y = startNode.Y + step[iStep, 1];

        // check range
        if (rawNeighbor.X < 0 || rawNeighbor.X >= _dXMax ||
            rawNeighbor.Y < 0 || rawNeighbor.Y >= _dYMax)
        {
          continue;
        }

        startNode.CurrentStep = iStep;

        NodeInfo neighbor;

        if (IsFixed(rawNeighbor))
        { continue; }

        if (nodes.TryGetValue(rawNeighbor, out neighbor) == false)
        {
          neighbor = rawNeighbor;
          nodes.Add(neighbor, neighbor);
        }

        double cost = Cost(startNode, neighbor);
        double costTotal = startNode.Cost + cost;
        if (neighbor == rawNeighbor || costTotal < neighbor.Cost)
        {
          if (neighbor != rawNeighbor)
          {
            costs.Remove(neighbor);
          }
          neighbor.Cost = costTotal;
          neighbor.FromNode = startNode;
          costs.Add(neighbor, costTotal);
        }
      }
    }


    public LeastCostPath(Step step, IBox box, double dx)
    {
      if (step == null)
      { step = new Step16(); }

      _step = step;
      _stepDistance = new StepDistance(_step);
      _stepAngle = new StepAngle(_step);

      _dX0 = box.Min.X;
      _dDx = dx;
      _dY0 = box.Max.Y;
      _dDy = -dx;

      _dXMax = (int)((box.Max.X - box.Min.X) / dx + 1);
      _dYMax = (int)((box.Max.Y - box.Min.Y) / dx + 1);
    }

    public void CalcCost(Point start,
      DoubleGrid heightGrid, DoubleGrid velocityGrid,
      string costGridName, string dirGridName)
    {
      CalcCost(start, heightGrid, velocityGrid, costGridName, dirGridName, false);
    }

    public void CalcCost(Point start,
      DoubleGrid heightGrid, DoubleGrid velocityGrid,
      string costGridName, string dirGridName, bool invers)
    {
      DoubleGrid costGrid;
      IntGrid dirGrid;

      CalcCost(start, heightGrid, velocityGrid, out costGrid, out dirGrid, invers);

      costGrid.Save(costGridName);
      dirGrid.Save(dirGridName);
    }

    public void CalcCost(Point start,
      DoubleGrid heightGrid, DoubleGrid velocityGrid,
      out DoubleGrid costGrid, out IntGrid dirGrid)
    {
      CalcCost(start, heightGrid, velocityGrid, out costGrid, out dirGrid, false);
    }

    public void CalcCost(Point start, Point end,
      DoubleGrid heightGrid, DoubleGrid velocityGrid,
      out DoubleGrid costGrid, out IntGrid dirGrid)
    {
      CalcCost(start, new Point[] { end }, heightGrid, velocityGrid,
        out costGrid, out dirGrid);
    }

    public void CalcCost(Point start, IList<Point> endList,
      DoubleGrid heightGrid, DoubleGrid velocityGrid,
      out DoubleGrid costGrid, out IntGrid dirGrid)
    {
      List<int[]> stopList = new List<int[]>(endList.Count);
      foreach (Point end in endList)
      {

        int[] stop = new int[] {
          (int)((end.X - _dX0) / _dDx),
          (int)((end.Y - _dY0) / _dDy)
        };
        stopList.Add(stop);
      }

      CalcCost(start, heightGrid, velocityGrid, out costGrid, out dirGrid, false,
        stopList);
    }

    public void CalcCost(Point start,
      DoubleGrid heightGrid, DoubleGrid velocityGrid,
      out DoubleGrid costGrid, out IntGrid dirGrid, bool invers)
    {
      CalcCost(start, heightGrid, velocityGrid, out costGrid, out dirGrid, invers, null);
    }

    private void CalcCost(Point start,
      DoubleGrid heightGrid, DoubleGrid velocityGrid,
      out DoubleGrid costGrid, out IntGrid dirGrid, bool invers,
      IList<int[]> stop)
    {
      try
      {
        if (velocityGrid is VelocityGrid)
        { ((VelocityGrid)velocityGrid).LockBits(); }

        costGrid = new DoubleGrid(_dXMax, _dYMax, typeof(short),
          _dX0, _dY0, _dDx, 0, 0.1);
        dirGrid = new IntGrid(_dXMax, _dYMax, typeof(sbyte), _dX0, _dY0, _dDx);

        NodeInfo startField;
        _costGrid = costGrid;
        _dirGrid = dirGrid;
        _heightGrid = heightGrid;
        _veloGrid = velocityGrid;

        _invers = invers;

        _stopList = stop;

        startField = new NodeInfo(this);
        startField.X = (int)((start.X - _dX0) / _dDx);
        startField.Y = (int)((start.Y - _dY0) / _dDy);
        //startField.Height = heightGrid.Value(start.X, start.Y, EGridInterpolation.bilinear);
        //startField.Velocity = velocityGrid.Value(start.X, start.Y, EGridInterpolation.bilinear);
        //startField.Cost = 0;
        startField.IdDir = -2;

        _currentStep = 0;
        _maxSteps = _dXMax * _dYMax;

        BuildTree(startField);

        if (Status != null)
        { Status(this, new StatusEventArgs(costGrid, dirGrid, -1, -1, _maxSteps, _maxSteps)); }
      }
      finally
      {
        if (velocityGrid is VelocityGrid)
        { ((VelocityGrid)velocityGrid).UnlockBits(); }
      }
    }

    public StepAngle StepAngle
    {
      get { return _stepAngle; }
    }

    public Polyline Path(IntGrid dirGrid, Point end)
    {
      int ix;
      int iy;
      dirGrid.Extent.GetNearest(end, out ix, out iy);
      Polyline path = new Polyline();

      int dir = dirGrid[ix, iy];
      path.AddFirst(dirGrid.Extent.CellCenter(ix, iy));
      while (dir > 0)
      {
        dir = dir - 1;
        ix -= _step[dir, 0];
        iy -= _step[dir, 1];

        dir = dirGrid[ix, iy];
        path.AddFirst(dirGrid.Extent.CellCenter(ix, iy));
      }
      return path;
    }

    public bool IsFixed(NodeInfo node)
    {
      return _dirGrid[node.X, node.Y] != 0;
    }
    protected bool PreAssign(NodeInfo next)
    {
      _costGrid[next.X, next.Y] = next.Cost;
      _dirGrid[next.X, next.Y] = next.IdDir + 1;

      if (Status != null)
      {
        StatusEventArgs args = new StatusEventArgs(_costGrid, _dirGrid,
          next.X, next.Y, _currentStep, _maxSteps);
        Status(this, args);
      }
      _currentStep++;

      return !Stop(_stopList, next);
    }

    protected double Cost(NodeInfo from, NodeInfo to)
    {
      double h0, h1, v0, v1, s0, s1;
      if (_invers)
      {
        CostAt(from, out h1, out v1);
        s1 = from.Slope2;
        CostAt(to, out h0, out v0);
        s0 = to.Slope2;
      }
      else
      {
        CostAt(from, out h0, out v0);
        s0 = from.Slope2;
        CostAt(to, out h1, out v1);
        s1 = to.Slope2;
      }
      double cost = StepCost(h0, v0, s0, h1, v1, s1, _stepDistance[from.CurrentStep] * _dDx);

      return cost;
    }

    private void CostAt(NodeInfo node, out double h, out double v)
    {
      h = node.Height;
      v = node.Velocity;
    }

    private bool Stop(IList<int[]> stop, NodeInfo field)
    {
      if (stop == null)
      { return false; }

      int i = stop.Count - 1;
      while (i >= 0)
      {
        int[] s = stop[i];
        if (field.X == s[0] && field.Y == s[1])
        { stop.RemoveAt(i); }
        i--;
      }

      return (stop.Count == 0);
    }
  }
}

namespace GridTest1
{
  public abstract class Step
  {
    private readonly int[,] _step;

    private readonly int _nSteps;

    public Step(int[,] step)
    {
      _step = step;
      _nSteps = step.GetLength(0);
    }

    public int this[int i, int dir]
    {
      get { return _step[i, dir]; }
    }

    public int Count
    {
      get { return _nSteps; }
    }

  }

  public class Step16 : Step
  {
    private static readonly int[,] _step = new int[,]
    {         {-1, 2},        { 1, 2},
      {-2, 1},{-1, 1},{ 0, 1},{ 1, 1},{ 2, 1},
              {-1, 0},        { 1, 0},
      {-2,-1},{-1,-1},{ 0,-1},{ 1,-1},{ 2,-1},
              {-1,-2},        { 1,-2}
    };
    public Step16():
      base(_step)
    {  }
  }

  public class Step8 : Step
  {
    private static readonly int[,] _step = new int[,]
    {         
              {-1, 1},{ 0, 1},{ 1, 1},
              {-1, 0},        { 1, 0},
              {-1,-1},{ 0,-1},{ 1,-1}
    };
    public Step8()
      :
      base(_step)
    { }
  }

  public class Step4 : Step
  {
    private static readonly int[,] _step = new int[,]
    {         
                      { 0, 1},
              {-1, 0},        { 1, 0},
                      { 0,-1}
    };
    public Step4()
      : base(_step)
    { }
  }

  public class StepDistance
  {
    private double[] _stepDistValues = null;

    public StepDistance(Step step)
    {
      int n = step.Count;
      _stepDistValues = new double[n];
      for (int i = 0; i < n; i++)
      {
        _stepDistValues[i] =
          Math.Sqrt(step[i, 0] * step[i, 0] + step[i, 1] * step[i, 1]);
      }
    }
    public double this[int i]
    {
      get { return _stepDistValues[i]; }
    }
  }

  public class StepAngle
  {
    private double[] _stepAngleValues = null;

    public StepAngle(Step step)
    {
      int n = step.Count;
      _stepAngleValues = new double[n];
      for (int i = 0; i < n; i++)
      {
        _stepAngleValues[i] = Math.Atan2(step[i, 1], step[i, 0]);
      }
    }
    public double this[int i]
    {
      get { return _stepAngleValues[i]; }
    }

    public DoubleGrid this[IntGrid grd]
    {
      get { return new DoubleGrid(_stepAngleValues, grd); }
    }
  }

  /*
  public class StatusEventArgs
  {
    private DoubleGrid _costGrid;
    private IntGrid _dirGrid;

    private int _currentStep;
    private int _totalStep;

    private int _currentX;
    private int _currentY;

    private bool _cancel;

    public StatusEventArgs(DoubleGrid costGrid, IntGrid dirGrid,
      int ix, int iy, int currentStep, int totalStep)
    {
      _costGrid = costGrid;
      _dirGrid = dirGrid;

      _currentX = ix;
      _currentY = iy;

      _currentStep = currentStep;
      _totalStep = totalStep;
    }
    public DoubleGrid CostGrid
    { get { return _costGrid; } }
    public IntGrid DirGrid
    { get { return _dirGrid; } }

    public int CurrentStep
    { get { return _currentStep; } }
    public int TotalStep
    { get { return _totalStep; } }

    public int CurrentX
    { get { return _currentX; } }
    public int CurrentY
    { get { return _currentY; } }

    public bool Cancel
    {
      get { return _cancel; }
      set { _cancel = value; }
    }
  }

  public delegate double StepCostHandler(double h0, double v0, double h1, double v1,
  double slope2, double dist);
  public delegate void StatusEventHandler(object sender, StatusEventArgs args);
*/

  public class LeastCostPath : Geometry.Network.LeastCostPath<LeastCostPath.NodeInfo>
  {
    public class NodeInfo : Geometry.Network.LeastCostPath<LeastCostPath.NodeInfo>.NodeInfoBase
    {
      public int X, Y;
      LeastCostPath _env;
      private double _tx, _ty;
      private double _height;
      private double _velocity;
      private double _slope2;
      public int IdDir;
      public int CurrentStep;

      public NodeInfo(LeastCostPath environment)
      {
        _env = environment;
        _velocity = -1;
        _slope2 = -1;
      }

      public override NodeInfo FromNode
      {
        get { return base.FromNode; }
        set
        {
          IdDir = value.CurrentStep;
        }
      }

      public double Height
      {
        get
        {
          if (_velocity < 0)
          { GetHeightVelocity(); }
          return _height;
        }
      }

      public double Velocity
      {
        get
        {
          if (_velocity < 0)
          { GetHeightVelocity(); }
          return _velocity;
        }
      }

      public double Slope2
      {
        get
        {
          if (_slope2 < 0)
          {
            if (_velocity < 0)
            {
              _tx = _env._dX0 + X * _env._dDx;
              _ty = _env._dY0 + Y * _env._dDy;
            }

            _slope2 = _env._heightGrid.Slope2(_tx, _ty);
          }

          return _slope2;
        }
      }
      private void GetHeightVelocity()
      {
        if (_slope2 < 0)
        {
          _tx = _env._dX0 + X * _env._dDx;
          _ty = _env._dY0 + Y * _env._dDy;
        }

        _height = _env._heightGrid.Value(_tx, _ty);
        _velocity = _env._veloGrid.Value(_tx, _ty);
      }

      public override int CompareWithoutCostTo(NodeInfo other)
      {
        if (X < other.X)
        { return -1; }
        else if (X > other.X)
        { return 1; }
        else if (Y < other.Y)
        { return -1; }
        else if (Y > other.Y)
        { return 1; }

        return 0;
        //int i;
        //i = X.CompareTo(other.X);
        //if (i != 0)
        //{ return i; }

        //i = Y.CompareTo(other.Y);

        //return i;
      }

      public override IEnumerable<NodeInfo> RawNeighbors
      {
        get
        {
          Step step = _env._step;
          int n = step.Count;
          for (int iStep = 0; iStep < n; iStep++)
          {
            NodeInfo neighbor = new NodeInfo(_env);

            neighbor.X = X + step[iStep, 0];
            neighbor.Y = Y + step[iStep, 1];

            // check range
            if (neighbor.X < 0 || neighbor.X >= _env._dXMax ||
                neighbor.Y < 0 || neighbor.Y >= _env._dYMax)
            {
              continue;
            }

            CurrentStep = iStep;
            yield return neighbor;
          }
        }
      }
    }

    public StepCostHandler StepCost;
    public event StatusEventHandler Status;

    private Step _step;
    private StepDistance _stepDistance;
    private StepAngle _stepAngle;

    private int _dXMax, _dYMax;
    private double _dX0, _dY0, _dDx, _dDy;

    private DoubleGrid _costGrid;
    private IntGrid _dirGrid;
    private DoubleGrid _veloGrid;
    private DoubleGrid _heightGrid;

    private bool _invers;

    private IList<int[]> _stopList;
    private int _currentStep;
    private int _maxSteps;

    public LeastCostPath(Step step, IBox box, double dx)
    {
      if (step == null)
      { step = new Step16(); }

      _step = step;
      _stepDistance = new StepDistance(_step);
      _stepAngle = new StepAngle(_step);

      _dX0 = box.Min.X;
      _dDx = dx;
      _dY0 = box.Max.Y;
      _dDy = -dx;

      _dXMax = (int)((box.Max.X - box.Min.X) / dx + 1);
      _dYMax = (int)((box.Max.Y - box.Min.Y) / dx + 1);
    }

    public void CalcCost(Point start,
      DoubleGrid heightGrid, DoubleGrid velocityGrid,
      string costGridName, string dirGridName)
    {
      CalcCost(start, heightGrid, velocityGrid, costGridName, dirGridName, false);
    }

    public void CalcCost(Point start,
      DoubleGrid heightGrid, DoubleGrid velocityGrid,
      string costGridName, string dirGridName, bool invers)
    {
      DoubleGrid costGrid;
      IntGrid dirGrid;

      CalcCost(start, heightGrid, velocityGrid, out costGrid, out dirGrid, invers);

      costGrid.Save(costGridName);
      dirGrid.Save(dirGridName);
    }

    public void CalcCost(Point start,
      DoubleGrid heightGrid, DoubleGrid velocityGrid,
      out DoubleGrid costGrid, out IntGrid dirGrid)
    {
      CalcCost(start, heightGrid, velocityGrid, out costGrid, out dirGrid, false);
    }

    public void CalcCost(Point start, Point end,
      DoubleGrid heightGrid, DoubleGrid velocityGrid,
      out DoubleGrid costGrid, out IntGrid dirGrid)
    {
      CalcCost(start, new Point[] { end }, heightGrid, velocityGrid,
        out costGrid, out dirGrid);
    }

    public void CalcCost(Point start, IList<Point> endList,
      DoubleGrid heightGrid, DoubleGrid velocityGrid,
      out DoubleGrid costGrid, out IntGrid dirGrid)
    {
      List<int[]> stopList = new List<int[]>(endList.Count);
      foreach (Point end in endList)
      {

        int[] stop = new int[] {
          (int)((end.X - _dX0) / _dDx),
          (int)((end.Y - _dY0) / _dDy)
        };
        stopList.Add(stop);
      }

      CalcCost(start, heightGrid, velocityGrid, out costGrid, out dirGrid, false,
        stopList);
    }

    public void CalcCost(Point start,
      DoubleGrid heightGrid, DoubleGrid velocityGrid,
      out DoubleGrid costGrid, out IntGrid dirGrid, bool invers)
    {
      CalcCost(start, heightGrid, velocityGrid, out costGrid, out dirGrid, invers, null);
    }

    private void CalcCost(Point start,
      DoubleGrid heightGrid, DoubleGrid velocityGrid,
      out DoubleGrid costGrid, out IntGrid dirGrid, bool invers,
      IList<int[]> stop)
    {
      try
      {
        if (velocityGrid is VelocityGrid)
        { ((VelocityGrid)velocityGrid).LockBits(); }

        costGrid = new DoubleGrid(_dXMax, _dYMax, typeof(short),
          _dX0, _dY0, _dDx, 0, 0.1);
        dirGrid = new IntGrid(_dXMax, _dYMax, typeof(sbyte), _dX0, _dY0, _dDx);

        NodeInfo startField;
        _costGrid = costGrid;
        _dirGrid = dirGrid;
        _heightGrid = heightGrid;
        _veloGrid = velocityGrid;

        _invers = invers;

        _stopList = stop;

        startField = new NodeInfo(this);
        startField.X = (int)((start.X - _dX0) / _dDx);
        startField.Y = (int)((start.Y - _dY0) / _dDy);
        //startField.Height = heightGrid.Value(start.X, start.Y, EGridInterpolation.bilinear);
        //startField.Velocity = velocityGrid.Value(start.X, start.Y, EGridInterpolation.bilinear);
        //startField.Cost = 0;
        startField.IdDir = -2;

        _currentStep = 0;
        _maxSteps = _dXMax * _dYMax;

        BuildTree(startField);

        if (Status != null)
        { Status(this, new StatusEventArgs(costGrid, dirGrid, -1, -1, _maxSteps, _maxSteps)); }
      }
      finally
      {
        if (velocityGrid is VelocityGrid)
        { ((VelocityGrid)velocityGrid).UnlockBits(); }
      }
    }

    public StepAngle StepAngle
    {
      get { return _stepAngle; }
    }

    public Polyline Path(IntGrid dirGrid, Point end)
    {
      int ix;
      int iy;
      dirGrid.Extent.GetNearest(end, out ix, out iy);
      Polyline path = new Polyline();

      int dir = dirGrid[ix, iy];
      path.AddFirst(dirGrid.Extent.CellCenter(ix, iy));
      while (dir > 0)
      {
        dir = dir - 1;
        ix -= _step[dir, 0];
        iy -= _step[dir, 1];

        dir = dirGrid[ix, iy];
        path.AddFirst(dirGrid.Extent.CellCenter(ix, iy));
      }
      return path;
    }

    public override bool IsFixed(NodeInfo node)
    {
      return _dirGrid[node.X, node.Y] != 0;
    }
    protected override bool PreAssign(NodeInfo next)
    {
      _costGrid[next.X, next.Y] = next.Cost;
      _dirGrid[next.X, next.Y] = next.IdDir + 1;

      if (Status != null)
      {
        StatusEventArgs args = new StatusEventArgs(_costGrid, _dirGrid,
          next.X, next.Y, _currentStep, _maxSteps);
        Status(this, args);
      }
      _currentStep++;

      return !Stop(_stopList, next);
    }

    protected override double Cost(NodeInfo from, NodeInfo to)
    {
      double h0, h1, v0, v1, s0, s1;
      if (_invers)
      {
        CostAt(from, out h1, out v1);
        s1 = from.Slope2;
        CostAt(to, out h0, out v0);
        s0 = to.Slope2;
      }
      else
      {
        CostAt(from, out h0, out v0);
        s0 = from.Slope2;
        CostAt(to, out h1, out v1);
        s1 = to.Slope2;
      }
      double cost = StepCost(h0, v0, s0, h1, v1, s1, _stepDistance[from.CurrentStep] * _dDx);

      return cost;
    }

    private void CostAt(NodeInfo node, out double h, out double v)
    {
      h = node.Height;
      v = node.Velocity;
    }

    private bool Stop(IList<int[]> stop, NodeInfo field)
    {
      if (stop == null)
      { return false; }

      int i = stop.Count - 1;
      while (i >= 0)
      {
        int[] s = stop[i];
        if (field.X == s[0] && field.Y == s[1])
        { stop.RemoveAt(i); }
        i--;
      }

      return (stop.Count == 0);
    }
  }
}
