using System;
using System.Collections.Generic;

namespace Mathe.Geom
{
  /// <summary>
  /// Summary description for BoxTree.
  /// </summary>
  public class BoxTree<T>
  {
    public class BoxEnumerator : IEnumerator<TileEntry>
    {
      BoxTree<T> _tree;
      BoxTile _currentTile;
      Box _currentBox;
      IBox _searchBox;
      IEnumerator<TileEntry> _currentList;

      public BoxEnumerator(BoxTree<T> tree, IBox searchBox)
      {
        _tree = tree;
        _searchBox = searchBox;
        Reset();
      }
      public IBox SearchBox
      {
        get { return _searchBox; }
        set
        {
          if (_searchBox.RelationTo(value) != Relation.Contains)
          { throw new ArgumentException("new search box must be within current search box"); }
          _searchBox = value;
        }
      }
      #region IEnumerator Members

      public void Reset()
      {
        _currentTile = _tree._mainTile;
        _currentList = _tree.GetTileEnumerator(this, _currentTile.ElemList);
        _currentList.Reset();
        if (_tree._mainBox == null)
        { _currentBox = null; }
        else
        { _currentBox = _tree._mainBox.Clone(); }
      }

      public TileEntry Current
      {
        get
        {
          if (_currentBox != null &&
            _currentBox.RelationTo(_currentList.Current.Box) != Relation.Contains)
          { throw new InvalidProgramException("Error in software design assumption"); }
          return _currentList.Current;
        }
      }
      object System.Collections.IEnumerator.Current
      { get { return Current; } }

      public void Dispose()
      { }

      public bool MoveNext()
      {
        while (_currentList.MoveNext())
        {
          if (_searchBox == null)
          { return true; }
          if (Current.Box.RelationTo(_searchBox) != Relation.Disjoint)
          { return true; }
        }

        if (_currentBox == null)
        { return false; }

        while (_currentTile != null)
        {
          _currentTile = _tree.NextSelTile(_currentTile, _currentBox, _searchBox);
          if (_currentTile == null)
          { continue; }

          _currentList = _currentTile.ElemList.GetEnumerator();
          _currentList.Reset();

          while (_currentList.MoveNext())
          {
            if (_searchBox == null)
            { return true; }
            if (Current.Box.RelationTo(_searchBox) != Relation.Disjoint)
            { return true; }
          }
        }

        return false;
      }

      #endregion
    }

    public class BoxEnumerable : IEnumerable<TileEntry>
    {
      private BoxTree<T> _pBoxTree;
      private IBox _pBox;

      public BoxEnumerable(BoxTree<T> tree, IBox box)
      {
        _pBoxTree = tree;
        _pBox = box;
      }
      #region IEnumerable Members

      public BoxEnumerator GetEnumerator()
      { return new BoxEnumerator(_pBoxTree, _pBox); }
      IEnumerator<TileEntry> IEnumerable<TileEntry>.GetEnumerator()
      { return GetEnumerator(); }
      System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
      { return GetEnumerator(); }

      #endregion
    }

    public class TileEntry
    {
      private IBox _pBox;
      private T _pValue;
      public TileEntry(IBox box, T value)
      {
        _pBox = box;
        _pValue = value;
      }
      public IBox Box
      {
        get
        { return _pBox; }
      }
      public T Value
      {
        get
        { return _pValue; }
      }
    }

    internal class BoxTile
    {
      private BoxTile _parent;
      private BoxTile _child0;
      private BoxTile _child1;
      private int _splitDimension;
      private int _count = 0;
      private bool _bSplitting = false;

      private double _minSplitDim;
      private double _maxSplitDim;
      private double _minParentDim;
      private double _maxParentDim;

      private List<TileEntry> _elemList = new List<TileEntry>();

      public double MinInParentSplitDim
      {
        get
        { return _minParentDim; }
        set
        { _minParentDim = value; }
      }
      public double MaxInParentSplitDim
      {
        get
        { return _maxParentDim; }
        set
        { _maxParentDim = value; }
      }

      public int Count
      {
        get
        { return _count; }
        set
        { _count = value; }
      }

      public double MinInSplitDim
      {
        get
        { return _minSplitDim; }
        set
        { _minSplitDim = value; }
      }
      public double MaxInSplitDim
      {
        get
        { return _maxSplitDim; }
        set
        { _maxSplitDim = value; }
      }

      public List<TileEntry> ElemList
      {
        get
        { return _elemList; }
        set
        { _elemList = value; }
      }
      public BoxTile Parent
      {
        get
        { return _parent; }
        set
        { _parent = value; }
      }
      public BoxTile Child0
      {
        get { return _child0; }
      }
      public void InitChildren(BoxTile child0, BoxTile child1, double u0, double u, int counter0, int size, int denominator)
      {
        _child0 = child0;
        _child0.Parent = this;
        _child0.MinInParentSplitDim = Position(u0, u, counter0, denominator); //== _minSplitDim;
        _child0.MaxInParentSplitDim = Position(u0, u, counter0 + 2 * size, denominator); //(_maxSplitDim + _minSplitDim) / 2.0;

        // double d;
        _child1 = child1;
        _child1.Parent = this;
        // d = (_maxSplitDim - _minSplitDim) / 4.0;
        _child1.MinInParentSplitDim = Position(u0, u, counter0 + size, denominator);    //_minSplitDim + d;
        _child1.MaxInParentSplitDim = Position(u0, u, counter0 + 3 * size, denominator);//_maxSplitDim - d;

        _count = _count + _child0.Count + _child1.Count;
      }
      public BoxTile Child1
      {
        get { return _child1; }
      }
      public int SplitDimension
      {
        get
        { return _splitDimension; }
        set
        { _splitDimension = value; }
      }
      /// <summary>
      /// used to indicate that splitting is running. 
      /// if true, do not propagate added elements to parent count
      /// </summary>
      private bool Splitting
      {
        set
        { _bSplitting = value; }
      }

      /// <summary>
      /// Split the tile along the largest dimesion of box
      /// Assumption: extent represents the actual extent of the tile (see method BoxExtent in BoxTree)
      /// </summary>
      public void Split(Box unit, IList<int> counter0, IList<int> denominator, IList<int> mainCounter, IList<int> mainSize)
      {
        int splitDim;
        int dim = counter0.Count;
        double d0;
        double tx1, ux0, ux1;

        List<TileEntry> valueList;

        if ((_child0 == null) == false) throw new InvalidProgramException("child0 == null");
        if ((_child1 == null) == false) throw new InvalidProgramException("child1 == null");

        d0 = (unit.Max[0] - unit.Min[0]) * mainSize[0] / denominator[0];
        splitDim = 0;
        for (int i = 1; i < dim; i++)
        {
          double d = (unit.Max[i] - unit.Min[i]) * mainSize[i] / denominator[i];
          if (d0 < d)
          {
            d0 = d;
            splitDim = i;
          }
        }
        double u0 = unit.Min[splitDim];
        double u = unit.Max[splitDim] - unit.Min[splitDim];

        //TODO: verify
        _splitDimension = splitDim;
        int c0 = 2 * mainSize[splitDim] * counter0[splitDim] + mainCounter[splitDim] * denominator[splitDim];
        _minSplitDim = Position(u0, u, c0, denominator[splitDim]);
        _maxSplitDim = Position(u0, u, c0 + 4 * mainSize[splitDim], denominator[splitDim]);

        InitChildren(new BoxTile(), new BoxTile(),
          u0, u, c0, mainSize[splitDim], denominator[splitDim]);

        valueList = _elemList;
        _elemList = new List<TileEntry>();

        _child0.Splitting = true;
        _child1.Splitting = true;
        Splitting = true;

        tx1 = _child0.MaxInParentSplitDim;

        ux0 = _child1.MinInParentSplitDim;
        ux1 = _child1.MaxInParentSplitDim;

        for (int i = 0; i < valueList.Count; i++)
        {
          TileEntry entry = valueList[i];
          double dMin = entry.Box.Min[splitDim];
          double dMax = entry.Box.Max[splitDim];
          if (dMin < ux0)
          {
            if (dMax < tx1)
            { _child0.Add(entry); }
            else
            { Add(entry); } // TODO: do not add directly to mElemList (sorting in future);
          }
          else
          {
            //if ((dMin <= tx1) == false)
            //{
            //  BoxTile child = this;
            //  BoxTile parent = _parent;
            //  while (parent != null)
            //  {
            //    if (parent.SplitDimension == splitDim && parent.Child0 == child)
            //    { throw new InvalidProgramException("dMin <= tx1"); }
            //    child = parent;
            //    parent = parent.Parent;
            //  }
            //}
            if (dMax < ux1)
            { _child1.Add(entry); }
            else
            { Add(entry); }
          }
        }

        _child0.Splitting = false;
        _child1.Splitting = false;
        Splitting = false;
        _count = valueList.Count;

      }

      public TileEntry this[int index]
      {
        get
        {
          if (index < _elemList.Count)
          { return _elemList[index]; }
          else
          {
            index -= _elemList.Count;
            if (index < _child0.Count)
            { return _child0[index]; }

            index -= _child0.Count;
            return _child1[index];
          }
        }
      }
      public void Add(IBox box, T value)
      {
        Add(new TileEntry(box, value));
      }
      internal void Add(TileEntry entry)
      {
        // TODO: implement sorting
        _elemList.Add(entry);

        _count++;
        if (_bSplitting)
        { return; }

        BoxTile parent = Parent;
        while (parent != null)
        {
          parent.Count++;
          parent = parent.Parent;
        }
      }
      public void AddRange(List<TileEntry> elemList)
      {
        // TODO: implement sorting
        _elemList.AddRange(elemList);
      }
      public void Insert(int index, TileEntry entry)
      {
        _count++;
        _elemList.Insert(index, entry);

        BoxTile parent = Parent;
        while (parent != null)
        {
          parent.Count++;
          parent = parent.Parent;
        }
      }
      public void Remove(TileEntry entry)
      {
        _elemList.Remove(entry);

        _count--;
        BoxTile parent = Parent;
        while (parent != null)
        {
          parent.Count--;
          parent = parent.Parent;
        }
      }
      public void Collapse()
      {
        if ((_child0.Child0 == null) == false) throw new InvalidProgramException("_child0.child0 == null");
        if ((_child1.Child0 == null) == false) throw new InvalidProgramException("_child1.child0 == null");

        AddRange(_child0.ElemList);
        AddRange(_child1.ElemList);

        _child0 = null;
        _child1 = null;
      }

      internal bool VerifyPointTile(BoxTree<T> tree)
      {
        IBox extent = tree.TileExtent(this);
        foreach (TileEntry entry in _elemList)
        {
          if (extent.RelationTo(entry.Box) != Relation.Contains)
          {
            return false;
          }
        }
        return true;
      }
    }


    private int _dimension;
    private Box _unitBox;
    private Box _mainBox;
    private int[] _mainCounter;
    private int[] _mainSize;

    // private Box mSearchBox;
    private int _maxElemPerTile;
    private int _nElemJoin;

    private BoxTile _mainTile;

    private bool _dynamic;

    public BoxTree(int dimension)
    {
      Dimension = dimension;

      _mainTile = new BoxTile();
      _maxElemPerTile = 64;
      _dynamic = true;
    }

    public BoxTree(int dimension, int nElem, bool dynamic)
    {
      Dimension = dimension;
      _mainTile = new BoxTile();
      _maxElemPerTile = nElem;
      _dynamic = dynamic;
    }

    public void Init(int dimension, Box box, int maxElemPerTile, int nElemJoin)
    {
      Dimension = dimension;
      _maxElemPerTile = maxElemPerTile;
      _nElemJoin = nElemJoin;

      _mainCounter = new int[dimension];
      _mainSize = new int[dimension];

      if (box != null)
      {
        _unitBox = box.Clone();
        _mainBox = _unitBox.Clone();
        for (int iDim = 0; iDim < dimension; iDim++)
        {
          _mainBox.Max[iDim] += _mainBox.Max[iDim] - _mainBox.Min[iDim];
          _mainSize[iDim] = 2;
        }
      }
      else
      {
        _unitBox = null;
        _mainBox = null;
      }
    }
    public void Init(int dimension, int maxElemPerTile, int nElemJoin)
    {
      Init(dimension, null, maxElemPerTile, nElemJoin);
    }
    public void Init(int dimension, int maxElemPerTile)
    {
      Init(dimension, null, maxElemPerTile, maxElemPerTile / 2);
    }
    public void Init(Box box, int maxElemPerTile, int nElemJoin)
    {
      Init(box.Dimension, box, maxElemPerTile, nElemJoin);
    }
    public void Init(Box box, int maxElemPerTile)
    {
      Init(box.Dimension, box, maxElemPerTile, maxElemPerTile / 2);
    }

    /// <summary>
    /// Enlarge Region. Remark: when data is load, numerical problems may occure
    /// </summary>
    /// <param name="directedDimension">directed dimension to enlarge: (-dim -1) : add negativ (dim + 1) : add positiv </param>
    /// <returns>new parent tile</returns>
    private BoxTile Enlarge(int directedDimension)
    {
      int dimension;
      dimension = Math.Abs(directedDimension) - 1;

      BoxTile newTile = new BoxTile();

      int size = _mainSize[dimension];

      _mainSize[dimension] *= 2;

      double u0 = _unitBox.Min[dimension];
      double u = _unitBox.Max[dimension] - _unitBox.Min[dimension];
      if (directedDimension > 0)
      {
        _mainBox.Max[dimension] = Position(u0, u, _mainCounter[dimension] + 2 * size, 1);

        newTile.InitChildren(_mainTile, new BoxTile(),
          u0, u, _mainCounter[dimension], size / 2, 1);
      }
      else
      {
        if ((size & 1) != 0)
        { throw new InvalidProgramException(string.Format("Error in software design assumption size {0} must be divisible by 2", size)); }
        _mainCounter[dimension] -= size / 2;

        _mainBox.Min[dimension] = Position(u0, u, _mainCounter[dimension], 1);
        _mainBox.Max[dimension] = Position(u0, u, _mainCounter[dimension] + 2 * size, 1);

        newTile.InitChildren(new BoxTile(), _mainTile,
          u0, u, _mainCounter[dimension], size / 2, 1);
      }
      newTile.SplitDimension = dimension;
      _mainTile = newTile;

      return newTile;
    }

    private static double Position(double u0, double u, int counter, int denominator)
    {
      while (denominator > 1 && (counter & 1) == 0)
      {
        denominator /= 2;
        counter /= 2;
      }
      double pos = u0 + counter * u / denominator;
      return pos;
    }

    public void InitSize(IEnumerable<IGeometry> data)
    {
      InitSize((System.Collections.IEnumerable)data);
    }

    private void InitSize()
    {
      InitSize(_mainTile.ElemList);
    }

    private void InitSize(System.Collections.IEnumerable data)
    {
      _unitBox = null;
      foreach (object pRow in data)
      {
        IBox box;
        if (pRow is IGeometry)
        { box = ((IGeometry)pRow).Extent; }
        else
        {
          //Debug.Assert(pRow is TileEntry);
          box = ((TileEntry)pRow).Box;
        }
        if (_unitBox == null)
        {
          _unitBox = new Box(Point.Create(box.Min), Point.Create(box.Max));
          continue;
        }

        for (int iDim = 0; iDim < Dimension; iDim++)
        {
          if (box.Min[iDim] < _unitBox.Min[iDim])
          { _unitBox.Min[iDim] = box.Min[iDim]; }
          if (box.Max[iDim] > _unitBox.Max[iDim])
          { _unitBox.Max[iDim] = box.Max[iDim]; }
        }
      }

      // double the size of each dimension
      _mainBox = _unitBox.Clone();
      _mainCounter = new int[Dimension];
      _mainSize = new int[Dimension];
      for (int iDim = 0; iDim < Dimension; iDim++)
      {
        _mainBox.Max[iDim] = 2 * _unitBox.Max[iDim] - _unitBox.Min[iDim];
        _mainCounter[iDim] = 0;
        _mainSize[iDim] = 2;
      }
    }

    private BoxTile StartFindAddTile(IBox geomExtent, bool dynamic)
    {
      IList<int> counter0 = null;
      IList<int> denominator = null;

      if (dynamic && _unitBox != null)
      {
        GetPositions(out counter0, out denominator);
      }

      return FindAddTile(_mainTile, geomExtent, dynamic, counter0, denominator);
    }

    private void GetPositions(out IList<int> counter0, out IList<int> denominator)
    {
      counter0 = new int[Dimension];
      denominator = new int[Dimension];
      for (int iDim = 0; iDim < Dimension; iDim++)
      {
        denominator[iDim] = 4;
      }
    }

    /// <summary>
    /// find tile where geomExtent fits inside
    /// </summary>
    private BoxTile FindAddTile(BoxTile parent, IBox geomExtent, bool dynamic,
      IList<int> counter0, IList<int> denominator)
    {
      BoxTile child0 = parent.Child0;
      if (child0 != null)
      {
        BoxTile child1 = parent.Child1;
        int divDim = parent.SplitDimension;

        if (geomExtent.Min[divDim] < child1.MinInParentSplitDim)
        {
          if (geomExtent.Max[divDim] < child0.MaxInParentSplitDim)
          {
            if (dynamic)
            {
              counter0[divDim] *= 2;
              denominator[divDim] *= 2;
            }
            return FindAddTile(child0, geomExtent, dynamic, counter0, denominator);
          }
          else
          { return parent; }
        }
        else
        {
          if (geomExtent.Max[divDim] < child1.MaxInParentSplitDim)
          {
            if (dynamic)
            {
              counter0[divDim] += counter0[divDim] + 1;
              denominator[divDim] *= 2;
            }
            return FindAddTile(child1, geomExtent, dynamic, counter0, denominator);
          }
          else
          { return parent; }
        }
      }
      else
      { // tile has no children
        if (dynamic == false || parent.ElemList.Count < _maxElemPerTile - 1)
        { return parent; }
        else
        { // split tile and try again
          if (_unitBox == null)
          {
            InitSize();
            VerifyExtent(geomExtent);

            return StartFindAddTile(geomExtent, true);
          }
          else
          {
            parent.Split(_unitBox, counter0, denominator, _mainCounter, _mainSize);
            return FindAddTile(parent, geomExtent, false, counter0, denominator);
          }
        }
      }
    }

    /// <summary>
    /// Get the Extent of Tile
    /// </summary>
    private Box TileExtent(BoxTile tile)
    {
      Box box = _mainBox.Clone();
      BoxTile child = tile;
      BoxTile parent = child.Parent;
      double d = -1;
      double dMax = 0;

      // proceed up so long until it is verified that all dimensions of the 
      // box are set to the correct extent
      while (parent != null && d <= dMax)
      {
        int splitDim = parent.SplitDimension;

        d = child.MaxInParentSplitDim - child.MinInParentSplitDim;
        if (box.Min[splitDim] < child.MinInParentSplitDim)
        { box.Min[splitDim] = child.MinInParentSplitDim; }
        if (box.Max[splitDim] > child.MaxInParentSplitDim)
        { box.Max[splitDim] = child.MaxInParentSplitDim; }

        if (dMax == 0)
        {
          dMax = parent.MaxInSplitDim - parent.MinInSplitDim;
          // dMax was the largest extent when splitting the parent of tile,
          // maybe together with others.
          // So if d > dMax, it must be dimension, that was split already
        }

        child = parent;
        parent = parent.Parent;
      }
      return box;
    }


    public TileEntry this[int index]
    {
      get
      { return _mainTile[index]; }
    }
    public int Add(IBox box, T value)
    {
      BoxTile addTile;

      VerifyExtent(box);

      addTile = StartFindAddTile(box, _dynamic);
      addTile.Add(box, value);

      return _mainTile.Count - 1;
    }

    private void VerifyExtent(IBox box)
    {
      if (_mainBox != null)
      {
        int dim = _mainBox.ExceedDimension(box);
        while (dim != 0)
        {
          Enlarge(dim);
          dim = _mainBox.ExceedDimension(box);
        }
      }
    }

    internal bool VerifyPointTree()
    {
      return VerifyTile(_mainTile);
    }
    private bool VerifyTile(BoxTile tile)
    {
      if (tile == null)
      { return true; }

      if (VerifyTile(tile.Child0) == false)
      {
        return false;
      }
      if (VerifyTile(tile.Child1) == false)
      {
        return false;
      }

      return tile.VerifyPointTile(this);
    }

    /// <summary>
    /// Inserts a geometry at position.
    /// </summary>
    public void Insert(int index, TileEntry entry)
    {
      // TODO , only valid if just one tail exists
      _mainTile.Insert(index, entry);
    }

    public void Remove(TileEntry entry)
    {
      BoxTile tile = StartFindAddTile(entry.Box, false);
      tile.Remove(entry);

      if (_dynamic && tile.Child0 == null && tile.Parent != null &&
        tile.Parent.Child0.ElemList.Count + tile.Parent.Child1.ElemList.Count +
        tile.Parent.ElemList.Count < _nElemJoin)
      { tile.Parent.Collapse(); }
    }


    /// <summary>
    /// 
    /// </summary>
    /// <param name="currentTile"></param>
    /// <param name="currentBox">begin: box of current tile; end: box of return tile</param>
    /// <param name="extent">extent which must intersect or touch next tile</param>
    /// <returns></returns>
    private BoxTile NextSelTile(BoxTile currentTile, Box currentBox, IBox extent)
    {
      int splitDim;
      BoxTile child;

      splitDim = currentTile.SplitDimension;
      child = currentTile.Child0;
      if (child != null)
      {
        if ((currentBox.Min[splitDim] == child.MinInParentSplitDim) == false)
        { throw new InvalidProgramException("Error in software design assumption: SplitDim"); }
        currentBox.Max[splitDim] = child.MaxInParentSplitDim;
        if (extent == null || currentBox.Intersects(extent))
        { return child; }
      }

      while (currentTile != null)
      {
        if (child != currentTile.Child1)
        {
          child = currentTile.Child1;
          currentBox.Min[splitDim] = child.MinInParentSplitDim;
          currentBox.Max[splitDim] = child.MaxInParentSplitDim;
          if (extent == null || currentBox.Intersects(extent))
          { return child; }
        }

        // reset box to current extent
        if (child != null)
        {
          currentBox.Min[splitDim] = currentTile.MinInSplitDim;
          currentBox.Max[splitDim] = currentTile.MaxInSplitDim;
        }

        // one step up
        child = currentTile;
        currentTile = currentTile.Parent;
        if (currentTile != null)
        { splitDim = currentTile.SplitDimension; }
      }
      // fr genuegend unterteilte Tiles (sonst ev. Min.X == 0 o.ae.):
      // Debug.Assert(currentBox.Equals(mMainBox)); 
      return null;
    }

    public BoxEnumerable Search(IBox box)
    {
      return new BoxEnumerable(this, box);
    }

    public TileEntry Near(IBox box)
    {
      if (_mainBox == null)
      {
        if (_mainTile.ElemList.Count == 0)
        { return null; }
      }
      return Near(box, _mainTile);
    }

    private TileEntry Near(IBox box, BoxTile parent)
    {
      if (parent.ElemList.Count > 0)
      { return parent.ElemList[0]; }

      int divDim;

      BoxTile child0 = parent.Child0;
      if (child0 == null)
      { return null; }

      BoxTile child1 = parent.Child1;
      divDim = parent.SplitDimension;

      if (box.Min[divDim] < child1.MinInParentSplitDim)
      {
        TileEntry entry = Near(box, child0);
        if (entry != null)
        { return entry; }
        else
        { return Near(box, child1); }
      }
      else
      {
        TileEntry entry = Near(box, child1);
        if (entry != null)
        { return entry; }
        else
        { return Near(box, child0); }
      }
    }

    #region IGeometry Members

    public int Dimension
    {
      get { return _dimension; }
      set
      {
        _dimension = value;
      }
    }

    public int Topology
    {
      get { return Dimension; }
    }

    public Box Extent
    {
      get { return _mainBox; }
    }

    public bool Intersects(Box box)
    { return Extent.Intersects(box); }

    #endregion

    #region IList Members

    public bool Contains(TileEntry value)
    {
      BoxTile tile = StartFindAddTile(value.Box, false);
      if (tile == null)
      { return false; }
      return tile.ElemList.Contains(value);
    }

    public void Clear()
    {
      _mainTile = null;
    }

    public bool IsFixedSize
    {
      get { return false; }
    }

    #endregion

    #region ICollection Members

    public int Count
    {
      get { return _mainTile.Count; }
    }

    #endregion

    public void Sort(IComparer<TileEntry> comparer)
    {
      Sort(_mainTile, comparer);
    }

    private void Sort(BoxTile tile, IComparer<TileEntry> comparer)
    {
      if (tile == null)
      { return; }

      tile.ElemList.Sort(comparer);

      Sort(tile.Child0, comparer);
      Sort(tile.Child1, comparer);
    }

    protected virtual IEnumerator<TileEntry> GetTileEnumerator(
      BoxEnumerator enumerator, List<TileEntry> list)
    {
      return list.GetEnumerator();
    }
  }
}
