using System;
using System.Collections.Generic;
using PntOp = Mathe.Geom.PointOperator;

namespace Mathe.Geom
{
  public static class GeometryOperator
  {
    #region nested classes

    private class LinearIntersect : IGeometry
    {
      private IPoint _fullOrig;
      private OrthogonalSystem _fullSystem;
      private int _xTopology;
      private IPoint _offset;
      private Axis _offsetAxis;
      private IList<Axis> _commonAxes;

      private IList<double> _vectorFactors;

      private IPoint _near;

      public LinearIntersect(IPoint fullOffset, OrthogonalSystem fullSystem, int xTopology,
        IPoint offset, Axis offsetAxis, IList<Axis> commonAxes)
      {
        _fullOrig = fullOffset;
        _fullSystem = fullSystem;
        _xTopology = xTopology;
        _offset = offset;
        _offsetAxis = offsetAxis;
        _commonAxes = commonAxes;
      }

      public static LinearIntersect Create(
        IPoint xOrig, OrthogonalSystem xSystem, IPoint yOrig, IList<IPoint> yAxes)
      {
        OrthogonalSystem fullSystem = xSystem;
        int xTopology = xSystem.Axes.Count;

        Point offset = PntOp.Sub(yOrig, xOrig);

        List<Axis> nullAxes = null;

        foreach (IPoint yAxis in yAxes)
        {
          Axis axis = fullSystem.GetOrthogonal(yAxis);
          if (axis.Length2 > 0)
          { fullSystem.Axes.Add(axis); }
          else
          {
            if (nullAxes == null)
            { nullAxes = new List<Axis>(); }
            nullAxes.Add(axis);
          }
        }

        Axis offsetAxis = fullSystem.GetOrthogonal(offset);

        LinearIntersect intersect = new LinearIntersect(xOrig, fullSystem, xTopology,
          offset, offsetAxis, nullAxes);
        return intersect;
      }

      public int Dimension
      {
        get { return _offsetAxis.Point.Dimension; }
      }
      public int Topology
      {
        get
        {
          if (_commonAxes == null)
          { return 0; }
          return _commonAxes.Count;
        }
      }
      public IBox Extent
      {
        get { throw new NotImplementedException("TODO: get extent of "); }
      }
      public IGeometry Border
      {
        get
        {
          if (Topology == 0)
          { return null; }
          throw new NotImplementedException("TODO: get border of ");
        }
      }
      public bool EqualGeometry(IGeometry other)
      {
        throw new NotImplementedException("TODO: project ");
      }
      public IGeometry Project(IProjection prj)
      {
        throw new NotImplementedException("TODO: project ");
      }
      public bool Intersects(IGeometry geometry)
      {
        throw new NotImplementedException("TODO: intersects ");
      }
      public GeometryCollection Intersection(IGeometry geometry)
      {
        throw new NotImplementedException("TODO: intersects ");
      }
      public bool IsWithin(IPoint point)
      {
        throw new NotImplementedException("TODO: intersects ");
      }
      public IGeometry Split(IGeometry[] border)
      {
        throw new NotImplementedException("TODO: intersects ");
      }

      public double Offset2
      {
        get { return _offsetAxis.Length2; }
      }

      private static bool _optimzeWarned = false;
      public IPoint Near
      {
        get
        {
          if (_near == null)
          {
            if (_vectorFactors == null)
            {
              if (_optimzeWarned == false)
              {
                System.Diagnostics.Debug.Assert(false, "TODO: Optimize");
                _optimzeWarned = true;
              }
            }

            IList<double> factors = VectorFactors;
            Point near = PntOp.Add(_fullOrig, _offset);
            for (int i = factors.Count - 1; i >= _xTopology; i--)
            {
              Point p = _fullSystem.OrigVector(i);
              near = near - factors[i] * p;
            }
            _near = near;
          }
          return _near;
        }
      }

      public bool IsInside(params bool[] points)
      {
        int n = points.Length;
        if (n != 2)
        { throw new InvalidOperationException("TODO: Implementation for non-Point-Intersections"); }

        for (int i = 0; i < n; i++)
        {
          if (IsInside(i, points[i]) == false)
          { return false; }
        }
        return true;
      }

      /// <summary>
      /// 
      /// </summary>
      /// <param name="part"></param>
      /// <param name="points">
      /// if true : returns (intersections lies inside the convex hull of the points)
      /// if false: returns (intersections lies inside the convex hull of the axes)
      /// </param>
      /// <returns></returns>
      public bool IsInside(int part, bool points)
      {
        if (_commonAxes != null && _commonAxes.Count > 0)
        { throw new NotImplementedException("TODO: Implementation for non-Point-Intersections"); }

        IList<double> factors = VectorFactors;

        bool inside;
        if (part == 0)
        {
          inside = IsInside(factors, 0, _xTopology, points);
        }
        else if (part == 1)
        {
          inside = IsInside(factors, _xTopology, factors.Count, points);
        }
        else
        { throw new NotImplementedException("TODO: Implementation for intersections with more than 2 parts"); }
        return inside;
      }

      private bool IsInside(IList<double> factors, int i0, int i1, bool points)
      {
        double sum = 0;
        for (int i = i0; i < i1; i++)
        {
          double f = factors[i];
          if (i0 > 0)
          { f = -f; }
          if (0 > f || f > 1)
          { return false; }
          sum += f;
        }

        if (points && sum > 1)
        { return false; }

        return true;
      }

      public IList<double> VectorFactors
      {
        get
        {
          if (_vectorFactors == null)
          { _vectorFactors = _fullSystem.VectorFactors(_offsetAxis.Factors); }
          return _vectorFactors;
        }
      }

      public IPoint Parameter(int index, IBox box)
      {
        Point d = PntOp.Sub(box.Max, box.Min);
        if (index == 0)
        {
          IPoint p = Point.Create(_xTopology);
          for (int i = 0; i < _xTopology; i++)
          {
            p[i] = box.Min[i] + d[i] * VectorFactors[i];
          }
          return p;
        }
        else if (index == 1)
        {
          int n = VectorFactors.Count - _xTopology;
          IPoint p = Point.Create(n);
          for (int i = 0; i < n; i++)
          {
            p[i] = box.Min[i] - d[i] * VectorFactors[i + _xTopology];
          }
          return p;
        }
        else
        { throw new InvalidOperationException(""); }
      }
    }

    private class PartEnumerable : IEnumerable<IList<IPoint>>
    {
      private class Enumerator : IEnumerator<IList<IPoint>>
      {
        private PartEnumerable _parent;

        private IEnumerator<PartEnumerable> _listEnumerator;
        private Enumerator _partEnumerator;
        private bool _first;

        public Enumerator(PartEnumerable enumerable)
        {
          _parent = enumerable;
          Reset();
        }

        public void Reset()
        {
          int nDim = _parent._dimensions.Count;
          if (nDim == 0)
          {
            _listEnumerator = null;
            _first = true;
            return;
          }

          int nPoints = _parent._points.Count;
          IPoint c = _parent._points[nPoints - 1];
          IList<PartEnumerable> childEnumerables
            = new List<PartEnumerable>(2 * nDim);
          for (int iDim = 0; iDim < nDim; iDim++)
          {
            int dim = _parent._dimensions[iDim];
            List<int> childDims = new List<int>(nDim - 1);
            for (int i = 0; i < nDim; i++)
            {
              if (i != iDim)
              { childDims.Add(_parent._dimensions[i]); }
            }
            foreach (double d in new double[] { _parent._range.Min[dim], _parent._range.Max[dim] })
            {
              Point p = Point.Create(c);
              p[dim] = d;

              List<IPoint> childPoints = new List<IPoint>(_parent._points);
              childPoints.Add(p);

              PartEnumerable childEnum = new PartEnumerable(childPoints, childDims, _parent._range);
              childEnumerables.Add(childEnum);
            }
          }
          _listEnumerator = childEnumerables.GetEnumerator();
        }

        object System.Collections.IEnumerator.Current
        { get { return Current; } }
        public IList<IPoint> Current
        {
          get
          {
            if (_listEnumerator == null)
            {
              return _parent._points;
            }
            else
            {
              return _partEnumerator.Current;
            }
          }
        }

        public void Dispose()
        { }


        public bool MoveNext()
        {
          if (_listEnumerator == null)
          {
            if (_first)
            {
              _first = false;
              return true;
            }
            return false;
          }

          while (_partEnumerator == null || _partEnumerator.MoveNext() == false)
          {
            bool next = _listEnumerator.MoveNext();
            if (next == false)
            { return false; }

            PartEnumerable partEnumerable = _listEnumerator.Current;
            _partEnumerator = partEnumerable.GetEnumerator_();
          }
          return true;
        }

      }

      private IBox _range;
      private IList<IPoint> _points;
      private IList<int> _dimensions;

      private Point _center;

      public PartEnumerable(IBox range)
      {
        _range = range;

        _center = 0.5 * PntOp.Add(_range.Min, _range.Max);
        int dim = _center.Dimension;
        _dimensions = new int[dim];
        for (int i = 0; i < dim; i++)
        { _dimensions[i] = i; }

        _points = new List<IPoint>();
        _points.Add(_center);
      }

      private PartEnumerable(IList<IPoint> points, IList<int> dimensions, IBox range)
      {
        _points = points;
        _dimensions = dimensions;
        _range = range;
      }

      public IPoint Center
      {
        get { return _center; }
      }

      System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
      { return GetEnumerator(); }
      public IEnumerator<IList<IPoint>> GetEnumerator()
      { return GetEnumerator_(); }
      private Enumerator GetEnumerator_()
      {
        return new Enumerator(this);
      }
    }

    private class ParamInfo
    {
      private IParamGeometry _geometry;
      private IPoint _min;
      private IPoint _delta;
      private SortedList<IPoint, IPoint> _pointsAt;

      public ParamInfo(IParamGeometry geometry)
      {
        _geometry = geometry;
      }

      public IParamGeometry BaseGeometry
      {
        get { return _geometry; }
      }

      public int Topology
      {
        get { return _geometry.Topology; }
      }

      /// <summary>
      /// 
      /// </summary>
      /// <param name="param">Parameter of ParamInfo, not of base geometry</param>
      /// <returns></returns>
      public IPoint CachedPointAt(IPoint param)
      {
        IPoint pointAt;

        if (_pointsAt == null)
        { _pointsAt = new SortedList<IPoint, IPoint>(new PointComparer()); }

        if (_pointsAt.TryGetValue(param, out pointAt) == false)
        {
          IPoint paramAt = GeometryParamAt(param);
          pointAt = _geometry.PointAt(paramAt);
          _pointsAt.Add(param, pointAt);
        }

        return pointAt;
      }

      private IPoint Min
      {
        get
        {
          if (_min == null)
          { _min = _geometry.ParameterRange.Min; }
          return _min;
        }
      }
      private IPoint Delta
      {
        get
        {
          if (_delta == null)
          { _delta = PntOp.Sub(_geometry.ParameterRange.Max, Min); }
          return _delta;
        }
      }

      public IPoint GeometryParamAt(IPoint param)
      {
        IPoint paramAt = Point.Create(param);
        for (int iDim = 0; iDim < param.Topology; iDim++)
        {
          paramAt[iDim] = Min[iDim] + param[iDim] * Delta[iDim];
        }
        return paramAt;
      }
    }

    private class ParamPartList : List<ParamPart>
    {
      private class CornerInfo
      {
        IList<ParamPart> _neighbors;
        Dictionary<ParamPart, IList<ParamsRelation>> _relations;

        public CornerInfo()
        {
          _neighbors = new List<ParamPart>();
        }

        public IList<ParamPart> Neighbors
        {
          get { return _neighbors; }
        }
        public void Add(ParamPart boxNeighbor)
        {
          _neighbors.Add(boxNeighbor);
        }
        public void AddSplit(ParamPart intersect, IList<ParamsRelation> relations)
        {
          if (_relations == null)
          { _relations = new Dictionary<ParamPart, IList<ParamsRelation>>(); }

          _relations.Add(intersect, relations);
        }
        public bool ContainsSplit(ParamPart intersect)
        {
          if (_relations == null)
          { return false; }
          bool contains = _relations.ContainsKey(intersect);
          return contains;
        }
      }

      private ParamInfo _geometry;
      private IPoint _center;

      private SortedList<IPoint, CornerInfo> _corners;

      public ParamPartList(ParamInfo geometry, IPoint center)
      {
        _geometry = geometry;
        _center = center;
      }

      private SortedList<IPoint, CornerInfo> Corners
      {
        get
        {
          if (_corners == null)
          {
            CornerInfo cornerInfo;
            _corners = new SortedList<IPoint, CornerInfo>(new PointComparer());
            foreach (ParamPart part in this)
            {
              IPoint corner = part.BoxCorner;
              if (_corners.TryGetValue(corner, out cornerInfo) == false)
              {
                cornerInfo = new CornerInfo();
                _corners.Add(corner, cornerInfo);
              }
              cornerInfo.Add(part);
            }
          }
          return _corners;
        }
      }

      public IPoint Center
      { get { return _center; } }

      public void Split(IPoint cornerPoint, ParamPart yPart, List<ParamsRelation> parentList)
      {
        CornerInfo cornerInfo = Corners[cornerPoint];

        IList<ParamPart> subParts = ParamPart.CreateParts(_geometry, Point.Create(_center),
          Point.Create(cornerPoint));

        List<ParamsRelation> relations = new List<ParamsRelation>();
        foreach (ParamPart subPart in subParts)
        {
          ParamsRelation relation = new ParamsRelation(subPart, yPart, parentList);
          parentList.Add(relation);
        }
        cornerInfo.AddSplit(yPart, relations);
      }

      public bool IsSplit(IPoint cornerPoint, ParamPart yPart)
      {
        if (_corners == null)
        { return false; }

        CornerInfo corner = Corners[cornerPoint];

        return corner.ContainsSplit(yPart);
      }
    }

    private class ParamPart
    {
      private ParamInfo _geometry;
      private IList<IPoint> _cornerParams;
      private IBox _paramBox;

      private IPoint _origin;
      private List<IPoint> _axes;
      private OrthogonalSystem _orthoSystem;

      private ParamPartList _parentList;
      private double _maxOffset = -1;

      public static IList<ParamPart> CreateParts(ParamInfo x)
      {
        int dimension = x.Topology;
        IPoint min = x.BaseGeometry.ParameterRange.Min;
        IPoint max = x.BaseGeometry.ParameterRange.Max;

        if (x.BaseGeometry.NormedMaxOffset == 0)
        {
          IList<IPoint> corners = new List<IPoint>(dimension + 1);
          Point p;

          p = Point.Create(min);
          corners.Add(p);
          for (int i = 0; i < dimension; i++)
          {
            p = Point.Create(min);
            p[i] = max[i];
            corners.Add(p);
          }
          ParamPart y = new ParamPart(x, corners, null);
          IList<ParamPart> list = new ParamPart[] { y };
          return list;
        }

        return CreateParts(x, min, max);
      }

      public static IList<ParamPart> CreateParts(ParamInfo x, IPoint range0, IPoint range1)
      {
        IBox range = new Box(range0, range1, true);
        PartEnumerable enumPoints = new PartEnumerable(range);
        List<IList<IPoint>> parts = new List<IList<IPoint>>(enumPoints);

        ParamPartList parentList = new ParamPartList(x, enumPoints.Center);
        foreach (IList<IPoint> part in parts)
        {
          ParamPart xPart = new ParamPart(x, part, parentList);
          parentList.Add(xPart);
        }

        return parentList;
      }

      public ParamPart(ParamInfo geometry, IList<IPoint> cornerParameters,
      ParamPartList parentList)
      {
        _geometry = geometry;
        _cornerParams = cornerParameters;

        _parentList = parentList;
      }

      private ParamPart()
      { }

      public int Topology
      {
        get
        {
          if (_axes != null)
          {
            return _axes.Count;
          }
          else
          {
            return _cornerParams.Count - 1;
          }
        }
      }

      public ParamInfo ParamInfo
      {
        get { return _geometry; }
      }

      public ParamPartList ParentList
      {
        get { return _parentList; }
      }

      public double MaxOffset
      {
        get
        {
          if (_maxOffset < 0)
          {
            double normedOffset = _geometry.BaseGeometry.NormedMaxOffset;
            if (normedOffset == 0)
            { _maxOffset = 0; }
            else
            {
              IList<IPoint> axes = Axes;
              double d20 = 0;
              int n = axes.Count;
              for (int i0 = 0; i0 < n; i0++)
              {
                double d2;

                d2 = PntOp.OrigDist2(axes[i0]);
                if (d2 > d20)
                { d20 = d2; }

                for (int i1 = i0 + 1; i1 < n; i1++)
                {
                  d2 = PntOp.Dist2(axes[i0], axes[i1]);
                  if (d2 > d20)
                  { d20 = d2; }
                }
              }
              _maxOffset = d20 * normedOffset;
            }
          }
          return _maxOffset;
        }
      }

      private List<ParamPart> _borders;
      public IList<ParamPart> Borders
      {
        get
        {
          if (_borders == null)
          {
            int n = Axes.Count + 1;
            _borders = new List<ParamPart>(n);
            for (int j = n - 1; j >= 0; j--)
            {
              ParamPart border = new ParamPart();
              border._axes = new List<IPoint>(n - 2);
              if (j != n - 1)
              {
                border._origin = Origin;
                for (int i = 0; i < n - 1; i++)
                {
                  if (i != j)
                  {
                    border._axes.Add(Axes[i]);
                  }
                }
              }
              else
              {
                IPoint ax0 = _axes[0];
                border._origin = PntOp.Add(Origin, ax0);
                for (int i = 1; i < n - 1; i++)
                {
                  border._axes.Add(PntOp.Sub(Axes[i], ax0));
                }
              }

              _borders.Add(border);
            }
          }
          return _borders;
        }
      }

      public IPoint BoxCorner
      {
        get
        {
          int n = _cornerParams.Count - 1;
          IPoint p = _cornerParams[n];
          return p;
        }
      }

      public IBox ParamBox
      {
        get
        {
          if (_paramBox == null)
          {
            IPoint range0 = Point.Create(_cornerParams[0]);
            IPoint range1 = Point.Create(BoxCorner);
            IBox range = new Box(range0, range1, true);
            _paramBox = range;
          }
          return _paramBox;
        }
      }

      public LinearIntersect LinearIntersect(ParamPart other)
      {
        OrthogonalSystem fullSystem = OrthoSystem.Clone();
        LinearIntersect intersect =
          GeometryOperator.LinearIntersect.Create(Origin, fullSystem, other.Origin, other.Axes);
        return intersect;
      }

      public IBox GeometryBox()
      {
        IPoint min = ParamInfo.GeometryParamAt(ParamBox.Min);
        IPoint max = ParamInfo.GeometryParamAt(ParamBox.Max);
        Box box = new Box(min, max);
        return box;
      }

      private void InitSystem()
      {
        int n = _cornerParams.Count;
        _axes = new List<IPoint>(n - 1);
        for (int i = 0; i < n; i++)
        {
          IPoint corner = _geometry.CachedPointAt(_cornerParams[i]);
          if (i == 0)
          { _origin = corner; }
          else
          {
            Point axis = PntOp.Sub(corner, _origin);
            _axes.Add(axis);
          }
        }
      }

      private IPoint Origin
      {
        get
        {
          if (_origin == null)
          { InitSystem(); }
          return _origin;
        }
      }
      private IList<IPoint> Axes
      {
        get
        {
          if (_axes == null)
          { InitSystem(); }
          return _axes;
        }
      }

      private OrthogonalSystem OrthoSystem
      {
        get
        {
          if (_orthoSystem == null)
          { _orthoSystem = OrthogonalSystem.Create(Axes); }
          return _orthoSystem;
        }
      }


      public static ParamPart CreatePart(ParamPart part, Point center)
      {
        ParamPart p = new ParamPart();

        p._origin = center;
        p._axes = new List<IPoint>(part.Axes.Count);

        foreach (IPoint axis in part.Axes)
        {
          Point a = PntOp.Add(axis, part.Origin) - center;
          p._axes.Add(a);
        }
        return p;
      }
    }

    private class ParamsRelation
    {
      private ParamPart _x;
      private ParamPart _y;
      private List<ParamsRelation> _parentList;

      private LinearIntersect _linearIntersect;
      private ParamRelate _rel;

      public ParamsRelation(ParamPart x, ParamPart y, List<ParamsRelation> parentList)
      {
        _x = x;
        _y = y;
        _parentList = parentList;

        _rel = ParamRelate.Unknown;
      }

      public bool IsKnown()
      {
        if (_linearIntersect != null)
        { return true; }

        if (_x.ParentList != null && _x.ParentList.IsSplit(_x.BoxCorner, _y))
        { return true; }

        if (_y.ParentList != null && _y.ParentList.IsSplit(_y.BoxCorner, _x))
        { return true; }

        return false;
      }

      private IPoint _intersection;
      private ParamGeometryRelation _paramRelation;

      public IPoint Intersection
      {
        get
        {
          if (_intersection == null)
          { Solve(LinearIntersect); }
          return _intersection;
        }
      }

      private void Solve(LinearIntersect intersect)
      {
        if (_x.ParamInfo.BaseGeometry.IsLinear &&
          _y.ParamInfo.BaseGeometry.IsLinear)
        {
          _intersection = intersect.Near;
          _paramRelation = new ParamGeometryRelation(
            _x.ParamInfo.BaseGeometry, null, _y.ParamInfo.BaseGeometry, null,
            Point.Create(_x.ParamInfo.BaseGeometry.Dimension));
          _paramRelation.Intersection = _intersection;
          return;
        }

        IParamGeometry xGeom = _x.ParamInfo.BaseGeometry;
        IBox xBox = _x.GeometryBox();
        int xDim = xGeom.Topology;

        IParamGeometry yGeom = _y.ParamInfo.BaseGeometry;
        IBox yBox = _y.GeometryBox();
        int yDim = yGeom.Topology;

        IPoint xParam = intersect.Parameter(0, _x.ParamBox);
        xParam = _x.ParamInfo.GeometryParamAt(xParam);

        IPoint yParam = intersect.Parameter(1, _y.ParamBox);
        yParam = _y.ParamInfo.GeometryParamAt(yParam);

        IPoint xAt = xGeom.PointAt(xParam);
        IPoint yAt = yGeom.PointAt(yParam);
        Point d = PntOp.Sub(xAt, yAt);
        double d2 = d.OrigDist2();

        while (d2 > 1.0e-12)
        {
          OrthogonalSystem xSys = XSystem(xGeom, xParam, xBox, xAt);

          IList<IPoint> yApprox;
          ITangentParamGeometry yTan = yGeom as ITangentParamGeometry;
          if (yTan != null)
          {
            yApprox = yTan.TangentAt(yParam);
          }
          else
          { yApprox = Sekante(yGeom, yParam, yBox, yAt); }

          intersect = GeometryOperator.LinearIntersect.Create(xAt, xSys, yAt, yApprox);

          for (int j = 0; j < xDim; j++)
          {
            SetDeltaParam(xParam, j, intersect.VectorFactors[j], xBox);
          }
          for (int j = 0; j < yDim; j++)
          {
            SetDeltaParam(yParam, j, -intersect.VectorFactors[j + xDim], yBox);
          }

          xAt = xGeom.PointAt(xParam);
          yAt = yGeom.PointAt(yParam);
          d = PntOp.Sub(xAt, yAt);
          d2 = d.OrigDist2();
        }

        _intersection = xAt;
        _paramRelation = new ParamGeometryRelation(
          _x.ParamInfo.BaseGeometry, xParam, _y.ParamInfo.BaseGeometry, yParam,
          Point.Create(_x.ParamInfo.BaseGeometry.Dimension));
        _paramRelation.Intersection = _intersection;

      }

      public ParamGeometryRelation ParamRelation
      {
        get
        {
          if (_paramRelation == null)
          { Solve(LinearIntersect); }

          return _paramRelation;
        }
      }
      private static void SetDeltaParam(IPoint param, int index, double delta, IBox box)
      {
        double newParam = param[index] + delta;
        if (newParam > box.Max[index])
        { newParam = box.Max[index]; }
        else if (newParam < box.Min[index])
        { newParam = box.Min[index]; }
        param[index] = newParam;
      }

      private static OrthogonalSystem XSystem(IParamGeometry xGeom, IPoint xParam, IBox xBox, IPoint xAt)
      {
        IList<IPoint> xApprox;
        ITangentParamGeometry xTan = xGeom as ITangentParamGeometry;
        OrthogonalSystem xSys = null;
        if (xTan != null)
        {
          xApprox = xTan.TangentAt(xParam);
          xSys = new OrthogonalSystem(xApprox.Count);
          foreach (IPoint point in xApprox)
          {
            Axis newAxe = xSys.GetOrthogonal(point);
            if (newAxe.Length2 == 0)
            {
              xSys = null;
              break;
            }
            xSys.Axes.Add(newAxe);
          }
        }
        if (xSys == null)
        {
          xApprox = Sekante(xGeom, xParam, xBox, xAt);
          xSys = OrthogonalSystem.Create(xApprox);
        }
        return xSys;
      }

      private static IList<IPoint> Sekante(IParamGeometry geom, IPoint param, IBox box, IPoint at)
      {
        IList<IPoint> sek = new List<IPoint>();
        for (int i = 0; i < param.Dimension; i++)
        {
          double imax = box.Max[i];
          double idt = 0.1 * (imax - box.Min[i]);
          double iparam = param[i] + idt;
          if (iparam < imax)
          {
            idt = -idt;
            param[i] += idt;
          }
          else
          { param[i] = iparam; }

          try
          {
            IPoint ipoint = geom.PointAt(param);
            Point isek = (1 / idt) * PntOp.Sub(ipoint, at);
            sek.Add(isek);
          }
          finally
          {
            param[i] -= idt;
          }
        }
        return sek;
      }
      public LinearIntersect LinearIntersect
      {
        get
        {
          if (_linearIntersect == null)
          {
            if (_x.Topology >= _y.Topology)
            { _linearIntersect = _x.LinearIntersect(_y); }
            else
            { _linearIntersect = _y.LinearIntersect(_x); }
          }
          return _linearIntersect;
        }
      }

      /// <summary>
      /// Relation  of the x and y linear approximations
      /// </summary>
      public ParamRelate Relation
      {
        get
        {
          if (_rel == ParamRelate.Unknown)
          {
            bool xLinear = _x.ParamInfo.BaseGeometry.IsLinear;
            bool yLinear = _y.ParamInfo.BaseGeometry.IsLinear;

            LinearIntersect i = LinearIntersect;
            if (_x.MaxOffset == 0 && _y.MaxOffset == 0)
            {
              if (i.Offset2 == 0 && i.IsInside(!xLinear, !yLinear))
              { _rel = ParamRelate.Intersect; }
              else
              { _rel = ParamRelate.Disjoint; }
              return _rel;
            }
            else
            {
              if (i.Offset2 == 0 && i.IsInside(!xLinear, !yLinear))
              {
                if (_y.MaxOffset > 0)
                {
                  double n2 = _y.MaxOffset;
                  n2 = n2 * n2;
                  foreach (ParamPart border in _x.Borders)
                  {
                    LinearIntersect near = _y.LinearIntersect(border);
                    if (near.Offset2 < n2 && near.IsInside(!xLinear, !yLinear))
                    {
                      _rel = ParamRelate.Near;
                      return _rel;
                    }
                  }
                }
                if (_x.MaxOffset > 0)
                {
                  double n2 = _x.MaxOffset;
                  n2 = n2 * n2;
                  foreach (ParamPart border in _y.Borders)
                  {
                    LinearIntersect near = _x.LinearIntersect(border);
                    if (near.Offset2 < n2 && near.IsInside(!xLinear, !yLinear))
                    {
                      _rel = ParamRelate.Near;
                      return _rel;
                    }
                  }
                }
                _rel = ParamRelate.Intersect;
                return _rel;

              }
              double m2 = _x.MaxOffset + _y.MaxOffset;
              m2 = m2 * m2;
              foreach (ParamPart border in _x.Borders)
              {
                LinearIntersect near = _y.LinearIntersect(border);
                if (near.Offset2 < m2 && near.IsInside(!xLinear, !yLinear))
                {
                  _rel = ParamRelate.Near;
                  return _rel;
                }
              }
              foreach (ParamPart border in _y.Borders)
              {
                LinearIntersect near = _x.LinearIntersect(border);
                if (near.Offset2 < m2 && near.IsInside(!xLinear, !yLinear))
                {
                  _rel = ParamRelate.Near;
                  return _rel;
                }
              }
              _rel = ParamRelate.Disjoint;
            }
          }
          return _rel;
        }
      }

      public void Split()
      {
        double xMax = _x.MaxOffset;
        double yMax = _y.MaxOffset;

        if (xMax > yMax)
        { Split(_x, _y); }
        else
        { Split(_y, _x); }
      }

      public ParamPart X
      { get { return _x; } }
      public ParamPart Y
      { get { return _y; } }

      private void Split(ParamPart xPart, ParamPart yPart)
      {
        ParamPartList rangeList = xPart.ParentList;
        rangeList.Split(xPart.BoxCorner, yPart, _parentList);
      }

    }

    #endregion

    public static bool Intersects(IGeometry x, IGeometry y)
    {
      TrackOperatorProgress track = new TrackOperatorProgress();
      try
      {
        track.RelationFound += intersects_RelationFound;
        IList<ParamGeometryRelation> firstRelation = CreateRelations(x, y, track);
      }
      finally
      { track.RelationFound -= intersects_RelationFound; }

      return track.Cancel;
    }

    private static void intersects_RelationFound(object sender, ParamRelationCancelArgs args)
    {
      args.Cancel = true;
    }

    public static GeometryCollection Intersection(IGeometry x, IGeometry y)
    {
      IList<ParamGeometryRelation> list = CreateRelations(x, y, null);
      if (list == null)
      { return null; }

      GeometryCollection result = new GeometryCollection();
      foreach (ParamGeometryRelation rel in list)
      { result.Add(rel.Intersection); }

      return result;
    }

    public static IList<ParamGeometryRelation> CreateRelations(IGeometry x, IGeometry y, TrackOperatorProgress track)
    {
      Relation rel = x.Extent.RelationTo(y.Extent);
      return CreateRelations(x, y, track, rel);
    }

    public static IList<ParamGeometryRelation> CreateRelations(IGeometry x, IGeometry y,
      TrackOperatorProgress track, Relation extentRelation)
    {
      return CreateRelations(x, y, track, extentRelation, true);
    }

    private static IList<ParamGeometryRelation> CreateRelations(IGeometry x, IGeometry y,
      TrackOperatorProgress track, Relation extentRelation, bool createMethodExecuted)
    {
      if (extentRelation == Relation.Disjoint)
      { return null; }
      if (extentRelation == Relation.Contains)
      {
        if (x is IBox && track != null)
        { track.OnRelationFound(null, null); }
      }
      if (extentRelation == Relation.Within)
      {
        if (y is IBox && track != null)
        { track.OnRelationFound(null, null); }
      }

      if (track != null && track.Cancel)
      { return null; }

      if (IsMultipart(x))
      {
        if (IsMultipart(y))
        {
          IBox xExtent = x.Extent;
          IBox yExtent = y.Extent;

          if (xExtent.MaxExtent() > yExtent.MaxExtent())
          {
            return MpRelations(x, y, track);
          }
          else
          {
            return MpRelations(y, x, track);
          }
        }

        // y is single part
        if (extentRelation == Relation.Within)
        { return IntersectReduce(y, x, track, false); }
        else
        { return MpRelations(x, y, track); }
      }
      if (IsMultipart(y))
      {
        if (extentRelation == Relation.Contains)
        { return IntersectReduce(x, y, track, createMethodExecuted); }
        else
        { return MpRelations(y, x, track); }
      }
      // no multiparts
      if (x.Topology >= y.Topology)
      { return IntersectReduce(x, y, track, createMethodExecuted); }
      else
      { return IntersectReduce(y, x, track, false); }
    }

    private static IList<ParamGeometryRelation> MpRelations(
      IGeometry x, IGeometry y, TrackOperatorProgress track)
    {
      List<ParamGeometryRelation> relations = null;
      IMultipartGeometry xMulti = (IMultipartGeometry)x;

      IBox yExtent = y.Extent;
      foreach (IGeometry xPart in xMulti.Subparts())
      {
        if (track != null && track.Cancel)
        { return relations; }

        Relation extentRelation = xPart.Extent.RelationTo(yExtent);
        IList<ParamGeometryRelation> list = CreateRelations(xPart, y,
          track, extentRelation, false);
        if (list != null)
        {
          if (relations == null)
          { relations = new List<ParamGeometryRelation>(); }
          foreach (ParamGeometryRelation rel in list)
          {
            rel.AddParent(x, xPart);
            relations.Add(rel);
          }
        }
      }
      return relations;
    }

    private static IList<ParamGeometryRelation> IntersectReduce(
      IGeometry x, IGeometry y, TrackOperatorProgress track, bool createMethodExecuted)
    {
      if (!(x.Topology >= y.Topology)) throw new InvalidProgramException("Topology");
      if (!(IsMultipart(x) == false)) throw new InvalidProgramException("multipart"); ;

      if (x.Topology < x.Dimension)
      {
        if (IsMultipart(y))
        { return MpRelations(y, x, track); }

        IParamGeometry xParam = x as IParamGeometry;
        if (xParam == null && x is IBox)
        { xParam = new ParamBox((IBox)x); }

        IParamGeometry yParam = y as IParamGeometry;
        if (yParam == null && y is IBox)
        { yParam = new ParamBox((IBox)y); }

        if (xParam == null || yParam == null)
        {
          throw new NotImplementedException("Base geometry parts are not typeof " +
          typeof(IParamGeometry));
        }

        if (createMethodExecuted)
        {
          IList<ParamGeometryRelation> list = CreateRelations(xParam, yParam, track);
          return list;
        }

        {
          IList<ParamGeometryRelation> list = xParam.CreateRelations(yParam, track);
          return list;
        }
      }
      else
      {
        IList<ParamGeometryRelation> list = CreateRelations(x.Border, y, track);
        return list;
      }
    }

    public static bool IsMultipart(IGeometry geom)
    {
      IMultipartGeometry multi = geom as IMultipartGeometry;
      return multi != null && multi.HasSubparts;
    }

    private class ParamBox : ITangentParamGeometry
    {
      private IBox _box;
      private IBox _paramRange;

      public ParamBox(IBox box)
      { _box = box; }

      public IList<IPoint> TangentAt(IPoint parameters)
      {
        List<IPoint> tangents = new List<IPoint>(parameters.Dimension);

        int d = Dimension;

        for (int i = 0; i < d; i++)
        {
          double o = _box.Max[i] - _box.Min[i];
          if (o > 0)
          {
            Point p = Point.Create(d);
            p[i] = o;
            tangents.Add(p);
          }
        }

        return tangents;
      }

      public IBox ParameterRange
      {
        get
        {
          if (_paramRange == null)
          {
            int t = Topology;
            Point min = Point.Create(t);
            Point max = Point.Create(t);
            for (int i = 0; i < t; i++)
            { max[i] = 1; }

            _paramRange = new Box(min, max);
          }
          return _paramRange;
        }
      }

      public double NormedMaxOffset
      {
        get { return 0; }
      }

      public IPoint PointAt(IPoint parameters)
      {
        int d = Dimension;
        int t = 0;

        Point at = Point.Create(_box.Min);
        for (int i = 0; i < d; i++)
        {
          double o = _box.Max[i] - _box.Min[i];
          if (o > 0)
          {
            at[i] += o * parameters[t];
            t++;
          }
        }

        return at;
      }

      public bool IsLinear
      { get { return true; } }

      public ParamRelate Relation(IBox paramBox)
      {
        int n = paramBox.Dimension;
        ParamRelate rel = ParamRelate.Intersect;
        for (int i = 0; i < n; i++)
        {
          double min = paramBox.Min[i];
          double max = paramBox.Max[i];
          if (min > 1 || max < 0)
          { return ParamRelate.Disjoint; }
          if (min < 0 || max > 1)
          { rel = ParamRelate.Near; }
        }
        return rel;
      }

      public IList<ParamGeometryRelation> CreateRelations(IParamGeometry other,
        TrackOperatorProgress trackProgress)
      {
        return GeometryOperator.CreateRelations(this, other, trackProgress);
      }

      public int Dimension
      { get { return _box.Dimension; } }

      public int Topology
      { get { return _box.Topology; } }

      public bool IsWithin(IPoint point)
      { return _box.IsWithin(point); }

      public IBox Extent
      { get { return _box; } }

      public IGeometry Border
      { get { return _box.Border; } }

      public IGeometry Project(IProjection projection)
      { return _box.Project(projection); }

      public bool EqualGeometry(IGeometry other)
      { return _box == other; }
    }

    private static IList<ParamGeometryRelation> CreateRelations(
      IParamGeometry x, IParamGeometry y, TrackOperatorProgress track)
    {
      ParamInfo xInfo = new ParamInfo(x);
      ParamInfo yInfo = new ParamInfo(y);
      IList<ParamPart> xParts = ParamPart.CreateParts(xInfo);
      IList<ParamPart> yParts = ParamPart.CreateParts(yInfo);

      List<ParamGeometryRelation> result = null;

      List<ParamsRelation> relations = new List<ParamsRelation>();
      foreach (ParamPart xPart in xParts)
      {
        foreach (ParamPart yPart in yParts)
        {
          ParamsRelation rel = new ParamsRelation(xPart, yPart, relations);
          relations.Add(rel);
        }
      }

      for (int i = 0; i < relations.Count; i++)
      {
        if (track != null && track.Cancel)
        { return result; }

        ParamsRelation rel = relations[i];
        if (rel.IsKnown())
        { continue; }

        ParamRelate r = rel.Relation;
        if (r == ParamRelate.Disjoint)
        { continue; }

        if (r == ParamRelate.Intersect)
        {
          ParamRelate xRel = x.Relation(rel.X.ParamBox);
          if (xRel == ParamRelate.Disjoint)
          { continue; }
          ParamRelate yRel = y.Relation(rel.Y.ParamBox);
          if (yRel == ParamRelate.Disjoint)
          { continue; }

          if (xRel == ParamRelate.Intersect && yRel == ParamRelate.Intersect)
          {
            if (result == null)
            { result = new List<ParamGeometryRelation>(); }
            result.Add(rel.ParamRelation);

            if (track != null)
            { track.OnRelationFound(track, rel.ParamRelation); }
            continue;
          }

          if (rel.X.ParamInfo.BaseGeometry.IsLinear == false ||
            rel.Y.ParamInfo.BaseGeometry.IsLinear == false)
          {
            rel.Split();
          }
          continue;
        }

        if (r != ParamRelate.Near)
        { throw new InvalidOperationException("Unhandled relation " + r); }

        rel.Split();
      }

      return result;
    }
  }
}
