using System;
using System.Diagnostics;

namespace Mathe.Geom
{
  public partial class Mesh
  {
    public abstract class Tri : IGeometry
    {
      private Point2D _center = null;
      private double _radius2 = 0;

      private uint _processed;

      public bool EqualGeometry(IGeometry other)
      { return this == other; }

      public abstract MeshLine this[int i] { get; }

      public Point2D Center
      {
        get
        {
          if (_center == null)
          { CalcCenterRadius(); }
          return _center;
        }
        protected set { _center = value; }
      }
      public double Radius2
      {
        get
        {
          if (_radius2 <= 0)
          { CalcCenterRadius(); }
          return _radius2;
        }
        protected set { _radius2 = value; }
      }

      private void CalcCenterRadius()
      {
        IPoint p0 = this[0].Start;
        _center = Geometry.OutCircleCenter(p0, this[1].Start, this[2].Start);
        if (_center != null)
        { _radius2 = _center.Dist2(p0); }
        else
        { _radius2 = double.PositiveInfinity; }
      }

      public int GetLineIndex(MeshLine line)
      {
        for (int i = 0; i < 3; i++)
        {
          if (this[i].HasEqualBaseLine(line) &&
            this[i].IsReverse == line.IsReverse)
          { return i; }
        }
        return -1;
      }

      /// <summary>
      /// M2GetNextLine
      /// </summary>
      public abstract MeshLine GetNextLine(MeshLine line);

      /// <summary>
      /// M2GetPreviousLine
      /// </summary>
      public abstract MeshLine GetPreviousLine(MeshLine line);

      public IPoint GetPoint(int i)
      {
        return this[i].Start;
      }

      public int GetPointIndex(Point2D point)
      {
        if (GetPoint(0) == point) { return 0; }
        if (GetPoint(1) == point) { return 1; }
        if (GetPoint(2) == point) { return 2; }
        return -1;
      }
      #region IGeometry Members

      public int Dimension
      {
        get        { return 2; }
      }

      public int Topology
      {
        get { return 2; }
      }

      public bool IsWithin(IPoint point)
      {
        IPoint p;
        double x, y;

        IPoint p0 = this[0].Start;
        double dx = point.X - p0.X;
        double dy = point.Y - p0.Y;
    
        p = this[1].Start;
        x = p.X - p0.X;
        y = p.Y - p0.Y;
        double u = x * dx + y * dy;
        if (u < 0 || u > 1)
        { return false; }

        p = this[2].Start;
        x = p.X - p0.X;
        y = p.Y - p0.Y;
        double v = x * dx + y * dy;

        if (v < 0 || v > 1 || u + v > 1)
        { return false; }
        return true;
      }

      public abstract Box Extent { get; }
      IBox IGeometry.Extent
      { get { return Extent; } }

      public Polyline Border
      {
        get
        {
          return new Polyline(
            new IPoint[] { this[0].Start, this[1].Start, this[2].Start });
        }
      }
      IGeometry IGeometry.Border
      { get { return Border; } }

      IGeometry IGeometry.Project(IProjection projection)
      { return null; }

      public IGeometry Split(IGeometry[] border)
      {
        throw new NotImplementedException();
      }

      #endregion

      internal bool GetProcessed(uint processId)
      {
        return ((_processed & processId) != 0);
      }
      internal void SetProcessed(uint processId, bool processed)
      {
        if (processed)
        { _processed |= processId; }
        else
        { _processed = (_processed & processId) ^ _processed; }
      }

    }

    internal class _Tri : Tri
    {
      private _MeshLine[] _line;
      public _Tri(_MeshLine l0, _MeshLine l1, _MeshLine l2)
      {
        _line = new _MeshLine[] { l0, l1, l2 };
        l0._LeftTri = this;
        l1._LeftTri = this;
        l2._LeftTri = this;
      }

      public override MeshLine this[int i]
      {
        get
        { return _line[i]; }
      }
      public _MeshLine _GetLine(int i)
      {
        return _line[i];
      }
      public void _SetLine(int i, _MeshLine l)
      {
        _line[i] = l;
        l._LeftTri = this;
      }

      public _MeshLine _GetNextLine(MeshLine line)
      {
        int index = GetLineIndex(line);
        if (index < 0)
        { return null; }

        return _GetLine(index < 2 ? index + 1 : 0);
      }
      public override MeshLine GetNextLine(MeshLine line)
      {
        return _GetNextLine(line);
      }

      public _MeshLine _GetPreviousLine(MeshLine line)
      {
        int index = GetLineIndex(line);
        if (index < 0)
        { return null; }

        return _GetLine(index > 0 ? index - 1 : 2);
      }

      public override MeshLine GetPreviousLine(MeshLine line)
      {
        return _GetPreviousLine(line);
      }

      public void ClearCircle()
      {
        Center = null;
        Radius2 = 0;
      }

      public bool Validate()
      {
        //for (int i = 0; i < 3; i++)
        //{
        //  if (_line[i].LeftTri != this)
        //  { return false; }
        //  if (_line[i].End != _line[i < 2 ? i + 1 : 0].Start)
        //  { return false; }
        //  if (_line[i].Start == _line[i < 2 ? i + 1 : 0].Start)
        //  { return false; }
        //}

        return true;
      }

      public override Box Extent
      {
        get
        {
          IPoint p0 = Point.Create(_line[0].Start);
          IPoint p1 = Point.Create(p0);
          for (int i = 1; i < 3; i++)
          {
            IPoint p = _line[i].Start;
            if (p.X < p0.X)
            { p0.X = p.X; }
            else if (p.X > p1.X)
            { p1.X = p.X; }
            if (p.Y < p0.Y)
            { p0.Y = p.Y; }
            else if (p.Y > p1.Y)
            { p1.Y = p.Y; }
          }
          return new Box(p0, p1);
        }
      }
    }

    internal class MeshBaseLine
    {
      private MeshPoint _startPoint;
      private MeshPoint _endPoint;
      private _Tri _leftTri;
      private _Tri _rightTri;
      private object _tag;
      private bool _tagReverse;

      private uint _processed;

      public MeshBaseLine(MeshPoint startPoint, MeshPoint endPoint, object tag)
      {
        _startPoint = startPoint;
        _endPoint = endPoint;
        _tag = tag;
      }

      public MeshPoint StartPoint
      {
        get
        { return _startPoint; }
        set
        { _startPoint = value; }
      }

      public MeshPoint EndPoint
      {
        get
        { return _endPoint; }
        set
        { _endPoint = value; }
      }

      public _Tri LeftTri
      {
        get
        { return _leftTri; }
        set
        { _leftTri = value; }
      }

      public _Tri RightTri
      {
        get
        { return _rightTri; }
        set
        { _rightTri = value; }
      }

      public object GetTag(out bool isReverse)
      {
        isReverse = _tagReverse;
        return _tag;
      }

      public void SetTag(object tag, bool isReverse)
      {
        _tag = tag;
        _tagReverse = isReverse;
      }

      internal bool GetProcessed(uint processId)
      {
        return ((_processed & processId) != 0);
      }
      internal void SetProcessed(uint processId, bool processed)
      {
        if (processed)
        { _processed |= processId; }
        else
        { _processed = (_processed & processId) ^ _processed; }
      }
    }

    public abstract class MeshLine : Line
    {
      private bool _isReverse;

      public abstract Tri LeftTri { get; }
      public abstract Tri RightTri { get; }
      public abstract object GetTag(out bool isReverse);
      public abstract void SetTag(object tag, bool isReverse);

      public bool IsReverse
      {
        get { return _isReverse; }
        protected set { _isReverse = value; }
      }

      public abstract bool HasEqualBaseLine(MeshLine line);
      public abstract MeshLine Inverse();

      public static MeshLine operator -(MeshLine line)
      {
        return line.Inverse();
      }

      /// <summary>
      /// M2GetNextLine_
      /// </summary>
      public MeshLine GetNextTriLine()
      {
        Tri tri = LeftTri;
        if (tri == null)
        { return null; }
        return tri.GetNextLine(this);
      }

      /// <summary>
      /// M2GetPreviosLine_
      /// </summary>
      public MeshLine GetPreviousTriLine()
      {
        Tri tri = LeftTri;
        if (tri == null)
        { return null; }
        return tri.GetPreviousLine(this);
      }

      /// <summary>
      /// return next line after this starting at start point of this 
      /// in counterclockwise direction
      /// </summary>
      public MeshLine GetNextPointLine()
      {
        Tri tri = LeftTri;
        if (tri != null)
        { return -tri.GetPreviousLine(this); }

        MeshLine l = this;
        tri = RightTri;
        while (tri != null)
        {
          l = tri.GetNextLine(-l);
          tri = l.RightTri;
        }
        return l;
      }

      /// <summary>
      /// return next line after this starting at start point of this 
      /// in clockwise direction
      /// </summary>
      public MeshLine GetPreviousPointLine()
      {
        Tri tri = RightTri;
        if (tri != null)
        { return tri.GetNextLine(-this); }

        MeshLine l = this;
        tri = LeftTri;
        while (tri != null)
        {
          l = -tri.GetPreviousLine(l);
          tri = l.LeftTri;
        }
        return l;
      }

      public abstract void ChangeDiag(int rec);

      public bool OptionChangeDiag(int rec)
      {
        bool isReverse;
        if (GetTag(out isReverse) != null)
        { return false; }

        Tri lTri = LeftTri;

        int i0 = lTri.GetLineIndex(this);
        int i1 = i0 < 2 ? i0 + 1 : 0;

        IPoint left = lTri[i1].End;
        Tri rTri = RightTri;
        if (rTri == null)
        {
          if (double.IsPositiveInfinity(lTri.Radius2) == false)
          { return false; }
          ChangeDiag(rec);
          return true;
        }
        if (rTri.Center != null && rTri.Center.Dist2(left) >= rTri.Radius2)
        { return false; }

        ChangeDiag(rec);
        return true;
      }
      #region ILine Members
      public int Count
      {
        get
        { return 2; }
      }
      public IPoint this[int index]
      {
        get
        {
          if (index == 0)
          { return Start; }
          else if (index == 1)
          { return End; }
          else
          { throw new InvalidOperationException(string.Format("Index {0} out of Range", index)); }
        }
      }
      #endregion
    }


    internal class _MeshLine : MeshLine
    {
      private MeshBaseLine _baseLine;

      public _MeshLine(MeshPoint startPoint, MeshPoint endPoint, object tag, bool isReverse)
      {
        _baseLine = new MeshBaseLine(startPoint, endPoint, tag);
        IsReverse = isReverse;
      }
      public _MeshLine(MeshPoint startPoint, MeshPoint endPoint)
      {
        _baseLine = new MeshBaseLine(startPoint, endPoint, null);
        IsReverse = false;
      }
      public _MeshLine(MeshBaseLine baseLine, bool isReverse)
      {
        _baseLine = baseLine;
        IsReverse = isReverse;
      }

      public override MeshLine Inverse()
      {
        return new _MeshLine(BaseLine, !(IsReverse));
      }

      public MeshPoint _StartPoint
      {
        get
        {
          if (IsReverse == false)
          { return _baseLine.StartPoint; }
          else
          { return _baseLine.EndPoint; }
        }
        set
        {
          if (IsReverse == false)
          { _baseLine.StartPoint = value; }
          else
          { _baseLine.EndPoint = value; }
        }
      }

      public override IPoint Start
      {
        get { return _StartPoint.Point; }
      }

      public MeshPoint _EndPoint
      {
        get
        {
          if (IsReverse == false)
          { return _baseLine.EndPoint; }
          else
          { return _baseLine.StartPoint; }
        }
        set
        {
          if (IsReverse == false)
          { _baseLine.EndPoint = value; }
          else
          { _baseLine.StartPoint = value; }
        }
      }

      public override IPoint End
      {
        get { return _EndPoint.Point; }
      }


      public _Tri _LeftTri
      {
        get
        {
          if (IsReverse == false)
          { return _baseLine.LeftTri; }
          else
          { return _baseLine.RightTri; }
        }
        set
        {
          if (IsReverse == false)
          { _baseLine.LeftTri = value; }
          else
          { _baseLine.RightTri = value; }
        }
      }

      public override Tri LeftTri
      {
        get
        { return _LeftTri; }
      }

      public _Tri _RightTri
      {
        get
        {
          if (IsReverse == false)
          { return _baseLine.RightTri; }
          else
          { return _baseLine.LeftTri; }
        }
      }

      public override Tri RightTri
      {
        get
        { return _RightTri; }
      }

      public override object GetTag(out bool isReverse)
      {
        object tag = _baseLine.GetTag(out isReverse);
        isReverse = (IsReverse == isReverse);
        return tag;
      }
      public override void SetTag(object tag, bool isReverse)
      {
        isReverse = (isReverse == IsReverse);
        _baseLine.SetTag(tag, isReverse);
      }

      public MeshBaseLine BaseLine
      {
        get
        { return _baseLine; }
      }

      public override bool HasEqualBaseLine(MeshLine line)
      {
        _MeshLine l = line as _MeshLine;
        if (l == null)
        { return false; }
        return _baseLine == l.BaseLine;
      }


      public new _MeshLine GetNextTriLine()
      {
        _Tri tri = _LeftTri;
        if (tri == null)
        { return null; }
        return tri._GetNextLine(this);
      }

      public static _MeshLine operator -(_MeshLine line)
      {
        return new _MeshLine(line._baseLine, !(line.IsReverse));
      }

      /// <summary>
      /// M2GetNextLine_
      /// </summary>
      public _MeshLine _GetNextTriLine()
      {
        _Tri tri = _LeftTri;
        if (tri == null)
        { return null; }
        return tri._GetNextLine(this);
      }

      /// <summary>
      /// M2GetPreviosLine_
      /// </summary>
      public _MeshLine _GetPreviousTriLine()
      {
        _Tri tri = _LeftTri;
        if (tri == null)
        { return null; }
        return tri._GetPreviousLine(this);
      }

      /// <summary>
      /// return next line after this starting at start point of this 
      /// in counterclockwise direction
      /// </summary>
      public _MeshLine _GetNextPointLine()
      {
        _Tri tri = _LeftTri;
        if (tri != null)
        { return -tri._GetPreviousLine(this); }

        _MeshLine l = this;
        tri = _RightTri;
        while (tri != null)
        {
          l = tri._GetNextLine(-l);
          tri = l._RightTri;
        }
        return l;
      }

      /// <summary>
      /// return next line after this starting at start point of this 
      /// in clockwise direction
      /// </summary>
      public _MeshLine _GetPreviousPointLine()
      {
        _Tri tri = _RightTri;
        if (tri != null)
        { return tri._GetNextLine(-this); }

        _MeshLine l = this;
        tri = _LeftTri;
        while (tri != null)
        {
          l = -tri._GetPreviousLine(l);
          tri = l._LeftTri;
        }
        return l;
      }

      /// <summary>
      /// splits the meshline at point by creating a new meshline from start to point
      /// (return value) and replacing the start point of this by point
      /// </summary>
      /// <param name="point"></param>
      /// <returns>new meshline</returns>
      public _MeshLine Split(MeshPoint point)
      {
        _MeshLine l = new _MeshLine(_StartPoint, point);
        _StartPoint.IndexLine = l;
        _StartPoint = point;
        point.IndexLine = this;
        return l;
      }

      public override void ChangeDiag(int rec)
      {
        if (rec > 100)
        {}

        _Tri lTri = _LeftTri;
        _Tri rTri = _RightTri;
        int i0 = lTri.GetLineIndex(this);
        int i1 = (i0 < 2 ? i0 + 1 : 0);
        int i2 = (i1 < 2 ? i1 + 1 : 0);
        _MeshLine lL1 = lTri._GetLine(i1);
        _MeshLine lL2 = lTri._GetLine(i2);

        if (rTri == null)
        { // Drop line
          if (lL1._baseLine.LeftTri == lTri)
          { lL1._baseLine.LeftTri = null; }
          else
          {
            Debug.Assert(lL1._baseLine.RightTri == lTri);
            lL1._baseLine.RightTri = null;
          }
          if (lL2._baseLine.LeftTri == lTri)
          { lL2._baseLine.LeftTri = null; }
          else
          {
            Debug.Assert(lL2._baseLine.RightTri == lTri);
            lL2._baseLine.RightTri = null;
          }
          lL1._StartPoint.IndexLine = lL1;
          lL2._EndPoint.IndexLine = -lL2;

          return;
        }

        _MeshLine rL0 = -this;
        i0 = rTri.GetLineIndex(rL0);
        i1 = (i0 < 2 ? i0 + 1 : 0);
        i2 = (i1 < 2 ? i1 + 1 : 0);
        _MeshLine rL1 = rTri._GetLine(i1);
        _MeshLine rL2 = rTri._GetLine(i2);

        lTri.ClearCircle();
        rTri.ClearCircle();

        _StartPoint = rL2._StartPoint;
        _EndPoint = lL2._StartPoint;

        lTri._SetLine(0, rL2);
        lTri._SetLine(1, lL1);
        lTri._SetLine(2, rL0);

        rTri._SetLine(0, rL1);
        rTri._SetLine(1, this);
        rTri._SetLine(2, lL2);

        if (lTri.Validate() == false) throw new InvalidProgramException("Error near " + rL2.Start);
        if (rTri.Validate() == false) throw new InvalidProgramException("Error near " + rL1.Start);

        lL1._StartPoint.IndexLine = lL1;
        rL1._StartPoint.IndexLine = rL1;

        rL2.OptionChangeDiag(rec + 1);
        rL1.OptionChangeDiag(rec + 1);
      }
    }

    public class MeshPoint : IGeometry
    {
      private IPoint _point;
      private _MeshLine _indexLine;

      public MeshPoint(IPoint point)
      {
        _point = point;
      }
      public IPoint Point
      {
        get { return _point; }
      }

      public bool EqualGeometry(IGeometry other)
      { return this == other; }

      internal _MeshLine IndexLine
      {
        get { return _indexLine; }
        set
        {
          Debug.Assert(value._StartPoint == this);
          _indexLine = value;
        }
      }
      #region IGeometry Members

      public int Dimension
      {
        get
        { return _point.Dimension; }
      }

      public int Topology
      {
        get
        { return _point.Topology; }
      }

      public bool IsWithin(IPoint point)
      { return false; }

      public bool Intersects(IGeometry other)
      {
        return other.IsWithin(_point);
      }

      public IBox Extent
      {
        get
        { return _point.Extent; }
      }

      public IGeometry Border
      { get { return null; } }

      public IPoint Project(IProjection projection)
      {
        return _point.Project(projection);
      }
      IGeometry IGeometry.Project(IProjection projection)
      { return Project(projection); }

      #endregion
    }

    private class PointBoxTree : BoxTree<MeshPoint>
    {
      internal PointBoxTree()
        : base(2)
      { }
      internal PointBoxTree(int capacity)
        : base(2, capacity, true)
      { }
      internal int Add(MeshPoint point)
      {
        return Add(point.Point.Extent, point);
      }
      internal void Insert(int index, MeshPoint point)
      {
        Insert(index, new TileEntry(point.Point.Extent, point));
      }
    }
  }
}
