using System;
using System.Collections.Generic;
using System.Data;
using System.Text;
using Ocad;
using Ocad.StringParams;

namespace OCourse.Ext
{
  public class VariationBuilder
  {
    public event EventHandler HandlingSplit;

    public List<SectionList> BuildPermutations(SectionCollection course, int nPermutations)
    {
      PermutationBuilder builder = new PermutationBuilder(course);
      builder.HandlingSplit += builder_HandlingSplit;
      List<SectionList> permutations = builder.BuildPermutations(nPermutations);

      return permutations;
    }

    public List<SectionList> AnalyzeLeg(SectionCollection course, int leg)
    {
      LegBuilder builder = new LegBuilder(course, leg);

      List<SectionList> permuts = builder.Analyze();
      return permuts;
    }
    public List<SectionList> Analyze(SectionCollection course)
    {
      CourseBuilder builder = new CourseBuilder(course);

      List<SectionList> permuts = builder.Analyze();
      return permuts;
    }


    protected virtual void OnHandlingSplit(EventArgs e)
    {
      if (HandlingSplit != null)
      { HandlingSplit(this, e); }
    }
    private void builder_HandlingSplit(object sender, EventArgs e)
    {
      OnHandlingSplit(e);
    }

    public class WhereInfo
    {
      public static WhereInfo Failed = new WhereInfo(State.Failed);
      public static WhereInfo FulFilled = new WhereInfo(State.Fulfilled);

      public enum State { Fulfilled, Unknown, Failed }
      [ThreadStatic]
      private static DataView _whereView;
      [ThreadStatic]
      private static DataRow _exprRow;
      [ThreadStatic]
      private static DataColumn _exprColumn;

      private int _max;
      public int MaxControl { get { return _max; } }

      private State _state;
      private string _parsed;

      public WhereInfo(string where, SectionList permut)
      {
        string parsed;
        _state = GetState(where, permut, out parsed);

        _parsed = parsed;
      }
      private WhereInfo(State state)
      {
        _state = state;
      }

      public State GetState(SectionList sections)
      {
        if (_state == State.Failed)
        { return State.Failed; }
        if (_state == State.Fulfilled)
        { return State.Fulfilled; }

        if (sections.ControlsCount <= _max)
        { return State.Unknown; }

        string parsed;
        State state = GetState(_parsed, sections, out parsed);

        return state;
      }

      private State GetState(string where, SectionList sections, out string parse)
      {
        parse = where;
        if (string.IsNullOrEmpty(where))
        {
          return State.Fulfilled;
        }

        if (_whereView == null)
        {
          DataTable whereTbl = new DataTable();
          whereTbl.Rows.Add();
          whereTbl.AcceptChanges();
          _whereView = new DataView(whereTbl);
        }

        if (where.Contains("leg"))
        {
          int leg = 0;
          Control start = sections.Start;
          foreach (Control ctr in sections.Controls)
          {
            if (ctr == start)
            { leg++; }
          }
          where = where.Replace("leg", leg.ToString());
        }
        StringBuilder parsed = new StringBuilder();
        for (int idx = where.IndexOf("["); idx >= 0; idx = where.IndexOf("["))
        {
          int end = where.IndexOf("]");
          parsed.Append(where.Substring(0, end + 1));
          string ctrName = where.Substring(idx + 1, end - idx - 1);

          where = where.Substring(end + 1);

          if (_exprRow == null)
          {
            DataTable exprTbl = new DataTable();
            _exprColumn = exprTbl.Columns.Add("Expr", typeof(int), "1");
            _exprRow = exprTbl.NewRow();
            exprTbl.Rows.Add(_exprRow);
            _exprRow.AcceptChanges();
          }

          string ctrExpr = ctrName.Replace("#", sections.ControlsCount.ToString());
          _exprColumn.Expression = ctrExpr;

          int ctrIdx = (int)_exprRow[_exprColumn];
          if (ctrIdx < 0)
          {
            parse = parsed.ToString();
            return State.Failed;
          }

          if (ctrIdx <= sections.ControlsCount)
          {
            Control ctr = sections.GetControl(ctrIdx);
            parsed.Replace(string.Format("[{0}]", ctrName), string.Format("'{0}'", ctr.Name));
          }
          else
          {
            parsed.Replace(string.Format("[{0}]", ctrName), string.Format("[{0}]", ctrIdx));
            _max = Math.Max(_max, ctrIdx);
          }
        }
        parsed.Append(where);

        parse = parsed.ToString();
        if (_max > sections.ControlsCount)
        {
          return State.Unknown;
        }
        _whereView.RowFilter = parse;
        if (_whereView.Count == 1)
        {
          return State.Fulfilled;
        }
        else
        {
          return State.Failed;
        }
      }
    }

    private class PermutationBuilder : Builder
    {
      private SectionCollection _course;
      private Dictionary<Control, SplitControl> _splitControls;
      private int _splitId;
      private int _nPermutations;

      protected override Control StartControl { get { return (Control)_course.First.Value; } }
      protected override Control EndControl { get { return (Control)_course.Last.Value; } }
      private Dictionary<Control, SplitControl> SplitControls
      {
        get
        {
          if (_splitControls == null)
          {
            Dictionary<Control, SplitControl> splitControls = new Dictionary<Control, SplitControl>();
            foreach (SimpleSection section in Sections)
            {
              if (!splitControls.ContainsKey(section.From))
              { splitControls.Add(section.From, new SplitControl(this, section.From)); }
              if (!splitControls.ContainsKey(section.To))
              { splitControls.Add(section.To, new SplitControl(this, section.To)); }
            }

            _splitControls = splitControls;
          }
          return _splitControls;
        }
      }

      protected override IList<SimpleSection> GetSections()
      {
        IList<SimpleSection> sections = _course.GetAllSections();
        sections = MakeUniqueControls(sections);

        return sections;
      }

      public PermutationBuilder(SectionCollection course)
      {
        SectionCollection init = course;
        if (!(init.First.Value is Control))
        {
          init = course.Clone();
          init.AddFirst(new Control("++", ControlPar.TextDescKey));
        }
        if (!(init.Last.Value is Control))
        {
          init = course.Clone();
          init.AddLast(new Control("--", ControlPar.TextDescKey));
        }

        _course = init;
      }

      public List<SectionList> BuildPermutations(int nPermutations)
      {
        _nPermutations = nPermutations;
        List<SectionList> allPermutations = InitPermutations(nPermutations);

        Control from = StartControl;
        SplitControl startSplit = SplitControls[from];
        startSplit.Add(allPermutations);

        List<SplitControl> unhandledSplits = new List<SplitControl>();
        unhandledSplits.Add(startSplit);

        while (unhandledSplits.Count > 0)
        {
          OnHandlingSplit(new EventArgs());

          SplitControl fromSplit = unhandledSplits[0];
          List<SplitControl> endSplits = BuildPermutations(fromSplit);
          unhandledSplits.RemoveAt(0);
          foreach (SplitControl endSplit in endSplits)
          {
            if (!unhandledSplits.Contains(endSplit))
            { unhandledSplits.Add(endSplit); }
          }
          unhandledSplits.Sort(SplitControl.CompleteCompare);
        }

        return allPermutations;
      }

      private List<SectionList> InitPermutations(int nParticipants)
      {
        List<SectionList> permutations = new List<SectionList>();
        string countExpr = GetCountExpression(_course);
        for (int i = 0; i < nParticipants; i++)
        {
          SectionList participant = new SectionList(StartControl);
          WhereInfo countWhere = new WhereInfo(countExpr, participant);
          participant.Add(countWhere);

          permutations.Add(participant);
        }
        _splitId = 0;

        return permutations;
      }

      private List<SplitControl> BuildPermutations(SplitControl splitControl)
      {
        Dictionary<NextInfo, List<SectionList>> distributed = splitControl.GetDistributed();
        distributed = new Dictionary<NextInfo, List<SectionList>>(distributed);
        splitControl.Clear();

        List<SplitControl> endSplits = new List<SplitControl>();
        foreach (KeyValuePair<NextInfo, List<SectionList>> pair in distributed)
        {
          NextInfo next = pair.Key;
          SplitControl endControlSections = null;
          List<NextInfo> nextSections = NextSections[next];
          foreach (SectionList participant in pair.Value)
          {
            AddSections(participant, nextSections);
            WhereInfo.State state = participant.GetState();
            if (state == WhereInfo.State.Fulfilled)
            { continue; }

            if (endControlSections == null)
            {
              Control endControl = participant.GetControl(-1);
              SplitControl endSplit = SplitControls[endControl];
              if (!endSplits.Contains(endSplit))
              {
                _splitId++;
                endSplit.SplitId = _splitId;
                endSplits.Add(endSplit);
              }

              endControlSections = endSplit;
            }
            endControlSections.Add(participant);
          }
        }

        return endSplits;
      }

      private class Weight
      {
        public static int Compare(Weight x, Weight y)
        {
          int i = x.W.CompareTo(y.W);
          if (i != 0)
          { return i; }

          i = x.NextInfo.Value.Count.CompareTo(y.NextInfo.Value.Count);
          if (i != 0)
          { return i; }

          i = x.NextInfo.Key.Code.CompareTo(y.NextInfo.Key.Code);
          return i;
        }
        public double W;
        public KeyValuePair<NextInfo, List<SectionList>> NextInfo;
        public Weight(double w, KeyValuePair<NextInfo, List<SectionList>> next)
        {
          W = w;
          NextInfo = next;
        }
      }

      private class Participant
      {
        public SectionList Sections;
        public Dictionary<char, NextWhere> Candidates;
        public Participant(SectionList sections, Dictionary<char, NextWhere> candidates)
        {
          Sections = sections;
          Candidates = candidates;
        }
      }

      protected class SplitControl
      {
        public readonly PermutationBuilder Parent;
        public readonly Control Control;
        public int SplitId;
        private readonly List<SectionList> _unhandled;
        private readonly Dictionary<char, List<Participant>> _variants;
        private readonly Dictionary<NextInfo, List<SectionList>> _distributed;


        public SplitControl(PermutationBuilder parent, Control control)
        {
          Parent = parent;
          Control = control;
          _unhandled = new List<SectionList>();

          _variants = new Dictionary<char, List<Participant>>();
          _distributed = new Dictionary<NextInfo, List<SectionList>>();
        }

        public static int CompleteCompare(SplitControl x, SplitControl y)
        {
          int i;
          Dictionary<NextInfo, List<SectionList>> xDistr = x.GetDistributed();
          Dictionary<NextInfo, List<SectionList>> yDistr = y.GetDistributed();
          double allParts = x.Parent._nPermutations;

          int xSplit = xDistr.Count;
          int xParts = ParticipantCount(xDistr);
          int ySplit = yDistr.Count;
          int yParts = ParticipantCount(yDistr);
          double tx = Math.Round(allParts / xParts) * xSplit;
          double ty = Math.Round(allParts / yParts) * ySplit;

          i = -tx.CompareTo(ty);
          if (i != 0)
          { return i; }

          i = x.SplitId.CompareTo(y.SplitId);
          return i;
        }

        private static int ParticipantCount(Dictionary<NextInfo, List<SectionList>> distr)
        {
          int n = 0;
          foreach (List<SectionList> list in distr.Values)
          {
            n += list.Count;
          }
          return n;
        }
        public void Clear()
        {
          _unhandled.Clear();
          _variants.Clear();
          _distributed.Clear();

          SplitId = 0;
        }
        public void Add(SectionList permutations)
        {
          _unhandled.Add(permutations);
        }
        public void Add(IEnumerable<SectionList> permutations)
        {
          _unhandled.AddRange(permutations);
        }
        public Dictionary<NextInfo, List<SectionList>> GetDistributed()
        {
          if (_unhandled.Count == 0)
          {
            return _distributed;
          }

          Control from = Control;
          NextInfoList nextInfos;
          if (!Parent.NextInfos.TryGetValue(from, out nextInfos))
          {
            from = Parent.StartControl;
            foreach (SectionList sections in _unhandled)
            { sections.Add(Parent.StartNext); }
            nextInfos = Parent.NextInfos[from];
          }

          List<Participant> participants = AssembleOptions(from);

          foreach (NextInfo next in nextInfos.List)
          {
            if (_variants.ContainsKey(next.Code) && !_distributed.ContainsKey(next))
            { _distributed.Add(next, new List<SectionList>(participants.Count)); }
          }

          foreach (Participant participant in participants)
          {
            Weight wMin = null;
            foreach (KeyValuePair<NextInfo, List<SectionList>> pair in _distributed)
            {
              char code = pair.Key.Code;
              if (!participant.Candidates.ContainsKey(code))
              { continue; }

              List<SectionList> preSet = pair.Value;
              double commonWeight = GetCommonWeight(participant.Sections, preSet);
              Weight w = new Weight(commonWeight, pair);
              if (wMin == null || Weight.Compare(w, wMin) < 0)
              { wMin = w; }
            }

            wMin.NextInfo.Value.Add(participant.Sections);
          }
          return _distributed;
        }

        private List<Participant> AssembleOptions(Control from)
        {
          NextInfoList nextInfos = Parent.NextInfos[from];
          List<Participant> participants = new List<Participant>(_unhandled.Count);

          foreach (SectionList part in _unhandled)
          {
            Dictionary<char, NextWhere> candidates = Parent.GetPossibleNexts(part, from, nextInfos.List, true);
            Participant participant = new Participant(part, candidates);
            participants.Add(participant);
            foreach (char c in candidates.Keys)
            {
              List<Participant> varParticipants;
              if (!_variants.TryGetValue(c, out varParticipants))
              {
                varParticipants = new List<Participant>(_unhandled.Count);
                _variants.Add(c, varParticipants);
              }
              varParticipants.Add(participant);
            }
          }

          _unhandled.Clear();
          return participants;
        }

        private double GetCommonWeight(SectionList sections, IList<SectionList> compare)
        {
          return 0;
        }
      }
    }

    private class LegBuilder : Builder
    {
      private SectionCollection _course;
      private int _leg;

      protected override Control StartControl { get { return (Control)_course.First.Value; } }
      protected override Control EndControl { get { return (Control)_course.Last.Value; } }

      public LegBuilder(SectionCollection course, int leg)
      {
        _course = course;
        _leg = leg;
      }

      public List<SectionList> Analyze()
      {
        SectionList permut = new SectionList(StartControl);
        List<SectionList> validList = new List<SectionList>();

        for (int i = 1; i < _leg; i++)
        { permut.Add(StartNext); }

        GetVariations(permut, StartControl, false, false, validList);

        foreach (SectionList permutation in validList)
        {
          permutation.SetLeg(_leg);
        }
        return validList;

      }

      protected override IList<SimpleSection> GetSections()
      {
        IList<SimpleSection> sections = GetSections(_course, _leg);
        return sections;
      }
      private static List<SimpleSection> GetSections(SectionCollection course, int leg)
      {
        List<SimpleSection> simples = new SectionsBuilder().GetSections(course, leg);

        List<SimpleSection> sections = new List<SimpleSection>(simples.Count);
        Dictionary<string, Control> controls = new Dictionary<string, Control>();
        foreach (SimpleSection section in simples)
        {
          Control from = GetUniqueControl(controls, section.From);
          Control to = GetUniqueControl(controls, section.To);

          SimpleSection add = new SimpleSection(from, to);
          add.SetWhere(section.Where);
          sections.Add(add);
        }
        return sections;
      }
    }

    private class CourseBuilder : Builder
    {
      private SectionCollection _course;

      protected override Control StartControl { get { return (Control)_course.First.Value; } }
      protected override Control EndControl { get { return (Control)_course.Last.Value; } }

      public CourseBuilder(SectionCollection course)
      {
        _course = course;
      }

      public List<SectionList> Analyze()
      {
        SectionList permut = new SectionList(StartControl);
        string countExpr = GetCountExpression(_course);
        WhereInfo countWhere = new WhereInfo(countExpr, permut);
        permut.Add(countWhere);
        List<SectionList> validList = new List<SectionList>();

        GetVariations(permut, StartControl, true, false, validList);

        return validList;

      }

      protected override IList<SimpleSection> GetSections()
      {
        IList<SimpleSection> sections = GetSections(_course);
        return sections;
      }
      private static List<SimpleSection> GetSections(SectionCollection course)
      {
        List<SimpleSection> simples = course.GetAllSections();

        List<SimpleSection> sections = new List<SimpleSection>(simples.Count);
        Dictionary<string, Control> controls = new Dictionary<string, Control>();
        foreach (SimpleSection section in simples)
        {
          Control from = GetUniqueControl(controls, section.From);
          Control to = GetUniqueControl(controls, section.To);

          SimpleSection add = new SimpleSection(from, to);
          add.SetWhere(section.Where);
          sections.Add(add);
        }
        return sections;
      }
    }

    private abstract class Builder
    {
      public event EventHandler HandlingSplit;

      private IList<SimpleSection> _sections;
      private Dictionary<Control, NextInfoList> _nextInfos;
      private Dictionary<SimpleSection, int> _sectionCount;
      private Dictionary<NextInfo, List<NextInfo>> _nextSections;
      private NextInfo _startNext;

      protected abstract Control StartControl { get;}
      protected abstract Control EndControl { get; }
      protected abstract IList<SimpleSection> GetSections();
      protected IList<SimpleSection> Sections
      {
        get
        {
          if (_sections == null)
          {
            _sections = GetSections();
          }
          return _sections;
        }
      }
      protected Dictionary<Control, NextInfoList> NextInfos
      {
        get
        {
          if (_nextInfos == null)
          {
            IList<SimpleSection> sections = Sections;
            Dictionary<Control, NextInfoList> nextInfos = AssembleSections(sections);
            foreach (NextInfoList nextInfo in nextInfos.Values)
            { nextInfo.AssureCodes(); }

            _nextInfos = nextInfos;
          }
          return _nextInfos;
        }
      }
      private Dictionary<SimpleSection, int> SectionCount
      {
        get
        {
          if (_sectionCount == null)
          {
            Dictionary<SimpleSection, int> dict = new Dictionary<SimpleSection, int>(new SimpleSection.EqualControlsComparer());
            foreach (SimpleSection section in Sections)
            {
              if (!dict.ContainsKey(section))
              {
                dict.Add(section, 1);
              }
              else
              {
                dict[section]++;
              }
            }
            _sectionCount = dict;
          }
          return _sectionCount;
        }
      }
      protected NextInfo StartNext
      {
        get
        {
          if (_startNext == null)
          { _startNext = new NextInfo(StartControl); }
          return _startNext;
        }
      }
      protected Dictionary<NextInfo, List<NextInfo>> NextSections
      {
        get
        {
          if (_nextSections == null)
          {
            Dictionary<NextInfo, List<NextInfo>> nextSections = new Dictionary<NextInfo, List<NextInfo>>(Sections.Count);
            foreach (NextInfoList nextList in NextInfos.Values)
            {
              foreach (NextInfo next in nextList.List)
              {
                List<NextInfo> list = new List<NextInfo>();
                NextInfo n = next;
                while (n != null)
                {
                  list.Add(n);

                  NextInfoList nl;
                  if (!NextInfos.TryGetValue(n.Control, out nl))
                  { n = null; }
                  else
                  {
                    n = nl.List[0];
                    if (n.Code != 0)
                    { n = null; }
                  }
                }
                nextSections.Add(next, list);
              }
            }
            _nextSections = nextSections;
          }
          return _nextSections;
        }
      }

      protected virtual void OnHandlingSplit(EventArgs e)
      {
        if (HandlingSplit != null)
        { HandlingSplit(this, e); }
      }

      protected string GetCountExpression(SectionCollection course)
      {
        int nSections = course.LegCount() + Sections.Count - 1;
        string expr = string.Format("[{0}] = '{1}'", nSections, EndControl.Name);
        return expr;
      }

      protected void AddSections(SectionList sections, List<NextInfo> nextSections)
      {
        Control from = sections.GetControl(-1);
        foreach (NextInfo section in nextSections)
        {
          WhereInfo where = GetWhere(sections, from, section);
          sections.Add(section);
          sections.Add(where);
          from = section.Control;
        }
      }

      protected Dictionary<char, NextWhere> GetPossibleNexts(SectionList pre,
        Control from, IList<NextInfo> nexts, bool verifyFull)
      {
        Dictionary<char, NextWhere> possibles = new Dictionary<char, NextWhere>();
        List<char> invalids = new List<char>();
        foreach (NextInfo next in nexts)
        {
          if (possibles.ContainsKey(next.Code))
          { continue; }

          WhereInfo where = GetWhere(pre, from, next);
          WhereInfo.State state = where.GetState(pre);
          if (state == WhereInfo.State.Failed)
          {
            possibles.Add(next.Code, null);
            invalids.Add(next.Code);
            continue;
          }
          else if (state == WhereInfo.State.Fulfilled)
          { where = null; }

          if (verifyFull)
          {
            if (!ExistsVariation(pre, next, where))
            {
              possibles.Add(next.Code, null);
              invalids.Add(next.Code);
              continue;
            }
          }
          possibles.Add(next.Code, new NextWhere(next, where));
          if (next.Code == 0)
          {
            break;
          }
        }
        foreach (char code in invalids)
        { possibles.Remove(code); }
        return possibles;
      }

      private bool ExistsVariation(SectionList variation, NextInfo next, WhereInfo where)
      {
        SectionList candidate = variation.Clone();
        List<SectionList> sectionsList = new List<SectionList>();

        GetVariations(candidate, next, where, true, true, sectionsList);
        bool exists = sectionsList.Count > 0;
        return exists;
      }

      protected void GetVariations(SectionList variation, Control start,
        bool allLegs, bool oneOnly, List<SectionList> valids)
      {
        Dictionary<char, NextWhere> variations = GetPossibleNexts(variation,
          start, NextInfos[start].List, false);

        foreach (NextWhere next in variations.Values)
        {
          if (variations.Count > 1)
          {
            SectionList clone = variation.Clone();
            GetVariations(clone, next.Next, next.Where, allLegs, oneOnly, valids);
            if (oneOnly && valids.Count > 0)
            { return; }
          }
          else
          {
            GetVariations(variation, next.Next, next.Where, allLegs, oneOnly, valids);
          }
        }


      }

      protected void GetVariations(SectionList variation, NextInfo current, WhereInfo currentWhere,
        bool allLegs, bool oneOnly, List<SectionList> valids)
      {
        do
        {
          variation.Add(current);
          variation.Add(currentWhere);
          WhereInfo.State state = variation.GetState();
          if (state == WhereInfo.State.Failed)
          { return; }

          NextInfoList nextList;
          if (NextInfos.TryGetValue(current.Control, out nextList) == false) // last Control
          {
            if (allLegs == false)
            {
              if (state == WhereInfo.State.Fulfilled)
              { valids.Add(variation); }
              return;
            }
            current = null;
            nextList = NextInfos[StartControl];
          }

          Dictionary<char, NextWhere> variations = GetPossibleNexts(variation,
            current != null ? current.Control : StartControl, nextList.List, false);

          bool finished = true;
          foreach (NextWhere next in variations.Values)
          {
            finished = false;

            if (current == null)
            {
              current = StartNext;
              variation.Add(current);
            }

            if (variations.Count > 1)
            {
              SectionList clone = variation.Clone();
              GetVariations(clone, next.Next, next.Where, allLegs, oneOnly, valids);
              if (oneOnly && valids.Count > 0)
              { return; }
            }
            else
            {
              current = next.Next;
              currentWhere = next.Where;
            }
          }
          if (variations.Count > 1)
          {
            return;
          }
          if (finished)
          {
            WhereInfo.State statex = variation.GetState();
            if (statex == WhereInfo.State.Fulfilled)
            {
              valids.Add(variation);
            }
            return;
          }
        } while (true);
      }

      private WhereInfo GetWhere(SectionList sections, Control from, NextInfo next)
      {
        int nSections = sections.ControlsCount - 1;
        if (nSections == 0)
        {
          return WhereInfo.FulFilled;
        }
        WhereInfo where = WhereInfo.FulFilled;
        if (next.Where != null)
        {
          where = new WhereInfo(next.Where, sections);
          WhereInfo.State state = where.GetState(sections);
          if (state == WhereInfo.State.Failed)
          { return where; }
        }
        if (from == null) from = sections.Start;
        SimpleSection newSection = new SimpleSection(from, next.Control);
        if (sections.CountExisting(from, next.Control) >= SectionCount[newSection])
        {
          return WhereInfo.Failed;
        }
        return where;
      }

      private static Dictionary<Control, NextInfoList> AssembleSections(IList<SimpleSection> sections)
      {
        Dictionary<Control, NextInfoList> dict = new Dictionary<Control, NextInfoList>();
        foreach (SimpleSection section in sections)
        {
          Control pre = section.From;

          NextInfoList nextList;
          if (dict.TryGetValue(pre, out nextList) == false)
          {
            nextList = new NextInfoList();
            dict.Add(pre, nextList);
          }
          nextList.Add(section);
        }
        return dict;
      }

      protected static IList<SimpleSection> MakeUniqueControls(IList<SimpleSection> sections)
      {
        Dictionary<string, Control> controls = new Dictionary<string, Control>();
        List<SimpleSection> unique = new List<SimpleSection>(sections.Count);
        foreach (SimpleSection section in sections)
        {
          Control from = GetUniqueControl(controls, section.From);
          Control to = GetUniqueControl(controls, section.To);
          SimpleSection add = new SimpleSection(from, to);
          add.SetWhere(section.Where);
          unique.Add(add);
        }
        return unique;
      }

      protected static Control GetUniqueControl(Dictionary<string, Control> controls, Control control)
      {
        Control unique;
        if (controls.TryGetValue(control.Name, out unique) == false)
        {
          unique = control;
          controls.Add(unique.Name, unique);
        }
        return unique;
      }

      protected class NextWhere
      {
        public readonly NextInfo Next;
        public readonly WhereInfo Where;

        public NextWhere(NextInfo next, WhereInfo where)
        {
          Next = next;
          Where = where;
        }
      }
    }
  }

  public class SectionList
  {
    private Control _start;
    private readonly List<VariationBuilder.WhereInfo> Wheres = new List<VariationBuilder.WhereInfo>();
    private readonly List<NextInfo> _nextInfos = new List<NextInfo>();

    private string _legString;

    public SectionList(Control start)
    {
      _start = start;
    }
    public Control Start
    { get { return _start; } }

    public int ControlsCount
    {
      get { return 1 + _nextInfos.Count; }
    }
    public string GetName()
    {
      StringBuilder sb = new StringBuilder();
      sb.AppendFormat("{0}", _legString);
      foreach (NextInfo next in _nextInfos)
      {
        if (next.Code != 0)
        { sb.Append(next.Code); }
      }
      return sb.ToString();
    }
    public override string ToString()
    {
      StringBuilder sb = new StringBuilder();
      foreach (Control control in Controls)
      {
        sb.AppendFormat("{0} ", control.Name);
      }
      return sb.ToString();
    }

    public IEnumerable<Control> Controls
    {
      get
      {
        yield return _start;
        {
          foreach (NextInfo next in _nextInfos)
          { yield return next.Control; }
        }
      }
    }
    public IEnumerable<NextInfo> NextInfos
    {
      get
      {
        yield return new NextInfo(_start);
        {
          foreach (NextInfo next in _nextInfos)
          { yield return next; }
        }
      }
    }

    public SectionList Clone()
    {
      SectionList clone = new SectionList(_start);
      clone._nextInfos.AddRange(_nextInfos);
      clone.Wheres.AddRange(Wheres);
      return clone;
    }
    public IList<Course> Parts
    {
      get { throw new NotImplementedException(); }
    }

    public VariationBuilder.WhereInfo.State GetState()
    {
      if (Wheres.Count == 0)
      { return VariationBuilder.WhereInfo.State.Fulfilled; }

      List<VariationBuilder.WhereInfo> remove = new List<VariationBuilder.WhereInfo>();
      foreach (VariationBuilder.WhereInfo where in Wheres)
      {
        VariationBuilder.WhereInfo.State state = where.GetState(this);
        if (state == VariationBuilder.WhereInfo.State.Failed)
        { return VariationBuilder.WhereInfo.State.Failed; }

        if (state == VariationBuilder.WhereInfo.State.Unknown)
        { continue; }

        if (state == VariationBuilder.WhereInfo.State.Fulfilled)
        { remove.Add(where); }
      }
      foreach (VariationBuilder.WhereInfo rem in remove)
      { Wheres.Remove(rem); }

      if (Wheres.Count == 0)
      { return VariationBuilder.WhereInfo.State.Fulfilled; }

      return VariationBuilder.WhereInfo.State.Unknown;
    }

    public Control GetControl(int i)
    {
      if (i < 0)
      { i = _nextInfos.Count + 1 + i; }

      if (i == 0) return Start;
      return _nextInfos[i - 1].Control;
    }
    public NextInfo Add(NextInfo nextInfo)
    {
      _nextInfos.Add(nextInfo);
      return nextInfo;
    }

    public VariationBuilder.WhereInfo Add(VariationBuilder.WhereInfo where)
    {
      if (where == null)
      { return null; }
      Wheres.Add(where);
      return where;
    }

    public int CountExisting(Control from, Control to)
    {
      int count = 0;
      Control c1 = Start;
      foreach (NextInfo section in _nextInfos)
      {
        Control c0 = c1;
        c1 = section.Control;
        if (c0 == from && c1 == to)
        {
          count++;
        }
      }
      return count;
    }

    public void SetLeg(int leg)
    {
      _legString = leg.ToString();
    }
  }

}
