using System;
using System.Collections.Generic;
using System.Data;
using System.IO;
using System.Windows.Forms;
using Mathe.Geom;
using Grid;
using LeastCostPathUI;
using Ocad;
using OCourse.Ext;
using OCourse.Route;
using Tracking;
using Control = Ocad.Control;

namespace OCourse.Gui
{
  public partial class WdgOCourse : Form
  {
    private delegate void ShowGridInContext(LeastCostPath path, StatusEventArgs args);
    private delegate void InitHandler(WdgOCourse wdg);
    private delegate void DrawCourseHandler();

    private static event InitHandler Init;
    private event ShowGridInContext showProgress;
    private event DrawCourseHandler drawCourse;

    private enum VelocityType { Auto, GrayScale, Colors }

    private const string LegColumn = "Leg";
    private const string MeanPrefix = "_Mean ";

    private RouteCalculator _routeCalc;

    private Course _course;
    private Dictionary<SectionInfo, SectionList> _fullInfo;

    private SectionCollection _part;

    private Setup _setup;
    private Dictionary<string, VelocityType> _veloTypes;

    CombinationTable _selectedRoute;
    private bool _suspend;

    public WdgOCourse()
    {
      InitializeComponent();

      _veloTypes = new Dictionary<string, VelocityType>();
      _veloTypes.Add("Auto", VelocityType.Auto);
      _veloTypes.Add("From Gray value", VelocityType.GrayScale);
      _veloTypes.Add("From Color", VelocityType.Colors);

      lstVelo.Items.Clear();
      foreach (string type in _veloTypes.Keys)
      { lstVelo.Items.Add(type); }
      lstVelo.SelectedIndex = 0;

      _selectedRoute = new CombinationTable();

      cntConfig.Resolution = 4;
      cntSection.PartChanged += cntSection_PartChanged;

      if (Init != null)
      { Init(this); }
    }

    public static void SetCourseList(string courseFile, ComboBox lstCourse)
    {
      string oldCourse = lstCourse.Text;
      lstCourse.Items.Clear();
      foreach (string course in Utils.GetCourseList(courseFile))
      {
        lstCourse.Items.Add(course);
      }
      lstCourse.Items.Add("");
      lstCourse.Text = oldCourse;
    }

    private void AddItem(CostInfo leg, string info, SectionList combination)
    {
      _selectedRoute.AddRow(
        info, (leg.Direct + 10 * leg.Climb) / 1000.0, leg.Direct / 1000.0,
        leg.Climb, (leg.OptimalLength + 10.0 * leg.Climb) / 1000.0, leg.OptimalLength / 1000.0,
        leg.OptimalCost, combination);

      dgvInfo.DataSource = _selectedRoute;
    }

    private Dictionary<SectionInfo, SectionList> CalcCourse(SectionCollection course, Setup setup)
    {
      InitHeightAndMap();

      double resol = Resolution;
      Dictionary<SectionInfo, SectionList> courseInfo = RouteCalculator.CalcCourse(course, resol, setup);
      return courseInfo;
    }

    private Dictionary<SectionInfo, SectionList> CalcCourse(IList<SectionList> permuts, Setup setup)
    {
      InitHeightAndMap();

      double resol = Resolution;
      Dictionary<SectionInfo, SectionList> courseInfo = RouteCalculator.GetCourseInfo(permuts, resol, setup);
      return courseInfo;
    }


    private RouteCalculator InitHeightAndMap()
    {
      Map map = null;
      DoubleGrid heightGrid = null;
      Polyline nearLine = null;
      double corridorWidth = 0;
      bool isNew = false;
      if (_routeCalc != null)
      {
        map = _routeCalc.Map;
        heightGrid = _routeCalc.HeightGrid;
        nearLine = _routeCalc.NearLine;
        corridorWidth = _routeCalc.CorridorWidth;
        if (cntConfig.Step.Count != _routeCalc.Step.Count)
        { isNew = true; }
        if (cntConfig.StepCost != _routeCalc.StepCost)
        { isNew = true; }
      }
      if (cntConfig.Map != map)
      {
        isNew = true;
        map = cntConfig.Map;
      }
      if (cntConfig.HeightGrid != heightGrid)
      {
        isNew = true;
        heightGrid = cntConfig.HeightGrid;
      }
      if (cntConfig.NearLine != nearLine)
      {
        isNew = true;
        nearLine = cntConfig.NearLine;
      }
      if (Math.Abs(cntConfig.CorridorWidth - corridorWidth) > 0.001)
      {
        isNew = true;
        corridorWidth = cntConfig.CorridorWidth;
      }

      if (isNew || _routeCalc == null)
      {
        if (_routeCalc != null)
        { _routeCalc.StatusChanged -= routeCalc_Status; }

        _routeCalc = new RouteCalculator(map, heightGrid,
          cntConfig.Step, cntConfig.StepCost, nearLine, corridorWidth);
        _routeCalc.StatusChanged += routeCalc_Status;
      }

      return _routeCalc;
    }

    private void routeCalc_Status(object sender, StatusEventArgs args)
    {
      if (sender is string)
      {
        lblProgress.Text = (string)sender;
      }
      if (sender == null && args == null)
      {
        Application.DoEvents();
      }
      if (args != null && (args.CurrentStep & 1023) == 0)
      {
        int l = lblProgress.Text.LastIndexOf(':');
        lblProgress.Text = lblProgress.Text.Substring(0, l + 1) +
                           args.CurrentStep + " of (max) " + args.TotalStep;

        //ShowInContext((GridTest.LeastCostPath)sender, args);
        if (showProgress != null)
        { showProgress((LeastCostPath)sender, args); }
        Application.DoEvents();
      }
    }

    internal double Resolution
    {
      get { return cntConfig.Resolution; }
    }
    internal RouteCalculator RouteCalculator
    {
      get
      {
        InitHeightAndMap();
        return _routeCalc;
      }
    }

    private void txtCourse_Validated(object sender, EventArgs e)
    {
      if (string.IsNullOrEmpty(txtCourse.Text))
      { return; }
      SetCourseList();
    }

    private void btnCourse_Click(object sender, EventArgs e)
    {
      dlgOpen.FileName = txtCourse.Text;
      dlgOpen.Filter = "*.ocd|*.ocd";
      if (dlgOpen.ShowDialog(this) != DialogResult.OK)
      { return; }

      txtCourse.Text = dlgOpen.FileName;
      SetCourseList();
    }
    private void SetCourseList()
    {
      try
      {
        _suspend = true;
        SetCourseList(txtCourse.Text, lstCourse);
      }
      finally
      {
        _suspend = false;
      }
    }

    private void btnBackCalc_Click(object sender, EventArgs e)
    {
      try
      {
        btnBackCalc.Enabled = false;
        InitHeightAndMap();
        SetCourseList();

        double resolution = Resolution;
        RouteCalculator.CalcEvent(txtCourse.Text, resolution);
      }
      finally
      { btnBackCalc.Enabled = true; }
    }

    private void lstCourse_SelectedIndexChanged(object sender, EventArgs e)
    {
      if (_suspend)
      { return; }

      Course course;
      using (OcadReader reader = OcadReader.Open(txtCourse.Text))
      {
        _setup = reader.ReadSetup();
        course = reader.ReadCourse(lstCourse.Text);
      }

      SetVariations(course);
      cntSection.SetCourse(course);

      _selectedRoute.Clear();
      _part = null;
      Dictionary<SectionInfo, SectionList> info = CalcCourse(course, _setup);
      SetInfo(info);

      _course = course;
      _part = course;

      _fullInfo = info;
    }

    private void SetVariations(SectionCollection sections, IList<int> legs, DataTable tbl)
    {
      foreach (ISection section in sections)
      {
        if (section is Variation)
        {
          Variation v = (Variation)section;
          foreach (Variation.Branch b in v.Branches)
          {
            SetVariations(b, b.Legs, tbl);
          }
        }
        else if (section is Fork)
        {
          DataColumn col = tbl.Columns.Add("F" + tbl.Columns.Count);
          col.ExtendedProperties["Fork"] = section;
          foreach (DataRow row in tbl.Rows)
          { row[col] = "-"; }
          char v = 'A';
          foreach (int leg in legs)
          {
            tbl.Rows[leg - 1][col] = v.ToString();
            v++;
          }
        }
      }
    }

    private void SetVariations(Course course)
    {
      int nLegs = course.LegCount();
      DataTable tbl = new DataTable("Gugus");
      tbl.Columns.Add(LegColumn);
      List<int> legs = new List<int>(nLegs);
      for (int i = 1; i <= nLegs; i++)
      {
        tbl.Rows.Add("L" + i);
        legs.Add(i);
      }
      SetVariations(course, legs, tbl);
      tbl.AcceptChanges();

      bool origSuspend = _suspend;
      try
      {
        _suspend = true;
        dgvVars.SuspendLayout();
        dgvVars.Columns.Clear();
        dgvVars.AutoGenerateColumns = false;
        dgvVars.RowHeadersVisible = false;

        foreach (DataColumn col in tbl.Columns)
        {
          DataGridViewTextBoxColumn txt = new DataGridViewTextBoxColumn();
          txt.Tag = col.ExtendedProperties["Fork"];
          txt.DataPropertyName = col.ColumnName;
          txt.HeaderText = col.ColumnName;
          txt.AutoSizeMode = DataGridViewAutoSizeColumnMode.AllCells;
          dgvVars.Columns.Add(txt);
        }

        DataView view = new DataView(tbl);
        view.AllowEdit = false;
        view.AllowDelete = false;
        view.AllowNew = false;
        dgvVars.DataSource = view;
      }
      finally
      {
        dgvVars.ResumeLayout();
        _suspend = origSuspend;
      }
    }

    private void SetInfo(Dictionary<SectionInfo, SectionList> info)
    {
      if (info.Count > 1)
      {
        Dictionary<int, SectionInfo> sumList = new Dictionary<int, SectionInfo>();
        Dictionary<int, int> countList = new Dictionary<int, int>();
        foreach (SectionInfo sectionInfo in info.Keys)
        {
          int leg;
          int.TryParse(sectionInfo.Name.Substring(0, 1), out leg);

          SectionInfo sum;
          if (sumList.ContainsKey(leg) == false)
          {
            sum = new SectionInfo(null, null, 0, 0, 0, 0);
            sumList.Add(leg, sum);
            countList.Add(leg, 0);
          }
          else
          {
            sum = sumList[leg];
          }

          sumList[leg] = sum + sectionInfo;
          countList[leg]++;
        }
        foreach (KeyValuePair<int, SectionInfo> pair in sumList)
        {
          SectionInfo sum = pair.Value;
          sum = (1.0 / countList[pair.Key]) * sum;
          AddItem(sum, MeanPrefix + pair.Key, null);
        }
      }
      foreach (KeyValuePair<SectionInfo, SectionList> pair in info)
      {
        AddItem(pair.Key, pair.Key.Name, pair.Value);
      }
    }

    private void cntSection_PartChanged(object sender, IList<SectionList> list)
    {
      Dictionary<SectionInfo, SectionList> info;
      if (list != null)
      {
        Setup setup;
        using (OcadReader reader = OcadReader.Open(txtCourse.Text))
        {
          setup = reader.ReadSetup();
        }

        info = CalcCourse(list, setup);
        _part = null;
      }
      else
      {
        info = _fullInfo;
        _part = _course;
      }
      _selectedRoute.Clear();
      SetInfo(info);

      if (cntSection.StartControl != null && cntSection.EndControl != null)
      {
        btnCalcSelection.Enabled = true;
        btnCalcSelection.Text = string.Format("Calc {0}-{1}",
                                              cntSection.StartControl.Name,
                                              cntSection.EndControl.Name);
      }
      else
      {
        btnCalcSelection.Enabled = false;
      }
    }

    private void chkOnTop_CheckedChanged(object sender, EventArgs e)
    {
      TopMost = chkOnTop.Checked;
    }

    private void btnExport_Click(object sender, EventArgs e)
    {
      Shape.ShapeWriter writer = null;
      DataTable schema = new DataTable();
      schema.Columns.Add("Shape", typeof(Polyline));
      schema.Columns.Add(new DBase.DBaseColumn("From", DBase.ColumnType.String, 5, 0));
      schema.Columns.Add(new DBase.DBaseColumn("To", DBase.ColumnType.String, 5, 0));
      schema.Columns.Add(new DBase.DBaseColumn("Resol", DBase.ColumnType.Double, 8, 2));
      schema.Columns.Add(new DBase.DBaseColumn("Climb", DBase.ColumnType.Double, 8, 2));
      schema.Columns.Add(new DBase.DBaseColumn("Optimal", DBase.ColumnType.Double, 10, 2));
      schema.Columns.Add(new DBase.DBaseColumn("Cost", DBase.ColumnType.Double, 10, 2));

      dlgSave.Filter = "*.shp | *.shp";
      if (dlgSave.ShowDialog(this) != DialogResult.OK)
      {
        return;
      }
      try
      {
        writer = new Shape.ShapeWriter(dlgSave.FileName, Shape.ShapeType.LineZ, schema);
        RouteCalculator origCalcList = _routeCalc;
        SortedDictionary<RouteInfo, RouteInfo> rawList;
        try
        {
          _routeCalc = null;
          RouteCalculator.InitRawCourses(txtCourse.Text);
          rawList = _routeCalc.RouteInfoList;
        }
        finally
        {
          _routeCalc = origCalcList;
        }
        double resolution = Resolution;
        foreach (RouteInfo route in rawList.Keys)
        {
          route.Resolution = resolution;
          RouteInfo calcRoute;
          if (RouteCalculator.RouteInfoList.TryGetValue(route, out calcRoute))
          {
            writer.Write(calcRoute.Route,
              new object[] { 
                route.From.Name, route.To.Name, route.Resolution,
                calcRoute.Climb, calcRoute.OptimalLength, calcRoute.OptimalCost
              });
          }
        }
      }
      finally
      {
        if (writer != null)
        { writer.Close(); }
      }
    }

    private void btnImport_Click(object sender, EventArgs e)
    {
      dlgOpen.Filter = "*.shp | *.shp";
      if (dlgOpen.ShowDialog() != DialogResult.OK)
      { return; }

      InitHeightAndMap();

      Shape.ShapeReader reader = null;

      try
      {
        reader = new Shape.ShapeReader(dlgOpen.FileName);
        foreach (DataRow row in reader)
        {
          Polyline route = (Polyline)row["Shape"];
          string from = (string)row["From"];
          string to = (string)row["To"];
          double resol = (double)row["Resol"];
          double climb = (double)row["Climb"];
          double optimal = (double)row["Optimal"];
          double cost = (double)row["Cost"];

          Control ctrl0 = new Control();
          Control ctrl1 = new Control();

          ctrl0.Name = from.Trim();
          ctrl1.Name = to.Trim();
          IPoint start = route.Points.First.Value;
          IPoint end = route.Points.Last.Value;
          RouteInfo info = new RouteInfo(ctrl0, ctrl1, start, end,
            resol, Math.Sqrt(PointOperator.Dist2(start, end)), climb, route, optimal, cost);

          RouteInfo tVal;
          if (RouteCalculator.RouteInfoList.TryGetValue(info, out tVal) == false)
          { RouteCalculator.RouteInfoList.Add(info, info); }
        }
      }
      finally
      {
        if (reader != null)
        { reader.Close(); }
      }
    }

    private void dgvInfo_CurrentCellChanged(object sender, EventArgs e)
    {
      string course = "Mean";
      SectionList comb = null;
      if (dgvInfo.CurrentRow != null)
      {
        DataRowView view = (DataRowView)dgvInfo.CurrentRow.DataBoundItem;
        CombinationTable.Row row = (CombinationTable.Row)view.Row;
        course = row.Name;

        if (!row.IsNull(CombinationTable.CombinationColumn.Name))
        {
          comb = row.Combination;
        }
      }

      cntSection.ShowCombination(comb);
    }

    private void dgvVars_SelectionChanged(object sender, EventArgs e)
    {
      if (_suspend)
      { return; }
      if (drawCourse != null)
      { drawCourse(); }
    }

    private void btnCalcSelection_Click(object sender, EventArgs e)
    {
      WdgOutput wdg = new WdgOutput();
      Control cStart = cntSection.StartControl;
      Control cEnd = cntSection.EndControl;
      wdg.Text = string.Format("Calculate {0} - {1}", cStart.Name, cEnd.Name);

      Setup setup;
      using (OcadReader reader = OcadReader.Open(txtCourse.Text))
      { setup = reader.ReadSetup(); }

      IPoint point;

      point = cStart.GetPoint();
      IPoint start = point.Project(setup.Map2Prj);
      point = cEnd.GetPoint();
      IPoint end = point.Project(setup.Map2Prj);

      double l = Math.Sqrt(PointOperator.Dist2(start, end)) / 2.0;
      Box box = RouteCalculator.GetBox(start, end, l);

      if (cntConfig.CorridorMode)
      {
        Polyline nearline = Track.GpxToPolyline(cntConfig.GpxFile);
        wdg.Init(RouteCalculator.HeightGrid, RouteCalculator.Map, cntConfig.Resolution, cntConfig.Step, cntConfig.StepCost, nearline, cntConfig.CorridorWidth);
      }
      else
      {
        wdg.Init(RouteCalculator.HeightGrid, RouteCalculator.Map, cntConfig.Resolution, cntConfig.Step, cntConfig.StepCost, null, 0);
      }

      wdg.SetExtent(box);
      wdg.SetStart(start);
      wdg.SetEnd(end);
      wdg.SetNames(cStart.Name, cEnd.Name);
      wdg.SetMainPath(Path.GetDirectoryName(dlgOpen.FileName));

      wdg.ShowDialog(this);
    }

    private void btnRelay_Click(object sender, EventArgs e)
    {
      WdgRelay wdg = new WdgRelay();
      List<RouteInfo> infos = null;
      if (_routeCalc != null)
      { infos = new List<RouteInfo>(_routeCalc.RouteInfoList.Keys); }
      wdg.Init(txtCourse.Text, infos);
      wdg.ShowDialog();
    }

    private WdgPermutations _wdgVariations;
    private WdgPermutations WdgVariations
    {
      get
      {
        if (_wdgVariations == null || _wdgVariations.IsDisposed)
        { _wdgVariations = new WdgPermutations(); }
        return _wdgVariations;
      }
    }
    private void btnPermutations_Click(object sender, EventArgs e)
    {
      WdgPermutations wdg = WdgVariations;

      wdg.SetData(txtCourse.Text, this);

      if (wdg.Visible == false)
      {
        wdg.Show(this);
      }
    }

    private void btnTrack_Click(object sender, EventArgs e)
    {
      string error = TryRunTrack();
      if (error != null)
      {
        MessageBox.Show(error);
      }
    }

    private string TryRunTrack()
    {
      {
        WdgTrack wdg = new WdgTrack();

        wdg.SetData(txtCourse.Text, this);
        wdg.ShowDialog();
      }

      if (dgvInfo.CurrentRow == null)
      { return "No Course selected"; }
      if (dgvInfo.CurrentRow.DataBoundItem == null)
      { return "No Course selected"; }

      DataRowView vRow = (DataRowView)dgvInfo.CurrentRow.DataBoundItem;
      CombinationTable.Row row = (CombinationTable.Row)vRow.Row;
      string comb = row.Name;
      if (comb.StartsWith(MeanPrefix))
      { return string.Format("Invalid Course {0} selected", comb); }

      Setup setup;
      Course course;
      using (OcadReader reader = OcadReader.Open(txtCourse.Text))
      {
        setup = reader.ReadSetup();
        course = reader.ReadCourse(lstCourse.Text);
      }
      List<Control> controls = course.GetCombination(comb);
      List<Control> prjControls = new List<Control>(controls.Count);

      foreach (Control control in controls)
      {
        if (control.Element == null)
        { continue; }
        IPoint p = control.Element.Geometry as IPoint;
        if (p == null)
        { continue; }

        control.Element.Geometry = control.Element.Geometry.Project(setup.Map2Prj);
        prjControls.Add(control);
      }
      {
        WdgTrack wdg = new WdgTrack();

        wdg.SetData(txtCourse.Text, this);
        wdg.ShowDialog();
      }

      return null;
    }

    private void btnCreateScripts_Click(object sender, EventArgs e)
    {
      dlgSave.Filter = "*.bat | *.bat";
      if (dlgSave.ShowDialog(this) != DialogResult.OK)
      { return; }

      InitHeightAndMap();
      SetCourseList();

      double resolution = Resolution;
      RouteCalculator.ScriptEvent(txtCourse.Text, cntConfig.HeightName,
        cntConfig.MapName, resolution, dlgSave.FileName);
    }

  }
}