
using System;
using System.Collections.Generic;
using Grid;
using Mathe.Geom;
using Ocad;
using OCourse.Ext;

namespace OCourse.Route
{
  public class RouteCalculator
  {
    #region nested classes
    private class SectionComparer : IComparer<RouteInfo>
    {
      public int Compare(RouteInfo x, RouteInfo y)
      {
        int i;
        i = x.Start.X.CompareTo(y.Start.X);
        if (i != 0) return i;
        i = x.Start.Y.CompareTo(y.Start.Y);
        if (i != 0) return i;
        i = x.End.X.CompareTo(y.End.X);
        if (i != 0) return i;
        i = x.End.Y.CompareTo(y.End.Y);
        if (i != 0) return i;
        i = x.Resolution.CompareTo(y.Resolution);
        if (i != 0) return i;

        return 0;
      }
    }
    #endregion

    public StatusEventHandler StatusChanged;

    private readonly Map _map;
    private readonly DoubleGrid _heightGrid;
    private readonly Step _step;
    private readonly StepCostHandler _stepCost;

    private readonly Polyline _nearLine;
    private readonly double _corridorWidth;


    private readonly SortedDictionary<RouteInfo, RouteInfo> _calcList;

    public RouteCalculator(Map map, DoubleGrid heightGrid,
      Step step, StepCostHandler stepCost, Polyline nearLine,
      double corridorWidth)
    {
      _map = map;
      _heightGrid = heightGrid;
      _step = step;
      _stepCost = stepCost;
      _nearLine = nearLine;
      _corridorWidth = corridorWidth;

      _calcList = new SortedDictionary<RouteInfo, RouteInfo>(new SectionComparer());
    }

    public Map Map
    { get { return _map; } }
    public DoubleGrid HeightGrid
    { get { return _heightGrid; } }
    public Step Step
    { get { return _step; } }
    public StepCostHandler StepCost
    { get { return _stepCost; } }

    public Polyline NearLine
    { get { return _nearLine; } }
    public double CorridorWidth
    { get { return _corridorWidth; } }

    internal SortedDictionary<RouteInfo, RouteInfo> RouteInfoList
    {
      get { return _calcList; }
    }

    public void CalcEvent(string courseFile, double resolution)
    {
      InitRawCourses(courseFile);
      List<RouteInfo> addList = GetRoutes(resolution);

      foreach (RouteInfo info in addList)
      {
        RouteInfoList.Add(info, info);
      }
    }

    public void ScriptEvent(string courseFile, string heightGrid, string map, double resolution, string resultBat)
    {
      InitRawCourses(courseFile);

      using (System.IO.TextWriter w = new System.IO.StreamWriter(resultBat))
      {
        SortedDictionary<IPoint, List<RouteInfo>> starts = new SortedDictionary<IPoint, List<RouteInfo>>(new PointComparer());
        SortedDictionary<IPoint, List<RouteInfo>> ends = new SortedDictionary<IPoint, List<RouteInfo>>(new PointComparer());
        foreach (RouteInfo info in _calcList.Keys)
        {
          List<RouteInfo> startList;
          if (starts.TryGetValue(info.Start, out startList) == false)
          {
            startList = new List<RouteInfo>();
            starts.Add(info.Start, startList);
          }
          startList.Add(info);

          List<RouteInfo> endList;
          if (ends.TryGetValue(info.End, out endList) == false)
          {
            endList = new List<RouteInfo>();
            ends.Add(info.End, endList);
          }
          endList.Add(info);
        }
        string lcpPath = typeof(LeastCostPathUI.IO).Assembly.Location;
        w.WriteLine("-- Start files scripts");
        foreach (KeyValuePair<IPoint, List<RouteInfo>> pair in starts)
        {
          IBox extent = GetBox(pair.Value, new List<IPoint>());
          Round(extent, resolution);
          w.WriteLine("{0} -h {1} -v {2} -E {3} {4} {5} {6} -r {7} -s {8} {9} -sc {10}_c.grd -sd {10}_d.grd",
            lcpPath, heightGrid, map,
            extent.Min.X, extent.Min.Y, extent.Max.X, extent.Max.Y,
            resolution, pair.Key.X, pair.Key.Y,
            pair.Value[0].From.Name);
        }
        w.WriteLine("-- End files scripts");
        foreach (KeyValuePair<IPoint, List<RouteInfo>> pair in ends)
        {
          IBox extent = GetBox(pair.Value, new List<IPoint>());
          Round(extent, resolution);
          w.WriteLine("{0} -h {1} -v {2} -E {3} {4} {5} {6} -r {7} -e {8} {9} -ec _{10}c.grd -ed _{10}d.grd",
            lcpPath, heightGrid, map,
            extent.Min.X, extent.Min.Y, extent.Max.X, extent.Max.Y,
            resolution, pair.Key.X, pair.Key.Y,
            pair.Value[0].To.Name);
        }

        w.WriteLine("-- route files script");
        foreach (RouteInfo route in _calcList.Keys)
        {
          IBox extent = GetBox(new RouteInfo[] { route }, new List<IPoint>());
          Round(extent, resolution);
          w.WriteLine("{0} -E {1} {2} {3} {4} -r {5} -sc {6}_c.grd -sd {6}_d.grd -ec _{7}c.grd -ed _{7}d.grd -C {6}_{7}.tif -rg {6}_{7}r.tif",
            lcpPath,
            extent.Min.X, extent.Min.Y, extent.Max.X, extent.Max.Y,
            resolution,
            route.From.Name, route.To.Name);
        }
      }
    }

    private void Round(IBox extent, double resolution)
    {
      extent.Min.X = Math.Floor(extent.Min.X / resolution) * resolution;
      extent.Min.Y = Math.Floor(extent.Min.Y / resolution) * resolution;

      extent.Max.X = Math.Ceiling(extent.Max.X / resolution) * resolution;
      extent.Max.Y = Math.Ceiling(extent.Max.Y / resolution) * resolution;
    }

    public Dictionary<SectionInfo, SectionList> CalcCourse(SectionCollection course,
      double resol, Setup setup)
    {
      int legs = course.LegCount();
      VariationBuilder builder = new VariationBuilder();
      List<SectionList> allPermuts = new List<SectionList>();
      if (legs > 1)
      {
        for (int i = 1; i <= legs; i++)
        {
          allPermuts.AddRange(builder.AnalyzeLeg(course, i));
        }
      }
      else
      {
        allPermuts.AddRange(builder.Analyze(course));
      }

      Dictionary<SectionInfo, SectionList> courseInfos = GetCourseInfo(allPermuts, resol, setup);
      return courseInfos;
    }

    public Dictionary<SectionInfo, SectionList> GetCourseInfo(IList<SectionList> permuts, double resol, Setup setup)
    {
      Dictionary<SectionInfo, SectionList> courseInfo = new Dictionary<SectionInfo, SectionList>(permuts.Count);
      foreach (SectionList permut in permuts)
      {
        List<Control> controls = new List<Control>(permut.Controls);
        Control from = null;
        IPoint toP = null;
        SectionInfo sum = null;
        foreach (Control to in controls)
        {
          IPoint fromP = toP;

          if (to.Code == Ocad.StringParams.CoursePar.TextBlockKey)
          { continue; }

          IPoint point = to.GetPoint();
          toP = point.Project(setup.Map2Prj);

          if (from != null)
          {
            SectionInfo part = CalcSection(from, to, fromP, toP, resol,
                                           from.Name + "->" + to.Name);
            if (sum == null)
            {
              sum = part;
            }
            else
            {
              sum = sum + part;
            }
          }
          from = to;
        }
        if (sum != null)
        {
          sum.Name = permut.GetName();
          courseInfo.Add(sum, permut);
        }
      }
      return courseInfo;
    }

    public Box GetBox(IPoint start, IPoint end, double l)
    {
      Box box = new Box(Point.Create(start), Point.Create(end), true);

      box.Min.X -= l;
      box.Min.Y -= l;
      box.Max.X += l;
      box.Max.Y += l;

      box.Min.X = Math.Max(box.Min.X, _heightGrid.Extent.Extent.Min.X);
      box.Min.Y = Math.Max(box.Min.Y, _heightGrid.Extent.Extent.Min.Y);
      box.Max.X = Math.Min(box.Max.X, _heightGrid.Extent.Extent.Max.X);
      box.Max.Y = Math.Min(box.Max.Y, _heightGrid.Extent.Extent.Max.Y);
      return box;
    }

    internal List<RouteInfo> GetRoutes(double resolution)
    {
      RouteInfo start = null;
      List<RouteInfo> calcList = new List<RouteInfo>();
      List<RouteInfo> addList = new List<RouteInfo>();
      foreach (RouteInfo info in _calcList.Keys)
      {
        if (info.Resolution > 0)
        { continue; }
        if (_calcList.ContainsKey(new RouteInfo(info.From, info.To, info.Start, info.End,
                                                resolution, 0, 0, null, 0, 0)))
        { continue; }

        if (start == null ||
            start.Start.X != info.Start.X || start.Start.Y != info.Start.Y)
        {
          if (start != null)
          {
            addList.AddRange(CalcRoutes(calcList, resolution));
          }
          calcList.Clear();

          start = info;
        }

        calcList.Add(info);
      }
      addList.AddRange(CalcRoutes(calcList, resolution));
      return addList;
    }

    private List<RouteInfo> CalcRoutes(List<RouteInfo> calcList,
      double resolution)
    {
      if (_heightGrid == null || _map == null || calcList.Count == 0)
      {
        return new List<RouteInfo>();
      }
      OnStatusChanged(calcList[0].From.Name + " : ");

      DoubleGrid costGrid;
      IntGrid dirGrid;

      List<IPoint> endList = new List<IPoint>();

      Box box = GetBox(calcList, endList);

      //GridTest.LeastCostPath path = new GridTest.LeastCostPath(new GridTest.Step16(), box, resolution);
      LeastCostPath path = new LeastCostPath(box, resolution, _step);
      path.StepCost = _stepCost;
      path.Status += routeCalc_Status;

      path.CalcCost(calcList[0].Start, endList, _heightGrid, _map, out costGrid, out dirGrid);
      List<RouteInfo> result = new List<RouteInfo>();
      foreach (RouteInfo info in calcList)
      {
        double dh;
        double optimal;
        double cost = costGrid.Value(info.End.X, info.End.Y);
        Polyline route = GetRoute(path, dirGrid, costGrid, info.Start, info.End, out dh, out optimal);

        RouteInfo add = new RouteInfo(
          info.From, info.To, info.Start, info.End, resolution,
          Math.Sqrt(PointOperator.Dist2(info.Start, info.End)), dh, route, optimal, cost);
        result.Add(add);

        OnStatusChanged(null);
      }
      return result;
    }

    private static Box GetBox(IList<RouteInfo> calcList, List<IPoint> endList)
    {
      Box box = new Box(Point.Create(calcList[0].Start),
                        Point.Create(calcList[0].End), true);
      foreach (RouteInfo info in calcList)
      {
        endList.Add(info.End);
        double l = Math.Sqrt(PointOperator.Dist2(info.Start, info.End)) / 2.0;
        Box infoBox = new Box(Point.Create(info.Start), Point.Create(info.End), true);

        infoBox.Min.X -= l;
        infoBox.Min.Y -= l;
        infoBox.Max.X += l;
        infoBox.Max.Y += l;

        box.Include(infoBox);
      }
      return box;
    }

    private SectionInfo CalcSection(Control from, Control to,
      IPoint start, IPoint end, double resol, string section)
    {
      RouteInfo info;
      RouteInfo existingInfo;
      if (_calcList != null && _calcList.TryGetValue(new RouteInfo(from, to,
        start, end, resol, 0, 0, null, 0, 0), out info))
      {
        return info;
      }
      if (_heightGrid == null || _map == null || resol < 0)
      {
        info = new RouteInfo(from, to, start, end, resol,
          Math.Sqrt(PointOperator.Dist2(start, end)), 0, null, 0, 0);

        if (_calcList != null && _calcList.TryGetValue(info, out existingInfo) == false)
        { _calcList.Add(info, info); }

        return info;
      }

      OnStatusChanged(section + " : ");

      double l = Math.Sqrt(PointOperator.Dist2(start, end)) / 2.0;
      Box box = GetBox(start, end, l);

      IntGrid dirGrid;
      DoubleGrid costGrid;

      //GridTest.LeastCostPath path = new GridTest.LeastCostPath(new GridTest.Step16(), box, resolution);
      LeastCostPath path = new LeastCostPath(box, resol, _step);
      path.StepCost = _stepCost;
      path.Status += path_Status;


      DateTime t0 = DateTime.Now;
      if (_nearLine != null)
      {
        path.CalcCost(start, end, _nearLine, _corridorWidth, _heightGrid, _map, out costGrid, out dirGrid);
      }
      else
      {
        path.CalcCost(start, end, _heightGrid, _map, out costGrid, out dirGrid);
      }


      DateTime t1 = DateTime.Now;
      TimeSpan dt = t1 - t0;

      double dh;
      double length;
      double cost = costGrid.Value(end.X, end.Y);
      Polyline route = GetRoute(path, dirGrid, costGrid, start, end, out dh, out length);

      OnStatusChanged(null);

      info = new RouteInfo(from, to, start, end, resol, l * 2, dh, route, length, cost);
      if (_calcList.TryGetValue(info, out existingInfo) == false)
      { _calcList.Add(info, info); }
      else
      { }
      return info;
    }

    //private Polyline GetRoute(GridTest.LeastCostPath path, 
    private Polyline GetRoute(LeastCostPath path,
      IntGrid dir, DoubleGrid costGrid, IPoint start, IPoint end, out double climb, out double optimal)
    {
      Polyline line = LeastCostPath.GetPath(dir, path.Step, costGrid, end);

      line.Points.First.Value.X = start.X;
      line.Points.First.Value.Y = start.Y;

      line.Points.Last.Value.X = end.X;
      line.Points.Last.Value.Y = end.Y;

      climb = -1;
      double h0 = 0;
      foreach (IPoint p1 in line.Points)
      {
        double h1 = _heightGrid.Value(p1.X, p1.Y, EGridInterpolation.bilinear);
        if (climb >= 0 && h0 < h1)
        {
          climb += h1 - h0;
        }
        else if (climb < 0)
        { climb = 0; }
        h0 = h1;
      }
      if (line.Dimension != 2)
      { throw new InvalidProgramException(string.Format("Expected dimension 2, got {0}", line.Dimension)); }

      optimal = line.Project(Mathe.Geom.Geometry.ToXY).Length();

      return line;
    }

    public void InitRawCourses(string courseFile)
    {
      IList<string> courseList = Utils.GetCourseList(courseFile);
      using (OcadReader reader = OcadReader.Open(courseFile))
      {
        Setup setup = reader.ReadSetup();

        foreach (string courseName in courseList)
        {
          Course course = reader.ReadCourse(courseName);
          CalcCourse(course, -1, setup);
        }
      }
    }

    private void path_Status(object sender, StatusEventArgs args)
    {
      OnStatusChanged(sender, args);
    }

    protected void OnStatusChanged(object sender, StatusEventArgs args)
    {
      if (StatusChanged != null)
      { StatusChanged(sender, args); }
    }

    private void OnStatusChanged(string msg)
    {
      if (StatusChanged != null)
      { StatusChanged(msg, null); }
    }

    private void routeCalc_Status(object sender, StatusEventArgs args)
    {
      OnStatusChanged(sender, args);
    }
  }
}
