using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Data.Common;
using Mathe.Data;
using Mathe.Geom;

namespace Ocad.Data
{
  public class OcadConnection : DbBaseConnection
  {
    public event CancelEventHandler StartingOperation;
    public event EventHandler CommittingOperation;
    public event EventHandler AbortingOperation;

    private OcadTransaction _trans;
    private string _connection;
    private Pool _pool;

    public static readonly string FieldShape = "Shape";
    public static readonly string FieldSymbol = "Symbol";
    public static readonly string FieldAngle = "Angle";

    private static Dictionary<string, Pool> _poolDict;

    internal class Pool
    {
      private string _connectionString;
      public Pool(string connectionString)
      {
        _connectionString = connectionString;
      }
      public List<OcadConnection> Connections = new List<OcadConnection>();
      public Dictionary<string, OcadInfo> Infos = new Dictionary<string, OcadInfo>();

      public OcadInfo GetInfo(string name)
      {
        OcadInfo info;
        if (Infos.TryGetValue(name, out info) == false)
        {
          string fullName = System.IO.Path.Combine(_connectionString, name);
          info = new OcadInfo(fullName);
          Infos.Add(name, info);
        }
        return info;
      }
    }

    public OcadConnection(string connectionString)
    {
      _connection = connectionString;
      if (_poolDict == null)
      { _poolDict = new Dictionary<string, Pool>(); }

      Pool pool;
      if (_poolDict.TryGetValue(_connection, out pool) == false)
      {
        pool = new Pool(_connection);
        _poolDict.Add(_connection, pool);
      }
      pool.Connections.Add(this);
      _pool = pool;
    }

    public override void Close()
    { }
    public override void Open()
    { }

    protected override void Dispose(bool disposing)
    {
      _pool.Connections.Remove(this);
      base.Dispose(disposing);
    }

    public override string ConnectionString
    {
      get { return _connection; }
      set { _connection = value; }
    }

    internal OcadInfo GetReader(string name)
    {
      OcadInfo ocad = _pool.GetInfo(name);
      return ocad;
    }
    public override IBox GetExtent(string tableName)
    {
      OcadInfo info = _pool.GetInfo(tableName);
      IBox extent = info.GetExtent();
      return extent;
    }

    public override void ChangeDatabase(string databaseName)
    { throw new Exception("The method or operation is not implemented."); }
    public override string Database
    {
      get { throw new Exception("The method or operation is not implemented."); }
    }
    public override string DataSource
    {
      get { throw new Exception("The method or operation is not implemented."); }
    }
    public override string ServerVersion
    {
      get { throw new Exception("The method or operation is not implemented."); }
    }
    public override ConnectionState State
    {
      get { return ConnectionState.Open; }
    }

    public override SchemaTable GetTableSchema(string tableName)
    {
      SchemaTable schema = new SchemaTable();

      schema.AddSchemaColumn(FieldShape, typeof(IGeometry));
      schema.AddSchemaColumn(FieldSymbol, typeof(int));
      schema.AddSchemaColumn(FieldAngle, typeof(double));

      return schema;
    }

    public new OcadTransaction BeginTransaction()
    {
      return (OcadTransaction)base.BeginTransaction();
    }
    public new OcadTransaction BeginTransaction(IsolationLevel isolationLevel)
    {
      return (OcadTransaction)base.BeginTransaction(isolationLevel);
    }
    protected override DbTransaction BeginDbTransaction(IsolationLevel isolationLevel)
    {
      _trans = OcadTransaction.Start(this, isolationLevel);
      return _trans;
    }

    public new OcadCommand CreateCommand()
    {
      return (OcadCommand)CreateDbCommand();
    }
    protected override DbCommand CreateDbCommand()
    {
      return new OcadCommand(this);
    }

    internal void OnStartingOperation(CancelEventArgs args)
    {
      if (StartingOperation != null)
      { StartingOperation(this, args); }
    }

    internal void OnCommittingOperation(EventArgs args)
    {
      if (CommittingOperation != null)
      { CommittingOperation(this, args); }
    }

    internal void OnAbortingOperation(EventArgs args)
    {
      if (AbortingOperation != null)
      { AbortingOperation(this, args); }
    }

    public new OcadAdapter CreateAdapter()
    {
      return (OcadAdapter)base.CreateAdapter();
    }

    protected override DbBaseAdapter CreateDbAdapter()
    {
      OcadAdapter apt = new OcadAdapter(this);
      return apt;
    }

  }
}
