using System;
using System.IO;
using System.Collections.Generic;
using Mathe.Geom;

namespace Ocad
{
  /// <summary>
  /// Summary description for OcadWriter.
  /// </summary>
  public abstract class OcadWriter : IDisposable
  {
    public delegate T ElementIndexDlg<T>(ElementIndex element);

    protected EndianWriter _writer;
    protected FileParam _param;
    protected OcadReader _ocdReader;
    protected EndianReader _reader;
    protected Setup _setup;

    protected abstract void Init();

    public OcadReader Reader
    {
      get
      {
        if (_ocdReader == null)
        { Init(); }
        return _ocdReader;
      }
    }

    public Setup Setup
    {
      get
      {
        Init();
        return _setup;
      }
    }

    /// <summary>
    /// assumption : index has all values set except for pos
    /// </summary>
    /// <param name="index"></param>
    public int Append(ElementIndex index)
    {
      Init();
      long p, p0;
      int i;
      p = _param.FirstIndexBlock;
      p0 = p;

      int indexNr = 0;

      while (p > 0)
      {
        p0 = p;
        _writer.BaseStream.Seek(p, SeekOrigin.Begin);
        p = _reader.ReadInt32();

        if (p > 0)
        {
          indexNr += FileParam.N_IN_ELEM_BLOCK;
        }
      }
      if (p0 > 0)
      {
        p = p0 + 4;
        _reader.BaseStream.Seek(p + 16, SeekOrigin.Begin); /* index->pos */
        i = 0;
        while (_reader.ReadInt32() != 0 && i < FileParam.N_IN_ELEM_BLOCK)
        {
          p += _param.INDEX_LENGTH;
          _reader.BaseStream.Seek(p + 16, SeekOrigin.Begin);
          i++;
        }
        indexNr = indexNr + i;
      }
      else
      { // empty ocad file
        i = FileParam.N_IN_ELEM_BLOCK;
        indexNr = 0;
      }
      if (i == FileParam.N_IN_ELEM_BLOCK)
      { /* new index block */
        _writer.BaseStream.Seek(0, SeekOrigin.End);
        p = _writer.BaseStream.Position;
        _writer.Write((int)0);
        int n = FileParam.N_IN_ELEM_BLOCK * _param.INDEX_LENGTH;
        for (int j = 0; j < n; j++)
        { _writer.Write((byte)0); }
        if (p0 > 0)
        {
          _writer.BaseStream.Seek(p0, SeekOrigin.Begin);
        }
        else
        {
          _writer.BaseStream.Seek(12, SeekOrigin.Begin); /* first Index Block position */
          _param.FirstIndexBlock = (int)p;
        }
        _writer.Write((int)p);
        p += 4;

        index.Position = (int)p + n;
      }
      else
      {
        _writer.BaseStream.Seek(0, SeekOrigin.End);
        index.Position = (int)_writer.BaseStream.Position;
      }
      _writer.BaseStream.Seek(p, SeekOrigin.Begin);

      Write(index);

      return indexNr;
    }

    public int Append(Element element)
    {
      if (element == null)
      { return -1; }

      Init();
      IGeometry prj = null;
      if (element.IsGeometryProjected)
      {
        prj = element.Geometry;
        element.Geometry = element.Geometry.Project(_setup.Prj2Map);
        element.IsGeometryProjected = false;
      }
      ElementIndex index = element.GetIndex();

      int pos = Append(index);

      _writer.BaseStream.Seek(index.Position, SeekOrigin.Begin);
      Write(element);

      if (prj != null)
      {
        element.Geometry = prj;
        element.IsGeometryProjected = true;
      }

      return pos;
    }

    protected virtual void Write(ElementIndex index)
    {
      Coord pCoord = Coord.Create(index.Box.Min, 0);
      _writer.Write((int)pCoord.Ox);
      _writer.Write((int)pCoord.Oy);
      pCoord = Coord.Create(index.Box.Max, 0);
      _writer.Write((int)pCoord.Ox);
      _writer.Write((int)pCoord.Oy);
    }

    protected abstract void Write(Element element);
    protected abstract void WriteElementHeader(Element element);
    protected abstract void WriteElementSymbol(int symbol);

    protected void Write(IGeometry geometry)
    {
      Coord pCoord;

      if (geometry is Point)
      {
        Point p = geometry as Point;
        pCoord = Coord.Create(p, Coord.Flags.none);
        _writer.Write((int)pCoord.Ox);
        _writer.Write((int)pCoord.Oy);
      }
      else if (geometry is PointCollection)
      {
        foreach (Point p in (geometry as PointCollection))
        {
          pCoord = Coord.Create(p, Coord.Flags.none);
          _writer.Write((int)pCoord.Ox);
          _writer.Write((int)pCoord.Oy);
        }
      }
      else if (geometry is Polyline)
      {
        Write((Polyline)geometry, false);
      }
      else if (geometry is Area)
      {
        Area area = geometry as Area;
        bool bIsInnerRing = false;
        foreach (Polyline pLine in area.Border)
        {
          Write(pLine, bIsInnerRing);
          bIsInnerRing = true;
        }
      }
      else
      { throw new Exception("Unhandled geometry type " + geometry.GetType()); }
    }

    public void Write(Symbol.SymbolGraphics graphics)
    {
      if (graphics == null)
      { return; }

      _writer.Write((short)graphics.Type);
      ushort iFlags = 0;
      if (graphics.RoundEnds)
      { iFlags |= 1; }
      _writer.Write((short)iFlags);
      _writer.Write((short)graphics.Color);

      _writer.Write((short)graphics.LineWidth);
      _writer.Write((short)graphics.Diameter);

      _writer.Write((short)Element.PointCount(graphics.Geometry));
      _writer.Write((short)0); // reserved
      _writer.Write((short)0); // reserved

      Write(graphics.Geometry);
    }

    private void Write(Polyline polyline, bool isInnerRing)
    {
      Coord pCoord;

      if (polyline.Points.Count < 1)
      { return; }

      if (isInnerRing)
      { pCoord = Coord.Create(polyline.Points.First.Value, Coord.Flags.firstHolePoint); }
      else
      { pCoord = Coord.Create(polyline.Points.First.Value, Coord.Flags.none); }
      _writer.Write((int)pCoord.Ox);
      _writer.Write((int)pCoord.Oy);

      foreach (Curve pSeg in polyline.Segments)
      {
        if (pSeg is Bezier)
        {
          Bezier pBezier = pSeg as Bezier;

          pCoord = Coord.Create(pBezier.P1, Coord.Flags.firstBezierPoint);
          _writer.Write((int)pCoord.Ox);
          _writer.Write((int)pCoord.Oy);

          pCoord = Coord.Create(pBezier.P2, Coord.Flags.secondBezierPoint);
          _writer.Write((int)pCoord.Ox);
          _writer.Write((int)pCoord.Oy);
        }

        pCoord = Coord.Create(pSeg.End, Coord.Flags.none);
        _writer.Write((int)pCoord.Ox);
        _writer.Write((int)pCoord.Oy);
      }

    }


    public void WriteHeader(FileParam data)
    {
      _writer.BaseStream.Seek(0, SeekOrigin.Begin);

      _writer.Write((short)FileParam.OCAD_MARK);
      _writer.Write((short)data.SectionMark);
      _writer.Write((short)data.Version);
      _writer.Write((short)data.SubVersion);
      _writer.Write((int)data.FirstSymbolBlock);
      _writer.Write((int)data.FirstIndexBlock);
      _writer.Write((int)data.SetupPosition);
      _writer.Write((int)data.SetupSize);
      _writer.Write((int)data.InfoPosition);
      _writer.Write((int)data.InfoSize);
      _writer.Write((int)data.FirstStrIdxBlock);
      _writer.Write((int)data.FileNamePosition);
      _writer.Write((int)data.FileNameSize);
      _writer.Write((int)0); // reserved 4
    }

    public void WriteSymData(SymbolData sym)
    {
      _writer.BaseStream.Seek(FileParam.HEADER_LENGTH, SeekOrigin.Begin);

      _writer.Write((short)sym.ColorCount);
      _writer.Write((short)sym.ColorSeparationCount);
      _writer.Write((short)sym.CyanFreq);
      _writer.Write((short)sym.CyanAngle);
      _writer.Write((short)sym.MagentaFreq);
      _writer.Write((short)sym.MagentaAngle);
      _writer.Write((short)sym.YellowFreq);
      _writer.Write((short)sym.YellowAngle);
      _writer.Write((short)sym.BlackFreq);
      _writer.Write((short)sym.BlackAngle);
      _writer.Write((short)0); // reserved 1
      _writer.Write((short)0); // reserved 2
    }

    public void WriteSetup(Setup setup, int length)
    {
      int i, l;

      l = 0;
      _writer.Write((int)setup.Offset.X); l += 4;
      _writer.Write((int)setup.Offset.Y); l += 4;

      _writer.Write((double)setup.GridDistance); l += 8;

      _writer.Write((short)setup.WorkMode); l += 2;
      _writer.Write((short)setup.LineMode); l += 2;
      _writer.Write((short)setup.EditMode); l += 2;
      _writer.Write((short)setup.Symbol); l += 2;
      _writer.Write((double)setup.Scale); l += 8;

      _writer.Write((double)setup.PrjTrans.X); l += 8;
      _writer.Write((double)setup.PrjTrans.Y); l += 8;
      _writer.Write((double)setup.PrjRotation); l += 8;
      _writer.Write((double)setup.PrjGrid); l += 8;

      for (i = l; i < length; i++)
      { _writer.Write((char)0); }
    }

    public int WriteDefaultHeaderV7(double scale)
    {
      FileParam data = new FileParamV8();
      SymbolData sym = new SymbolData();
      Setup setup = new Setup();
      int i, j;

      setup.Offset.X = 0;
      setup.Offset.Y = 0;
      setup.GridDistance = 500;
      setup.WorkMode = WorkMode.curve;
      setup.LineMode = LineMode.curve;
      setup.EditMode = EditMode.editPoint;
      setup.Symbol = 0;
      setup.Scale = scale;
      setup.PrjTrans.X = 0;
      setup.PrjTrans.Y = 0;
      setup.PrjRotation = 0;
      setup.PrjGrid = 1000;

      sym.ColorCount = 0;
      sym.ColorSeparationCount = 0;
      sym.CyanFreq = 1500;
      sym.CyanAngle = 150;
      sym.MagentaFreq = 1500;
      sym.MagentaAngle = 750;
      sym.YellowFreq = 1500;
      sym.YellowAngle = 0;
      sym.BlackFreq = 1500;
      sym.BlackAngle = 450;

      data.SectionMark = 7;
      data.Version = 7;
      data.SubVersion = 90;
      data.FirstSymbolBlock = 0;
      data.FirstIndexBlock = 0;

      data.SetupPosition = FileParam.HEADER_LENGTH + FileParam.SYM_BLOCK_LENGTH +
        FileParam.MAX_COLORS * FileParam.COLORINFO_LENGTH +
        FileParam.MAX_SEPS * FileParam.COLORSEP_LENGTH;
      data.SetupSize = 1348;

      data.FirstSymbolBlock = data.SetupPosition + data.SetupSize;
      data.FirstIndexBlock = data.FirstSymbolBlock +
        (FileParam.N_IN_SYMBOL_BLOCK + 1) * 4; // (sizeof(int))

      data.InfoPosition = 0;
      data.InfoSize = 0;
      data.FirstStrIdxBlock = 0;

      WriteHeader(data);

      WriteSymData(sym);

      /* colors */
      for (i = 0; i < FileParam.MAX_COLORS; i++)
      {
        for (j = 0; j < FileParam.COLORINFO_LENGTH; j++)
        { _writer.Write((char)0); }
      }

      /* separations */
      for (i = 0; i < FileParam.MAX_SEPS; i++)
      {
        for (j = 0; j < FileParam.COLORSEP_LENGTH; j++)
        { _writer.Write((char)0); }
      }

      WriteSetup(setup, data.SetupSize);

      /* symbol indices */
      _writer.Write((int)0);
      for (i = 0; i < FileParam.N_IN_SYMBOL_BLOCK; i++)
      {
        _writer.Write((int)0);
      }

      /* elements */
      return 0;
    }

    public void Dispose()
    {
      if (_writer != null)
      { _writer.Close(); }
      _writer = null;
    }

    public void Close()
    {
      _writer.Close();
    }

    public int DeleteElements(ElementIndexDlg<bool> checkDelete)
    {
      Init();

      int n = 0;
      ElementIndex pIndex;
      int iIndex = 0;
      pIndex = _ocdReader.ReadIndex(iIndex);
      while (pIndex != null)
      {
        long pos = _writer.BaseStream.Position;
        if (checkDelete(pIndex))
        {
          pIndex.Status = ElementIndex.StatusDeleted;
          _writer.BaseStream.Seek(pos - _param.INDEX_LENGTH, SeekOrigin.Begin);
          Write(pIndex);

          n++;
        }

        iIndex++;
        pIndex = _ocdReader.ReadIndex(iIndex);
      }
      return n;
    }

    public int ChangeSymbols(ElementIndexDlg<int> changeSymbol)
    {
      Init();

      int n = 0;
      ElementIndex pIndex;
      int iIndex = 0;
      pIndex = _ocdReader.ReadIndex(iIndex);
      List<ElementIndex> changedElements = new List<ElementIndex>();
      while (pIndex != null)
      {
        long pos = _writer.BaseStream.Position;
        int newSymbol = changeSymbol(pIndex);
        if (newSymbol != pIndex.Symbol)
        {
          pIndex.Symbol = newSymbol;
          _writer.BaseStream.Seek(pos - _param.INDEX_LENGTH, SeekOrigin.Begin);
          Write(pIndex);

          changedElements.Add(pIndex);
          n++;
        }

        iIndex++;
        pIndex = _ocdReader.ReadIndex(iIndex);
      }

      foreach (ElementIndex index in changedElements)
      {
        _writer.BaseStream.Seek(index.Position, SeekOrigin.Begin);
        // Assumption: Symbol and ObjectType is at the start position!
        WriteElementSymbol(index.Symbol);
        _writer.Write((byte)index.ObjectType);
      }
      return n;
    }

  }
  #region OCAD8
  public class Ocad8Writer : OcadWriter
  {
    public Ocad8Writer(Stream ocadStream)
    {
      _writer = new EndianWriter(ocadStream);
      _param = new FileParamV8();
    }

    protected override void Init()
    {
      if (_ocdReader == null)
      {
        _ocdReader = new Ocad8Reader(_writer.BaseStream);
        _reader = new EndianReader(_writer.BaseStream, true);
        _param = _ocdReader.FileParam;
        _setup = _ocdReader.ReadSetup();
      }
    }

    protected override void Write(ElementIndex index)
    {
      base.Write(index);

      _writer.Write((int)index.Position);
      _writer.Write((short)index.Length);
      _writer.Write((short)index.Symbol);
    }

    protected override void Write(Element element)
    {
      WriteElementHeader(element);

      Write(element.Geometry);

      if (element.Text != "")
      {
        if (element.UnicodeText)
        { _writer.WriteUnicodeString(element.Text); }
        else
        {
          _writer.Write(element.Text);
          _writer.Write((char)0);
        }
      }
    }

    protected override void WriteElementHeader(Element element)
    {
      ElementV8 elem8 = element as ElementV8;
      WriteElementSymbol(element.Symbol);
      _writer.Write((byte)element.Type);
      _writer.Write((byte)(element.UnicodeText ? 1 : 0));
      _writer.Write((short)element.PointCount());
      _writer.Write((short)element.TextCount());
      _writer.Write((short)(element.Angle * 180 / Math.PI)); // 1 Degrees
      _writer.Write((short)0);
      if (elem8 != null)
      { _writer.Write((int)elem8.ReservedHeight); }
      else
      { _writer.Write((int)0); }
      for (int i = 0; i < 16; i++)
      { _writer.Write(' '); }
    }

    protected override void WriteElementSymbol(int symbol)
    {
      _writer.Write((short)symbol);
    }
  }
  #endregion

  #region OCAD9
  public class Ocad9Writer : OcadWriter
  {
    private int _removeNIndex;

    private StringType _removeType;

    private StringType _overwriteType;

    private StringParamIndex _overwriteIndex;

    public static Ocad9Writer AppendTo(string file)
    {
      return AppendTo(new FileStream(file, FileMode.Open,
         FileAccess.ReadWrite));
    }
    public static Ocad9Writer AppendTo(Stream ocadStream)
    {
      Ocad9Writer ocd = new Ocad9Writer();
      ocd._writer = new EndianWriter(ocadStream);
      ocd._param = new FileParamV9();
      return ocd;
    }

    protected override void Init()
    {
      if (_ocdReader == null)
      {
        _ocdReader = new Ocad9Reader(_writer.BaseStream);
        _reader = new EndianReader(_writer.BaseStream, System.Text.Encoding.UTF7, true);
        _param = _ocdReader.FileParam;
        _setup = _ocdReader.ReadSetup();
      }
    }

    public void DeleteElements(IList<int> symbols)
    {
      Init();
      List<int> symSort = null;
      if (symbols != null)
      {
        symSort = new List<int>(symbols);
        symSort.Sort();
      }

      ElementIndex pIndex;
      int iIndex = 0;
      pIndex = _ocdReader.ReadIndex(iIndex);
      while (pIndex != null)
      {
        if (symSort == null || symSort.BinarySearch(pIndex.Symbol) >= 0)
        {
          pIndex.Status = ElementIndex.StatusDeleted;
          _writer.BaseStream.Seek(-_param.INDEX_LENGTH, SeekOrigin.Current);
          Write(pIndex);
        }
        iIndex++;
        pIndex = _ocdReader.ReadIndex(iIndex);
      }
    }

    public void Remove(StringType type)
    {
      Init();
      _removeNIndex = 0;
      _removeType = type;
      _ocdReader.StringIndexRead += Remove_StringIndexRead;
      _ocdReader.ReadStringParamIndices();
      _ocdReader.StringIndexRead -= Remove_StringIndexRead;
    }

    private void Remove_StringIndexRead(object sender, StringIndexEventArgs args)
    {
      if (args.Index.Type != _removeType)
      { return; }

      _writer.BaseStream.Seek(-8, SeekOrigin.Current);

      _writer.Write((int)-1);

      _writer.BaseStream.Seek(4, SeekOrigin.Current);

      //if (args.Index.Type == _removeType)
      //{
      //  _removeNIndex++;
      //}
      //else if (_removeNIndex > 0)
      //{
      //  long iPos = _reader.BaseStream.Position;
      //  _writer.BaseStream.Seek(-(_removeNIndex + 1) * 16, SeekOrigin.Current);

      //  _writer.Write((int)args.Index.FilePosition);
      //  _writer.Write((int)args.Index.Size);
      //  _writer.Write((int)args.Index.Type);
      //  _writer.Write((int)args.Index.ElemNummer);

      //  _reader.BaseStream.Seek(iPos, SeekOrigin.Begin);
      //}

      //if (_removeNIndex > 0)
      //{
      //  _writer.BaseStream.Seek(-16, SeekOrigin.Current);

      //  _writer.Write((int)0);
      //  _writer.Write((int)0);
      //  _writer.Write((int)0);
      //  _writer.Write((int)0);
      //}
    }

    public void Append(StringType type, int elemNummer, string stringParam)
    {
      Init();
      _reader.BaseStream.Seek(0, SeekOrigin.End);
      int iPos = (int)_reader.BaseStream.Position;

      StringParamIndex overwriteIndex = new StringParamIndex();
      overwriteIndex.Type = type;
      overwriteIndex.ElemNummer = elemNummer;
      overwriteIndex.Size = stringParam.Length / 64 * 64 + 64; // for some funny reason
      Append(overwriteIndex);

      _reader.BaseStream.Seek(overwriteIndex.FilePosition, SeekOrigin.Begin);
      int n = stringParam.Length;
      for (int i = 0; i < n; i++)
      { _writer.Write((byte)stringParam[i]); }

      for (int i = 0; i < 64; i++) // same reason as above
      { _writer.Write((byte)0); }
    }

    public void Overwrite(StringType type, int elemNummer, string stringParam)
    {
      Init();
      _reader.BaseStream.Seek(0, SeekOrigin.End);
      int iPos = (int)_reader.BaseStream.Position;

      _overwriteType = type;
      _overwriteIndex = null;

      _ocdReader.StringIndexRead += Overwrite_StringIndexRead;
      _ocdReader.ReadStringParamIndices();
      _ocdReader.StringIndexRead -= Overwrite_StringIndexRead;

      if (_overwriteIndex == null)
      {
        _overwriteIndex = new StringParamIndex();
        _overwriteIndex.Type = type;
        _overwriteIndex.ElemNummer = elemNummer;
        _overwriteIndex.Size = stringParam.Length;
        Append(_overwriteIndex);
      }
      else
      {
        _reader.BaseStream.Seek(-16, SeekOrigin.Current);
        _overwriteIndex.Size = stringParam.Length;
        _overwriteIndex.FilePosition = iPos;

        _writer.Write((int)_overwriteIndex.FilePosition);
        _writer.Write((int)_overwriteIndex.Size);
        _writer.Write((int)_overwriteIndex.Type);
        _writer.Write((int)_overwriteIndex.ElemNummer);
      }

      _reader.BaseStream.Seek(_overwriteIndex.FilePosition, SeekOrigin.Begin);
      int n = stringParam.Length;
      for (int i = 0; i < n; i++)
      { _writer.Write((byte)stringParam[i]); }
    }

    private void Overwrite_StringIndexRead(object sender, StringIndexEventArgs args)
    {
      if (args.Index.Type == _overwriteType)
      {
        args.Cancel = true;
        _overwriteIndex = args.Index;
      }
    }

    private void Append(StringParamIndex index)
    {
      Init();
      long p, p0;
      int i;
      p = _param.FirstStrIdxBlock;
      p0 = p;
      while (p > 0)
      {
        p0 = p;
        _writer.BaseStream.Seek(p, SeekOrigin.Begin);
        p = _reader.ReadInt32();
      }
      if (p0 > 0)
      {
        p = p0 + 4;
        _reader.BaseStream.Seek(p, SeekOrigin.Begin); /* index->pos */
        StringParamIndex pIndex = new StringParamIndex();
        pIndex.FilePosition = 1;
        i = 0;
        while (pIndex.FilePosition > 0 && i < FileParam.N_IN_STRINGPARAM_BLOCK)
        {
          pIndex.FilePosition = _reader.ReadInt32();
          pIndex.Size = _reader.ReadInt32();
          pIndex.Type = (StringType)_reader.ReadInt32();
          pIndex.ElemNummer = _reader.ReadInt32();

          p += 16;
          i++;
        }
      }
      else
      { // empty ocad file
        i = FileParam.N_IN_STRINGPARAM_BLOCK;
      }
      if (i == FileParam.N_IN_STRINGPARAM_BLOCK)
      { /* new index block */
        _writer.BaseStream.Seek(0, SeekOrigin.End);
        p = _writer.BaseStream.Position;
        _writer.Write((int)0);
        int n = FileParam.N_IN_STRINGPARAM_BLOCK * 16;
        for (int j = 0; j < n; j++)
        { _writer.Write((byte)0); }
        if (p0 > 0)
        {
          _writer.BaseStream.Seek(p0, SeekOrigin.Begin);
        }
        else
        {
          _writer.BaseStream.Seek(32, SeekOrigin.Begin); /* first String Index Block position */
          _param.FirstStrIdxBlock = (int)p;
        }
        _writer.Write((int)p);
        p += 4;

        index.FilePosition = (int)p + n;
      }
      else
      {
        _writer.BaseStream.Seek(0, SeekOrigin.End);
        index.FilePosition = (int)_writer.BaseStream.Position;
        p -= 16;
      }
      _writer.BaseStream.Seek(p, SeekOrigin.Begin);

      _writer.Write((int)index.FilePosition);
      _writer.Write((int)index.Size);
      _writer.Write((int)index.Type);
      _writer.Write((int)index.ElemNummer);
    }

    protected override void Write(ElementIndex index)
    {
      base.Write(index);

      _writer.Write((int)index.Position);
      _writer.Write((int)index.Length);
      _writer.Write((int)index.Symbol);

      _writer.Write((sbyte)index.ObjectType);
      _writer.Write((sbyte)0); // reserved
      _writer.Write((sbyte)index.Status);
      _writer.Write((sbyte)index.ViewType);
      _writer.Write((short)index.Color);
      _writer.Write((short)0); // reserved
      _writer.Write((short)index.ImportLayer);
      _writer.Write((short)0); // reserved
    }

    public static void Write(Element element, Stream stream)
    {
      Ocad9Writer writer = AppendTo(stream);
      writer.Write(element);
    }
    protected override void Write(Element element)
    {
      WriteElementHeader(element);

      Write(element.Geometry);

      if (element.Text != "")
      {
        _writer.WriteUnicodeString(element.Text);
        int n = ElementV9.TextCount(element.Text);
        for (int i = element.Text.Length * 2; i < n; i++)
        { _writer.BaseStream.WriteByte(0); }
      }
    }

    protected override void WriteElementHeader(Element element)
    {
      ElementV9 elem9 = element as ElementV9;

      WriteElementSymbol(element.Symbol);
      _writer.Write((byte)element.Type);
      _writer.Write((byte)0);

      _writer.Write((short)(element.Angle * 1800 / Math.PI)); // 0.1 Degrees

      _writer.Write((int)element.PointCount());
      int nText = 0;
      if (element.Text.Length > 0)
      { nText = ElementV9.TextCount(element.Text) / 8; }
      _writer.Write((short)nText);
      _writer.Write((short)0); // reserved

      if (elem9 != null)
      {
        _writer.Write((int)elem9.Color);
        _writer.Write((short)elem9.LineWidth);
        _writer.Write((short)elem9.Flags);
      }
      else
      {
        _writer.Write((int)0);
        _writer.Write((short)0);
        _writer.Write((short)0);
      }

      _writer.Write((double)0); // reserved
      _writer.Write((double)0); // reserved
    }

    protected override void WriteElementSymbol(int symbol)
    {
      _writer.Write((int)symbol);
    }

    public void WriteSymbolStatus(Symbol.SymbolStatus status, int symbolStartPosition)
    {
      _reader.BaseStream.Seek(symbolStartPosition + 11, SeekOrigin.Begin);
      _writer.Write((byte)status);
    }

    public void SymbolsSetState(IList<int> symbols, Symbol.SymbolStatus state)
    {
      Init();
      List<int> sortSymbols = null;
      if (symbols != null)
      {
        sortSymbols = new List<int>(symbols);
        sortSymbols.Sort();
      }

      List<int> symPos = new List<int>();

      int iSymbol = 0;
      for (int posIndex = _ocdReader.ReadSymbolPosition(iSymbol);
        posIndex > 0; posIndex = _ocdReader.ReadSymbolPosition(iSymbol))
      {
        symPos.Add(posIndex);

        iSymbol++;
      }

      foreach (int posIndex in symPos)
      {
        Symbol.BaseSymbol symbol = _ocdReader.ReadSymbol(posIndex);
        if (symbols == null || sortSymbols.BinarySearch(symbol.Number) >= 0)
        {
          WriteSymbolStatus(state, posIndex);
        }
      }
    }
  }
  #endregion
}
