using System;
using System.Collections;
using System.Diagnostics;

namespace Geometry
{
  /// <summary>
  /// Summary description for Point.
  /// </summary>
  public abstract class Point: IPoint 
  {
    protected double[] mdVector;
    protected Point(int dimension)
    {
      mdVector = new double[dimension];
    }
    /// <summary>
    /// returns coordinate of dimension index
    /// </summary>
    public double this[int index]
    {
      get
      { return mdVector[index]; }
      set
      { mdVector[index] = value; }
    }

    public double X
    {
      get
      { return mdVector[0]; }
      set
      { mdVector[0] = value; }
    }
    public double Y
    {
      get
      { return mdVector[1]; }
      set
      { mdVector[1] = value; }
    }

    public abstract Box Extent { get; }
    IBox IGeometry.Extent
    { get { return this.Extent; } }

    public IGeometry Border
    { get {return null;} }

    public bool Intersects(IBox box)
    {
      return box.IsWithin(this);
    }
    public bool Intersects(IGeometry other)
    {
      if (other.Topology > this.Topology)
      { return other.Intersects(this); }

      if (other is IMultipartGeometry)
      {
        IMultipartGeometry pMulti = (IMultipartGeometry) other;
        if (pMulti.HasSubparts)
        {
          foreach (IGeometry pSubPart in pMulti.Subparts(this.Extent))
          {
            if (this.Intersects(pSubPart))
            { return true; }
          }
          return false;
        }
      }

      IPoint pOther = (IPoint) other;
      int iMax = Math.Max(this.Dimension, other.Dimension);
      for (int iDim = 0; iDim < iMax; iDim++)
      {
        if (this[iDim] != pOther[iDim])
        { return false; }
      }
      return true;
    }
    public bool Intersects(IBox box,int[] dimensionList)
    {
      return box.IsWithin(this,dimensionList);
    }

    public Point Project(IProjection projection)
    {
      IPoint p = projection.Project(this);

      if (p is Point)
      { return (Point) p; }
      else
      { return Point.Create(p); }
    }
    IGeometry IGeometry.Project(IProjection projection)
    { return Project(projection); }

    public Point Clone()
    {
      Point p = new Vector(this.Dimension);
      for (int i = 0; i < this.Dimension; i++)
      { p[i] = this[i]; }
      return p;
    }
    public int Topology
    { get { return 0; } }
    public abstract int Dimension { get; }

    public static Point Create(int dim)
    {
      if (dim == 2)
      { return new Point2D(); }
      else if (dim == 3)
      { return new Point3D(); }
      else 
      { return new Vector(dim); }
    }

    public static Point Create(IPoint point)
    {
      int iDim = point.Dimension;
      Point p = Point.Create(iDim);
      for (int i = 0; i < iDim; i++)
      { p[i] = point[i]; }
      return p;
    }
    public double Dist2(Point point)
    { return Dist2(point,Math.Min(Dimension,point.Dimension)); }
    public double Dist2(Point point,int dimension)
    {
      double dDist2 = 0;
      for (int i = 0; i < dimension; i++)
      { 
        double d = this[i] - point[i];
        dDist2 += d * d;
      }
      return dDist2;
    }

    public double OrigDist2()
    { return OrigDist2(Dimension); }
    public double OrigDist2(int dimension)
    {
      double dDist2 = 0;
      for (int i = 0; i < dimension; i++)
      { 
        double d = this[i];
        dDist2 += d * d;
      }
      return dDist2;
    }

    public static Point operator + (Point p0, Point p1)
    {
      int iDim = Math.Min(p0.Dimension,p1.Dimension);
      Point pSum = Point.Create(iDim);
      for (int i = 0; i < iDim; i++)
      { pSum[i] = p0[i] + p1[i]; }

      return pSum;
    }

    public static Point operator - (Point p0, Point p1)
    {
      int iDim = Math.Min(p0.Dimension,p1.Dimension);
      Point pDiff = Point.Create(iDim);
      for (int i = 0; i < iDim; i++)
      { pDiff[i] = p0[i] - p1[i]; }

      return pDiff;
    }

    /// <summary>
    /// scalar product
    /// </summary>
    public static Point operator * (double f, Point p)
    {
      int iDim = p.Dimension;
      Point pSkalar = Point.Create(iDim);
      for (int i = 0; i < iDim; i++)
      { pSkalar[i] = f * p[i]; }

      return pSkalar;
    }

    /// <summary>
    /// Vector product
    /// </summary>
    public double VectorProduct(Point p1)
    {
      return this.X * p1.Y - this.Y * p1.X;
    }

  }

  public class Point2D:Point 
  {
    [System.Diagnostics.DebuggerStepThrough()]
    public Point2D():base(2)
    { }
    public Point2D(double x,double y):base(2)
    {
      mdVector[0] = x;
      mdVector[1] = y;
    }
    protected Point2D(int dimension):base(dimension) {}

    public override bool Equals(object obj)
    {
      Point2D cmpr = obj as Point2D;
      if (cmpr == null) 
      { return false; }
      if (cmpr.Dimension != this.Dimension) 
      { return false; }
      return mdVector[0] == cmpr.X && mdVector[1] == cmpr.Y;
    }
    public override int GetHashCode()
    {
      return mdVector[0].GetHashCode() ^ mdVector[1].GetHashCode() ;
    }


    public new double X  
    {
      get
      { return mdVector[0]; } 
      set
      { mdVector[0] = value; }
    }
 
    public new double Y  
    {  
      get
      { return mdVector[1]; } 
      set
      { mdVector[1] = value; }
    }

    #region IGeometry Members

    public override int Dimension
    {
      get { return 2;}
    }

    public override Box Extent 
    {
      get { return new Box(this,this); }
    }

    public new Point2D Clone()
    {
      return new Point2D(mdVector[0],mdVector[1]);
    }

    #endregion
  }

  public class Point3D:Point2D
  {
    public Point3D():base(3)
    { }
    public Point3D(double x,double y,double z):base(3)
    {
      mdVector[0] = x;
      mdVector[1] = y;
      mdVector[2] = z;
    }
    protected Point3D(int dimension):base(dimension)
    { }

    public override bool Equals(object obj)
    {
      Point3D cmpr = obj as Point3D;
      if (cmpr == null) 
      { return false; }
      if (cmpr.Dimension != this.Dimension) 
      { return false; }
      return mdVector[0] == cmpr.X && mdVector[1] == cmpr.Y && mdVector[2] == cmpr.Z;
    }
    public override int GetHashCode()
    {
      return base.GetHashCode();
      //    return mdVector[0].GetHashCode() ^ mdVector[1].GetHashCode() ^ mdVector[2].GetHashCode() ;
    }

    public override string ToString()
    {
      return string.Format("{0};{1};{2}",X,Y,Z);
    }

    #region IGeometry Members

    public double Z  
    {  
      get { return mdVector[2]; } 
      set { mdVector[2] = value; }
    }

    public override int Dimension
    {
      get { return 3;}
    }

    public override Box Extent 
    {
      get { return new Box(this,this); }
    }

    public new Point3D Clone()
    {
      return new Point3D(mdVector[0],mdVector[1],mdVector[2]);
    }

    #endregion
  }

  /// <summary>
  /// Summary description for Point.
  /// </summary>
  public class Vector:Point3D
  {
    private int miDimension;

    public new Point Clone()
    {
      return new Vector((double[]) mdVector.Clone());
    }

    public Vector(int dimension):base(dimension)
    {
      mdVector = new double[dimension];
      miDimension = dimension;
    }
    public Vector(double[] coordinates):base(0)
    {
      mdVector = coordinates;
      miDimension = coordinates.Length;
    }
    public override bool Equals(object obj)
    {
      Point cmpr = obj as Point;
      if (cmpr == null) 
      { return false; }
      if (cmpr.Dimension != miDimension) 
      { return false; }
      for (int i = 0; i < miDimension; i++) 
      {
        if (cmpr[i] != mdVector[i]) 
        { return false; }
      }
      return true;
    }

    public override int GetHashCode()
    {
      if (miDimension == 1) 
      { return mdVector[0].GetHashCode(); }
      else 
      {
        return (mdVector[0].GetHashCode() ^ mdVector[1].GetHashCode()) ;
      }
    }

    public static Vector operator + (Vector v0, Vector v1)
    {
      Debug.Assert(v0.Dimension == v1.Dimension);
      int iDim = v0.Dimension;
      Vector v = new Vector(iDim);
      for (int i = 0; i < iDim; i++)
      { v[i] = v0[i] + v1[i]; }
      return v;
    }
    public static Vector operator - (Vector v0, Vector v1)
    {
      Debug.Assert(v0.Dimension == v1.Dimension);
      int iDim = v0.Dimension;
      Vector v = new Vector(iDim);
      for (int i = 0; i < iDim; i++)
      { v[i] = v0[i] - v1[i]; }
      return v;
    }

    public double Dist2(Vector v)
    {
      Debug.Assert(Dimension == v.Dimension);
      double d2 = 0;
      double d;
      for (int i = 0; i < Dimension; i++)
      { 
        d = this[i] - v[i];
        d2 += d * d;
      }
      return d2;
    }

    public override string ToString()
    {
      if (Dimension > 2)
      { return string.Format("{0};{1};{2}",X,Y,Z); }
      else if (Dimension > 1)
      { return string.Format("{0};{1}",X,Y); }
      else if (Dimension > 0)
      { return X.ToString(); }
      else
      { return "null"; }
    }

    #region IGeometry Members

    public override int Dimension
    {
      get { return miDimension; }
    }

    #endregion

  }
}
