using System;
using System.Collections;
using System.Diagnostics;

namespace Geometry
{
  /// <summary>
  /// Summary description for PointCollection.
  /// </summary>
  public class PointCollection : CollectionBase, IGeometry, IMultipartGeometry
  {
    public PointCollection()
    {	}

    public Point this[int index]
    {
      get
      { return (Point) InnerList[index]; }
      set
      { InnerList[index] = value; }
    }

    public int Add(Point point)
    {
      return InnerList.Add(point);
    }

    public void AddRange(Point[] points)
    {
      InnerList.AddRange(points);
    }
    internal void AddRange(ICollection points)
    {
      InnerList.AddRange(points);
    }

    public IGeometry Border
    { get { return null; } }

    public int Topology
    { get { return 0; } }

    public int Dimension
    {
      get
      {
        int iDim = 0;
        foreach (Point p in this.InnerList)
        { iDim = Math.Max(iDim,p.Dimension); }
        return iDim;
      }
    }

    public Box Extent
    {
      get
      {
        Box pBox = null;
        foreach (Point p in this.InnerList)
        {
          if (pBox == null)
          { pBox = p.Extent.Clone(); }
          else
          { pBox.Include(p); }
        }

        return pBox;
      }
    }

    IBox IGeometry.Extent
    { get { return this.Extent; } }

    public PointCollection Project(IProjection projection)
    {
      PointCollection pPrjList = new PointCollection();
      foreach (Point p in this.InnerList)
      { pPrjList.Add(p.Project(projection)); }

      return pPrjList;
    }
    IGeometry IGeometry.Project(IProjection projection)
    { return this.Project(projection); }

    public bool Intersects(IBox box)
    {
      foreach (Point p in this.Subparts(box))
      { 
        Debug.Assert(p != null,"Dummy usage of p for the compiler");
        return true; 
      }
      return false;
    }

    public bool Intersects(IGeometry geometry)
    {
      if (geometry.Topology > this.Topology)
      { return geometry.Intersects(this); }

      IBox pBoxOther = geometry.Extent;

      foreach (Point p in this.InnerList)
      {
        if (p.Intersects(pBoxOther) == false)
        { continue; }

        if (p.Intersects(geometry))
        { return true; }
      }
      return false;
    }

    #region IMultipartGeometry Members

    public bool HasSubparts
    {
      get
      { return true;	}
    }

    public IEnumerable Subparts()
    {
      return this;
    }

    #region class MyBoxEnumerator
    private class MyBoxEnumerator : IEnumerator
    {
      private PointCollection mpPoints;
      private IBox mpBox;
      private IEnumerator mpEnum;
      private Point mpPoint;

      public MyBoxEnumerator(PointCollection points,IBox box)
      {
        mpPoints = points;
        mpBox = box;
        mpEnum = mpPoints.GetEnumerator();
        Reset();
      }
      #region IEnumerator Members

      public void Reset()
      {
        mpEnum.Reset();
      }

      public Point Current
      {
        get
        {	return mpPoint; }
      }
      object IEnumerator.Current
      { get { return this.Current; } }

      public bool MoveNext()
      {
        do
        {
          if (mpEnum.MoveNext() == false)
          { return false; }
          mpPoint = (Point) mpEnum.Current;
          if (mpPoint.Intersects(mpBox))
          { return true; }
        }
        while (true);
      }

      #endregion
    }
    #endregion
    #region class MyBoxEnumerable
    private class MyBoxEnumerable : IEnumerable
    {
      private PointCollection mpPoints;
      private IBox mpBox;

      public MyBoxEnumerable(PointCollection points,IBox box)
      {
        mpPoints = points;
        mpBox = box;
      }

      public IEnumerator GetEnumerator()
      {	return new MyBoxEnumerator(mpPoints,mpBox);	}
    }
    #endregion
    public IEnumerable Subparts(IBox box)
    {
      return new MyBoxEnumerable(this,box);
    }

    #endregion
  }
}
