using System;
using System.Collections;

namespace Geometry
{
  /// <summary>
  /// Summary description for Line2D.
  /// </summary>
  public class Polyline: IGeometry, IMultipartGeometry
  {
    private PointCollection mpPointList;
    private SegmentList mpSegmentList;
    private InnerLineCollection mpInnerLineList = null;
    private Box mpExtent = null;

    public Polyline()
    {
      mpPointList = new PointCollection();
      mpSegmentList = new SegmentList(this);
    }

    public Polyline(Point[] pointList)
    {
      mpPointList = new PointCollection();
      mpPointList.AddRange(pointList);
      mpSegmentList = new SegmentList(this);
    }

    private Polyline(ICollection pointList)
    {
      mpPointList = new PointCollection();
      mpPointList.AddRange(pointList);
      mpSegmentList = new SegmentList(this);
    }

    internal InnerLineCollection SegmentParts
    { // used from SegmentList
      get
      { return mpInnerLineList; }
    }

    public SegmentList Segments
    {
      get
      { return mpSegmentList; }
    }

    public int Add(Point point)
    {
      int i = mpPointList.Add(point);
      if (mpInnerLineList != null)
      { mpInnerLineList.Add(null); }
      return i;
    }

    public int Add(Line segment)
    {
      int i = mpPointList.Count - 1;
      if (i < 0)
      { mpPointList.Add(segment.Start); }
      else
      {
        if (mpPointList[i].Equals(segment.Start) == false) // TODO implement Equals
        { throw new System.Exception("segment does not fit to the end of polyline"); }
      }
      mpPointList.Add(segment.End);

      if (mpInnerLineList == null && segment.InnerLine() != null)
      {
        mpInnerLineList = new InnerLineCollection();
        for (int iSeg = 0; iSeg < i; i++)
        { mpInnerLineList.Add(null); }
      }
      if (mpInnerLineList != null)
      {	mpInnerLineList.Add(segment.InnerLine()); }

      return i + 1;
    }

    public Polyline Clone()
    {
      return new Polyline(mpPointList);
    }

    public Polyline SubPart(int start,int length)
    {
      Polyline pSubpart = new Polyline();
      int iEnd = start + length;
      for (int i = start; i < iEnd; i++)
      { pSubpart.Add(this.Segments[i]); }
      return pSubpart;
    }

    public PointCollection Points
    {
      get
      { return mpPointList; }
    }

    #region IGeometry Members

    public int Dimension
    {
      get { return 2;}
    }

    public int Topology
    {
      get
      { return 1; }
    }

    public Box Extent
    {
      get 
      { 
        if (mpExtent == null)
        {
          foreach (Line pSeg in this.Segments)
          {
            if (mpExtent == null)
            { mpExtent = pSeg.Extent.Clone(); }
            else
            { mpExtent.Include(pSeg.Extent); }
          }
        }
        return mpExtent;
      }
    }
    IBox IGeometry.Extent
    {	get { return this.Extent;	} }

    public PointCollection Border
    {
      get
      { 
        int iNPoints = mpPointList.Count;
        if (iNPoints == 0)
        { return null; }
        Point pStart = mpPointList[0];
        Point pEnd = mpPointList[iNPoints - 1];
        if (pStart.Intersects(pEnd))
        { return new PointCollection(); }
        else
        {
          PointCollection pBorder = new PointCollection();
          pBorder.AddRange(new Point[] { pStart, pEnd} );
          return pBorder;
        }
      }
    }
    IGeometry IGeometry.Border
    { get { return this.Border;} }

    public bool Intersects(IBox box)
    {
      Box pBox = Extent.Clone();
      if (box.IsWithin(pBox))
      { return true; }
      if (box.Intersects(pBox) == false)
      { return false; }
      foreach (Line pSeg in this.Segments)
      {
        pBox = pSeg.Extent;
        if (box.IsWithin(pBox))
        { return true; }
        if (box.Intersects(pBox))
        {
          if (pSeg.Intersects(box))
          { return true; }
        }
      }
      return false;
    }

    public bool Intersects(IGeometry other)
    {
      if (other.Topology > this.Topology)
      { return other.Intersects(this); }

      IBox pOtherBox = other.Extent;
      int iNSeg = this.Segments.Count - 1;

      for (int iSeg = 0; iSeg < iNSeg; iSeg++)
      {
        Line pSeg = this.Segments[iSeg];
        Box pSegBox = pSeg.Extent;
        if (pSegBox.Intersects(pOtherBox) == false)
        { continue; }

        if (pSeg.Intersects(other))
        { return true; }
      }
      return false;
    }
    public bool Intersects_(Polyline other)
    {
      Box pBox = other.Extent;
      int iNSeg0 = this.Segments.Count - 1;
      int iNSeg1 = other.Segments.Count - 1;
      for (int iSeg0 = 0; iSeg0 <= iNSeg0; iSeg0++)
      {
        Line pSeg0 = this.Segments[iSeg0];
        Box pSeg0Box = pSeg0.Extent;
        if (pSeg0Box.Intersects(pBox))
        {
          for (int iSeg1 = 0; iSeg1 <= iNSeg1; iSeg1++)
          {
            Line pSeg1 = other.Segments[iSeg1];
            if (pSeg0Box.Intersects(pSeg1.Extent))
            {
              double u = -1;
              double v =  0;
              pSeg0.Intersect(pSeg1,ref u,ref v);

              if (0 <= u && u <= 1 && 0 <= v && v <= 1)
              { // check for "touches"
                if (iSeg0 == 0 && u == 0)
                { continue; }
                if (iSeg0 == iNSeg0 && Math.Abs(u - 1) < 1e-8)
                { continue; }
                if (iSeg1 == 0 && v == 0)
                { continue; }
                if (iSeg1 == iNSeg1 && Math.Abs(v - 1) < 1e-8)
                { continue; }

                return true; 
              }
            }
          }
        }
      }
      return false;
    }

    public Polyline Project(IProjection projection)
    {
      Polyline pPoly = new Polyline();
      foreach (Line line in this.Segments)
      { pPoly.Add(line.Project(projection)); }
      return pPoly;
    }
    IGeometry IGeometry.Project(IProjection projection)
    { return Project(projection); }

    #endregion

    #region IMultipartGeometry Members

    public bool HasSubparts
    {
      get
      { return true;	}
    }

    public IEnumerable Subparts()
    {
      return mpSegmentList;
    }

    #region class MyBoxEnumerator
    private class MyBoxEnumerator : IEnumerator
    {
      private SegmentList mpSegments;
      private IBox mpBox;
      private IEnumerator mpEnum;
      private Line mpLine;

      public MyBoxEnumerator(SegmentList segments,IBox box)
      {
        mpSegments = segments;
        mpBox = box;
        mpEnum = mpSegments.GetEnumerator();
        Reset();
      }
      #region IEnumerator Members

      public void Reset()
      {
        mpEnum.Reset();
      }

      public Line Current
      {
        get
        {	return mpLine; }
      }
      object IEnumerator.Current
      { get { return this.Current; } }

      public bool MoveNext()
      {
        do
        {
          if (mpEnum.MoveNext() == false)
          { return false; }
          mpLine = (Line) mpEnum.Current;
          if (mpLine.Intersects(mpBox))
          { return true; }
        }
        while (true);
      }

      #endregion
    }
    #endregion
    #region class MyBoxEnumerable
    private class MyBoxEnumerable : IEnumerable
    {
      private SegmentList mpSegments;
      private IBox mpBox;

      public MyBoxEnumerable(SegmentList segments,IBox box)
      {
        mpSegments = segments;
        mpBox = box;
      }

      public IEnumerator GetEnumerator()
      {	return new MyBoxEnumerator(mpSegments,mpBox);	}
    }
    #endregion
    IEnumerable IMultipartGeometry.Subparts(IBox box)
    {
      return new MyBoxEnumerable(mpSegmentList,box);
    }

    #endregion
  }
}
